#define COLOR2
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <stdlib.h>
#include "paint.h"

static Widget transpProtCoMenu;	/* transparent & protected colors menu */
static Widget UDCoMenuW;	/* user defined color remap */

typedef struct
{
 int rectW, rectH, width, height, rows, columns, gap;
 Boolean initialize;
 int activeCell;
 unsigned char *colArr;
 int *colNo;
 Widget DAW;
} colMatrixStruct;

static colMatrixStruct *activeCM, *passiveCM;

static struct
{
 colMatrixStruct transpCM, protCM;
} prTrArea;

static struct
{
 unsigned char red[MAX_CELLS + 2], green[MAX_CELLS + 2], blue[MAX_CELLS + 2];
 unsigned char colArr[MAX_CELLS], remap[MAX_CELLS];
 int colorNo, unused, CRType, BWTreshold;
 Boolean undo, calcDone;
 Widget previewW, BWScaleW;
 colMatrixStruct colMatrix;
} colRed;

typedef enum { untouched, modified } statusType;

static struct
{
 unsigned char red[MAX_CELLS], green[MAX_CELLS], blue[MAX_CELLS];
 unsigned char selCol[MAX_CELLS], remap[MAX_CELLS], mapCol[MAX_CELLS];
 unsigned char selAreaCol[MAX_CELLS];
 int selColNo, mapColNo, selAreaColNo;
 statusType status;
 Boolean undo, calcDone;
 colMatrixStruct selCM, mapCM, selAreaCM;
 Widget scrollWinW, scrollBX, scrollBY, previewB;
 Rectangle clipWin;
} colRemap;

static void exposeMapWin();
static void setNewMapWinSize();
/* draw rectangle around active cell */
static void drawCellRect(Widget w, int colInd, unsigned long color, colMatrixStruct * CM)
{
 XGCValues values;

 if (CM->columns == 0)
   return;
 values.line_width = CM->gap;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
 XSetForeground(xinfo.dpy, xinfo.gc, color);
 XDrawRectangle(xinfo.dpy, XtWindow (w), xinfo.gc,
  (colInd % CM->columns)*(CM->rectW + CM->gap) + CM->gap / 2,
  (colInd / CM->columns)*(CM->rectH + CM->gap) + CM->gap / 2, CM->rectW + CM->gap, CM->rectH + CM->gap);

 XSetForeground (xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
 values.line_width = ApData.drawinfo.lineWidth;
 if (values.line_width == 1)
   values.line_width--;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
}

/* process drawing area input callback */
static void processInput(Widget w, XtPointer data, XEvent * event)
{
  colMatrixStruct *CM = (colMatrixStruct *) data;

  if ((event->type == ButtonPress) && (event->xbutton.button == Button1))
    {
      if (activeCM->DAW == w)	/*color selection inside area */
	{
	  short selCol, row, col;

	  row = event->xbutton.y / (CM->rectH + CM->gap);
	  col = event->xbutton.x / (CM->rectW + CM->gap);
	  selCol = row * CM->columns + col;

	  if (selCol < *CM->colNo)
	    {
	      /*unselect old active cell than select new one */
	      drawCellRect (w, CM->activeCell, GrayBckGround, CM);
	      CM->activeCell = selCol;
	      drawCellRect (w, selCol, BlackPixel (xinfo.dpy, xinfo.screenNo),
			    CM);
	    }
	}
      else /*change active area */
	{
	  colMatrixStruct *tmp;

	  drawCellRect (activeCM->DAW, activeCM->activeCell,
			GrayBckGround, CM);
	  activeCM->activeCell = 0;

	  tmp = activeCM;
	  activeCM = passiveCM;
	  passiveCM = tmp;
	  drawCellRect (w, activeCM->activeCell,
			BlackPixel (xinfo.dpy, xinfo.screenNo), CM);
	}
    }
  else /* manage popup menu */
  if ((event->type == ButtonPress) && (event->xbutton.button == Button3))
    {
      XmMenuPosition (transpProtCoMenu, (XButtonPressedEvent *) (event));
      XtManageChild (transpProtCoMenu);
    }
}

/* redraw drawing areas of transparent and protected colors */
static void redrawTPDA(Widget w, colMatrixStruct * CM)
{
 Window win = XtWindow (w);
 int colNo, i;
 unsigned char *colors;
 int x, y;

 if (CM->colNo == NULL)
   return;
 colNo = *CM->colNo;
 colors = CM->colArr;

 /*resize occured */
 if (CM->initialize)
   {
     XtVaGetValues (w, XmNwidth, &CM->width, XmNheight, &CM->height, NULL);

     CM->rectW = CM->rectH = 25;
     CM->gap = 5;

     CM->rows = CM->height / (CM->rectH + CM->gap);
     CM->columns = CM->width / (CM->rectW + CM->gap);

     CM->initialize = False;
   }
 /* fill background with color */
 if (XtIsRealized (w))
   {
     XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
     XFillRectangle (xinfo.dpy, win, xinfo.gc, 0, 0,
		     CM->width, CM->height);
   }
 x = y = CM->gap;
 for (i = 0; i < colNo; i++)
   {
     if (xinfo.trueColor)
       XSetForeground (xinfo.dpy, xinfo.gc,
		       xinfo.pixelTC[xinfo.allocCells[colors[i]].pixel]);
     else
       XSetForeground (xinfo.dpy, xinfo.gc, xinfo.allocCells[colors[i]].pixel);
     if ((x + CM->rectW) > CM->width)
       {
	 y += CM->rectH + CM->gap;
	 x = CM->gap;
       }
     XFillRectangle (xinfo.dpy, win, xinfo.gc, x, y, CM->rectW, CM->rectH);
     x += CM->rectW + CM->gap;
   }
 if (activeCM == CM)
  drawCellRect(w, CM->activeCell, xinfo.black, CM);
 XSetForeground(xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
}

/* add selected color to array of active colors */
DCOLOR2 Boolean addColor2Array(unsigned char color, Boolean checkPassive, Boolean soundOn)
{
 int i;
 Boolean dontAdd = False, result = False;

 if ((checkPassive) && (passiveCM))
   for (i = 0; i < *passiveCM->colNo; i++)
     if (passiveCM->colArr[i] == color)
       dontAdd = True;
 for (i = 0; i < *activeCM->colNo; i++)
   if (activeCM->colArr[i] == color)
     dontAdd = True;
 if (dontAdd)
   {
     if (soundOn)
       XBell (xinfo.dpy, 0);
   }
 else
   {
     result = True;
     activeCM->colArr[(*activeCM->colNo)++] = color;
     redrawTPDA (activeCM->DAW, activeCM);
     redrawColSelCB (NULL, "1", NULL);
   }
 return(result);
}

/* delelete one or many entries in transparent or protected colors array */
static void deleteTranspProtColor(colMatrixStruct * CM, int howMany)
{
 if (howMany > 1) /* delete all */
   {
     *CM->colNo = 0;
     CM->activeCell = 0;
   }
 else /* delete selected color */
   {
     int i, selCo = CM->activeCell;

     if (selCo < (*CM->colNo - 1))
       for (i = selCo; i < *CM->colNo; i++)
	 CM->colArr[i] = CM->colArr[i + 1];

     else if ((selCo == (*CM->colNo - 1)) && (selCo > 0))
       CM->activeCell--;

     (*CM->colNo)--;
   }
 XSetForeground(xinfo.dpy, xinfo.gc, GrayBckGround);

 XFillRectangle(xinfo.dpy, XtWindow (CM->DAW), xinfo.gc, 0, 0,
  CM->width, CM->height);
 redrawTPDA(CM->DAW, CM);
}

/* invert selected colors in a region */
static void invertSelectedColors(Boolean checkPassive)
{
  int i, j, oldNo, newNo;
  Boolean dontAdd;
  unsigned char color, oldTbl[MAX_CELLS];

  newNo = 0;
  oldNo = *activeCM->colNo;
  for (i = 0; i < oldNo; i++)
    oldTbl[i] = activeCM->colArr[i];
  for (j = 0; j < ApData.ColorNo; j++)
    {
      color = j;
      dontAdd = False;
      if ((passiveCM) && (checkPassive))
	for (i = 0; i < *passiveCM->colNo; i++)
	  if (passiveCM->colArr[i] == color)
	    {
	      dontAdd = True;
	      break;
	    }
      for (i = 0; i < oldNo; i++)
	if (oldTbl[i] == color)
	  {
	    dontAdd = True;
	    break;
	  }
      if (!dontAdd)
	activeCM->colArr[newNo++] = color;
    }
  *activeCM->colNo = newNo;
  activeCM->activeCell = 0;
  redrawTPDA (activeCM->DAW, activeCM);
  redrawColSelCB (NULL, "1", NULL);
}

/* calculate color transformation table */
DCOLOR2 void calcColTransfTbl()
{
 int i, j;
 unsigned char ind;
 Boolean found;

 for (i = 0; i < MAX_CELLS; i++)
   for (j = 0; j < MAX_CELLS; j++)
     ApData.transfTbl[i][j] = (unsigned char) i;

 /* set protected colors */
 for (i = 0; i < ApData.ProtColNo; i++)
   {
     ind = xinfo.allocCells[ApData.ProtectColors[i]].pixel;
     for (j = 0; j < MAX_CELLS; j++)
       ApData.transfTbl[j][ind] = ind;
   }
 /* set transparent colors */
 for (i = 0; i < ApData.TranspColNo; i++)
   {
     found = False;
     for (j = 0; j < ApData.ProtColNo; j++)
        if (ApData.ProtectColors[j] == ApData.TranspColors[i])
	  {
	   found = True;
	   break;
	  }
     if (found) continue;
     ind = xinfo.allocCells[ApData.TranspColors[i]].pixel;
     for (j = 0; j < MAX_CELLS; j++)
       ApData.transfTbl[ind][j] = (unsigned char) j;
   }
}

/* reduce number of colors, temporarily or really */
static void reduceColorNo(Boolean restore)
{
  int i, j, minInd;
  unsigned char selCol[3], remCol[3], rcolor;
  unsigned long minDist, dist;

  if (restore)
    {
      if (colRed.CRType == 2)
	{
	  colRed.red[0] = colRed.red[MAX_CELLS];
	  colRed.blue[0] = colRed.blue[MAX_CELLS];
	  colRed.green[0] = colRed.green[MAX_CELLS];
	  colRed.red[1] = colRed.red[MAX_CELLS + 1];
	  colRed.blue[1] = colRed.blue[MAX_CELLS + 1];
	  colRed.green[1] = colRed.green[MAX_CELLS + 1];
	}
      for (i = 0; i < ApData.ColorNo; i++)
	{
	  xinfo.red[i] = colRed.red[i];
	  xinfo.green[i] = colRed.green[i];
	  xinfo.blue[i] = colRed.blue[i];
	  xinfo.allocCells[i].red = colRed.red[i] << 8;
	  xinfo.allocCells[i].green = colRed.green[i] << 8;
	  xinfo.allocCells[i].blue = colRed.blue[i] << 8;
	}
      colRed.calcDone = False;
    }
  else
    {
      for (i = 0; i < MAX_CELLS; i++)
	colRed.remap[i] = i;

      if (colRed.CRType == 2)
	{
	  colRed.red[MAX_CELLS] = colRed.red[0];
	  colRed.blue[MAX_CELLS] = colRed.blue[0];
	  colRed.green[MAX_CELLS] = colRed.green[0];
	  colRed.red[MAX_CELLS + 1] = colRed.red[1];
	  colRed.blue[MAX_CELLS + 1] = colRed.blue[1];
	  colRed.green[MAX_CELLS + 1] = colRed.green[1];
	  colRed.red[0] = colRed.green[0] = colRed.blue[0] = 0;
	  colRed.red[1] = colRed.green[1] = colRed.blue[1] = 255;
	  XtVaGetValues (colRed.BWScaleW, XmNvalue, &colRed.BWTreshold, NULL);
	  colRed.colorNo = 2;
	  colRed.colArr[0] = 0;
	  colRed.colArr[1] = 1;
	}
      if (colRed.colorNo < 2)
	return;
      for (i = 0; i < ApData.ColorNo; i++)
	{
	  if (colRed.CRType == 1)
	    {
	      for (j = 0; j < colRed.colorNo; j++)
		if (colRed.colArr[j] == i)
		  break;
	      if (j != colRed.colorNo)
		continue;
	      minInd = 0;
	      minDist = 0xfffff;
	      remCol[RED] = colRed.red[i];
	      remCol[GREEN] = colRed.green[i];
	      remCol[BLUE] = colRed.blue[i];
	      for (j = 0; j < colRed.colorNo; j++)
		{
		  rcolor = colRed.colArr[j];
		  selCol[RED] = colRed.red[rcolor];
		  selCol[GREEN] = colRed.green[rcolor];
		  selCol[BLUE] = colRed.blue[rcolor];
		  dist = getDistance (&selCol[0], &remCol[0]);
		  if (dist < minDist)
		    {
		      minDist = dist;
		      minInd = rcolor;
		    }
		}
	    }
	  else
	    /* convert to BW according to the threshold value */
	    {
	      minDist = (222 * xinfo.red[i] + 707 * xinfo.green[i] +
			 71 * xinfo.blue[i]) / 1000;
	      if (minDist > colRed.BWTreshold)
		minInd = 1;
	      else
		minInd = 0;
	    }
	  colRed.remap[xinfo.allocCells[i].pixel] = xinfo.allocCells[minInd].pixel;
	  xinfo.red[i] = colRed.red[minInd];
	  xinfo.green[i] = colRed.green[minInd];
	  xinfo.blue[i] = colRed.blue[minInd];

	  xinfo.allocCells[i].red = colRed.red[minInd] << 8;
	  xinfo.allocCells[i].green = colRed.green[minInd] << 8;
	  xinfo.allocCells[i].blue = colRed.blue[minInd] << 8;
	}
      colRed.calcDone = True;
    }
  if (xinfo.trueColor)
    calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
     &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
  else
   XStoreColors(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[0], ApData.ColorNo);
  if (xinfo.trueColor)
   {
    refreshDrawingArea(True, True);
    refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
   }
}

/* remap the picture according to the confirmed selected colors */
static void confirmColorReduction()
{
 int i, j, k;
 unsigned char sortArr[MAX_CELLS];

 if (colRed.undo)
   setLabelString(colRed.previewW, getTxtI8N (PREVIEW_MSG));

 XtUnmanageChild(XtParent (colRed.colMatrix.DAW));

 if (colRed.calcDone == False)
   reduceColorNo (False);
 if (colRed.colorNo == 0)
   return;
 remapData((unsigned char *) ApData.img->data,
  &colRed.remap[0], ApData.img->bytes_per_line * ApData.img->height);

 /* copy remained colors to the low indexes of xinfo structure */
 memset (&sortArr[0], 0, MAX_CELLS);
 for (j = 0; j < colRed.colorNo; j++)
   sortArr[colRed.colArr[j]] = 1;

 if (colRed.CRType == 1)
  {
   k = 0;
   for (j = 0; j < colRed.colorNo; j++)
    {
     XColor tmp;

     i = colRed.colArr[j];
     if (i < colRed.colorNo)
       continue;
     while ((sortArr[k]) && (k < MAX_CELLS))
       k++;
     memcpy (&tmp, &xinfo.allocCells[i], sizeof (XColor));
     memcpy (&xinfo.allocCells[i], &xinfo.allocCells[k], sizeof (XColor));
     memcpy (&xinfo.allocCells[k], &tmp, sizeof (XColor));
     SWAP (xinfo.red[k], xinfo.red[i]);
     SWAP (xinfo.green[k], xinfo.green[i]);
     SWAP (xinfo.blue[k], xinfo.blue[i]);
     sortArr[k] = 1;
     k++;
    }
  }
 else /* we forced remap to the first two values - so set color accordingly */
  {
   xinfo.red[0] = xinfo.green[0] = xinfo.blue[0] = 0;
   xinfo.red[1] = xinfo.green[1] = xinfo.blue[1] = 255;

   xinfo.allocCells[0].red = xinfo.allocCells[0].green =
     xinfo.allocCells[0].blue = 0;
   xinfo.allocCells[1].red = xinfo.allocCells[1].green =
     xinfo.allocCells[1].blue = 255 << 8;
  }

 freeColors (ApData.ColorNo - colRed.colorNo);
 if (xinfo.trueColor)
   calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
    &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
 else
   XStoreColors (xinfo.dpy, xinfo.cmap, &xinfo.allocCells[0], colRed.colorNo);
 ApData.ColorNo = colRed.colorNo;
 initColorSelectionArea (ApData.ColorNo, 0, 0);
 ColorSelectConseq(Button1, 0);
 ColorSelectConseq(Button4, 1);

 if (Shm.alloc)
   XSync (xinfo.dpy, False);

 refreshDrawingArea(True, True);
 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
  zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);

 ApData.pic_change = True;
 if (Select.selected)
  Select.colorCons = False;
}

#define TPC_OK 1
#define TPC_CANCEL 2
#define TPC_RC_OK 3
#define TPC_RC_CANCEL 4
#define TPC_EXPOSE 5
#define TPC_RESIZE 6
#define TPC_MAP 7
#define TPC_UNMAP 8
#define TPC_INIT_TRANSP 9
#define TPC_INIT_PROT 10
#define TPC_INIT_MENU 11
#define TPC_DEL_ONE 12
#define TPC_DEL_ALL 13
#define TPC_INIT_CR 14
#define TPC_MAP_CR 15
#define TPC_CR_PREVIEW 16
#define TPC_CR_SCALE_INIT 17
#define TPC_CR_COLOR 18
#define TPC_CR_BW 19
#define TPC_CR_SCALE_CH_VAL 20
#define TPC_NOT_ON_SEL_COLORS 21
#define TPC_EXPOSE_TRANSP 22
#define TPC_EXPOSE_PROT 23
#define TPC_CR_EXPOSE 24
#define TPC_RESIZE_TRANSP 25
#define TPC_RESIZE_PROT 26
#define TPC_CR_RESIZE 27

DCOLOR2 void transpProtColCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static Widget dialogW=NULL;
 static unsigned char oldProtCol[MAX_CELLS], oldTransCol[MAX_CELLS];
 static unsigned char oldProtColNo, oldTransColNo;
 int prm = atoi ((char *) clientData), i;

 switch (prm)
  {
   case TPC_EXPOSE_TRANSP: redrawTPDA(w, &prTrArea.transpCM); break;
   case TPC_EXPOSE_PROT: redrawTPDA(w, &prTrArea.protCM); break;
   case TPC_CR_EXPOSE: redrawTPDA(w, &colRed.colMatrix); break;
   case TPC_RESIZE_TRANSP:
     prTrArea.transpCM.initialize = True;
     redrawTPDA(w, &prTrArea.transpCM);
     break;
   case TPC_RESIZE_PROT:
     prTrArea.protCM.initialize = True;
     redrawTPDA(w, &prTrArea.protCM);
     break;
   case TPC_CR_RESIZE:
     colRed.colMatrix.initialize = True;
     redrawTPDA(w, &colRed.colMatrix);
     break;
   case TPC_OK:
     if (dialogW)
      if (widgetIsViewable(dialogW))
       XtUnmanageChild(dialogW);
     calcColTransfTbl();
     if ((ApData.ProtColNo != 0) || (ApData.TranspColNo != 0))
       ApData.specialPutImg = True;
     else
       ApData.specialPutImg = False;

     if ((Select.selected) && (Select.rect_on))
       {
	 drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
	 /* first copy original part of picture to zoom pixmap */
	 refreshZoomWin(opBounds.x, opBounds.y, opBounds.width,
			 opBounds.height, True);
	 LastDrawOnAllWin(False);
       }
     else if (Undo.tool == pasteMulA)
       LastDrawOnAllWin (False);
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_CANCEL:
     XtUnmanageChild(dialogW);

     ApData.ProtColNo = oldProtColNo;
     ApData.TranspColNo = oldTransColNo;
     for (i = 0; i < MAX_CELLS; i++)
       {
	 ApData.ProtectColors[i] = oldProtCol[i];
	 ApData.TranspColors[i] = oldTransCol[i];
       }
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_DEL_ONE:
     deleteTranspProtColor (activeCM, 1);
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_DEL_ALL:
     deleteTranspProtColor (activeCM, 100);
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_MAP:
     oldProtColNo = (unsigned char) ApData.ProtColNo;
     oldTransColNo = (unsigned char) ApData.TranspColNo;

     for (i = 0; i < MAX_CELLS; i++)
       {
	 oldProtCol[i] = ApData.ProtectColors[i];
	 oldTransCol[i] = ApData.TranspColors[i];
       }
     prTrArea.protCM.colNo = &ApData.ProtColNo;
     prTrArea.protCM.colArr = ApData.ProtectColors;
     prTrArea.transpCM.colNo = &ApData.TranspColNo;
     prTrArea.transpCM.colArr = ApData.TranspColors;

     activeCM = &prTrArea.transpCM;
     passiveCM = &prTrArea.protCM;

     dialogW = w;
     XtRemoveEventHandler (colSelW, ButtonPressMask, FALSE,
			   (XtEventHandler) ColorSelectPress, 0);
     XtAddEventHandler (colSelW, ButtonPressMask, FALSE,
			(XtEventHandler) ColorSelectPress, (XtPointer) 1);
     break;
   case TPC_UNMAP:
     XtRemoveEventHandler(colSelW, ButtonPressMask, FALSE,
			 (XtEventHandler) ColorSelectPress, (XtPointer) 1);
     XtAddEventHandler(colSelW, ButtonPressMask, FALSE, (XtEventHandler) ColorSelectPress, 0);
     break;
   case TPC_INIT_TRANSP:
     prTrArea.transpCM.DAW = w;
     prTrArea.transpCM.colNo = NULL;
     prTrArea.transpCM.initialize = True;
     XtAddEventHandler (w, ButtonPressMask, FALSE,
       (XtEventHandler) processInput, (XtPointer) &prTrArea.transpCM);
     break;
   case TPC_INIT_PROT:
     prTrArea.protCM.DAW = w;
     prTrArea.protCM.initialize = True;
     prTrArea.protCM.colNo = NULL;
     XtAddEventHandler (w, ButtonPressMask, FALSE,
       (XtEventHandler) processInput, (XtPointer) &prTrArea.protCM);
     break;
   case TPC_INIT_MENU: transpProtCoMenu = w; break;
   case TPC_INIT_CR:
     colRed.colMatrix.DAW = w;
     colRed.colMatrix.colNo = &colRed.colorNo;
     colRed.colMatrix.colArr = colRed.colArr;
     colRed.colMatrix.initialize = True;
     XtAddEventHandler (w, ButtonPressMask, FALSE,
	    (XtEventHandler) processInput, (XtPointer) & colRed.colMatrix);
     break;
   case TPC_CR_PREVIEW:
     colRed.previewW = w;
     if (colRed.undo)
      {
       reduceColorNo(True);
       setLabelString(w, getTxtI8N (PREVIEW_MSG));
      }
     else
      {
       setLabelString(w, getTxtI8N (UNDO_MSG));
       reduceColorNo(False);
      }
     colRed.undo = !colRed.undo;
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_RC_CANCEL:
     if (colRed.calcDone)
       reduceColorNo (True);
     if (colRed.undo)
       setLabelString(colRed.previewW, getTxtI8N (PREVIEW_MSG));
     XtUnmanageChild (XtParent (colRed.colMatrix.DAW));
     if (xinfo.trueColor)
      drawColorPalette();
     break;
   case TPC_RC_OK:
     confirmColorReduction();
     break;
   case TPC_MAP_CR:
     finalDraw (ApData.pixy);
     disableUndo ();

     colRed.colMatrix.activeCell = 0;
     colRed.colorNo = colRed.unused = 0;
     colRed.previewW = 0;
     activeCM = &colRed.colMatrix;
     passiveCM = NULL;

     memcpy(&colRed.red[0], &xinfo.red[0], sizeof (unsigned char) * MAX_CELLS);
     memcpy(&colRed.green[0], &xinfo.green[0], sizeof (unsigned char) * MAX_CELLS);
     memcpy(&colRed.blue[0], &xinfo.blue[0], sizeof (unsigned char) * MAX_CELLS);
     colRed.undo = colRed.calcDone = False;
     XtRemoveEventHandler(colSelW, ButtonPressMask, FALSE, (XtEventHandler) ColorSelectPress, 0);
     XtAddEventHandler(colSelW, ButtonPressMask, FALSE, (XtEventHandler) ColorSelectPress, (XtPointer) 1);
     Select.rect_on = False;
     ApData.tool = Undo.tool = none;
     toolBarCB(curveDW, "0", NULL);
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_CR_SCALE_INIT:
     colRed.BWScaleW = w;
     colRed.CRType = 1;
     break;
   case TPC_CR_SCALE_CH_VAL:
     if (colRed.previewW == 0)
       return;
     if (colRed.calcDone)
       reduceColorNo(True);
     reduceColorNo(False);
     if (colRed.undo == False)
       setLabelString(colRed.previewW, getTxtI8N (UNDO_MSG));
     colRed.undo = True;
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_CR_COLOR:
     if ((colRed.CRType != 1) && (colRed.calcDone))
       {
	reduceColorNo (True);
	setLabelString(colRed.previewW, getTxtI8N (PREVIEW_MSG));
       }
     colRed.CRType = 1;
     colRed.colorNo = 0;
     redrawTPDA(colRed.colMatrix.DAW, &colRed.colMatrix);
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_CR_BW:
     if ((colRed.CRType != 2) && (colRed.calcDone))
       {
	reduceColorNo(True);
	setLabelString(colRed.previewW, getTxtI8N (PREVIEW_MSG));
       }
     colRed.CRType = 2;
     colRed.colorNo = 0;
     redrawColSelCB (NULL, "1", NULL);
     break;
   case TPC_NOT_ON_SEL_COLORS: invertSelectedColors(False); break;
  }
}

/* check whether the image column is all of the same color */
static Boolean oneColorColumnRow(XImage * img, int colRo, int start, int stop,
 Pixel color, Boolean column)
{
 int i = start, bpl = img->bytes_per_line;

 if (column)
  {
   stop = min (img->height, stop);
   colRo = min (img->width, colRo);
   while ((i < stop) && (img->data[bpl * i + colRo] == color))
     i++;
  }
 else
  {
   stop = min (img->width, stop);
   colRo = min (img->height, colRo);
   while ((i < stop) && (img->data[bpl * colRo + i] == color))
     i++;
  }
 return(i == stop);
}

/* cut out color border from the selected area */
DCOLOR2 void cutOutColorBorder()
{
 int width, height, x, y, xOff, yOff;
 XImage *img;
 Boolean OK;

 if (ApData.viewMode != normalView)
   {
     XBell (xinfo.dpy, 0);
     return;
   }

 if (Select.selected)
   if ((Select.height < 2) || (Select.width < 2))
     return;
 finalDraw (ApData.pixy);
 disableUndo ();

 setToolEventHandler(none, XC_left_ptr);
 highlightButton(NULL);
 if (XtIsRealized(zoomDrawAreaW))
  drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW),
   zoomDraw.factorX, zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
 ApData.tool = none;
 if ((Select.selected) && (Select.rect_on))
   {
     if (Select.imgAllocated)
       undoXImage (&Select.img);
     Select.imgAllocated = False;
     drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW),
	zoomDraw.factorX, zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
   }
 else
   {
     Select.x = Select.y = 0;
     Select.width = ApData.width;
     Select.height = ApData.height;
     Select.npoints = 0;
     setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);
   }
 Select.rectangular = True;

 img = ApData.img;
 xOff = Select.x;
 yOff = Select.y;
 x = xOff;
 y = yOff;
 width = Select.width + xOff;
 height = Select.height + yOff;

 OK = True;
 while (OK && (x < width))
  OK = oneColorColumnRow(img, x++, y, height, xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, True);
 x--;
 OK = True;
 while (OK && (width > x))
  OK = oneColorColumnRow (img, width-- - 1, y, height, xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, True);
 width++;
 OK = True;
 while (OK && (y < height))
  OK = oneColorColumnRow (img, y++, x, width, xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, False);
 y--;
 OK = True;
 while (OK && (height > y))
  OK = oneColorColumnRow (img, height-- - 1, x, width, xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, False);
 height++;

 ApData.tool = Undo.tool = selectA;

 Select.npoints = 0;

 Select.x = x;
 Select.y = y;
 Select.width = width - x;
 Select.height = height - y;

 Select.selected = Select.rect_on = True;

 setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);

 drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
 drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW),
  zoomDraw.factorX, zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);

 XtVaSetValues (CopyBW, XmNsensitive, True, NULL);
 XtVaSetValues (CutBW, XmNsensitive, True, NULL);
 enableUndo (True);
 setToolEventHandler (pasteA, 0);
}

/* create a multibit pixel from red, green, blue for visuals with  more than 8bpp */
DCOLOR2 unsigned long eight2MoreBits(int red, int green, int blue)
{
 unsigned long result;

 result = 0;
 if (xinfo.redShift >= 0)
   result = (red << xinfo.redShift) & xinfo.visualInfo->red_mask;
 else
   result = (red >> xinfo.redShift) & xinfo.visualInfo->red_mask;
 if (xinfo.greenShift >= 0)
   result |= (green << xinfo.greenShift) & xinfo.visualInfo->green_mask;
 else
   result |= (green >> xinfo.greenShift) & xinfo.visualInfo->green_mask;
 if (xinfo.blueShift >= 0)
   result |= (blue << xinfo.blueShift) & xinfo.visualInfo->blue_mask;
 else
   result |= (blue >> -xinfo.blueShift) & xinfo.visualInfo->blue_mask;
 return (result);
}


/* calculate true color pixel values - in visual with more than 8 bpp and
   store it in xinfo.pixelTC array */
DCOLOR2 void calcTrueColorPixels(XColor * colCells, unsigned long *TCpixel,
 unsigned char *red, unsigned char *green, unsigned char *blue)
{
 int i;

 for (i = 0; i < MAX_CELLS; i++)
  TCpixel[colCells[i].pixel] = eight2MoreBits (red[i], green[i], blue[i]);
}

/* converts a rectangle x,y,width, height of 8 bit XImage to true color XImage.
   The size of the converted rectangle can be magnified by magnX, magnY and is
   always put on coordinates 0,0 of the imgTC. Caller must allocate the true color XImage. */
DCOLOR2 void convertImg2TrueColor(XImage * img8, XImage * imgTC, int x, int y,
 unsigned int width, unsigned int height, unsigned int magnX, unsigned int magnY)
{
 Rectangle img1R, img2R, tmpR, resR;
 int newWidth, newHeight, i, j;
 Boolean LSBboth;
 unsigned char *p, *r;
 unsigned short *r2;
 unsigned long pixelTC, *r4;

 if ((img8 == NULL) || (imgTC == NULL))
   return;
 if ((img8->depth != 8) || (imgTC->depth < 9))
   return;
 magnX = MAX (1, magnX);
 magnY = MAX (1, magnY);

 /* first do clipping on the size */
 resR.x = x;
 resR.y = y;
 resR.width = width;
 resR.height = height;

 img2R.x = img2R.y = 0;
 img2R.width = img8->width;
 img2R.height = img8->height;
 if (!intersectRect (&img2R, &resR, &img1R))
   return;

 img2R.x = img2R.y = tmpR.x = tmpR.y = 0;
 tmpR.width = magnX * img1R.width;
 tmpR.height = magnY * img1R.height;
 img2R.width = imgTC->width;
 img2R.height = imgTC->height;
 if (!intersectRect (&img2R, &tmpR, &resR))
   return;

 newWidth = resR.width / magnX;
 newHeight = resR.height / magnY;

 LSBboth = ((imgTC->bitmap_bit_order==LSBFirst) && (imgTC->byte_order==LSBFirst));
 if (!LSBboth)
   return;

 for (j = 0; j < newHeight; j++)
  {
   p = &img8->data[img8->bytes_per_line * (j + img1R.y) + img1R.x];
   r = &imgTC->data[imgTC->bytes_per_line * j];
   switch (imgTC->bits_per_pixel)
    {
     case 32:
      r4 = (unsigned long *) r;
      for (i = newWidth; i; i--) /*for (i = 0; i < newWidth; i++)*/
       *r4++ = xinfo.pixelTC[*p++];
      break;
     case 16:
      r2 = (unsigned short *) r;
      for (i = newWidth; i; i--)
       *r2++ = (unsigned short) xinfo.pixelTC[*p++];
      break;
     case 24:
      for (i = newWidth; i; i--)
       {
 	pixelTC = xinfo.pixelTC[*p++];
	*r++ = (unsigned char) pixelTC;
	*r++ = (unsigned char) (pixelTC >> 8);
	*r++ = (unsigned char) (pixelTC >> 16);
       }
    }
  }
}

/* get palette index for pixel pix in XColor array of colNo allocated colors */
static int getColorIndex(Pixel pix, XColor * cArr, int colNo)
{
 int i, res;

 res = -1;
 for (i = 0; i < colNo; i++)
  if (cArr[i].pixel == pix)
    {
     res = i;
     break;
    }
 return(res);
}

/* init all the structures for color remapping */
static void initUDcolorRemap()
{
 int i, ind;
 unsigned long hist[MAX_CELLS];
 Pixel pix;
 Rectangle rect, wholeA, selectR;
 XImage *img;

 highlightButton (NULL);
 finalDraw(ApData.pixy);
 disableUndo();
 setToolEventHandler(none, 0);

 colRemap.selColNo = colRemap.mapColNo = 0;
 for (i = 0; i < MAX_CELLS; i++)
  colRemap.selCol[0] = colRemap.mapCol[0] = 0;

 if (!Select.selected)
  {
   disableSelection(True);
   Select.x = Select.y = 0;
   Select.width = ApData.width;
   Select.height = ApData.height;
   Select.selected = Select.rect_on = Select.rectangular = True;
   setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);
  }
 else
  {
   Rectangle oldSel;
   Boolean oldrect;
   int oldnpts;

   oldSel.x = Select.x;
   oldSel.y = Select.y;
   oldSel.width = Select.width;
   oldSel.height = Select.height;
   oldrect = Select.rectangular;
   oldnpts = Select.npoints;
   disableSelection (True);
   Select.x = oldSel.x;
   Select.y = oldSel.y;
   Select.width = oldSel.width;
   Select.rectangular = oldrect;
   Select.height = oldSel.height; /* boundary checking !!!! */
   Select.npoints = oldnpts;
   if (Select.rectangular)
    setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);
   else
    Select.reg = XPolygonRegion(Select.points, Select.npoints, WindingRule);
   Select.rect_on = Select.selected = True;
  }
 selectR.x = Select.x;
 selectR.y = Select.y;
 selectR.width = (unsigned int) Select.width;
 selectR.height = (unsigned int) Select.height;

 wholeA.x = wholeA.y = 0;
 wholeA.width = (unsigned int) ApData.width;
 wholeA.height = (unsigned int) ApData.height;

 if (intersectRect(&selectR, &wholeA, &rect))
  {
   img = SubImage(ApData.img, rect.x, rect.y, rect.width, rect.height);
   if (checkXImage(img))
    {
     ApData.tool = Undo.tool = mapColorsA;

     Select.width = opBounds.width = rect.width;
     Select.height = opBounds.height = rect.height;
     Select.x = opBounds.x = Undo.points[Undo.npoints].x = rect.x;
     Select.y = opBounds.y = Undo.points[Undo.npoints++].y = rect.y;

     enableUndo (False);
     Undo.img = img;
     Undo.imgAllocated = True;
     memcpy (&Undo.area, &rect, sizeof (Rectangle));
     ApData.pic_change = True;
     opBounds.x -= 2;
     opBounds.y -= 2;
     opBounds.width += 4;
     opBounds.height += 4;
    }
  }

 initHist (&hist[0]);
 addToHistFromRegion (&hist[0], (unsigned char *) ApData.img->data, Select.x, Select.y,
  Select.width, Select.height, ApData.img->bytes_per_line, Select.reg, 0, 0);

 for (i = 0; i < MAX_CELLS; i++)
  if (hist[i] > 0)
   colRemap.selCol[colRemap.selColNo++] = i;
 for (i = 0; i < colRemap.selColNo; i++)
  {
   pix = colRemap.selCol[i];
   ind = getColorIndex (pix, &xinfo.allocCells[0], maxColors);
   if (ind > -1)
    colRemap.selCol[i] = ind;
   else
    colRemap.selCol[i] = 0; /* error - illegal index! */
  }
 for (i = 0; i < colRemap.selColNo; i++)
   colRemap.selAreaCol[i] = colRemap.selCol[i];
 colRemap.selAreaColNo = colRemap.selColNo;

 colRemap.selCM.initialize = True;
 colRemap.selCM.colArr = colRemap.selCol;
 colRemap.selCM.colNo = &colRemap.selColNo;
 colRemap.selCM.activeCell = 0;

 colRemap.mapCM.initialize = True;
 colRemap.mapCM.colArr = colRemap.mapCol;
 colRemap.mapCM.colNo = &colRemap.mapColNo;
 colRemap.mapCM.activeCell = 0;
 colRemap.mapCM.rectW = colRemap.mapCM.rectH = 25;
 colRemap.mapCM.gap = 5;

 for (i = 0; i < MAX_CELLS; i++)
   colRemap.remap[i] = i;
 activeCM = &colRemap.mapCM;
 setNewMapWinSize ();

 redrawTPDA (colRemap.selCM.DAW, &colRemap.selCM);

 colRemap.selAreaCM.initialize = True;
 colRemap.selAreaCM.colArr = colRemap.selAreaCol;
 colRemap.selAreaCM.colNo = &colRemap.selAreaColNo;
 passiveCM = &colRemap.selAreaCM;
 colRemap.status = untouched;

 setLabelString(colRemap.previewB, getTxtI8N (PREVIEW_MSG));

 XtRemoveEventHandler(colSelW, ButtonPressMask, FALSE,
  (XtEventHandler) ColorSelectPress, 0);
 XtAddEventHandler(colSelW, ButtonPressMask, FALSE,
  (XtEventHandler) ColorSelectPress, (XtPointer) 2);
}

/* find out which color index in mapped color array was selected */
static int whichMapColInd(row, col)
{
 int i, result;
 unsigned char destColor = 0, mappedColor = 0;

 result = 0;
 col -= 2; /* beause first two rows are for color and arrow */
 if ((col < 0) || (row >= colRemap.mapColNo))
   return (-1);
 destColor = colRemap.mapCol[row];
 for (i = 0; i < MAX_CELLS; i++)
   {
     if ((colRemap.remap[i] == destColor) && (destColor != i))
       {
	 if (mappedColor == col)
	   {
	     result = i;
	     break;
	   }
	 else
	   mappedColor++;
       }
   }
 return(result);
}

/* process drawing area input callback */
static void processInputUDCM(Widget w, XtPointer data, XEvent * event)
{
  colMatrixStruct *CM = (colMatrixStruct *) data;

  if ((event->type == ButtonPress) && (event->xbutton.button == Button1))
    {
      int row, col, selCol, colInd;

      row = event->xbutton.y / (CM->rectH + CM->gap);
      col = event->xbutton.x / (CM->rectW + CM->gap);
      selCol = row * CM->columns + col;
      selCol = MIN (MAX_CELLS, selCol);
      colInd = CM->colArr[selCol];

      if (CM == &colRemap.selCM)
	{
	  if ((colRemap.mapColNo <= 0) || (selCol >= colRemap.selColNo))
	    return;

	  CM->activeCell = selCol;
	  deleteTranspProtColor (CM, 1);
	  /* map color */
	  colRemap.remap[colInd] = colRemap.mapCol[colRemap.mapCM.activeCell];
	  setNewMapWinSize ();
	  exposeMapWin ();
	}
      else if (CM == &colRemap.mapCM)
	{
	  selCol = whichMapColInd (row, col);
	  if (selCol < 0)
	    {
	      if (row < *CM->colNo)
		{
		  drawCellRect (CM->DAW, CM->activeCell * CM->columns, GrayBckGround,
				CM);
		  CM->activeCell = row;
		  drawCellRect (CM->DAW, CM->activeCell * CM->columns, xinfo.black,
				CM);
		}
	    }
	  else
	    {
	      colRemap.remap[selCol] = selCol;
	      activeCM = &colRemap.selCM;
	      addColor2Array (selCol, False, False);
	      drawCellRect (activeCM->DAW, activeCM->activeCell, GrayBckGround,
			    activeCM);
	      activeCM = CM;
	      setNewMapWinSize ();
	    }
	  exposeMapWin ();
	}
    }
  else
    /* manage popup menu */
    if ((event->type == ButtonPress) && (event->xbutton.button == Button3) &&
	(w == colRemap.mapCM.DAW))
    {
      XmMenuPosition (UDCoMenuW, (XButtonPressedEvent *) (event));
      XtManageChild (UDCoMenuW);
    }
}

/* delete one or all colors ponders */
static void deletePivotColors()
{
 int colInd, i, mapped;
 Boolean found;

 drawCellRect (activeCM->DAW, activeCM->activeCell, GrayBckGround, activeCM);
 /* delete pivot color and all its mapping colors !!! */

 mapped = 0;
 colInd = colRemap.mapCM.colArr[colRemap.mapCM.activeCell];
 for (i = 0; i < MAX_CELLS; i++)
   if ((colRemap.remap[i] == colInd) && (colInd != i))
     {
       colRemap.remap[i] = i;
       activeCM = &colRemap.selCM;
       addColor2Array (i, False, False);
       activeCM = &colRemap.mapCM;
       mapped++;
     }
 for (i = 0; i < colRemap.selAreaColNo; i++)
   if ((colRemap.selAreaCol[i] == colInd) && (mapped == 0))
     {

       activeCM = &colRemap.selCM;
       addColor2Array (colInd, False, False);
       activeCM = &colRemap.mapCM;
       break;
     }
 found = False;
 colRemap.mapColNo--;
 for (i = 0; i < colRemap.mapColNo; i++)
   {
     if (colInd == colRemap.mapCol[i])
       found = True;
     if (found)
       colRemap.mapCol[i] = colRemap.mapCol[i + 1];
   }
 colRemap.mapCM.activeCell = MIN (colRemap.mapColNo - 1,
				  colRemap.mapCM.activeCell);
 colRemap.mapCM.activeCell = MAX (0, colRemap.mapCM.activeCell);
 setNewMapWinSize ();
 exposeMapWin ();
 redrawTPDA(colRemap.selCM.DAW, &colRemap.selCM);
}

/* expose procedure for window with mapped colors */
static void exposeMapWin()
{
 int i, j, startRow, startCol, rows, cols, x, y, cellW, cellH, currClmn;
 unsigned char currCol;
 Window win;
 XGCValues values;

 getClipWinDims (colRemap.scrollWinW, colRemap.scrollBX, colRemap.scrollBY,
		 0xffff, 0xffff, &colRemap.clipWin);
 win = XtWindow (colRemap.mapCM.DAW);
 cellW = colRemap.mapCM.rectW + colRemap.mapCM.gap;
 cellH = colRemap.mapCM.rectH + colRemap.mapCM.gap;
 startCol = colRemap.clipWin.x / cellW;
 startRow = colRemap.clipWin.y / cellH;
 cols = MIN ((colRemap.clipWin.width + cellW - 1)/ cellW,
		 colRemap.mapCM.columns + 1);
 rows = (colRemap.clipWin.height + cellH - 1)/ cellH;
 rows = MIN (colRemap.mapColNo + 1, rows);

 XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
 XFillRectangle (xinfo.dpy, win, xinfo.gc, colRemap.clipWin.x, colRemap.clipWin.y,
		 colRemap.clipWin.width, colRemap.clipWin.height);

 y = startRow * cellH + colRemap.mapCM.gap;
 if (colRemap.mapColNo <= 0)
   return;
 values.line_width = 2;
 XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);

 for (j = 0; j < rows; j++)
   {
     x = startCol * cellW + colRemap.mapCM.gap;
     currCol = colRemap.mapCol[j];
     currClmn = 2;

     if (startCol == 0)
       {
	 if (xinfo.trueColor)
	   XSetForeground (xinfo.dpy, xinfo.gc,
			   xinfo.pixelTC[xinfo.allocCells[currCol].pixel]);
	 else
	   XSetForeground (xinfo.dpy, xinfo.gc, xinfo.allocCells[currCol].pixel);
	 XFillRectangle (xinfo.dpy, win, xinfo.gc, x, y,
			 colRemap.mapCM.rectW, colRemap.mapCM.rectH);
	 x += cellW;
       }
     if (startCol < 2)
       {
	 /* arrow */
	 XSetForeground (xinfo.dpy, xinfo.gc, xinfo.black);
	 XDrawLine (xinfo.dpy, win, xinfo.gc, x, y + colRemap.mapCM.rectH / 2,
		    x + colRemap.mapCM.rectW, y + colRemap.mapCM.rectH / 2);
	 XDrawLine (xinfo.dpy, win, xinfo.gc, x, y + colRemap.mapCM.rectH / 2,
	       x + colRemap.mapCM.rectW / 2, y + colRemap.mapCM.rectH / 4);
	 XDrawLine (xinfo.dpy, win, xinfo.gc, x, y + colRemap.mapCM.rectH / 2,
	   x + colRemap.mapCM.rectW / 2, y + 3 * colRemap.mapCM.rectH / 4);
	 x += cellW;
       }
     for (i = 0; i < MAX_CELLS; i++)
       if ((currCol == colRemap.remap[i]) && (i != colRemap.remap[i]))
	 {
	   if (currClmn >= startCol)
	     {
	       if (xinfo.trueColor)
		 XSetForeground (xinfo.dpy, xinfo.gc,
				 xinfo.pixelTC[xinfo.allocCells[i].pixel]);
	       else
		 XSetForeground (xinfo.dpy, xinfo.gc, xinfo.allocCells[i].pixel);
	       XFillRectangle (xinfo.dpy, win, xinfo.gc, x, y,
			       colRemap.mapCM.rectW, colRemap.mapCM.rectH);
	       x += cellW;
	     }
	   currClmn++;
	   if (x > (colRemap.clipWin.width + colRemap.clipWin.x))
	     break;
	 }
     y += cellH;
   }
 if (colRemap.mapCM.colNo)
   drawCellRect (colRemap.mapCM.DAW,
	   colRemap.mapCM.activeCell * colRemap.mapCM.columns, xinfo.black,
		 &colRemap.mapCM);
 XSetForeground(xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
 values.line_width = ApData.drawinfo.lineWidth;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
}

/* after each adding or removing of color to/from the map window resize the
   drawing area if necessary */
static void setNewMapWinSize()
{
 int i, j, row, column, maxX;

 row = colRemap.mapCM.rows = MAX (colRemap.mapColNo, 1);
 maxX = 0;
 for (i = 0; i < colRemap.mapColNo; i++)
   {
     column = 0;
     for (j = 0; j < MAX_CELLS; j++)
       if ((colRemap.mapCol[i] == colRemap.remap[j]) &&
	   (colRemap.mapCol[i] != j))
	 column++;
     maxX = MAX (column, maxX);
   }
 maxX += 2;
 colRemap.mapCM.width = maxX * (colRemap.mapCM.rectW + colRemap.mapCM.gap) +
   colRemap.mapCM.gap;
 colRemap.mapCM.height = row * (colRemap.mapCM.rectH + colRemap.mapCM.gap) +
   colRemap.mapCM.gap;
 colRemap.mapCM.columns = maxX;
 XtVaSetValues(colRemap.mapCM.DAW, XmNwidth, colRemap.mapCM.width,
  XmNheight, colRemap.mapCM.height, NULL);
}

/* add pivot color to the array */
DCOLOR2 Boolean addPivotColor(unsigned char color)
{
 int i;
 Boolean dontAdd = False, result = False;

 activeCM = &colRemap.mapCM;
 for (i = 0; i < *activeCM->colNo; i++)
   if (activeCM->colArr[i] == color)
     dontAdd = True;
 if (dontAdd)
   XBell (xinfo.dpy, 0);
 else
   {
     result = True;
     activeCM->colArr[(*activeCM->colNo)++] = color;
     setNewMapWinSize ();
     activeCM->activeCell = activeCM->rows - 1;
     if (activeCM->activeCell < 0)
       activeCM->activeCell = 0;
     exposeMapWin ();
     redrawColSelCB (NULL, "1", NULL);
   }
 return(result);
}

static void remapArea()
{
 unsigned char mapTbl[MAX_CELLS];
 int i;

 for (i = 0; i < MAX_CELLS; i++)
  mapTbl[i] = i;
 for (i = 0; i < MAX_CELLS; i++)
  mapTbl[xinfo.allocCells[i].pixel] = xinfo.allocCells[colRemap.remap[i]].pixel;

 remapImgArea(ApData.img, &mapTbl[0], Select.x, Select.y,
  Select.width, Select.height, Select.reg);
 colRemap.status = modified;
 LastDrawOnAllWin(False);
}

static void undoColorRemap()
{
 CopyAreaImg(Undo.img, ApData.img, NULL, 0, 0, 0, 0,
  Undo.area.width, Undo.area.height, Undo.area.x, Undo.area.y);
 refreshDrawingArea (False, True);
 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
  zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
 colRemap.status = untouched;
 setLabelString(colRemap.previewB, getTxtI8N (PREVIEW_MSG));
}

#define UDCM_INIT_SEL_DA 1
#define UDCM_INIT_MAP_DA 3
#define UDCM_EXPOSE_SEL 4
#define UDCM_RESIZE_SEL 5
#define UDCM_EXPOSE_MAP 6
#define UDCM_MAP 7
#define UDCM_UNMAP 8
#define UDCM_INIT_PREVIEW 9
#define UDCM_OK 10
#define UDCM_PREVIEW 11
#define UDCM_CANCEL 12
#define UDCM_MENU_INIT 13
#define UDCM_MENU_DEL_ONE 14
#define UDCM_SCROLL_WIN_INIT 17

/* user defined color remapping function */
DCOLOR2 void UDcolRemapCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case UDCM_RESIZE_SEL:
     colRemap.selCM.initialize = True;
     redrawTPDA (colRemap.selCM.DAW, &colRemap.selCM);
     break;
   case UDCM_INIT_PREVIEW:
     colRemap.previewB = w;
     break;
   case UDCM_SCROLL_WIN_INIT:
     colRemap.scrollWinW = w;
     XtVaGetValues(w, XmNhorizontalScrollBar, &colRemap.scrollBX,
      XmNverticalScrollBar, &colRemap.scrollBY, NULL);
     break;
   case UDCM_MENU_INIT:
     UDCoMenuW = w;
     break;
   case UDCM_MENU_DEL_ONE:
     deletePivotColors ();
     break;
   case UDCM_OK:
     if (colRemap.status == untouched)
      remapArea();
     XtUnmanageChild(XtParent(XtParent(colRemap.selCM.DAW)));
     break;
   case UDCM_PREVIEW:
     if (colRemap.status == untouched)
       {
	remapArea();
	setLabelString(w, getTxtI8N (UNDO_MSG));
       }
     else
       undoColorRemap();
     break;
   case UDCM_CANCEL:
     if (colRemap.status == modified)
       undoColorRemap ();
     XtUnmanageChild (XtParent (XtParent (colRemap.selCM.DAW)));
     break;
   case UDCM_MAP:
     initUDcolorRemap ();
     break;
   case UDCM_INIT_SEL_DA:
     colRemap.selCM.DAW = w;
     colRemap.selCM.initialize = True;
     colRemap.selCM.activeCell = 0;
     XtAddEventHandler(w, ButtonPressMask, FALSE,
      (XtEventHandler) processInputUDCM, (XtPointer) & colRemap.selCM);
     break;
   case UDCM_INIT_MAP_DA:
     colRemap.mapCM.DAW = w;
     colRemap.mapCM.initialize = True;
     colRemap.mapCM.activeCell = 0;
     XtAddEventHandler(w, ButtonPressMask, FALSE,
      (XtEventHandler) processInputUDCM, (XtPointer) & colRemap.mapCM);
     break;
   case UDCM_EXPOSE_SEL:
     redrawTPDA (w, &colRemap.selCM);
     break;
   case UDCM_EXPOSE_MAP:
     exposeMapWin();
     break;
   case UDCM_UNMAP:
     XtRemoveEventHandler(colSelW, ButtonPressMask, FALSE,
      (XtEventHandler) ColorSelectPress, (XtPointer) 2);
     XtAddEventHandler(colSelW, ButtonPressMask, FALSE,
      (XtEventHandler) ColorSelectPress, 0);
     break;
  }
}
