#define COLORSIM
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/RowColumn.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "paint.h"

static int activeColor;	/* currently edited color */
static int colorModel;	/* currently used color model */
static int reqColors;	/* num of required colors */

static struct
{
 XColor BckColor, FrgColor;
 int BckColorInd, FrgColorInd;
 Boolean changeCol;
} UndoCol;

static void sortColors(Boolean sortOnly);
static void selectColorFromWindow(int status);
static void changeColor(); /* change RGB  or HSL values of currently edited color */

/* count the number of set bits in long value */
static int countBits(unsigned long value)
{
 int res;

 res = 0;
 while (value)
  {
   if (value & 1)
    res++;
   value >>= 1;
  }
 return(res);
}

/* count the number of zero bits from right to left in long value */
static int countRightZeros(unsigned long value)
{
 int res;

 res = 0;
 while (value)
  {
   if (value & 1)
    break;
   else
    res++;
   value >>= 1;
  }
 return(res);
}

/* draw rectangle and circle with both colors */
static void showSelectedColors(Widget w, int x, int y, unsigned int width, unsigned int height)
{
 XGCValues values;
 int xoff, yoff, wid, hei;
 Window win;

 if (XtIsRealized (w))
  {
   xoff = width/20; yoff=height/20;
   wid = (width*14)/20;
   hei = (height*14)/20;
   win=XtWindow(w);
   XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
   XFillRectangle (xinfo.dpy, win, xinfo.gc, x, y, width, height);
   XSetForeground (xinfo.dpy, xinfo.gc, xinfo.black);
   if (ApData.drawinfo.lineWidth > 1)
    {
     values.line_width = 0;
     XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
    }
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+2*xoff+x, yoff+y, wid+5*xoff+x, yoff+y);
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+2*xoff+x, yoff+y, wid+3*xoff+x, yoff-yoff/2+y);
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+2*xoff+x, yoff+y, wid+3*xoff+x, yoff+yoff/2+y);
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+5*xoff+x, yoff+y, wid+5*xoff+x, yoff*4+y);
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+5*xoff+x, yoff*4+y, wid+5*xoff-xoff/2+x, yoff*3+y);
   XDrawLine(xinfo.dpy, win, xinfo.gc, wid+5*xoff+x, yoff*4+y, wid+5*xoff+xoff/2+x, yoff*3+y);

   XSetForeground(xinfo.dpy, xinfo.gc, ApData.drawinfo.backgrndC);
   XFillRectangle(xinfo.dpy, win, xinfo.gc, 5*xoff+x, 5*yoff+y, wid, hei);

   XSetForeground (xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
   XFillRectangle (xinfo.dpy, win, xinfo.gc, xoff+x, yoff+y, wid, hei);
   XSetForeground (xinfo.dpy, xinfo.gc, xinfo.black);
   if (ApData.drawinfo.lineWidth > 1)
    {
     values.line_width = ApData.drawinfo.lineWidth;
     if (values.line_width == 1)
       values.line_width--;
     XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
    }
  }
}

/* un/draw frame around colInd color */
static void frameSelectedColor(int colInd, Boolean frame)
{
 int r, c, x, y, rect_w, rect_h;
 XGCValues values;

 if (!XtIsRealized (colSelW))
  return;
 values.line_width = 0;
 XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);

 if (colInd < 0)
   return;
 rect_w = ColorMatrix.r_width;
 rect_h = ColorMatrix.r_height;
 r = colInd / ColorMatrix.col;
 c = colInd - r * ColorMatrix.col;
 x = c * rect_w + ColorMatrix.gap;
 y = r * rect_h + ColorMatrix.gap;
 rect_w -= ColorMatrix.gap;
 rect_h -= ColorMatrix.gap;
 if (frame)
  {
   XSetForeground(xinfo.dpy, xinfo.gc, xinfo.black);
   XDrawRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
    x - 1, y - 1, rect_w + 1, rect_h + 1);
   XSetForeground(xinfo.dpy, xinfo.gc, xinfo.white);
   XDrawRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
    x - 2, y - 2, rect_w + 3, rect_h + 3);
  }
 else
  {
   XSetForeground(xinfo.dpy, xinfo.gc, GrayBckGround);
   XDrawRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
    x - 2, y - 2, rect_w + 3, rect_h + 3);
   XDrawRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
    x - 1, y - 1, rect_w + 1, rect_h + 1);
  }
 values.line_width = ApData.drawinfo.lineWidth;
 if (values.line_width == 1)
   values.line_width--;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
 XSetForeground(xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
}

/* draw single color in palette */
static void drawSinglePaletteColor(int ind)
{
 int colType, k, r, c, x, y, rect_w, rect_h;
 XGCValues values;

 values.line_width = 0;
 if (xinfo.trueColor)
  values.foreground = xinfo.pixelTC[xinfo.allocCells[ind].pixel];
 else
  values.foreground = xinfo.allocCells[ind].pixel;
 XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth | GCForeground, &values);

 rect_w = ColorMatrix.r_width;
 rect_h = ColorMatrix.r_height;
 r = ind / ColorMatrix.col;
 c = ind - r * ColorMatrix.col;
 x = c * rect_w + ColorMatrix.gap;
 y = r * rect_h + ColorMatrix.gap;
 rect_w -= ColorMatrix.gap;
 rect_h -= ColorMatrix.gap;
 colType = 0;
 for (k = 0; k < ApData.TranspColNo; k++)
  if (ApData.TranspColors[k] == ind)
   {
    colType = 1;
    break;
   }
 for (k = 0; k < ApData.ProtColNo; k++)
  if (ApData.ProtectColors[k] == ind)
   {
    if (colType == 0)
      colType = 2;
    else
      colType = 3;
    break;
   }
 switch (colType)
  {
   case 0:
     XFillRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc, x, y, rect_w, rect_h);
     break;
   case 1:
     XFillArc(xinfo.dpy, XtWindow (colSelW), xinfo.gc, x, y, rect_w, rect_h, 23040, 23040);
     break;
   case 2:
     XFillRectangle(xinfo.dpy, XtWindow (colSelW), xinfo.gc, x, y, rect_w, rect_h);
     XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
     XFillArc(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
      x + rect_w / 3, y + rect_h / 3, rect_w / 3, rect_h / 3, 23040, 23040);
     break;
   case 3:
     XFillArc(xinfo.dpy, XtWindow (colSelW), xinfo.gc, x, y, rect_w, rect_h, 23040, 23040);
     XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
     XFillArc(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
      x + rect_w / 3, y + rect_h / 3, rect_w / 3, rect_h / 3, 23040, 23040);
     break;
  }
 values.line_width = ApData.drawinfo.lineWidth;
 if (values.line_width == 1)
   values.line_width--;
 values.foreground = ApData.drawinfo.foregrndC;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth | GCForeground, &values);
}

/* draw color pallete */
DCOLORSIM void drawColorPalette()
{
 int i, j, r = ColorMatrix.row, c = ColorMatrix.col, ind = 0, x, y;
 int gap = ColorMatrix.gap, rect_w, rect_h;

 x = gap;
 y = gap;
 rect_w = ColorMatrix.r_width - gap;
 rect_h = ColorMatrix.r_height - gap;
 if (XtIsRealized (colSelW))
  {
   int k, colType;
   XGCValues values;

   values.line_width = 0;
   XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
   XSetForeground(xinfo.dpy, xinfo.gc, GrayBckGround);
   XFillRectangle(xinfo.dpy, XtWindow(colSelW), xinfo.gc, 0, 0, ColorMatrix.width, ColorMatrix.height);

   for (i = 1; i <= r; i++)
    {
     for (j = 1; j <= c; j++)
      {
       if (ind < ApData.ColorNo)
	 if (xinfo.trueColor)
	  XSetForeground(xinfo.dpy, xinfo.gc, xinfo.pixelTC[xinfo.allocCells[ind].pixel]);
	 else
	  XSetForeground(xinfo.dpy, xinfo.gc, xinfo.allocCells[ind].pixel);
       else /* extra empty rectangles with no function */
	XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
       colType = 0;
       for (k = 0; k < ApData.TranspColNo; k++)
	 if (ApData.TranspColors[k] == ind)
	  {
	   colType = 1;
	   break;
	  }
       for (k = 0; k < ApData.ProtColNo; k++)
	 if (ApData.ProtectColors[k] == ind)
	  {
	   colType = colType ? 3 : 2;
	   break;
	  }
       switch (colType)
	{
	 case 0:
	  XFillRectangle(xinfo.dpy, XtWindow(colSelW), xinfo.gc, x, y, rect_w, rect_h);
	  break;
	 case 1:
	  XFillArc(xinfo.dpy, XtWindow(colSelW), xinfo.gc, x, y, rect_w, rect_h, 23040, 23040);
	  break;
	 case 2:
	  XFillRectangle(xinfo.dpy, XtWindow(colSelW), xinfo.gc, x, y, rect_w, rect_h);
	  XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
	  XFillArc(xinfo.dpy, XtWindow (colSelW), xinfo.gc,
	     x + rect_w / 3, y + rect_h / 3, rect_w / 3, rect_h / 3, 23040, 23040);
	  break;
	 case 3:
	  XFillArc(xinfo.dpy, XtWindow(colSelW), xinfo.gc, x, y, rect_w, rect_h, 23040, 23040);
	  XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
	  XFillArc (xinfo.dpy, XtWindow (colSelW), xinfo.gc,
	     x + rect_w / 3, y + rect_h / 3, rect_w / 3, rect_h / 3, 23040, 23040);
	  break;
	}
       x += (rect_w + gap);
       ind++;
      }
     y += rect_h + gap;
     x = gap;
    }
   frameSelectedColor (activeColor, True);
   values.line_width = ApData.drawinfo.lineWidth;
   if (values.line_width == 1)
     values.line_width--;
   XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
   XSetForeground (xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
  }
}

/* expose color selection drawing area Callback */
DCOLORSIM void redrawColSelCB(Widget w, XtPointer clientData, XtPointer callData)
{
 XEvent event;
 Dimension width, height;

 switch (atoi((char *) clientData))
  {
   case 1:
    while (XCheckTypedWindowEvent(xinfo.dpy, XtWindow(colSelW), Expose, &event));
    drawColorPalette();
    showSelectedColors(colSelW, ColorMatrix.FBx, ColorMatrix.FBy, ColorMatrix.FBwidth, ColorMatrix.FBheight);
    break;
   case 2:
    XtVaGetValues(colSelW, XmNwidth, &width, XmNheight, &height, NULL);
    initColorSelectionArea(ApData.ColorNo, (int) width, (int) height);
    break;
   case 3:
    if (ColorMatrix.mainWinWidth == 0)
     {
      XtVaGetValues(w, XmNwidth, &ColorMatrix.mainWinWidth, NULL);
      ColorMatrix.mainWinHeight = ColorMatrix.mainWinWidth;
     }
    showSelectedColors(ColorMatrix.mainWinDAW, 0, 0, ColorMatrix.mainWinWidth, ColorMatrix.mainWinHeight);
    break;
  }
}

static void StoreColor(unsigned long ind, int colEntry, int red, int blue, int green)
{
 xinfo.allocCells[colEntry].red = (unsigned short) red;
 xinfo.allocCells[colEntry].green = (unsigned short) green;
 xinfo.allocCells[colEntry].blue = (unsigned short) blue;
 xinfo.allocCells[colEntry].flags = DoRed | DoGreen | DoBlue;
 xinfo.red[colEntry] = (unsigned short) red >> 8;
 xinfo.green[colEntry] = (unsigned short) green >> 8;
 xinfo.blue[colEntry] = (unsigned short) blue >> 8;
 xinfo.allocCells[colEntry].pixel = xinfo.pixelTC[colEntry] = ind;
 if (xinfo.trueColor)
   xinfo.pixelTC[xinfo.allocCells[colEntry].pixel] = eight2MoreBits (red >> 8, green >> 8, blue >> 8);
}

/* allocate all available colors, set random values for them and release all unneccessary colors afterwards */
/* for servers which have unassigned colors set to black - xor won't work */
DCOLORSIM int initAllAvailableColors()
{
 unsigned long plane_mask[1];	/* dummy variable for XAllocColorCells */
 unsigned long colorPV[MAX_CELLS]; /* pixel values of allocated colors */
 int i, freeColors;

 freeColors = maxColors;
 if (xinfo.trueColor)
  {
   xinfo.redBitsNo = countBits (xinfo.visualInfo->red_mask);
   xinfo.greenBitsNo = countBits (xinfo.visualInfo->green_mask);
   xinfo.blueBitsNo = countBits (xinfo.visualInfo->blue_mask);
   xinfo.redShift = countRightZeros (xinfo.visualInfo->red_mask);
   xinfo.greenShift = countRightZeros (xinfo.visualInfo->green_mask);
   xinfo.blueShift = countRightZeros (xinfo.visualInfo->blue_mask);
   xinfo.redShift -= 8 - xinfo.redBitsNo;
   xinfo.greenShift -= 8 - xinfo.greenBitsNo;
   xinfo.blueShift -= 8 - xinfo.blueBitsNo;
   for (i = 0; i < MAX_CELLS; i++)
     colorPV[i] = xinfo.allocCells[i].pixel = i;
  }
 else
  while (freeColors > 0)
   {
    if (XAllocColorCells(xinfo.dpy, xinfo.cmap, False, plane_mask, 0, colorPV, freeColors))
     break;
    freeColors--;
   }

 if (freeColors < 2)
   return (freeColors);
 StoreColor (colorPV[0], 0, 0, 0, 0);	/* black is the 1st color */
 StoreColor (colorPV[1], 1, 0xffff, 0xffff, 0xffff); /* white is the 2nd color */
 /* make a gradation scale in HSL */
 for (i = 2; i < freeColors; i++)
   {
     int r, g, b;

     HSLtoRGB (((i - 2) * 360) / (freeColors - 2), 600, 600, &r, &g, &b);
     StoreColor (colorPV[i], i, r << 8, g << 8, b << 8);
   }
 if (xinfo.trueColor)
   calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
    &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
 else
   XStoreColors (xinfo.dpy, xinfo.cmap, &xinfo.allocCells[0], freeColors);

 colorModel = MODEL_RGB;
 ApData.drawinfo.foregrndC = xinfo.allocCells[0].pixel;
 ApData.drawinfo.backgrndC = xinfo.allocCells[1].pixel;
 ColorMatrix.FrgColorInd = 0;
 ColorMatrix.BckColorInd = 1;
 memcpy(&UndoCol.BckColor, &xinfo.allocCells[0], sizeof(XColor));
 memcpy(&UndoCol.FrgColor, &xinfo.allocCells[1], sizeof(XColor));
 return(freeColors);
}

/* allocate additional colors and initialize them */
DCOLORSIM int allocateMoreColors(unsigned int colsWanted, int firstEmpty)
{
 unsigned long plane_mask[1];	/* dummy variable for XAllocColorCells */
 unsigned long colorPV[MAX_CELLS];
 int i, freeColors;

 freeColors = colsWanted > MAX_CELLS ? MAX_CELLS : colsWanted;
 if (xinfo.trueColor)
  {
   unsigned char existingPix[MAX_CELLS];
   int j;

   for (i = 0; i < MAX_CELLS; i++)
     existingPix[i] = 0;
   for (i = 0; i < firstEmpty; i++)
     existingPix[xinfo.allocCells[i].pixel] = 1;
   freeColors = MIN (MAX_CELLS - maxColors, freeColors);
   i = 0;
   /* take "unallocated pixel values */
   for (j = 0; j < MAX_CELLS; j++)
     {
       if (existingPix[j])
	 continue;
       colorPV[i] = j;
       i++;
       if (i == freeColors)
	 break;
     }
  }
 else
  while (freeColors > 0)
   {
    if (XAllocColorCells(xinfo.dpy, xinfo.cmap, False, plane_mask, 0, colorPV, freeColors))
     break;
    freeColors--;
   }
 srand(time(0));
 for (i = 0; i < freeColors; i++)
   StoreColor (colorPV[i], firstEmpty + i, rand () << 1, rand () << 1, rand () << 1);
 if (!xinfo.trueColor)
   XStoreColors(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[firstEmpty], freeColors);
 return (freeColors);
}

/* free some colors */
DCOLORSIM void freeColors(int col2free)
{
 unsigned long colorPV[MAX_CELLS];
 int i, lastOne, howMany;

 if (col2free <= 0)
  return;
 howMany = ((maxColors - col2free) > 1) ? col2free : (maxColors - 2);
 if (howMany == 0)
  return;
 lastOne = maxColors - 1;
 for (i = 0; i < howMany; i++)
  {
   colorPV[i] = xinfo.allocCells[lastOne].pixel;
   StoreColor (colorPV[i], lastOne, 0, 0, 0);
   lastOne--;
  }
 if (!xinfo.trueColor)
  {
   XStoreColors(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[lastOne + 1], howMany);
   XFreeColors(xinfo.dpy, xinfo.cmap, &colorPV[0], howMany, 0);
  }
 maxColors -= howMany;
}

static void updateInterfaceColor(int num)
{
 int ca, cb, cc, d, e, f;

 ca = xinfo.allocCells[num].red >> 8;
 cb = xinfo.allocCells[num].green >> 8;
 cc = xinfo.allocCells[num].blue >> 8;
 if (colorModel == MODEL_HSL)
  {
   d = ca;
   e = cb;
   f = cc;
   RGBtoHSL(d, e, f, &ca, &cb, &cc);
  }
 XmScaleSetValue(color_comp1W, ca);
 XmScaleSetValue(color_comp2W, cb);
 XmScaleSetValue(color_comp3W, cc);
}

/* set the colors according to the mouse click: left - select, middle swap, right copy */
/* use "phantom" mouse button 4 to set background color */
DCOLORSIM void ColorSelectConseq(unsigned int MouseButton, unsigned long ColInd)
{
 int i=0;

 switch (MouseButton)
  {
   case Button3:
    if (ColInd!=activeColor)
     {
      xinfo.allocCells[ColInd].red = xinfo.allocCells[activeColor].red;
      xinfo.allocCells[ColInd].green = xinfo.allocCells[activeColor].green;
      xinfo.allocCells[ColInd].blue = xinfo.allocCells[activeColor].blue;
      changeColor();
     }
    else /* toggle the protect state of this color */
     {
      while ((i<ApData.ProtColNo) && (ApData.ProtectColors[i] != ColInd))
       i++;
      if (i<ApData.ProtColNo) /* found, need to remove it from list */
       {
	if (ApData.ProtColNo)
         ApData.ProtectColors[i]=ApData.ProtectColors[ApData.ProtColNo -1];
	ApData.ProtColNo--;
       }
      else
       ApData.ProtectColors[ApData.ProtColNo++] = ColInd;
      transpProtColCB(mainW, "1", NULL);
     }
    
    break;
   case Button2:
    if (ColInd!=activeColor)
     {
      SWAP(xinfo.allocCells[ColInd].red, xinfo.allocCells[activeColor].red);
      SWAP(xinfo.allocCells[ColInd].green, xinfo.allocCells[activeColor].green);
      SWAP(xinfo.allocCells[ColInd].blue, xinfo.allocCells[activeColor].blue);

      SWAP(xinfo.red[ColInd], xinfo.red[activeColor]);
      SWAP(xinfo.green[ColInd], xinfo.green[activeColor]);
      SWAP(xinfo.blue[ColInd], xinfo.blue[activeColor]);
      SWAP(Select.red[ColInd], Select.red[activeColor]);
      SWAP(Select.green[ColInd], Select.green[activeColor]);
      SWAP(Select.blue[ColInd], Select.blue[activeColor]);
      updateInterfaceColor(activeColor);
      changeColor();
     }
    else   /* toggle the transparency state of this color */
     {
      while ((i<ApData.TranspColNo) && (ApData.TranspColors[i] != ColInd))
       i++;
      if (i<ApData.TranspColNo) /* found, need to remove it from list */
       {
	if (ApData.TranspColNo)
         ApData.TranspColors[i]=ApData.TranspColors[ApData.TranspColNo -1];
	ApData.TranspColNo--;
       }
      else
       ApData.TranspColors[ApData.TranspColNo++] = ColInd;
      transpProtColCB(mainW, "1", NULL);
     }
    break;
   case Button4:
    memcpy(&UndoCol.BckColor, &xinfo.allocCells[ColorMatrix.BckColorInd], sizeof (XColor));
    UndoCol.BckColorInd = ColorMatrix.BckColorInd;

    if (xinfo.trueColor)
      ApData.drawinfo.backgrndC = xinfo.pixelTC[xinfo.allocCells[ColInd].pixel];
    else
      ApData.drawinfo.backgrndC = xinfo.allocCells[ColInd].pixel;
    ColorMatrix.BckColorInd = ColInd;
    XSetBackground (xinfo.dpy, xinfo.gc, ApData.drawinfo.backgrndC);
    break;
  }

 updateInterfaceColor(ColInd);
 frameSelectedColor(activeColor, False);

 if (MouseButton < Button4)
  {
   memcpy(&UndoCol.FrgColor, &xinfo.allocCells[ColorMatrix.FrgColorInd], sizeof(XColor));
   UndoCol.FrgColorInd = ColorMatrix.FrgColorInd;
   if (xinfo.trueColor)
    ApData.drawinfo.foregrndC = xinfo.pixelTC[xinfo.allocCells[ColInd].pixel];
   else
    ApData.drawinfo.foregrndC = xinfo.allocCells[ColInd].pixel;
   ColorMatrix.FrgColorInd = activeColor = ColInd;
   XSetForeground (xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
   if (MouseButton != Button1)
    changeColor();
  }
 frameSelectedColor(activeColor, True);
 showSelectedColors(colSelW, ColorMatrix.FBx, ColorMatrix.FBy,
  ColorMatrix.FBwidth, ColorMatrix.FBheight);
 showSelectedColors(ColorMatrix.mainWinDAW, 0, 0, ColorMatrix.mainWinWidth,
  ColorMatrix.mainWinHeight);
}

static void showChosenColorCB(XtPointer clientData, XtIntervalId * id)
{
 static short num = 0;
 int col = (int) clientData;
 unsigned short red=0, green=0, blue=0;
 static Boolean zoomWinUp;

 switch (num)
  {
   case 0:
     num++;
     red = green = blue = 0;
     if (widgetIsViewable (zoomDrawAreaW))
       {
	 zoomWinUp = True;
	 XtPopdown(GetShellParent(zoomDrawAreaW));
       }
     else
       zoomWinUp = False;
     XtAddTimeOut(500, (XtTimerCallbackProc) *showChosenColorCB, clientData);
     break;
   case 1:
     num++;
     red = green = blue = 255;
     XtAddTimeOut(500, (XtTimerCallbackProc) *showChosenColorCB, clientData);
     break;
   case 2:
     num = 0;
     red = xinfo.red[col];
     green = xinfo.green[col];
     blue = xinfo.blue[col];
     if (zoomWinUp)
       XtPopup(GetShellParent(zoomDrawAreaW), XtGrabNone);
     zoomWinUp = False;
     break;
  }
 xinfo.allocCells[col].red = red << 8;
 xinfo.allocCells[col].green = green << 8;
 xinfo.allocCells[col].blue = blue << 8;

 if (xinfo.trueColor)
  {
   xinfo.pixelTC[xinfo.allocCells[col].pixel] = eight2MoreBits(red, green, blue);
   refreshDrawingArea(True, False);
  }
 else
  XStoreColor(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[col]);
}

/* Handle mouse button press inside color selection area */
DCOLORSIM void ColorSelectPress(Widget w, XtPointer clientData, XEvent * event)
{
 int x, y, r, c, ind;
 int data = (int) clientData;
 static Time lastEventTime = 0;

 if (event->type != ButtonPress)
  return;
 x = event->xbutton.x;
 y = event->xbutton.y;
 if (x < 0)
   x = 0;
 if (y < 0)
   y = 0;

 if ((x > ColorMatrix.width) || (y > ColorMatrix.height) || (w==ColorMatrix.mainWinDAW) )
  {
   unsigned long tmp;

   if ( (w==ColorMatrix.mainWinDAW) && ((y>20) || (x<50)))
    { initColSelCB(w, "4", NULL); return; }
   SWAP(ApData.drawinfo.backgrndC, ApData.drawinfo.foregrndC);
   tmp = ColorMatrix.FrgColorInd;
   ColorMatrix.FrgColorInd = ColorMatrix.BckColorInd;
   ColorMatrix.BckColorInd = tmp;
   activeColor = ColorMatrix.FrgColorInd;
   UndoCol.changeCol = True;
   drawColorPalette();
   ColorSelectConseq(Button1, activeColor);
   ColorSelectConseq(Button4, ColorMatrix.BckColorInd);
   return;
  }

 c = (int) ColorMatrix.r_width;
 r = (int) ColorMatrix.r_height;
 c = (x - ColorMatrix.gap) / c;
 r = (y - ColorMatrix.gap) / r;
 ind = r * ColorMatrix.col + c;
 if ((ind < ApData.ColorNo) && (c < ColorMatrix.col) && (r < ColorMatrix.row))
  {
   if ((data == 0) && (activeColor == ind) &&
    ((event->xbutton.time - lastEventTime) < 1000) && (event->xbutton.button == Button1))
    showChosenColorCB((XtPointer) ind, NULL);
   else
    {
     UndoCol.changeCol = True;
     frameSelectedColor(activeColor, False);
     ColorSelectConseq(event->xbutton.button, ind);
     activeColor = ind;
     if (data == 1)
      addColor2Array (ind, False, True);
     else if (data == 2)
      addPivotColor(ind);
    }
  }
  if (event->xbutton.button == Button1)
   lastEventTime = event->xbutton.time;
}

/* Event Handler for choosing the color from the picture */
static void ChoosePicColEH(Widget w, XtPointer clientData, XEvent * event)
{
 XImage *img;
 unsigned long ind, clr;
 int i, x, y;

 if (event->type == ButtonPress)
  {
   switch (event->xbutton.button)
    {
     case Button1:
     case Button2:
       x = event->xbutton.x;
       y = event->xbutton.y;
       if ((int) clientData == 1)
	 {
	   x /= zoomDraw.factorX;
	   y /= zoomDraw.factorY;
	   x *= zoomDraw.factorX;
	   y *= zoomDraw.factorY;
	 }
       img = XGetImage(xinfo.dpy, XtWindow(w), x, y, 1, 1, AllPlanes, ZPixmap);
       clr = XGetPixel(img, 0, 0);
       i = ind = 0;
       if (xinfo.trueColor)
	 while ((i < maxColors) && (xinfo.pixelTC[xinfo.allocCells[i].pixel] != clr))
	   i++;
       else
	 while ((i < maxColors) && (xinfo.allocCells[i].pixel != clr))
	   i++;
       if (i < maxColors)
	ind = i;
       ColorSelectConseq(Button1, ind);
       undoXImage (&img);
       break;
     case Button3:
       break;
    }
   selectColorFromWindow (1);
  }
}

/* Change active color */
static void selectColorFromWindow(int status)
{
 static XtEventHandler oldEH = NULL;
 static EventMask oldEM = 0;

 if (status)
  setExclEventHandler((XtEventHandler) oldEH, oldEM, True);
 else
  {
   oldEH = ApData.CurrEventHandler;
   oldEM = ApData.eventMask;
   setExclEventHandler ((XtEventHandler) ChoosePicColEH, ButtonPressMask, True);
  }
}


/* set the color model used in color editor: RGB or HSL */
static void setColorModel(int colorMod)
{
 int a, b, c, d, e, f;
 char *title1, *title2, *title3;

 colorModel = colorMod;

 a = xinfo.allocCells[activeColor].red >> 8;
 b = xinfo.allocCells[activeColor].green >> 8;
 c = xinfo.allocCells[activeColor].blue >> 8;

 if (colorModel == MODEL_HSL)
  {
   d = a;
   e = b;
   f = c;
   RGBtoHSL (d, e, f, &a, &b, &c);
   title1 = getTxtI8N (HUE_MSG);
   title2 = getTxtI8N (SATURATION_MSG);
   title3 = getTxtI8N (LIGHT_MSG);
  }
 else
  {
   title1 = getTxtI8N (RED_MSG);
   title2 = getTxtI8N (GREEN_MSG);
   title3 = getTxtI8N (BLUE_MSG);
  }
 /* this causes resize of main palette window ! */
 XtVaSetValues(color_comp1W, XmNvalue, a, NULL);
 setTitleString(color_comp1W, title1);
 XtVaSetValues(color_comp2W, XmNvalue, b, NULL);
 setTitleString(color_comp2W, title2);
 XtVaSetValues(color_comp3W, XmNvalue, c, NULL);
 setTitleString(color_comp3W, title3);
}

/*Restore previously selected colors */
DCOLORSIM void restoreColorCB (Widget w, XtPointer clientData, XtPointer callData)
{
 if (UndoCol.changeCol)
  {
   memcpy(&xinfo.allocCells[ColorMatrix.BckColorInd],
    &UndoCol.BckColor, sizeof(XColor));
   ColorMatrix.BckColorInd = UndoCol.BckColorInd;
   XStoreColor(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[ColorMatrix.BckColorInd]);
   memcpy(&xinfo.allocCells[ColorMatrix.FrgColorInd],
    &UndoCol.FrgColor, sizeof(XColor));
   ColorMatrix.FrgColorInd = UndoCol.FrgColorInd;
   XStoreColor(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[ColorMatrix.FrgColorInd]);
  }
}

/* change RGB  or HSL values of currently edited color */
static void changeColor()
{
 int a, b, c, d, e, f;

 UndoCol.changeCol = False;
 XtVaGetValues (color_comp1W, XmNvalue, &a, NULL);
 XtVaGetValues (color_comp2W, XmNvalue, &b, NULL);
 XtVaGetValues (color_comp3W, XmNvalue, &c, NULL);
 if (colorModel == MODEL_HSL)
  {
   d = a;
   e = b;
   f = c;
   HSLtoRGB(d, e, f, &a, &b, &c);
  }
 xinfo.allocCells[activeColor].red = (unsigned short) a << 8;
 xinfo.allocCells[activeColor].green = (unsigned short) b << 8;
 xinfo.allocCells[activeColor].blue = (unsigned short) c << 8;
 xinfo.red[activeColor] = Select.red[activeColor] = (unsigned char) a;
 xinfo.green[activeColor] = Select.green[activeColor] = (unsigned char) b;
 xinfo.blue[activeColor] = Select.blue[activeColor] = (unsigned char) c;
 if (xinfo.trueColor)
  {
   xinfo.pixelTC[xinfo.allocCells[activeColor].pixel] = eight2MoreBits(a, b, c);
   drawSinglePaletteColor (activeColor);
   ApData.drawinfo.foregrndC = xinfo.pixelTC[xinfo.allocCells[ColorMatrix.FrgColorInd].pixel];
   ApData.drawinfo.backgrndC = xinfo.pixelTC[xinfo.allocCells[ColorMatrix.BckColorInd].pixel];
   showSelectedColors(ColorMatrix.mainWinDAW, 0, 0, ColorMatrix.mainWinWidth, ColorMatrix.mainWinHeight);
  }
 else
  XStoreColor(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[activeColor]);
 ApData.pic_change = True;
 showSelectedColors(colSelW, ColorMatrix.FBx, ColorMatrix.FBy,
  ColorMatrix.FBwidth, ColorMatrix.FBheight);
 if (xinfo.trueColor)
  {
   refreshDrawingArea (True, True);
   refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
    zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
  }
}

/* create color selection area with colors colors */
DCOLORSIM void initColorSelectionArea(int color_no, int w, int h)
{
 int r, c, gap;
 int pxm_w, pxm_h;
 float rect_w, rect_h, area;
 int maxOffset = 10;

 if (((w != 0) && (h != 0)))	/* change of size */
  {
   ColorMatrix.aspectR = 1.0 * w / h;
   if (ColorMatrix.aspectR >= 1.0)
    {
     pxm_h = h;
     if (ColorMatrix.aspectR < 2.0)
      {
       ColorMatrix.aspectR = 2.0;
       if (h * 2 > (xinfo.dpyw * 8 / 10))
	 pxm_h = (xinfo.dpyw * 8 / 10) / 2;
      }
     pxm_w = pxm_h * ColorMatrix.aspectR - pxm_h;
     maxOffset = w / 10;
    }
   else
    {
     pxm_w = w;
     if (ColorMatrix.aspectR > 0.5)
      {
       ColorMatrix.aspectR = 0.5;
       if (w * 2 > (xinfo.dpyh * 7 / 10))
	 pxm_w = (xinfo.dpyh * 7 / 10) / 2;
      }
     pxm_h = pxm_w / ColorMatrix.aspectR - pxm_w;
     maxOffset = h / 10;
    }
  }
 else
  {
   pxm_w = ColorMatrix.width;
   pxm_h = ColorMatrix.height;
  }
 ApData.ColorNo = color_no;
 gap = 2;
 area = pxm_h * pxm_w;
 rect_w = rect_h = sqrt (area / color_no);
 r = (pxm_h - gap) / rect_h;
 c = (pxm_w - gap) / rect_w;
 if (ColorMatrix.aspectR > 1.0) /* horizontal layout */
  {
   if (r < 1)
     r = 1;
   while ((r * c) < color_no)
     c++;
   if ((r * (c - 1)) > color_no)
     c--;
  }
 else /* vertical layout */
  {
   if (c < 1)
     c = 1;
   while ((r * c) < color_no)
     r++;
   if ((c * (r - 1)) > color_no)
     r--;
  }
 rect_w = 1.0 * (pxm_w - gap) / c;
 rect_h = 1.0 * (pxm_h - gap) / r;
 rect_w -= gap;
 rect_h -= gap;
 ColorMatrix.r_width = rect_w + gap;
 ColorMatrix.r_height = rect_h + gap;
 ColorMatrix.row = r;
 ColorMatrix.col = c;
 ColorMatrix.gap = gap;
 if (ColorMatrix.aspectR > 1.0)
  {
   ColorMatrix.x_offs = w - pxm_w - pxm_h;
   ColorMatrix.y_offs = h - pxm_h;
  }
 else
  {
   ColorMatrix.x_offs = w - pxm_w;
   ColorMatrix.y_offs = h - pxm_h - pxm_w;
  }
 if ((ColorMatrix.x_offs < 0) || (ColorMatrix.x_offs > maxOffset))
   ColorMatrix.x_offs = 0;

 if ((ColorMatrix.y_offs < 0) || (ColorMatrix.y_offs > maxOffset))
   ColorMatrix.y_offs = 0;
 pxm_w += ColorMatrix.x_offs;
 pxm_h += ColorMatrix.y_offs;
 ColorMatrix.width = pxm_w;
 ColorMatrix.height = pxm_h;

 if (ColorMatrix.aspectR < 1.0)
  {
   ColorMatrix.FBx = 0;
   ColorMatrix.FBy = pxm_h;
   ColorMatrix.FBheight = ColorMatrix.FBwidth = pxm_w;
   pxm_h += ColorMatrix.FBheight;
  }
 else
  {
   ColorMatrix.FBx = ColorMatrix.width;
   ColorMatrix.FBy = 0;
   ColorMatrix.FBheight = ColorMatrix.FBwidth = pxm_h;
   pxm_w += ColorMatrix.FBwidth;
  }
 if (XtIsRealized(colSelW)) /* draw rectangle and circle with both colors */
  {
   drawColorPalette();
   showSelectedColors(colSelW, ColorMatrix.FBx, ColorMatrix.FBy, ColorMatrix.FBwidth, ColorMatrix.FBheight);
  }
}

#define COLORNO_INIT_LABEL 1
#define COLORNO_INIT_SCALE 2
#define COLORNO_OK 3
#define COLORNO_CONFIRMED 4
#define COLORNO_MAP 5

/* set number of active (used) colors */
DCOLORSIM void setColorNoCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static Widget scale;
 static int oldColorNo;
 static Widget label;
 int newColNo, prm = atoi ((char *) clientData);

 switch (prm)
   {
   case COLORNO_OK:
     XmScaleGetValue (scale, &newColNo);
     if (Select.selected)
       Select.colorCons = False;
     if (newColNo < oldColorNo)
      manageWarningDialog (getTxtI8N (COLOR_NUM_LESS_THEN_USED_COLORS_MSG),
       setColorNoCB, "4", getTxtI8N (YES_MSG), NULL, NULL, getTxtI8N (NO_MSG),
       NULL, NULL, getTxtI8N (HELP_MSG));
     else
      { /* number of colors used is smaller than the new number */
       if (maxColors < newColNo)
	{
	 maxColors += allocateMoreColors (newColNo - maxColors, maxColors);
	 if (maxColors < newColNo)
	  XBell (xinfo.dpy, 0);
	}
       else
	freeColors(maxColors - newColNo);
       XmScaleSetValue (scale, maxColors);
       initColorSelectionArea (maxColors, 0, 0);
       ColorSelectConseq(Button1, 1);
       ColorSelectConseq(Button4, 0);
       XtUnmanageChild(XtParent (scale));
      }
     break;
   case COLORNO_CONFIRMED:
     XmScaleGetValue (scale, &newColNo);
     XtUnmanageChild (XtParent (scale));
     ApData.ColorNo = newColNo;
     ApData.pic_change = True;
     sortColors (False);
     freeColors (maxColors - newColNo);
     initColorSelectionArea (newColNo, 0, 0);
     ColorSelectConseq (Button1, 1);
     ColorSelectConseq (Button4, 0);
     Select.rect_on = False;
     ApData.tool = Undo.tool = none;
     setToolEventHandler (ApData.tool, 0);
     highlightButton (NULL);
     break;
   case COLORNO_INIT_SCALE:
     XtVaSetValues (w, XmNminimum, 2, XmNmaximum, 256, NULL);
     XmScaleSetValue (w, 2);
     scale = w;
     break;
   case COLORNO_INIT_LABEL:
     setLabelString(w, getTxtI8N (NUMBER_OF_COLOR_USED_MSG));
     label = w;
     break;
   case COLORNO_MAP:
     {
      char buf[512];

      finalDraw (ApData.pixy);
      disableUndo ();
      if (ApData.ColorNo > 1)
	XmScaleSetValue (scale, ApData.ColorNo);
      else
	XmScaleSetValue (scale, 2);
      sortColors (False);
      sprintf (buf, "%s:\t%d", getTxtI8N (NUMBER_OF_COLOR_USED_MSG), reqColors);
      setLabelString(label, buf);
      oldColorNo = reqColors;
      if (oldColorNo < 2)
	oldColorNo = 2;
     }
     break;
   }
}

/* scans picture and sorts colors according to their use */
static void sortColors(Boolean sortOnly)
{
 XImage *img;
 int oldNumColors;
 unsigned char cmap[256 * 3], trTbl[256];
 unsigned long colHist[MAX_CELLS];
 int i;

 oldNumColors = ApData.ColorNo;
 finalDraw (ApData.pixy);
 disableUndo();
 img = ApData.img;
 for (i = 0; i < 256 * 3; cmap[i++] = 0);
 for (i = 0; i < maxColors; i++)
  {
   cmap[xinfo.allocCells[i].pixel * 3] = xinfo.allocCells[i].red >> 8;
   cmap[xinfo.allocCells[i].pixel * 3 + 1] = xinfo.allocCells[i].green >> 8;
   cmap[xinfo.allocCells[i].pixel * 3 + 2] = xinfo.allocCells[i].blue >> 8;
  }
 reqColors = ApData.ColorNo;
 initHist (&colHist[0]);
 addToHist (&colHist[0], (unsigned char *) img->data, 0, 0, img->width, img->height, img->bytes_per_line);
 quantize (img, ApData.ColorNo, &reqColors, cmap, xinfo.allocCells,
       &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0], trTbl, &colHist[0]);
 if (xinfo.trueColor)
  calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0], &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
 else
  XStoreColors (xinfo.dpy, xinfo.cmap, xinfo.allocCells, ApData.ColorNo);
 if (sortOnly)
  ApData.ColorNo = oldNumColors;
 drawColorPalette();
 showSelectedColors(colSelW, ColorMatrix.FBx, ColorMatrix.FBy, ColorMatrix.FBwidth, ColorMatrix.FBheight);
 showSelectedColors(ColorMatrix.mainWinDAW, 0, 0, ColorMatrix.mainWinWidth, ColorMatrix.mainWinHeight);
 ColorSelectConseq(Button1, 1);
 ColorSelectConseq(Button4, 0);
 ApData.TranspColNo = ApData.ProtColNo = 0;
 ApData.specialPutImg = False;
 refreshDrawingArea (True, True);
 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
  zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
}

static void makeGradationInPalete()
{
 int i, l, h, redDiff, greenDiff, blueDiff, diff;

 l = ColorMatrix.BckColorInd;
 h = ColorMatrix.FrgColorInd;
 if (l > h)
   SWAP (l, h);
 diff = h - l;
 if (diff < 2)
   return;

 finalDraw (ApData.pixy);
 disableUndo ();

 redDiff = (int) xinfo.red[h] - xinfo.red[l];
 greenDiff = (int) xinfo.green[h] - xinfo.green[l];
 blueDiff = (int) xinfo.blue[h] - xinfo.blue[l];

 Undo.points[Undo.npoints].x = l + 1;
 Undo.points[Undo.npoints++].y = (short) h;
 for (i = l + 1; i < h; i++)
  {
   Undo.points[Undo.npoints].x = xinfo.red[i];
   Undo.points[Undo.npoints++].y = xinfo.green[i];
   Undo.points[Undo.npoints].x = xinfo.blue[i];
   Undo.points[Undo.npoints++].y = 0;
   xinfo.red[i] = xinfo.red[l] + (redDiff * (i - l)) / diff;
   xinfo.green[i] = xinfo.green[l] + (greenDiff * (i - l)) / diff;
   xinfo.blue[i] = xinfo.blue[l] + (blueDiff * (i - l)) / diff;

   xinfo.allocCells[i].red = xinfo.red[i] << 8;
   xinfo.allocCells[i].green = xinfo.green[i] << 8;
   xinfo.allocCells[i].blue = xinfo.blue[i] << 8;
   xinfo.allocCells[i].flags = DoRed | DoGreen | DoBlue;
  }
 if (xinfo.trueColor)
  {
   calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
    &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
   refreshDrawingArea(True, True);
   refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
   drawColorPalette();
  }
 else
  XStoreColors(xinfo.dpy, xinfo.cmap, xinfo.allocCells, ApData.ColorNo);
 ColorSelectConseq(Button4, ColorMatrix.BckColorInd);
 ColorSelectConseq(Button1, ColorMatrix.FrgColorInd);
 ApData.tool = Undo.tool = makeGradation;
 enableUndo(True);
 ApData.pic_change = True;
}

static void convert2Gray()
{
 int i, colTmp;

 finalDraw(ApData.pixy);
 disableUndo();
 Undo.points[Undo.npoints].x = 0;
 Undo.points[Undo.npoints++].y = (short) ApData.ColorNo;
 for (i = 0; i < ApData.ColorNo; i++)
  {
   Undo.points[Undo.npoints].x = xinfo.red[i];
   Undo.points[Undo.npoints++].y = xinfo.green[i];
   Undo.points[Undo.npoints].x = xinfo.blue[i];
   Undo.points[Undo.npoints++].y = 0;
   colTmp = (222 * xinfo.red[i] + 707 * xinfo.green[i] + 71 * xinfo.blue[i]) / 1000;
   xinfo.red[i] = xinfo.green[i] = xinfo.blue[i] = (unsigned short) colTmp;
   xinfo.allocCells[i].red = xinfo.allocCells[i].green =
    xinfo.allocCells[i].blue = (unsigned short) colTmp << 8;
   xinfo.allocCells[i].flags = DoRed | DoGreen | DoBlue;
  }
 if (xinfo.trueColor)
   calcTrueColorPixels (&xinfo.allocCells[0], &xinfo.pixelTC[0],
			&xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
 else
   XStoreColors(xinfo.dpy, xinfo.cmap, xinfo.allocCells, ApData.ColorNo);
 ColorSelectConseq (Button4, ColorMatrix.BckColorInd);
 ColorSelectConseq (Button1, ColorMatrix.FrgColorInd);
 ApData.tool = Undo.tool = grayConv;
 enableUndo (True);
 ApData.pic_change = True;
 if (xinfo.trueColor)
  {
   refreshDrawingArea(True, True);
   refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
    zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
   drawColorPalette ();
  }
}

DCOLORSIM void colorToolsCB(Widget w, XtPointer clientData, XtPointer callData)
{
 switch (atoi((char *) clientData))
  {
   case 1: changeColor(); break;
   case 3: sortColors (True); break;
   case 4: convert2Gray(); break;
   case 5: makeGradationInPalete(); break;
   case 6: cutOutColorBorder(); break;
   case 8: setColorModel(0); break;
   case 9: setColorModel(1); break;
   case 10: selectColorFromWindow(0); break;
  }
}

#define INIT_COL_DA 1
#define INIT_COL_CHANGE_COL 3
#define INIT_COL_MANAGE 4
#define INIT_COL_CHANGE_COL_UNMAP 5
#define INIT_COL_DA_MAIN 6
/* init color selection widget */
DCOLORSIM void initColSelCB(Widget w, XtPointer clientData, XtPointer callData)
{
 switch (atoi((char *) clientData))
  {
   case INIT_COL_DA: /* do first time initialisation of color related stuff */
    calcColTransfTbl();
    initColorSelectionArea (ApData.ColorNo, 120 * 2, 120);
    XtAddEventHandler(w, ButtonPressMask, FALSE, (XtEventHandler) ColorSelectPress, 0);
    XtVaGetValues (w, XmNbackground, &GrayBckGround, NULL);
    break;
   case INIT_COL_DA_MAIN:
    ColorMatrix.mainWinDAW = w;
    XtAddEventHandler(w, ButtonPressMask, FALSE, (XtEventHandler) ColorSelectPress, 0);
    break;
   case INIT_COL_CHANGE_COL:
    {
     static Boolean FirstTime = True;

     if (FirstTime)
      setColorModel(0);
     FirstTime = False;
    }
    break;
   case INIT_COL_CHANGE_COL_UNMAP:
    if (xinfo.trueColor)
     {
      refreshDrawingArea (True, True);
      refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
       zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
     }
    break;
   case INIT_COL_MANAGE:
    if (XtIsManaged (colSelDialogW))
      XMapRaised (xinfo.dpy, XtWindow (XtParent (colSelDialogW)));
    else
      XtManageChild (colSelDialogW);
    break;
  }
}

DCOLORSIM void changeColorKbd(Window win, int xPos, int yPos, short button)
{
 XImage *img;
 int i = 0, ind = 0;
 unsigned long clr;

 img = XGetImage(xinfo.dpy, win, xPos, yPos, 1, 1, AllPlanes, ZPixmap);
 if (img == NULL)
   return;
 clr = XGetPixel (img, 0, 0);
 undoXImage (&img);
 if (xinfo.trueColor)
   while ((i < maxColors) && (xinfo.pixelTC[xinfo.allocCells[i].pixel] != clr))
     i++;
 else
   while ((i < maxColors) && (xinfo.allocCells[i].pixel != clr))
     i++;
 if (i < maxColors)
   ind = i;
 ColorSelectConseq(button, ind);
}
