#define CONTOUR
#include <Xm/Xm.h>
#include <Xm/ToggleB.h>
#include <stdlib.h>
#include "paint.h"

static void thinContours(XImage *img, int colInd, Rectangle *rect, Region reg);

#define CONTOUR_NONE		0
#define CONTOUR_THIN		1
#define CONTOUR_THICK		2
#define CONTOUR_AVERAGE 	3
#define CONTOUR_REMOVE		4
#define CONTOUR_PRUN		5
#define CONTOUR_DESPECKLE	6
#define CONTOUR_CREATE		7
#define CONTOUR_CREATE_LEFT	8
#define CONTOUR_CREATE_RIGHT	9
#define CONTOUR_CREATE_UP	10
#define CONTOUR_CREATE_DOWN	11
#define CONTOUR_THICK_LEFT	12
#define CONTOUR_THICK_RIGHT	13
#define CONTOUR_THICK_UP	14
#define CONTOUR_THICK_DOWN	15
#define CONTOUR_STITCH_X	16
#define CONTOUR_STITCH_Y	17

static unsigned long *changepos;
static unsigned char *changeval;
static int changecount;

static void setChangePosVal(int changecount, int pos, int val)
{
 changepos[changecount] = pos;
 changeval[changecount] = val;
}

static void flushChanges(unsigned char *data, unsigned char *protected)
{
 int i;

 for (i = 0; i < changecount; i++)
  if (!protected[data[changepos[i]]])
   data[changepos[i]] = changeval[i];
 changecount = 0;
}

static void processContourOperation(XImage *img, int frgColInd, int bckColInd, Rectangle *rect,
 Region reg, int contourOperation)
{
 int i, j, k, x, y, xdim, ydim, maxval, mody;
 int xlimit, ylimit;
 int bpl, pos, doit;
 unsigned char *data;
 unsigned char rmcol, cmpcol, max, posval=0;
 unsigned char value[10];
 unsigned char count[10];
 int numIter=0;
 int left, right, up, down, midx, midy, buffHalf;
 int doLeft, doRight, doUp, doDown, doLeftUp, doLeftDown, doRightUp, doRightDown;
 Boolean doOperation=FALSE, allOver=FALSE;
 Boolean doOperator[8];

 unsigned char myProtected[256];

 for (i=0; i<256; i++)
  myProtected[i]=FALSE;
 for (i=0; i<ApData.ProtColNo; i++)
  myProtected[ApData.ProtectColors[i]]=TRUE;
 changecount = 0;
 for (i = 0; i < 8; i++)
   doOperator[i] = TRUE;
 rmcol = (unsigned char) xinfo.allocCells[frgColInd].pixel;
 cmpcol = (unsigned char) xinfo.allocCells[bckColInd].pixel;
 /* these operations should not be performed on protected colors */
 if (((contourOperation == CONTOUR_PRUN) || (contourOperation == CONTOUR_REMOVE)) &&
  myProtected[rmcol])
  { XBell(xinfo.dpy, 0); return; }

 bpl = img->bytes_per_line;
 xdim = img->width;
 ydim = img->height;
 xlimit = rect->x + rect->width + xdim;
 ylimit = rect->y + rect->height + ydim;
 data = (unsigned char *) img->data;

 buffHalf = rect->width * 3;	/* allocate 6 rows of data */
 changepos = malloc (buffHalf * 2 * sizeof (unsigned long));
 changeval = malloc (buffHalf * 2);
 if ((!changepos) || (!changeval))
  {
   if (changepos)
    free(changepos);
   if (changeval)
    free(changeval);
   allOver = True;
   XBell(xinfo.dpy, 0);
  }
 while (!allOver)
  { /* handle y case separately, since we need to loop x, then y */
   if (contourOperation==CONTOUR_STITCH_Y)
    {
     for (x = rect->x + xdim; x < xlimit; x++)
      {
       int midx, mody;
       unsigned char upCol, downCol;

       midx = x % xdim;
       left = (x + xdim - 1) % xdim;
       right = (x + 1) % xdim;
 
        for (y = rect->y + ydim; y < ylimit; y++)
	 {
	  midy = (y % ydim)*bpl;
	  mody = y % ydim;
	  up = ((y + ydim - 1) % ydim)*bpl;
	  down = ((y + 1) % ydim)*bpl;
	  if ((reg) && (!XPointInRegion (reg, midx, mody)))
	   continue;
	  pos = midx + midy;
	  if (data[pos] == rmcol)
	   {
	    i=1;
	    upCol = data[midx + up];
	    // get the total length of the color effect
	    while ((data[midx + ((y+i)%ydim)*bpl] == rmcol) && ((y+i)<ylimit))
	     i++;
	    downCol = data[midx + ((y + i) % ydim)*bpl];
	    j = i>>1; // mid point for stitch 
	    // move the mid point by one, if on same position as left
	    if (data[left + (((y + j) % ydim))*bpl]==rmcol)
	     j++;
	    // now color the up part with the up color 
	    for (k = 0; k<j; k++)
	     setChangePosVal(changecount++, midx + ((y+k)%ydim)*bpl, upCol);
	    // color mid point with contour color 
	     setChangePosVal(changecount++, midx + ((y+j)%ydim)*bpl, rmcol);
	    // now color the down part with the down color 
	    for (k = j+1; k<i; k++)
	     setChangePosVal(changecount++, midx + ((y+k)%ydim)*bpl, downCol);
	    if (i>1)
	     y+=i-1;
	   }
	 }

/*
       midy = (y % ydim) * bpl;
       mody = y % ydim;
       up = ((y - 1) % ydim) * bpl;
       down = ((y + 1) % ydim) * bpl;
 
        for (y = rect->y + ydim; y < ylimit; y++)
	 {
	  midx = x % xdim;
	  if ((reg) && (!XPointInRegion (reg, midx, mody)))
	   continue;
	  pos = midx + midy;
	  if (data[pos] == rmcol)
	   {
	    i=1;
	    left = data[midy + (x - 1) % xdim];
	    // get the total length of the color effect
	    while ((data[midy + ((x+i)%xdim)] == rmcol) && ((x+i)<xlimit))
	     i++;
	    right = data[midy + (x + i) % xdim];
	    j = i>>1; // mid point for stitch 
	    // move the mid point by one, if on same position as above
	    if (data[up + ((x + j) % xdim)]==rmcol)
	     j++;
	    // now color the left part with the left color 
	    for (k = 0; k<j; k++)
	     setChangePosVal(changecount++, midy + ((x+k)%xdim), left);
	    // color mid point with contour color 
	     setChangePosVal(changecount++, midy + ((x+j)%xdim), rmcol);
	    // now color the right part with the right color 
	    for (k = j+1; k<i; k++)
	     setChangePosVal(changecount++, midy + ((x+k)%xdim), right);
	    if (i>1)
	     x+=i-1;
	   }
	 }

*/


	// must apply it immediately, or it does not work properly 
	flushChanges(data, myProtected);
        allOver = TRUE;
      }
    }
   else // all the other cases

   for (y = rect->y + ydim; y < ylimit; y++)
    {
     if (changecount > buffHalf)
      flushChanges(data, myProtected);
     midy = (y % ydim) * bpl;
     mody = y % ydim;
     up = ((y - 1) % ydim) * bpl;
     down = ((y + 1) % ydim) * bpl;

     switch (contourOperation)
      {
       case CONTOUR_THICK:
       case CONTOUR_THICK_LEFT:
       case CONTOUR_THICK_RIGHT:
       case CONTOUR_THICK_UP:
       case CONTOUR_THICK_DOWN:
	doLeft=doRight=doUp=doDown=(contourOperation == CONTOUR_THICK);
        if (contourOperation == CONTOUR_THICK_LEFT)
	 doLeft=TRUE;
        if (contourOperation == CONTOUR_THICK_RIGHT)
	 doRight=TRUE;
        if (contourOperation == CONTOUR_THICK_UP)
	 doUp=TRUE;
        if (contourOperation == CONTOUR_THICK_DOWN)
	 doDown=TRUE;
	for (x = rect->x + xdim; x < xlimit; x++)
	 {
	  doit = FALSE;
	  midx = x % xdim;
	  pos = midy + midx; 
	  if ((reg) && (!XPointInRegion(reg, midx, mody)))
	   continue;
	  if (data[pos] != rmcol)
	   {
	    doit |= doLeft && (data[midy + ((x + 1) % xdim)] == rmcol);
	    doit |= doRight && (data[midy + ((x - 1) % xdim)] == rmcol);
	    doit |= doUp && (data[down+midx] == rmcol);
	    doit |= doDown && (data[up+midx] == rmcol);
	   }
          if (doit)
	   setChangePosVal(changecount++, pos, rmcol);
	 }
	break;
       case CONTOUR_CREATE:
       case CONTOUR_CREATE_LEFT:
       case CONTOUR_CREATE_RIGHT:
       case CONTOUR_CREATE_UP:
       case CONTOUR_CREATE_DOWN:
	doLeft=doRight=doUp=doDown=doLeftUp=doLeftDown=doRightUp=doRightDown=(contourOperation == CONTOUR_CREATE);
	if (contourOperation == CONTOUR_CREATE_LEFT)
	 doLeft=TRUE;
	if (contourOperation == CONTOUR_CREATE_RIGHT)
	 doRight=TRUE;
	if (contourOperation == CONTOUR_CREATE_UP)
	 doUp=TRUE;
	if (contourOperation == CONTOUR_CREATE_DOWN)
	 doDown=TRUE;
	for (x = rect->x + xdim; x < xlimit; x++)
	 {
	  midx = x % xdim;
	  if ((reg) && (!XPointInRegion (reg, midx, mody)))
	    continue;
	  right = (x + 1) % xdim;
	  left = (x + xdim - 1) % xdim;
	  pos = midy + midx;
	  if (data[pos] != rmcol)
	   {
	    doit = FALSE;

            if (doLeft && ( (rmcol==cmpcol) || (data[right + midy]==cmpcol) ))
	     doit |= (data[right + midy] != rmcol) && (data[pos] != data[right + midy]);

            if (doRight && ( (rmcol==cmpcol) || (data[left + midy]==cmpcol) ))
	     doit |= (data[left + midy] != rmcol) && (data[pos] != data[left + midy]);

            if (doUp && ( (rmcol==cmpcol) || (data[down + midx]==cmpcol) ))
	     doit |= (data[down + midx] != rmcol) && (data[pos] != data[down + midx]);

            if (doDown && ( (rmcol==cmpcol) || (data[up + midx]==cmpcol) ))
	     doit |= (data[up + midx] != rmcol) && (data[pos] != data[up + midx]);

            if (doLeftDown && ( (rmcol==cmpcol) || (data[down + left]==cmpcol) ))
	     doit |= (data[down + left] != rmcol) && (data[pos] != data[down + left]);

            if (doRightUp && ( (rmcol==cmpcol) || (data[up + right]==cmpcol) ))
	     doit |= (data[up + right] != rmcol) && (data[pos] != data[up + right]);

            if (doRightDown && ( (rmcol==cmpcol) || (data[down + right]==cmpcol) ))
	     doit |= (data[down + right] != rmcol) && (data[pos] != data[down + right]);

            if (doLeftUp && ( (rmcol==cmpcol) || (data[up + left]==cmpcol) ))
	     doit |= (data[up + left] != rmcol) && (data[pos] != data[up + left]);

	    if (doit)
	     setChangePosVal(changecount++, pos, rmcol);
	   }
	 }
        allOver = TRUE;
	break;
       /* operators for prunning from Gonzales, Woods: Digital Image Processing, pp. 542 */
       case CONTOUR_PRUN:
	for (x = rect->x + xdim; x < xlimit; x++)
	  {
	    midx = x % xdim;
	    if ((reg) && (!XPointInRegion (reg, midx, mody)))
	      continue;
	    pos = midx + midy;
	    if (data[pos] == rmcol)
	      {
		left = (x + xdim - 1) % xdim;
		right = (x + 1) % xdim;
		switch (numIter % 8)
		 {
		  case 0:
		    if ((doOperation = (data[midy + left] == rmcol) && (data[up + midx] != rmcol) &&
			 (data[up + right] != rmcol) && (data[midy + right] != rmcol) &&
			 (data[down + right] != rmcol) && (data[down + midx] != rmcol)))
		      changeval[changecount] = data[midy + right];
		    break;
		  case 1:
		    if ((doOperation = (data[up + midx] == rmcol) && (data[midx + right] != rmcol) &&
			 (data[down + right] != rmcol) && (data[down + midx] != rmcol) &&
			 (data[down + left] != rmcol) && (data[midy + left] != rmcol)))
		      changeval[changecount] = data[down + midx];
		    break;
		  case 2:
		    if ((doOperation = (data[midy + right] == rmcol) && (data[down + midx] != rmcol) &&
			 (data[down + left] != rmcol) && (data[midy + left] != rmcol) &&
			 (data[up + left] != rmcol) && (data[up + midx] != rmcol)))
		      changeval[changecount] = data[midy + left];
		    break;
		  case 3:
		    if ((doOperation = (data[down + midx] == rmcol) && (data[midy + left] != rmcol) &&
			 (data[up + left] != rmcol) && (data[up + midx] != rmcol) &&
			 (data[up + right] != rmcol) && (data[midy + right] != rmcol)))
		      changeval[changecount] = data[up + midx];
		    break;
		  case 4:
		    if ((doOperation = (data[up + left] == rmcol) && (data[up + midx] != rmcol) &&
			 (data[up + right] != rmcol) && (data[midy + left] != rmcol) &&
			 (data[midy + right] != rmcol) && (data[down + left] != rmcol) &&
			 (data[down + midx] != rmcol) && (data[down + right] != rmcol)))
		      changeval[changecount] = data[down + right];
		    break;
		  case 5:
		    if ((doOperation = (data[up + right] == rmcol) && (data[up + left] != rmcol) &&
			 (data[up + midx] != rmcol) && (data[midy + left] != rmcol) &&
			 (data[midy + right] != rmcol) && (data[down + left] != rmcol) &&
			 (data[down + midx] != rmcol) && (data[down + right] != rmcol)))
		      changeval[changecount] = data[down + left];
		    break;
		  case 6:
		    if ((doOperation = (data[down + right] == rmcol) && (data[up + left] != rmcol) &&
			 (data[up + midx] != rmcol) && (data[up + right] != rmcol) &&
			 (data[midy + left] != rmcol) && (data[midy + right] != rmcol) &&
			 (data[down + left] != rmcol) && (data[down + midx] != rmcol)))
		      changeval[changecount] = data[up + left];
		    break;
		  case 7:
		    if ((doOperation = (data[down + left] == rmcol) && (data[up + left] != rmcol) &&
			 (data[up + midx] != rmcol) && (data[up + right] != rmcol) &&
			 (data[midy + left] != rmcol) && (data[midy + right] != rmcol) &&
			 (data[down + midx] != rmcol) && (data[down + right] != rmcol)))
		      changeval[changecount] = data[up + right];
		    break;
		 }
		if (doOperation)
		 changepos[changecount++] = pos;
	      }
	  }
	 break;
       case CONTOUR_AVERAGE:
	for (x = rect->x + xdim; x < xlimit; x++)
	 {
	  midx = x % xdim;
	  if ((reg) && (!XPointInRegion (reg, midx, mody)))
	    continue;
	  pos = midx + midy;
	  left = (x + xdim - 1) % xdim;
	  right = (x + 1) % xdim;
	  for (i = 0; i < 10; i++)
	    count[i] = 0;

	  for (j = 0; j < 9; j++)
	   {
	    switch (j)
	     {
	      case 0: posval = data[up + left]; break;
	      case 1: posval = data[up + midx]; break;
	      case 2: posval = data[up + right]; break;
	      case 3: posval = data[midy + left]; break;
	      case 4: posval = data[pos]; break;
	      case 5: posval = data[midy + right]; break;
	      case 6: posval = data[down + left]; break;
	      case 7: posval = data[down + midx]; break;
	      case 8: posval = data[down + right]; break;
	     }
	    i = 0;
	    while ((value[i] != posval) && (count[i] != 0))
	      i++;
	    value[i] = posval;
	    count[i]++;
	   }
	  maxval = value[0];
	  max = count[0];
	  i = 1;
	  while (count[i] > 0)
	   {
	    if (count[i] > max)
	     { max = count[i]; maxval = value[i]; }
	    i++;
	   }
	  if ((maxval != data[pos]) && (max > 2))
	   setChangePosVal(changecount++, pos, (unsigned char) maxval);
	 }
        allOver = TRUE;
	break;

       case CONTOUR_DESPECKLE:
	for (x = rect->x + xdim; x < xlimit; x++)
	  {
	   midx = x % xdim;
	   if ((reg) && (!XPointInRegion (reg, midx, mody)))
	     continue;
	   pos = midx + midy;
	   if (data[pos] == rmcol)
	    {
	     left = (x + xdim - 1) % xdim;
	     right = (x + 1) % xdim;
	     for (i = 0; i < 9; i++)
	      { value[i] = rmcol; count[i] = 0; }

	     doOperation = TRUE;
	     for (j = 0; j < 8; j++)
	       {
		switch (j)
		 {
		  case 0: posval = data[up + left]; break;
		  case 1: posval = data[up + midx]; break;
		  case 2: posval = data[up + right]; break;
		  case 3: posval = data[midy + left]; break;
		  case 4: posval = data[midy + right]; break;
		  case 5: posval = data[down + left]; break;
		  case 6: posval = data[down + midx]; break;
		  case 7: posval = data[down + right]; break;
		 }
		if (posval != rmcol)
		 {
		  i = 0;
		  while ((value[i] != posval) && (value[i] != rmcol))
		    i++;
		  value[i] = posval;
		  count[i]++;
		 }
		else
		 doOperation = FALSE;
	       }

	      if (doOperation)
	       {
		maxval = value[0];
		max = count[0];
		i = 1;
		while (value[i] != rmcol)
		 {
		  if (count[i] > max)
		    { max = count[i]; maxval = value[i]; }
		  i++;
		 }
		if (maxval != rmcol)
                 setChangePosVal(changecount++, pos, (unsigned char) maxval);
	       }
	    }
	  }
        allOver = TRUE;
	break;

       case CONTOUR_STITCH_X:
	for (x = rect->x + xdim; x < xlimit; x++)
	 {
	  midx = x % xdim;
	  if ((reg) && (!XPointInRegion (reg, midx, mody)))
	   continue;
	  pos = midx + midy;
	  if (data[pos] == rmcol)
	   {
	    i=1;
	    left = data[midy + (x + xdim - 1) % xdim];
	    /* get the total length of the color effect */
	    while ((data[midy + ((x+i)%xdim)] == rmcol) && ((x+i)<xlimit))
	     i++;
	    right = data[midy + (x + i) % xdim];
	    j = i>>1; /* mid point for stitch */
	    /* move the mid point by one, if on same position as above */
	    if (data[up + ((x + j) % xdim)]==rmcol)
	     j++;
	    /* now color the left part with the left color */
	    for (k = 0; k<j; k++)
	     setChangePosVal(changecount++, midy + ((x+k)%xdim), left);
	    /* color mid point with contour color */
	     setChangePosVal(changecount++, midy + ((x+j)%xdim), rmcol);
	    /* now color the right part with the right color */
	    for (k = j+1; k<i; k++)
	     setChangePosVal(changecount++, midy + ((x+k)%xdim), right);
	    if (i>1)
	     x+=i-1;
	   }
	 }
	/* must apply it immediately, or it does not work properly */
	flushChanges(data, myProtected);
        allOver = TRUE;
        break;

       case CONTOUR_REMOVE:
	for (x = rect->x + xdim; x < xlimit; x++)
	 {
	  midx = x % xdim;
	  if ((reg) && (!XPointInRegion(reg, midx, mody)))
	    continue;
	  pos = midx + midy;
	  if (data[pos] == rmcol)
	   {
	    left = (x + xdim - 1) % xdim;
	    right = (x + 1) % xdim;
	    for (i = 0; i < 9; i++)
	      { value[i] = rmcol; count[i] = 0; }
	    for (j = 0; j < 8; j++)
	      {
	       switch (j)
		{
		 case 0: posval = data[up + left]; break;
		 case 1: posval = data[up + midx]; break;
		 case 2: posval = data[up + right]; break;
		 case 3: posval = data[midy + left]; break;
		 case 4: posval = data[midy + right]; break;
		 case 5: posval = data[down + left]; break;
		 case 6: posval = data[down + midx]; break;
		 case 7: posval = data[down + right]; break;
		}
	       if (posval != rmcol)
		{
		 i = 0;
		 while ((value[i] != posval) && (value[i] != rmcol))
		  i++;
		 value[i] = posval;
		 count[i]++;
		}
	      }
	     maxval = value[0];
	     max = count[0];
	     i = 1;
	     while (value[i] != rmcol)
	      {
	       if (count[i] > max)
		{ max = count[i]; maxval = value[i]; }
	       i++;
	      }
	     if (maxval != rmcol)
              setChangePosVal(changecount++, pos, (unsigned char) maxval);
	    }
	  }
	 break;
	}
     }
   if ( (contourOperation == CONTOUR_THICK) ||
        (contourOperation == CONTOUR_THICK_LEFT) ||
        (contourOperation == CONTOUR_THICK_RIGHT) ||
        (contourOperation == CONTOUR_THICK_UP) ||
        (contourOperation == CONTOUR_THICK_DOWN) )
    if (numIter==(ApData.drawinfo.lineWidth-1))
     allOver=TRUE;

   if (contourOperation == CONTOUR_PRUN)
    {
     if (changecount == 0)
      doOperator[numIter % 8] = FALSE;
     else
      doOperator[numIter % 8] = TRUE;
     allOver = TRUE;
     for (i = 0; i < 8; i++)
      if (doOperator[i])
       allOver = FALSE;
    }
   if ((contourOperation == CONTOUR_REMOVE) && (changecount == 0))
    allOver = TRUE;

   numIter++;
   if (numIter >= 40)
    allOver = TRUE;
   flushChanges(data, myProtected);
  }
 if (changepos)
  free(changepos);
 if (changeval)
  free(changeval);
}

static void processContours(XtPointer clientData, XtIntervalId *id)
{
 XImage *img = NULL;
 Rectangle rect, wholeA, selectR;
 Boolean rectangular;
 int pointNo;
 int contourOperation = atoi ((char *) clientData);

 highlightButton(NULL);
 finalDraw(ApData.pixy);
 disableUndo();
 setToolEventHandler(none, XC_left_ptr);

 selectR.x = Select.x;
 selectR.y = Select.y;
 selectR.width = (unsigned int) Select.width;
 selectR.height = (unsigned int) Select.height;
 rectangular = Select.rectangular;
 pointNo = Select.npoints;
 disableSelection (False);
 Select.selected = Select.rect_on = True;
 Select.rectangular = rectangular;
 Select.npoints = pointNo;
 Select.x = selectR.x;
 Select.y = selectR.y;
 Select.width = selectR.width;
 Select.height = selectR.height;
 if (!rectangular)
  Select.reg = XPolygonRegion(Select.points, Select.npoints, WindingRule);

 wholeA.x = wholeA.y = 0;
 wholeA.width = (unsigned int) ApData.width;
 wholeA.height = (unsigned int) ApData.height;

 if (intersectRect (&selectR, &wholeA, &rect))
  {
   img = SubImage(ApData.img, rect.x, rect.y, rect.width, rect.height);
   if (checkXImage(img))
    {
     ApData.tool = Undo.tool = contourT;
     Select.width = opBounds.width = rect.width;
     Select.height = opBounds.height = rect.height;
     Select.x = opBounds.x = Undo.points[Undo.npoints].x = rect.x;
     Select.y = opBounds.y = Undo.points[Undo.npoints++].y = rect.y;

     if (!rectangular)
      {
       int i, maxX, maxY;

       XDestroyRegion (Select.reg);
       maxX = Select.x + Select.width - 1;
       maxY = Select.y + Select.height - 1;
       for (i = 0; i < Select.npoints; i++)
	{
	 Select.points[i].x = MAX(Select.x, Select.points[i].x);
	 Select.points[i].x = MIN(maxX, Select.points[i].x);
	 Select.points[i].y = MAX(Select.y, Select.points[i].y);
	 Select.points[i].y = MIN(maxY, Select.points[i].y);
	}
       Select.reg = XPolygonRegion(Select.points, Select.npoints, WindingRule);
      }

     if (contourOperation == CONTOUR_THIN)
      thinContours(ApData.img, xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, &rect, Select.reg);
     else
      processContourOperation(ApData.img, ColorMatrix.FrgColorInd, ColorMatrix.BckColorInd, &rect, Select.reg, contourOperation);

     enableUndo (False);
     Undo.img = img;
     Undo.imgAllocated = True;
     memcpy (&Undo.area, &rect, sizeof (Rectangle));
     ApData.pic_change = True;
     opBounds.x -= 2;
     opBounds.y -= 2;
     opBounds.width += 4;
     opBounds.height += 4;
    }
  }
 LastDrawOnAllWin(False);
 setToolEventHandler(Undo.tool, 0);
 XtVaSetValues(PasteBW, XmNsensitive, False, NULL);
 XtVaSetValues(CopyBW, XmNsensitive, True, NULL);
 XtVaSetValues(CutBW, XmNsensitive, True, NULL);
}

DCONTOUR void doContourCB(Widget w, XtPointer clientData, XtPointer callData)
{
 if (!Select.selected)
  selectCB(w, "2", NULL); /* select the whole image if nothing is selected */
 if (Select.selected)
  {
   if (ApData.viewMode != normalView)
    XmToggleButtonSetState (normalViewTBW, True, True);
   XtAddTimeOut(100, (XtTimerCallbackProc) *processContours, clientData);
  }
}

/* "Efficient Binary Image Thinning using Neighborhood Maps by Joseph M. Cychosz,
   in "Graphics Gems IV", Academic Press, 1994
   Direction masks:     N    S     W     E  */
static int masks[] = {0200, 0002, 0040, 0010};

/* True if pixel neighbor map indicates the pixel is 8-simple and
   not an end point and thus can be deleted.  The neighborhood
   map is defined as an integer of bits abcdefghi with a non-zero
   bit representing a non-zero pixel.  The bit assignment for the
   neighborhood is:  a b c
                     d e f
                     g h i */

static unsigned char delete[512] =
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,1,0,0,1,1,0,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,1,1,1,0,1,1,0,0,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,1,1,1,0,1,1,0,0,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1};

static void thinContours(XImage *img, int colInd, Rectangle *rect, Region reg)
{
#define PIX(y,x) data[bpl*(y)+(x)]
#define DEL_PIX(yy,xx,ii) \
 switch (ii) \
 { \
  case 0: \
	if ((yy) == ystart) \
	 PIX((yy),(xx)) = PIX(ypred,(xx)); \
	else \
	 PIX((yy),(xx)) = PIX((yy)-1,(xx)); \
	break; \
  case 1: \
	if ((yy) == yend) \
	 PIX((yy),(xx)) = PIX(ysucc,(xx)); \
	else \
	 PIX((yy),(xx)) = PIX((yy)+1,(xx)); \
	break; \
  case 2: \
	if ((xx) == xstart) \
	 PIX((yy),(xx)) = PIX((yy),xpred); \
	else \
	 PIX((yy),(xx)) = PIX((yy),(xx)-1); \
	break; \
  case 3: \
	if ((xx) == xend) \
	 PIX((yy),(xx)) = PIX((yy),xsucc); \
	else \
	 PIX((yy),(xx)) = PIX((yy),(xx)+1); \
 }
 unsigned char *data = (unsigned char *) img->data;
 unsigned int bpl = img->bytes_per_line;
 unsigned int xstart = rect->x;
 unsigned int ystart = rect->y;
 unsigned int xend = xstart + rect->width - 1;
 unsigned int yend = ystart + rect->height - 1;
 unsigned int xpred = (xstart + img->width - 1) % img->width;
 unsigned int ypred = (ystart + img->height - 1) % img->height;
 unsigned int xsucc = (xstart + (unsigned int) (rect->width)) % img->width;
 unsigned int ysucc = (ystart + (unsigned int) (rect->height)) % img->height;
 unsigned int j;

 unsigned int xsize;	// Image resolution
 unsigned int x, y;	// Pixel location
 unsigned int i;	// Pass index
 unsigned int pc = 0;	// Pass count
 unsigned int count = 1;// Deleted pixel count
 unsigned int p, q;	// Neighborhood maps of adjacent cells
 unsigned int *qb;	// Neighborhood maps of previous scanline
 unsigned int m;	// Deletion direction mask

 /* do not thin contour if it is a protected color */
 for (i=0; i<ApData.ProtColNo; i++)
  if (ApData.ProtectColors[i]==colInd)
   { XBell(xinfo.dpy, 0); return; }

 if ((img->width < 2) || (img->height < 2))
   return;

 xsize = rect->width;
 qb = (unsigned int *) malloc ((xsize + 1) * sizeof (unsigned int));

 while (count && pc < 10)
  {	/* Scan image while deletions */
   pc++;
   count = 0;
   for (i = 0; i < 4; i++)
    {
     m = masks[i];

     /* Build initial previous scan buffer */
     p = PIX (ypred, xpred) == colInd;
     for (j = 0, x = xstart; j < xsize; j++, x++)
       qb[j] = p = ((p << 1) & 0006) | (PIX (ypred, x) == colInd);
     qb[xsize] = ((p << 1) & 0006) | (PIX (ypred, xsucc) == colInd);

     q = qb[0];
     p = ((q << 2) & 0010) | (PIX (ystart, xpred) == colInd);
     for (j = 0, x = xstart; j < xsize; j++, x++)
      {
       q = qb[j];
       p = ((p << 1) & 0066) | ((q << 3) & 0010) | (PIX (ystart, x) == colInd);
       qb[j] = p;
      }
     q = qb[xsize];
     qb[xsize] = ((p << 1) & 0066) | ((q << 3) & 0010) | (PIX (ystart, xsucc) == colInd);

     /* Scan image for pixel deletion candidates */
     for (y = ystart; y < yend; y++)
      {
       q = qb[0];
       p = ((q << 2) & 0110) | (PIX (y + 1, xpred) == colInd);
       p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (y + 1, xstart) == colInd);
       qb[0] = p;
       for (j = 1, x = xstart; j < xsize; j++, x++)
	{
	 q = qb[j];
	 p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (y + 1, x + 1) == colInd);
	 qb[j] = p;
	 if ((reg) && (!XPointInRegion (reg, x, y)))
	    continue;
	 if (((p & m) == 0) && delete[p])
	   { count++; DEL_PIX(y, x, i) }
	}
       /* Process right edge pixel */
       if ((reg) && (!XPointInRegion (reg, x, y)))
	  continue;
       q = qb[xsize];
       p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (y + 1, xsucc) == colInd);
       qb[xsize] = p;
       if ((p & m) == 0 && delete[p])
	 { count++; DEL_PIX(y, xend, i) }
      }
     /* Process bottom scan line */
     q = qb[0];
     p = ((q << 2) & 0110) | (PIX (ysucc, xpred) == colInd);
     p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (ysucc, xstart) == colInd);

     for (j = 1, x = xstart; j < xsize; j++, x++)
      {
       if ((reg) && (!XPointInRegion (reg, x, yend)))
	continue;
       q = qb[j];
       p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (ysucc, x + 1) == colInd);
       if ((p & m) == 0 && delete[p])
	 { count++; DEL_PIX(yend, x, i) }
      }
     q = qb[xsize];
     p = ((p << 1) & 0666) | ((q << 3) & 0110) | (PIX (ysucc, xsucc) == colInd);
     if ((reg) && (!XPointInRegion (reg, xend, yend)))
      continue;
     if ((p & m) == 0 && delete[p])
      { count++; DEL_PIX(yend, xend, i) }
    }
  }
 free(qb);
#undef DEL_PIX
#undef PIX
}
