#define DRAW
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ScrollBar.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "paint.h"

static XPoint parElArcPts[3];	/* storage of P, Q, K points on arcs */
static unsigned short n_arcPar;	/* number of arc points set so far */
static int currWin;		/* current drawing area window */
static int zoomXF, zoomYF;	/* current zoom factor */
static int zoX, zoY;		/* offset when drawing in zoom window */
static Drawable drawable;	/* where to draw to */
static XPoint oldEventPt;	/* point of the last event */
static Boolean firstClick;
static int drawZoom = FALSE;

DDRAW void findMinMax(XPoint *pts, int numPts, short *minX, short *minY,
 short *maxX, short *maxY)
{
 int i;
 short tmpMinX, tmpMinY, tmpMaxX, tmpMaxY;

 tmpMinX = tmpMinY = 0x7fff;
 tmpMaxX = tmpMaxY = 0;
 for (i = 0; i < numPts; i++)
  {
   if (tmpMinX > pts->x)
    tmpMinX = pts->x;
   if (tmpMinY > pts->y)
    tmpMinY = pts->y;
   if (tmpMaxX < pts->x)
    tmpMaxX = pts->x;
   if (tmpMaxY < pts->y)
    tmpMaxY = pts->y;
   pts++;
  }
 pts -= numPts;
 *minX = tmpMinX;
 *minY = tmpMinY;
 *maxX = tmpMaxX;
 *maxY = tmpMaxY;
}

static void arcDrawOnce(XPoint newEventPt)
{
 int np, lnNo;

 np = Undo.npoints;
 XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
 if (!firstClick)
  {
   lnNo = Undo.points[np - 1].x;
   np -= lnNo + 1;
   drawLines(drawable, xinfo.gc, False, &Undo.points[np], lnNo,
    zoX / zoomXF, zoY / zoomYF, True);
  }

 Undo.points[np - 3].x = parElArcPts[2].x = newEventPt.x;
 Undo.points[np - 3].y = parElArcPts[2].y = newEventPt.y;
 lnNo = DrawElArc(Undo.points[np - 5].x, Undo.points[np - 5].y,
  Undo.points[np - 4].x, Undo.points[np - 4].y, Undo.points[np - 3].x, Undo.points[np - 3].y,
 &Undo.points[np]);
 drawLines(drawable, xinfo.gc, False, &Undo.points[np], lnNo, zoX/zoomXF, zoY/zoomYF, True);
 if (drawZoom)
  {
   if (firstClick)
    findMinMax(&Undo.points[np], lnNo, &opBounds.x, &opBounds.y, &opBounds.width, &opBounds.height);
   else
    findMinMax(&Undo.points[np - 2], lnNo + 2, &opBounds.x, &opBounds.y,
     &opBounds.width, &opBounds.height);
   findMinMax(&Undo.points[np], lnNo, &Undo.points[np - 2].x, &Undo.points[np - 2].y,
     &Undo.points[np - 1].x, &Undo.points[np - 1].y);
   opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth*2;
   opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth*2;
   opBounds.x -= ApData.drawinfo.lineWidth;
   opBounds.y -= ApData.drawinfo.lineWidth;
   XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
   refreshZoomWin (opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
  }
 Undo.npoints = np + lnNo + 1;
 Undo.points[Undo.npoints - 1].x = (short) lnNo;
}

static void HandleBtnMotion(Widget w, XtPointer unused, XEvent *event)
{
 int r1, r2, dx, dy, xl, yt, np;
 XPoint newEventPt;

 if (event->type == ButtonRelease)
   {
    int oldFG = ColorMatrix.FrgColorInd;

     if (event->xbutton.button == Button3)
       return;
     if (event->xbutton.button == Button2)
      ColorSelectConseq(Button1, ColorMatrix.BckColorInd);

     setExclEventHandler((XtEventHandler) HandleBtnPress, ButtonPressMask, True);
     XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
     Undo.tool = ApData.tool;
     memcpy(&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));

     XCopyGC(xinfo.dpy, xinfo.gc, GCFunction | GCLineStyle | GCLineWidth | GCForeground |
      GCBackground | GCJoinStyle | GCCapStyle, Undo.gc);

     enableUndo (True);

     ApData.pic_change = True;
     switch (ApData.tool)
       {
       case curveDraw:
	 if (Undo.npoints == 1) /* no motion -> draw line of lenght 1 */
	   {
	     memcpy (&Undo.points[1], &Undo.points[0], sizeof (XPoint));
	     Undo.npoints++;
	   }
	 findMinMax (Undo.points, Undo.npoints, &opBounds.x, &opBounds.y,
		     &opBounds.width, &opBounds.height);
	 opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	 opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	 opBounds.x -= ApData.drawinfo.lineWidth;
	 opBounds.y -= ApData.drawinfo.lineWidth;
	 break;

       case lineDraw:
       case rotateByLine:
	 XSetFunction (xinfo.dpy, xinfo.gc, GXinvert);
	 drawLines (drawable, xinfo.gc, False,
		    &Undo.points[0], 2, zoX / zoomXF, zoY / zoomYF, False);
	 XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
	 findMinMax (&Undo.points[0], 2, &opBounds.x, &opBounds.y,
		     &opBounds.width, &opBounds.height);
	 opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	 opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	 opBounds.x -= ApData.drawinfo.lineWidth;
	 opBounds.y -= ApData.drawinfo.lineWidth;
	 break;

       case rectDraw:
       case squareDraw:
	 np = Undo.npoints - 2;
	 opBounds.x = Undo.points[np].x;
	 opBounds.y = Undo.points[np].y;
	 opBounds.width = Undo.points[np + 1].x + 1;
	 opBounds.height = Undo.points[np + 1].y + 1;
	 break;
       case ellipseDraw:
       case circDraw:
	 np = Undo.npoints - 3;
	 opBounds.x = Undo.points[np].x;
	 opBounds.y = Undo.points[np].y;
	 opBounds.width = Undo.points[np + 1].x;
	 opBounds.height = Undo.points[np + 1].y;
	 break;
       case arcDraw:
	 {
	  int zofx = zoX / zoomXF, zofy = zoY / zoomYF, lnNo;

	  np = Undo.npoints;

	  lnNo = Undo.points[np - 1].x;
	  np -= lnNo + 1;
	  XSetFunction (xinfo.dpy, Undo.gc, GXinvert);
	  drawLines(drawable, Undo.gc, False, &Undo.points[np], lnNo, zofx, zofy, True);
	  XSetFunction (xinfo.dpy, Undo.gc, GXcopy);

	  findMinMax(&Undo.points[np], lnNo, &opBounds.x, &opBounds.y, &opBounds.width, &opBounds.height);
	  opBounds.width -= opBounds.x;
	  opBounds.height -= opBounds.y;
	  n_arcPar = 0;
	 }
	 break;
       default:
	 break;
       }
     {
      int lnW = ApData.drawinfo.lineWidth;

      segmentDrawOp();
      switch (ApData.tool)
       {
	case lineDraw:
	case curveDraw:
	case rectDraw:
	case squareDraw:
	case ellipseDraw:
	case circDraw:
	case arcDraw:
	 calcOpBounds();
	 break;
	case rotateByLine:
	 rotatePictureByLine();
	 return;
	 break;
	default:
	 if (lnW == 0)
	   lnW++;
	 opBounds.x -= (lnW + 1) / 2;
	 opBounds.y -= (lnW + 1) / 2;
	 opBounds.width += lnW + 1;
	 opBounds.height += lnW + 1;
	 break;
       }
      Undo.area.x = opBounds.x;
      Undo.area.y = opBounds.y;
      Undo.area.width = opBounds.width;
      Undo.area.height = opBounds.height;
      Undo.img = SubImage(ApData.img, opBounds.x, opBounds.y,
       opBounds.width, opBounds.height);
      if (Undo.img == NULL)
	disableUndo();
      else
	Undo.imgAllocated = True;
     }
    LastDrawOnAllWin(True);
    if (event->xbutton.button == Button2)
     ColorSelectConseq(Button1, oldFG);
   }
 else
   {
     /* check if mouse move inside zoom really is a new point */
     newEventPt.x = (event->xbutton.x + zoX) / zoomXF;
     newEventPt.y = (event->xbutton.y + zoY) / zoomYF;
     if ((oldEventPt.x == newEventPt.x) &&
	 (oldEventPt.y == newEventPt.y))
       return;

     XSetFunction (xinfo.dpy, xinfo.gc, GXinvert);
     switch (ApData.tool)
      {
       case curveDraw:
	 if (Undo.npoints >= Undo.allocPoints)
	   return;
	 XSetForeground(xinfo.dpy, xinfo.gc,
	  foregroundDraw ? ApData.drawinfo.foregrndC : ApData.drawinfo.backgrndC);
	 XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);
	 np = Undo.npoints;
	 if (ApData.tool == curveDraw)
	  {
	   memcpy(&Undo.points[np], &newEventPt, sizeof (XPoint));
	   XDrawLine(xinfo.dpy, drawable, xinfo.gc,
	    Undo.points[np - 1].x - zoX / zoomXF,
	    Undo.points[np - 1].y - zoY / zoomYF,
            Undo.points[np].x - zoX / zoomXF,
	    Undo.points[np].y - zoY / zoomYF);
	   findMinMax(&Undo.points[np - 1], 2, &opBounds.x, &opBounds.y,
	    &opBounds.width, &opBounds.height);
	   opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	   opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	   opBounds.x -= ApData.drawinfo.lineWidth;
	   opBounds.y -= ApData.drawinfo.lineWidth;
	  }
	 else
	  {
	   Undo.points[np].x = newEventPt.x;
	   Undo.points[np].y = newEventPt.y;
	   XFillRectangle(xinfo.dpy, drawable, xinfo.gc,
	    Undo.points[np].x - (ApData.drawinfo.lineWidth >> 1) - zoX/zoomXF,
	    Undo.points[np].y - (ApData.drawinfo.lineWidth >> 1) - zoY/zoomYF,
	    ApData.drawinfo.lineWidth, ApData.drawinfo.lineWidth);
	   opBounds.x = Undo.points[np].x - (ApData.drawinfo.lineWidth + 1)/2;
	   opBounds.y = Undo.points[np].y - (ApData.drawinfo.lineWidth + 1)/2;
	   opBounds.width = opBounds.height = ApData.drawinfo.lineWidth;
	  }
	 Undo.npoints++;
	 if (drawZoom)
	   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	 break;
       case lineDraw:
       case rotateByLine:
	 np = Undo.npoints - 2;
	 if (!firstClick)
	   drawLines(drawable, xinfo.gc, False,
	    &Undo.points[np], 2, zoX / zoomXF, zoY / zoomYF, False);
	 memcpy (&Undo.points[np + 2], &Undo.points[np + 1], sizeof (XPoint));
	 memcpy (&Undo.points[np + 1], &newEventPt, sizeof (XPoint));
	 drawLines(drawable, xinfo.gc, False,
	  &Undo.points[np], 2, zoX / zoomXF, zoY / zoomYF, False);
	 findMinMax(&Undo.points[np], 3, &opBounds.x, &opBounds.y,
	  &opBounds.width, &opBounds.height);
	 opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	 opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	 opBounds.x -= ApData.drawinfo.lineWidth;
	 opBounds.y -= ApData.drawinfo.lineWidth;
	 if (drawZoom)
	  {
	   XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);
	   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	   XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
	  }
	 break;

       case rectDraw:
       case squareDraw:
	 np = Undo.npoints - 2;
	 if (!firstClick)
	  DrawRectangle(drawable, xinfo.gc, ApData.drawinfo, &Undo.points[np],
	   zoX/zoomXF, zoY/zoomYF);

	 dx = ApData.orig_xy.x - newEventPt.x;
	 dy = ApData.orig_xy.y - newEventPt.y;
	 memcpy(&Undo.points[np + 2], &Undo.points[np], sizeof (XPoint));
	 memcpy(&Undo.points[np + 3], &Undo.points[np + 1], sizeof (XPoint));

	 if (ApData.tool == squareDraw)
	   {
	     if (abs (dx) >= abs (dy))
	       dy = (int) (SGN(dy)*abs(dx)/densityRatio());
	     else
	       {
		 dx = SGN (dx) * abs (dy);
		 dy = (int) (dy/densityRatio());
	       }
	   }
	 xl = ApData.orig_xy.x;
	 yt = ApData.orig_xy.y;
	 if (dx > 0)
	   xl -= dx;
	 if (dy > 0)
	   yt -= dy;

	 Undo.points[np].x = (short) xl;
	 Undo.points[np].y = (short) yt;
	 Undo.points[np + 1].x = (short) abs (dx);
	 Undo.points[np + 1].y = (short) abs (dy);
	 DrawRectangle(drawable, xinfo.gc, ApData.drawinfo, &Undo.points[np],
			zoX/zoomXF, zoY/zoomYF);
	 if (drawZoom)
	  {
	   opBounds.x = MIN(Undo.points[np + 2].x, Undo.points[np].x);
	   opBounds.y = MIN(Undo.points[np + 2].y, Undo.points[np].y);
	   opBounds.width = MAX(Undo.points[np + 3].x + Undo.points[np + 2].x,
	    Undo.points[np + 1].x + Undo.points[np].x);
	   opBounds.height = MAX(Undo.points[np + 3].y + Undo.points[np + 2].y,
	    Undo.points[np + 1].y + Undo.points[np].y);
	   opBounds.width -= opBounds.x;
	   opBounds.height -= opBounds.y;
	   opBounds.x -= ApData.drawinfo.lineWidth;
	   opBounds.y -= ApData.drawinfo.lineWidth;
	   opBounds.width += ApData.drawinfo.lineWidth * 2;
	   opBounds.height += ApData.drawinfo.lineWidth * 2;
	   XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
	   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	   XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
	  }
	 break;

       case circDraw:
       case ellipseDraw:
	 np = Undo.npoints - 4;
	 drawArc(drawable, xinfo.gc, ApData.drawinfo, &Undo.points[np + 1],
		  zoX/zoomXF, zoY/zoomYF);

	 dx = Undo.points[np].x - newEventPt.x;
	 dy = Undo.points[np].y - newEventPt.y;
	 if (ApData.tool == circDraw)
	   {
	     if (abs (dx) >= abs (dy))
	       dy = (int) (SGN(dy)*abs(dx)/densityRatio());
	     else
	       {
		 dx = SGN (dx) * abs (dy);
		 dy = (int) (dy/densityRatio());
	       }
	   }
	 r1 = abs ((int) (dx));
	 r2 = abs ((int) (dy));
	 Undo.points[np + 1].x = Undo.points[np].x - r1;
	 Undo.points[np + 1].y = Undo.points[np].y - r2;
	 Undo.points[np + 2].x = 2 * r1;
	 Undo.points[np + 2].y = 2 * r2;

	 drawArc(drawable, xinfo.gc, ApData.drawinfo, &Undo.points[np + 1],
		  zoX/zoomXF, zoY/zoomYF);
	 if (drawZoom)
	  {
	   opBounds.x = MIN(Undo.points[np + 4].x, Undo.points[np + 1].x);
	   opBounds.y = MIN(Undo.points[np + 4].y, Undo.points[np + 1].y);
	   opBounds.width = MAX(Undo.points[np + 5].x, Undo.points[np + 2].x);
	   opBounds.height = MAX(Undo.points[np + 5].y, Undo.points[np + 2].y);

	   opBounds.x -= ApData.drawinfo.lineWidth;
	   opBounds.y -= ApData.drawinfo.lineWidth;
	   opBounds.width += ApData.drawinfo.lineWidth * 2;
	   opBounds.height += ApData.drawinfo.lineWidth * 2;
	   XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
	   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	   XSetFunction (xinfo.dpy, xinfo.gc, GXinvert);
	   memcpy(&Undo.points[np + 4], &Undo.points[np + 1], 2 * sizeof (XPoint));
	  }
	 break;
       case arcDraw:
	 arcDrawOnce (newEventPt);
	 break;
       default:
	 break;
      }
     memcpy(&oldEventPt, &newEventPt, sizeof(XPoint));
     if (firstClick)
       firstClick = False;
   }
}

/* Button Press event handler for drawing routines */
static void DrawBtnPress(Widget w, XEvent * event)
{
  XGCValues values;

  oldEventPt.x = ApData.orig_xy.x = (event->xbutton.x + zoX) / zoomXF;
  oldEventPt.y = ApData.orig_xy.y = (event->xbutton.y + zoY) / zoomYF;

  memcpy (&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));
  memcpy (&Undo.points[0], &ApData.orig_xy, sizeof (XPoint));

  switch (ApData.tool)
    {
    case arcDraw:
    case curveDraw:
      values.cap_style = CapRound;
      values.join_style = JoinRound;
      break;
    default:
      values.cap_style = CapButt;
      values.join_style = JoinMiter;
      break;
    }
  XChangeGC (xinfo.dpy, xinfo.gc, GCCapStyle | GCJoinStyle, &values);

  switch (ApData.tool)
   {
    case curveDraw:
      XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
      if (ApData.tool == curveDraw)
	XDrawLine(xinfo.dpy, drawable, xinfo.gc,
	 Undo.points[0].x - zoX / zoomXF, Undo.points[0].y - zoY / zoomYF,
	 Undo.points[0].x - zoX / zoomXF, Undo.points[0].y - zoY / zoomYF);
      else
       {
	XFillRectangle(xinfo.dpy, drawable, xinfo.gc,
	 Undo.points[0].x - (int) (ApData.drawinfo.lineWidth >> 1) - zoX / zoomXF,
	 Undo.points[0].y - (int) (ApData.drawinfo.lineWidth >> 1) - zoY / zoomYF,
	 ApData.drawinfo.lineWidth, ApData.drawinfo.lineWidth);
       }
      Undo.npoints = 1;
      if (drawZoom)
       {
	opBounds.x = Undo.points[0].x - (int) ApData.drawinfo.lineWidth / 2;
	opBounds.y = Undo.points[0].y - (int) ApData.drawinfo.lineWidth / 2;
	opBounds.width = opBounds.height = ApData.drawinfo.lineWidth * 2;
	refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
       }
      break;
    case lineDraw:
    case rotateByLine:
      Undo.npoints = 2;
      memcpy (&Undo.points[1], &ApData.orig_xy, sizeof (XPoint));
      break;
    case rectDraw:
    case squareDraw:
      memset (&Undo.points[1], 0, sizeof (XPoint) * 4);
      Undo.npoints = 2;
      break;
    case circDraw:
    case ellipseDraw:
      memcpy (&Undo.points[1], &ApData.orig_xy, sizeof (XPoint));
      Undo.points[2].x = Undo.points[2].y = 0;
      Undo.points[3].x = 0;
      Undo.points[3].y = 23040;
      Undo.npoints = 4;
      memset (&Undo.points[4], 0, sizeof (XPoint) * 2);
      break;
    case arcDraw:
      Undo.npoints = 5;
      memcpy (&Undo.points[0], &parElArcPts[0], 2 * sizeof (XPoint));
      memcpy (&Undo.points[2], &ApData.orig_xy, sizeof (XPoint));
      arcDrawOnce (oldEventPt);
      firstClick = False;
      break;
    default:
      Undo.npoints = 0;
      break;
   }
  repeats.firstDraw = pointInRepeat (ApData.orig_xy.x, ApData.orig_xy.y, repeats.currNum);
  setExclEventHandler((XtEventHandler) HandleBtnMotion,
   PointerMotionMask | ButtonReleaseMask, True);
}

/* Handle mouse button1 press - drawing routines - inside drawing area */
static void BtnPress_1(Widget w, XEvent *event)
{
 finalDraw(ApData.pixy);
 opBounds.x = opBounds.y = opBounds.width = opBounds.height = 0;
 disableUndo();

 if (ApData.tool==fill)
  Select.rect_on = False;

 switch (ApData.tool)
  {
   case shiftHor:
   case shiftVer:
     {
       XPoint shiftPt, orR;
       int ind;

       XtRemoveEventHandler(w, ButtonPressMask, FALSE,
	(XtEventHandler) HandleBtnPress, (XtPointer) currWin);

       shiftPt.x = (event->xbutton.x + zoX) / zoomXF;
       shiftPt.y = (event->xbutton.y + zoY) / zoomYF;

       ind = pointInRepeat (shiftPt.x, shiftPt.y, repeats.currNum);
       if (ind == -1)
	 return;

       switch (ApData.viewMode)
	 {
	 case repeatView:
	   orR.x = repeats.crop.width;
	   orR.y = repeats.crop.height;
	   movePts (&shiftPt, 1, -repeats.r[ind].x, -repeats.r[ind].y);
	   flipPts (&shiftPt, 1, repeats.r[ind].rType, &orR);
	   break;
	 default:
	   break;
	 }
       if (ApData.tool == shiftHor)
	 ShiftPic (shiftPt.x, 0);
       else
	 ShiftPic (0, shiftPt.y);
       XtAddEventHandler (w, ButtonPressMask, FALSE,
	(XtEventHandler) HandleBtnPress, (XtPointer) currWin);
     }
     break;
   case fill:
     {
      XImage *patternImg = NULL;

      if ((Select.imgAllocated) && (ApData.drawinfo.fillType == patternFill))
	patternImg = Select.img;
      /* swap back/foreground color, so we fill with the right color with respect to left/middle mouse button */
      if ((!patternImg)&&(event->xbutton.button==Button2))
       SWAP(ColorMatrix.BckColorInd, ColorMatrix.FrgColorInd);
      fillArea((event->xbutton.x + zoX)/zoomXF, (event->xbutton.y + zoY)/zoomYF, patternImg);
      if ((!patternImg)&&(event->xbutton.button==Button2))
       SWAP(ColorMatrix.BckColorInd, ColorMatrix.FrgColorInd);
     }
     break;
   case curveDraw:
   case lineDraw:
   case rotateByLine:
   case rectDraw:
   case squareDraw:
   case circDraw:
   case ellipseDraw:
     DrawBtnPress(w, event);
     break;
   case arcDraw:
     parElArcPts[n_arcPar].x = (event->xbutton.x + zoX) / zoomXF;
     parElArcPts[n_arcPar++].y = (event->xbutton.y + zoY) / zoomYF;
     if (n_arcPar != 3)
       {			/*draw first two points as crosses */
	XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
	drawCross(XtWindow(w), xinfo.gc,
	 parElArcPts[n_arcPar - 1].x * zoomXF - zoX + zoomXF/2-1,
	 parElArcPts[n_arcPar - 1].y * zoomYF - zoY + zoomYF/2-1,
	 5 * (zoomXF + zoomYF) / 2);
	XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);
       }
     else /* undraw first two points as crosses */
       {
	XSetFunction (xinfo.dpy, xinfo.gc, GXinvert);
	drawCross(XtWindow (w), xinfo.gc,
	 parElArcPts[0].x * zoomXF - zoX + zoomXF/2-1,
	 parElArcPts[0].y * zoomYF - zoY + zoomYF/2-1,
	 5 * (zoomXF + zoomYF) / 2);
	drawCross(XtWindow (w), xinfo.gc,
	 parElArcPts[1].x * zoomXF - zoX + zoomXF/2-1,
	 parElArcPts[1].y * zoomYF - zoY + zoomYF/2-1,
	 5 * (zoomXF + zoomYF) / 2);
	DrawBtnPress(w, event);
       }
     break;
   case spray:
   case mix:
     sprayBtnPress(w, (XtPointer) zoomXF, event);
     break;
   default:
     break;
  }
}

static void displayParameterWindow(void)
{
 int val1, val2, val3, val4;

 if ( (ApData.tool==squareDraw) || (ApData.tool==rectDraw) || (ApData.tool==lineDraw) ||
      (ApData.tool==circDraw) || (ApData.tool==ellipseDraw) )
 {
  val1 = val2 = val3 = val4 = 0;
  finalDraw (ApData.pixy); disableUndo();
  initDrawPrmDlg(0, "x1", val1, 0, 0xffff);
  initDrawPrmDlg(1, "y1", ApData.height - 1 - val2, 0, 0xffff);
  initDrawPrmDlg(2, "x2", val3, 0, 0xffff);
  initDrawPrmDlg(3, "y2", ApData.height - 1 - val4, 0, 0xffff);
  XtManageChild(drawPrmW);
 }
 else
  XBell(xinfo.dpy, 0);
}

/* position and refresh zoom window to the mouse position in the main window */
DDRAW void centerZoomWindowToMouse(XEvent *event)
{
 int newX=0, newY=0, valX, valY, sliderSize, incr, pageIncr;

 newX = event->xbutton.x * zoomDraw.factorX;
 newY = event->xbutton.y * zoomDraw.factorY;
 newX -= zoomDraw.clipWin.width / 2;
 newY -= zoomDraw.clipWin.height / 2;
 newX = (newX < 0) ? 0 : newX;
 newY = (newY < 0) ? 0 : newY;

 /* need to do refresh in this weird way, otherwise we get two jumps - one for x */
 /* and one for y, which wastes CPU and introduces flicker */
 blockZoomRefreshes = TRUE;
 XmScrollBarGetValues(scrollX, &valX, &sliderSize, &incr, &pageIncr);
 valX = ((newX + sliderSize) > ApData.DAwidth * zoomDraw.factorX) ?
  ApData.DAwidth*zoomDraw.factorX - sliderSize : newX;
 XmScrollBarSetValues(scrollX, valX, sliderSize, incr, pageIncr, TRUE);

 XmScrollBarGetValues(scrollY, &valY, &sliderSize, &incr, &pageIncr);
 valY = ((newY + sliderSize) > ApData.DAheight * zoomDraw.factorY) ?
  ApData.DAheight*zoomDraw.factorY - sliderSize : newY;
 XmScrollBarSetValues(scrollY, valY, sliderSize, incr, pageIncr, TRUE);

 patchScrollBarPos(scrollX, scrollY, valX, valY);
 blockZoomRefreshes = FALSE;
 refreshZoomWin(0, 0, 32000, 32000, TRUE); /* clipping will take care of this... */

 setZoomFactorCB(mainW, "4", NULL);
}

static void HandleZoomMotion(Widget w, XtPointer unused, XEvent *event)
{
 centerZoomWindowToMouse(event);

 if (event->type == ButtonRelease)
  setExclEventHandler((XtEventHandler) HandleBtnPress, ButtonPressMask, True);
 else
  setExclEventHandler((XtEventHandler) HandleZoomMotion,
   PointerMotionMask | ButtonReleaseMask, TRUE);
}

/* Handle mouse button press inside drawing area w */
DDRAW void HandleBtnPress(Widget w, XtPointer unused, XEvent *event)
{
 if (w == zoomDrawAreaW)
  {
   currWin = 1;
   zoomXF = zoomDraw.factorX;
   zoomYF = zoomDraw.factorY;
   zoX = zoomDraw.clipWin.x;
   zoY = zoomDraw.clipWin.y;
   drawable = ApData.scrPix; /* ABC double buffer problem */
   drawZoom = TRUE;
  }
 else
  {
   currWin = zoX = zoY = 0;
   zoomXF = zoomYF = 1;
   drawable = XtWindow(w);
   drawZoom = FALSE;
  }
 if (event->type == ButtonPress)
  {
   switch (event->xbutton.button)
    {
     case Button1:
      firstClick = True;
      BtnPress_1(w, event);
      ApData.pic_change = True;
      foregroundDraw=TRUE;
      break;
     case Button3:
      if (w==drawAreaWID)
       {
	initDrawAreaCB(mainW, "9", NULL); /* map the zoom window */
	setExclEventHandler((XtEventHandler) HandleZoomMotion,
         PointerMotionMask | ButtonReleaseMask, TRUE);
       }
      else
       if (w == zoomDrawAreaW)
	setZoomFactorCB(w, "2", NULL); /* show zoomed area */
       else
        setExclEventHandler((XtEventHandler) HandleBtnPress, ButtonPressMask, True); 
      break;
     case Button2:
      firstClick = True;
      BtnPress_1(w, event);
      ApData.pic_change = True;
      foregroundDraw=FALSE;
      break;
    }
  }
}

/* Draw the last picture element to the drawable */
DDRAW void LastDraw(Drawable drawable)
{
 int np, i, zmOfx = 0, zmOfy = 0, zmXF = 1, zmYF = 1;
 GC gc;
 Boolean alreadyDrawn;	/* for cut, paste, move operations */

 alreadyDrawn = False;
 gc = Undo.gc;
 if (drawable == ApData.scrPix) /* ABC double buffer problem */
  {
   zmOfx = zoomDraw.clipWin11.x;
   zmOfy = zoomDraw.clipWin11.y;
  }
 else if (drawable == drawAreaWin) /* because I don't know how to get clip winresize callback  */
  getClipWinDims(clipWin, scrollBX, scrollBY, ApData.DAwidth, ApData.DAheight, &ApData.clipWin);
 else if (drawable == ApData.pixy)
  {
   XGCValues values;
   unsigned long valuemask;

   valuemask = GCFunction | GCLineStyle | GCLineWidth |
     GCJoinStyle | GCCapStyle;

   XGetGCValues (xinfo.dpy, Undo.gc, valuemask, &values);
   gc = xinfo.gc1bit;
   XChangeGC(xinfo.dpy, xinfo.gc1bit, valuemask, &values);
  }
 np = Undo.npoints;
 if (np != 0)
   switch (Undo.tool)
     {
     case mix:
     case spray:
       sprayWrapped(drawable, gc, &Undo.points[0], np, zmOfx, zmOfy);
       break;
     case rectDraw:
     case squareDraw:
       np -= 2;
       DrawRectangleWrapped(drawable, gc, Undo.drawinfo, &Undo.points[np], zmOfx, zmOfy);
       break;
     case curveDraw:
     case lineDraw:
       DrawLinesWrapped(drawable, gc, False,
	 &Undo.points[0], Undo.npoints, zmOfx, zmOfy, False);
       break;
     case polyDraw:
       DrawLinesWrapped (drawable, gc, Undo.drawinfo.fill,
		       &Undo.points[0], Undo.npoints, zmOfx, zmOfy, False);
       break;
     case arcDraw:
       {
	 int zofx = zmOfx / zmXF, zofy = zmOfy / zmYF, lnNo;

	 lnNo = Undo.points[np - 1].x;
	 np -= lnNo + 1;
	 DrawLinesWrapped(drawable, gc, Undo.drawinfo.fill, &Undo.points[np],
	  lnNo, zofx, zofy, True);
       }
       break;
     case circDraw:
     case ellipseDraw:
       DrawArcWrapped(drawable, gc, Undo.drawinfo, &Undo.points[np - 3], zmOfx, zmOfy);
       break;
     case drawText:
       {
	char *txt;

	txt = malloc (Undo.points[2].x + 1);
	for (i = 0; i < Undo.points[2].x; i++)
	  txt[i] = Undo.points[3 + i].x;
	txt[Undo.points[2].x] = 0;
	XDrawImageString(xinfo.dpy, drawable, Undo.gc, Undo.points[0].x - zmOfx,
	 Undo.points[0].y + Undo.points[2].y - zmOfy, txt, strlen (txt));
	free(txt);
       }
       break;
     case selectA:
     case selectPolygonA:
     case contourT:
     case mapColorsA:
       if (drawable == ApData.pixy)
        {
	 alreadyDrawn = True;
	 drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
	 drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW), zoomDraw.factorX,
	  zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
	}
       break;
     case pasteA:
     case moveA:
     case resizeA:
     case mirrorHor:
     case mirrorVer:
     case mirrorFlip:
     case rotateA:
	pasteMotifWrapped (drawable);
	if (drawable == ApData.pixy)
	  alreadyDrawn = True;
       break;
     case pasteMulA:
       {
	Rectangle boundR, clipR, zoomWinR, areaR, clipMotifR;
	int oldX = Select.x, oldY = Select.y;
	XPoint startPt;

	zoomWinR.x = zmOfx;
	zoomWinR.y = zmOfy;
	zoomWinR.width = zoomDraw.clipWin11.width + 1;
	zoomWinR.height = zoomDraw.clipWin11.height + 1;
	boundR.height = Select.height;
	boundR.width = Select.width;
	areaR.x = Undo.points[0].x;
	areaR.y = Undo.points[0].y;
	areaR.width = Undo.points[1].x;
	areaR.height = Undo.points[1].y;

	XSetFunction (xinfo.dpy, Select.moveGC, GXcopy);

	for (i = 2; i < Undo.npoints; i++)
	 {
	  moveBoundaryPts(0, Select.npoints, Undo.points[i].x - Select.x,
	   Undo.points[i].y - Select.y);
	  if (!Select.rectangular)
	   {
	    XDestroyRegion(Select.reg);
	    XSetClipMask (xinfo.dpy, Select.moveGC, None);

	    Select.reg = XPolygonRegion(Select.points, Select.npoints, WindingRule);
	    XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
	   }
	  Select.x = boundR.x = Undo.points[i].x;
	  Select.y = boundR.y = Undo.points[i].y;
	  intersectRect (&boundR, &areaR, &clipMotifR);
	  startPt.x = clipMotifR.x - boundR.x;
	  startPt.y = clipMotifR.y - boundR.y;
	  startPt.x = MAX (0, startPt.x);
	  startPt.y = MAX (0, startPt.y);
	  if (drawable == ApData.scrPix) /* ABC double buffer problem */
	   {
	    if (!intersectRect(&boundR, &zoomWinR, &clipR))
	      continue;
	    if (!Select.rectangular)
	      XSetClipOrigin (xinfo.dpy, Select.moveGC, -zmOfx, -zmOfy);
	    putXImageTP(ApData.img, Select.img, drawable, Select.moveGC,
	     0, 0, Select.width, Select.height, Select.x, Select.y,
	     Select.x - zmOfx, Select.y - zmOfy, Select.reg);
	   }
	  else
	   repeatCopyArea(drawable, Select.img, Select.moveGC, startPt.x, startPt.y,
	    clipMotifR.width, clipMotifR.height, Undo.points[i].x + startPt.x,
	    Undo.points[i].y + startPt.y, ApData.specialPutImg, normal,
	    Select.img->width, Select.img->height, 0, 0, Select.reg);
	 }
	moveBoundaryPts(0, Select.npoints, oldX - Select.x, oldY - Select.y);
	Select.x = oldX;
	Select.y = oldY;
       }
       if (!Select.rectangular)
	XSetClipOrigin (xinfo.dpy, Select.moveGC, 0, 0);
       if (drawable == ApData.pixy)
	alreadyDrawn = True;
       break;
     default:
      if (drawable == ApData.pixy)
       alreadyDrawn = True;
      break;
     }
 if (drawable == drawAreaWin)
  drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
 else if (drawable == ApData.pixy)
  {
   if (!alreadyDrawn)
    draw2Image(ApData.img, ApData.pixy, Shm.alloc, opBounds.x, opBounds.y, opBounds.width,
     opBounds.height, ColorMatrix.FrgColorInd);
   /* because of the shared memory syncronisation problem call XSync */
   if (Shm.alloc)
     XSync(xinfo.dpy, False);
  }
}

/* move points pointed by pointer xp by dx/dy */
DDRAW void movePts(XPoint * xp, int pointNo, int dx, int dy)
{
 int i;

 if ((dx) || (dy))
  for (i = 0; i < pointNo; i++)
   { xp[i].x += dx; xp[i].y += dy; }
}

/* refresh repeat view */
DDRAW void refreshRepeatView(Drawable drawable, Rectangle *clipWinR, Boolean wholeWindow,
 Boolean final_draw)
{
  int i;
  Rectangle resR, opR;
  XPoint b[2], wrapA;

  if ((wholeWindow) && (clipWinR))
    exposeRepeat(drawable, clipWinR->x, clipWinR->y,
     clipWinR->width, clipWinR->height, clipWinR->x, clipWinR->y, repeats.drawGrid, True);
  else
    {
      b[0].x = opBounds.x;
      b[0].y = opBounds.y;
      b[1].x = opBounds.x + opBounds.width;
      b[1].y = opBounds.y + opBounds.height;
      opR.width = opBounds.width;
      opR.height = opBounds.height;

      wrapA.x = repeats.crop.width;
      wrapA.y = repeats.crop.height;
      for (i = 0; i < repeats.currNum; i++)
	{
	  flipPts (&b[0], 2, repeats.r[i].rType, &wrapA);
	  movePts (&b[0], 2, repeats.r[i].x, repeats.r[i].y);
	  opR.x = MIN (b[0].x, b[1].x);
	  opR.y = MIN (b[0].y, b[1].y);
	  if (intersectRect (clipWinR, &opR, &resR))
	    exposeRepeat (drawable, resR.x, resR.y, resR.width, resR.height,
			  resR.x, resR.y, repeats.drawGrid, False);
	  movePts (&b[0], 2, -repeats.r[i].x, -repeats.r[i].y);
	  flipPts (&b[0], 2, repeats.r[i].rType, &wrapA);
	}
      if (final_draw)
	finalDraw (drawable);
    }
}

DDRAW void LastDrawOnAllWin(Boolean definitely)
{
 if (definitely == True)
   LastDraw(ApData.pixy);

 refreshDrawingArea(False, True);
 if (widgetIsViewable(zoomDrawAreaW))
  {
   Rectangle *zmPixR, srcR, clipR, resR;

   zmPixR = &zoomDraw.clipWin11;
   srcR.width = opBounds.width;
   srcR.height = opBounds.height;
   srcR.x = opBounds.x;
   srcR.y = opBounds.y;
   if (ApData.viewMode==repeatView)
    {
     if (whichRepeats (zmPixR, &srcR, &resR))
       { /* ABC double buffer problem */
	exposeRepeat(ApData.scrPix, resR.x, resR.y, resR.width, resR.height,
         resR.x - zmPixR->x, resR.y - zmPixR->y, repeats.drawGrid, True);
	refreshZoomWin(resR.x, resR.y, resR.width, resR.height, False);
       }
     }
    else
     if (intersectRect (zmPixR, &srcR, &clipR))
      {
       Boolean refreshFromSource;

       LastDraw(ApData.scrPix);  /* ABC double buffer problem */
       switch (ApData.tool)
	{
	 case selectA:
	 case selectPolygonA:
	 case mirrorHor:
	 case mirrorVer:
	 case mirrorFlip:
	 case moveA:
	 case pasteA:
	 case rotateA:
	 case resizeA:
	 case pasteMulA:
	   refreshFromSource = False;
	   break;
	 default:
	   refreshFromSource = True;
	   break;
	}
       refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, refreshFromSource);
      }
  }
}

/* Undo Callback */
DDRAW void undoCB(Widget w, XtPointer clientData, XtPointer callData)
{
 switch (Undo.tool)
   {
   case grayConv:
   case makeGradation:
     {
       int i, j = 1;
       for (i = Undo.points[0].x; i < Undo.points[0].y; i++)
	 {
	  xinfo.red[i] = Undo.points[j].x;
	  xinfo.green[i] = Undo.points[j++].y;
	  xinfo.blue[i] = Undo.points[j++].x;
	  xinfo.allocCells[i].flags = DoRed | DoGreen | DoBlue;
	  xinfo.allocCells[i].red = (unsigned short) xinfo.red[i] << 8;
	  xinfo.allocCells[i].green = (unsigned short) xinfo.green[i] << 8;
	  xinfo.allocCells[i].blue = (unsigned short) xinfo.blue[i] << 8;
	 }
       if (xinfo.trueColor)
	 {
	   calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
			   &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
	   if (Select.imgAllocated)
	     calcTrueColorPixels(&Select.allocCells[0], &Select.pixelTC[0],
			&Select.red[0], &Select.green[0], &Select.blue[0]);
	   drawColorPalette();
	   refreshDrawingArea(True, True);
	   refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
	    zoomDraw.clipWin11.width + 1, zoomDraw.clipWin11.height + 1, True);
	 }
       else
	 XStoreColors (xinfo.dpy, xinfo.cmap, xinfo.allocCells, ApData.ColorNo);
       ColorSelectConseq (Button4, ColorMatrix.BckColorInd);
       ColorSelectConseq (Button1, ColorMatrix.FrgColorInd);
     }
     break;
   case selectA:
   case selectPolygonA:
     drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
     Select.selected = Select.rect_on = False;
     ChangeCursor(drawAreaWin, XC_left_ptr, True);
     if (XtIsRealized(zoomDrawAreaW))
       ChangeCursor(XtWindow (zoomDrawAreaW), XC_left_ptr, True);
     setToolEventHandler(selectA, 0);
     break;
   case cutA:
     Undo.tool = pasteA;
     LastDrawOnAllWin(False);
     setToolEventHandler(pasteA, 0);
     disableUndo();
     break;
   case cutClearA:
     Undo.tool = pasteA;
     LastDrawOnAllWin(True);
     toolBarCB(curveDW, "0", NULL);
     disableUndo();
     break;
   case resizeA:
     undoScaleSelectionArea();
     break;
   case moveA:
     break;
   case mirrorHor:
     mirrorPictureCB(UndoBW, "1", NULL);
     break;
   case mirrorVer:
     mirrorPictureCB(UndoBW, "0", NULL);
     break;
   case mirrorFlip:
     mirrorPictureCB(UndoBW, "2", NULL);
     break;
   case shiftVer:
   case shiftHor:
     ShiftPic (Undo.points[Undo.npoints - 1].x, Undo.points[Undo.npoints - 1].y);
     disableUndo();
     break;
   case pasteMulA:
     disableUndo();
     opBounds.x = opBounds.y = opBounds.width = opBounds.height = 0;
     refreshDrawingArea(True, True);
     refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
     break;
   default:
     if (Undo.imgAllocated)
       CopyAreaImg(Undo.img, ApData.img, NULL, 0, 0, 0, 0,
	Undo.area.width, Undo.area.height, Undo.area.x, Undo.area.y);
     opBounds.x = opBounds.y = opBounds.width = opBounds.height = 0;
     refreshDrawingArea (True, True);
     refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
     if ((ApData.tool != replaceCol) && (ApData.tool != contourT) &&
       (Undo.tool != pasteA) && (Undo.tool != pasteMulA) && (Undo.tool !=
	swapColor) && (Undo.tool != mapColorsA))
       Select.selected = Select.rect_on = False;
     disableUndo();
     break;
   }
 XtVaSetValues(UndoBW, XmNsensitive, False, NULL);
 XtVaSetValues(undoZoomBW, XmNsensitive, False, NULL);
}

/* get text input and display it as selection */
static void getAndDisplayText()
{
 XFontStruct *newFont;
 char *txt;
 int i;

 finalDraw(ApData.pixy);
 disableUndo();
 memset (&Undo.points[0], 0, sizeof (XPoint) * 4);
 ApData.tool = none;
 setToolEventHandler (ApData.tool, 0);
 newFont = FontSel(mainW, PREF_FIXED, ApData.fontList);
 if ((newFont != NULL) && (Undo.points[1].x) && (Undo.points[1].y))
  {
   Pixmap pixy;

   if (ApData.currFont)
     XFreeFont (xinfo.dpy, ApData.currFont);
   ApData.currFont = newFont;
   XSetFont (xinfo.dpy, Undo.gc, newFont->fid);
   XSetFont (xinfo.dpy, xinfo.gc1bit, newFont->fid);

   disableSelection (True);

   Select.x = Select.y = 0;
   Select.width = Undo.points[1].x;
   Select.height = Undo.points[1].y;

   pixy = safeCreatePixmap(xinfo.dpy, RootWindow(xinfo.dpy, xinfo.screenNo),
    Select.width, Select.height, 1);
   if (!checkPixmap (pixy))
     return;
   Select.colorCons = Select.selected = Select.rectangular = True;

   Select.aspectR = 1.0 * Select.width / Select.height;
   setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);

   XtVaSetValues(PasteBW, XmNsensitive, True, NULL);

   memcpy(&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));
   XCopyGC(xinfo.dpy, xinfo.gc, GCForeground | GCBackground, Undo.gc);

   txt = malloc (Undo.points[2].x + 1);
   for (i = 0; i < Undo.points[2].x; i++)
     txt[i] = Undo.points[3 + i].x;
   txt[Undo.points[2].x] = 0;
   XSetForeground(xinfo.dpy, xinfo.gc1bit, 0);
   XFillRectangle(xinfo.dpy, pixy, xinfo.gc1bit, Select.x, Select.y,
		   Select.width, Select.height);
   XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);
   XDrawImageString(xinfo.dpy, pixy, xinfo.gc1bit,
		  Undo.points[0].x, Undo.points[2].y, txt, strlen (txt));
   free (txt);
   Select.img = doXImage(8, Select.width, Select.height);
   if (Select.img)
    {
     unsigned char fillChar = (unsigned char)
     xinfo.allocCells[ColorMatrix.BckColorInd].pixel;

     memset(&Select.img->data[0], fillChar,
       Select.img->bytes_per_line * Select.img->height);
     draw2Image(Select.img, pixy, False, 0, 0, Select.width,
       Select.height, ColorMatrix.FrgColorInd);
    }
   XFreePixmap(xinfo.dpy, pixy);
   Undo.npoints = 0;
   if (!checkXImage(Select.img))
     return;
   Select.imgAllocated = True;
   pastemoveCB(NULL, NULL, NULL);
   setButtonIconCB(unselectBW, "unselect.xbm", NULL);
   ApData.pic_change = True;
   enableUndo(False);
  }
}

/* flip fill type from solid to outline and back */
static void flipFillType()
{
 ApData.drawinfo.fill = !ApData.drawinfo.fill;

 if (ApData.drawinfo.fill)
  {
   setButtonIconCB(DRectBW, "rectanglefilled.xbm", NULL);
   setButtonIconCB(DSquareBW, "squarefilled.xbm", NULL);
   setButtonIconCB(DCircBW, "circlefilled.xbm", NULL);
   setButtonIconCB(DEllipseBW, "ellipsefilled.xbm", NULL);
   setButtonIconCB(DArcBW, "arcfilled.xbm", NULL);
   setButtonIconCB(DpolyDrawBW, "polyDrawFilled.xbm", NULL);
  }
 else
  {
   setButtonIconCB(DRectBW, "rectanglempty.xbm", NULL);
   setButtonIconCB(DSquareBW, "squarempty.xbm", NULL);
   setButtonIconCB(DCircBW, "circlempty.xbm", NULL);
   setButtonIconCB(DEllipseBW, "ellipsempty.xbm", NULL);
   setButtonIconCB(DArcBW, "arcempty.xbm", NULL);
   setButtonIconCB(DpolyDrawBW, "polyDraw.xbm", NULL);
  }
}

#define T_CURVE		0
#define T_LINE		2
#define T_POLY		3
#define T_RECTANGLE	4
#define T_SQUARE	5
#define T_CIRCLE	6
#define T_ELLIPSE	7
#define T_ARC		8
#define T_SPRAY		9
#define T_TEXT		11
#define T_FILL		12
#define T_MIX		13
#define T_PARAMETERS	20

/* Generic callback for buttons in toolbar */
DDRAW void toolBarCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int reason;

 if (ApData.viewMode == densityView)
  return;
 reason = atoi((char *) clientData);
 switch (reason)
  {
   case T_PARAMETERS: displayParameterWindow(); return; break;
   case T_CURVE: ApData.tool = curveDraw; break;
   case T_LINE: ApData.tool = lineDraw; break;
   case T_POLY:
    if (ApData.tool == polyDraw)
     flipFillType();
    else
     ApData.tool = polyDraw;
    break;
   case T_RECTANGLE:
    if (ApData.tool == rectDraw)
     flipFillType();
    else
     ApData.tool = rectDraw;
    break;
   case T_SQUARE:
    if (ApData.tool == squareDraw)
     flipFillType();
    else
     ApData.tool = squareDraw;
    break;
   case T_CIRCLE:
    if (ApData.tool == circDraw)
     flipFillType();
    else
     ApData.tool = circDraw;
    break;
   case T_ELLIPSE:
    if (ApData.tool == ellipseDraw)
     flipFillType();
    else
     ApData.tool = ellipseDraw;
    break;
   case T_ARC:
    if (ApData.tool == arcDraw)
     flipFillType();
    else
     ApData.tool = arcDraw;
    n_arcPar = 0;
    break;
   case T_SPRAY:
    sprayData.mixMode=FALSE;
    if (ApData.tool == spray)
     {
      initDrawPrmDlg(0, getTxtI8N(NUMBER_OF_POINTS_MSG), sprayData.sprayPts, 10, 500);
      initDrawPrmDlg(1, getTxtI8N(RADIUS_MSG), sprayData.radius, 2, 500);
      XtManageChild(drawPrmW);
     }
    else
     ApData.tool = spray;
    break;
   case T_MIX:
    sprayData.mixMode=TRUE;
    if (ApData.tool == mix)
     {
      initDrawPrmDlg(0, getTxtI8N(NUMBER_OF_POINTS_MSG), sprayData.sprayPts, 10, 500);
      initDrawPrmDlg(1, getTxtI8N(RADIUS_MSG), sprayData.radius, 2, 500);
      XtManageChild(drawPrmW);
     }
    else
     ApData.tool = mix;
    break;
   case T_TEXT: getAndDisplayText(); break;
   case T_FILL:
    ApData.tool = fill;
    if ((Select.imgAllocated) && (ApData.drawinfo.fillType == solidFill))
     {
      setButtonIconCB(fillBW, "fillPattern.xbm", NULL);
      ApData.drawinfo.fillType = patternFill;
     }
    else
     {
      setButtonIconCB(fillBW, "fill.xbm", NULL);
      ApData.drawinfo.fillType = solidFill;
     }
   break;
  }

 setToolEventHandler(ApData.tool, 0);
 if (reason!=T_TEXT)
  highlightButton(w);
}


DDRAW double densityRatio()
{ /* a little more complicated to prevent division by zero */
 return(max(0.001, (1.0*ApData.imageXdpi)/max(ApData.imageYdpi, 1)));
}

/* return size in mm or inches depending on selected measurement system */
DDRAW double getImageXmm()
{
 return((double) ((ApData.width*INCH_MM)/(ApData.imageXdpi*(metricSystem ? 1.0 : 25.4))));
}

/* return size in mm or inches depending on selected measurement system */
DDRAW double getImageYmm()
{
 return((double) ((ApData.height*INCH_MM)/(ApData.imageYdpi*(metricSystem ? 1.0 : 25.4))));
}

#define ASPECT_MM_X	0
#define ASPECT_MM_Y	1
#define ASPECT_THREAD_X	2
#define ASPECT_THREAD_Y	3
#define ASPECT_DPI_X	4
#define ASPECT_DPI_Y	5
#define ASPECT_GOT_MM	6
#define ASPECT_GOT_THREAD 7
#define ASPECT_GOT_DPI	8
#define ASPECT_MAP	9

/* set aspect ratio callback - for drawing squares and circles properly */
DDRAW void aspectRatioCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static Widget sizeXmmW, sizeYmmW, threadsXW, threadsYW, sizeXdpiW, sizeYdpiW;
 double minVal = 0.1, maxVal = 10000.0;

 switch (atoi((char *) clientData))
  {
   case ASPECT_MM_X: sizeXmmW = w; break;
   case ASPECT_MM_Y: sizeYmmW = w; break;
   case ASPECT_THREAD_X: threadsXW = w; break;
   case ASPECT_THREAD_Y: threadsYW = w; break;
   case ASPECT_DPI_X: sizeXdpiW = w; break;
   case ASPECT_DPI_Y: sizeYdpiW = w; break;
   case ASPECT_GOT_MM:
    ApData.imageXdpi = ApData.width/getDoubleFromWidget(sizeXmmW, minVal, maxVal, 3);
    if (metricSystem)
     ApData.imageXdpi *= INCH_MM;
    setDoubleToWidget(sizeXdpiW, ApData.imageXdpi, minVal, maxVal, 3);
    setDoubleToWidget(threadsXW, ApData.imageXdpi*10/INCH_MM, minVal, maxVal, 3);
    ApData.imageYdpi = ApData.height/getDoubleFromWidget(sizeYmmW, minVal, maxVal, 3);
    if (metricSystem)
     ApData.imageYdpi *= INCH_MM;
    setDoubleToWidget(sizeYdpiW, ApData.imageYdpi, minVal, maxVal, 3);
    setDoubleToWidget(threadsYW, ApData.imageYdpi*10/INCH_MM, minVal, maxVal, 3);
    break;
   case ASPECT_GOT_THREAD:
    ApData.imageXdpi = INCH_MM*getDoubleFromWidget(threadsXW, minVal, maxVal, 3)/10;
    setDoubleToWidget(sizeXdpiW, ApData.imageXdpi, minVal, maxVal, 3);
    setDoubleToWidget(sizeXmmW, getImageXmm(), minVal, maxVal, 3);
    ApData.imageYdpi = INCH_MM*getDoubleFromWidget(threadsYW, minVal, maxVal, 3)/10;
    setDoubleToWidget(sizeYdpiW, ApData.imageYdpi, minVal, maxVal, 3);
    setDoubleToWidget(sizeYmmW, getImageYmm(), minVal, maxVal, 3);
    break;
   case ASPECT_GOT_DPI:
    ApData.imageXdpi = getDoubleFromWidget(sizeXdpiW, minVal, maxVal, 3);
    setDoubleToWidget(sizeXmmW, getImageXmm(), minVal, maxVal, 3);
    setDoubleToWidget(threadsXW, ApData.imageXdpi*10/INCH_MM, minVal, maxVal, 3);
    ApData.imageYdpi = getDoubleFromWidget(sizeYdpiW, minVal, maxVal, 3);
    setDoubleToWidget(sizeYmmW, getImageYmm(), minVal, maxVal, 3);
    setDoubleToWidget(threadsYW, ApData.imageYdpi*10/INCH_MM, minVal, maxVal, 3);
    break;
  case ASPECT_MAP:
    setDoubleToWidget(sizeXdpiW, ApData.imageXdpi, minVal, maxVal, 3);
    setDoubleToWidget(sizeXmmW, getImageXmm(), minVal, maxVal, 3);
    setDoubleToWidget(threadsXW, ApData.imageXdpi*10/INCH_MM, minVal, maxVal, 3);
    setDoubleToWidget(sizeYdpiW, ApData.imageYdpi, minVal, maxVal, 3);
    setDoubleToWidget(sizeYmmW, getImageYmm(), minVal, maxVal, 3);
    setDoubleToWidget(threadsYW, ApData.imageYdpi*10/INCH_MM, minVal, maxVal, 3);
    break;
  }
}

/* set/remove appropriate event handler and cursor shape according to selected tool */
DDRAW void setToolEventHandler(Sel_tool toolUsed, unsigned int curshape)
{
 unsigned int shape = curshape;

 if (ApData.CurrEventHandler != NULL)
   {
    XtRemoveEventHandler(drawAreaWID, ApData.eventMask, FALSE,
     (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 0);
    if (XtIsRealized (zoomDrawAreaW))
      XtRemoveEventHandler(zoomDrawAreaW, ApData.eventMask, FALSE,
       (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 1);
   }
 switch (toolUsed)
  {
   case none:
   case contourT:
     ApData.CurrEventHandler = NULL;
     ApData.eventMask = 0;
     break;
   case selectA:
   case selectPolygonA:
     ApData.CurrEventHandler = (XtEventHandler) SelectBtnPress;
     ApData.eventMask = ButtonPressMask;
     break;
   case mirrorHor:
   case mirrorVer:
   case mirrorFlip:
   case pasteA:
     ApData.CurrEventHandler = (XtEventHandler) handleSelection;
     ApData.eventMask = PointerMotionMask | ButtonPressMask;
     break;
   case polyDraw:
     ApData.CurrEventHandler = (XtEventHandler) polygonDraw;
     ApData.eventMask = PointerMotionMask | ButtonPressMask | ButtonReleaseMask;
     break;
   case drawText:
     ApData.CurrEventHandler = (XtEventHandler) moveString;
     ApData.eventMask = PointerMotionMask | ButtonPressMask | ButtonReleaseMask;
     break;
   default:
     ApData.CurrEventHandler = (XtEventHandler) HandleBtnPress;
     ApData.eventMask = ButtonPressMask;
     break;
  }
 if (ApData.CurrEventHandler != NULL)
  {
   XtAddEventHandler(drawAreaWID, ApData.eventMask, FALSE,
      (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 0);
   if (XtIsRealized (zoomDrawAreaW))
     XtAddEventHandler(zoomDrawAreaW, ApData.eventMask, FALSE,
      (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 1);
  }
 /* set cursor shape */
 if (curshape == 0)
   switch (toolUsed)
    {
     case curveDraw:
     case lineDraw:
     case rotateByLine:
     case polyDraw:
     case rectDraw:
     case squareDraw:
     case circDraw:
     case ellipseDraw:
       shape = XC_crosshair;
       break;
     case shiftHor:
       shape = XC_sb_h_double_arrow;
       break;
     case shiftVer:
       shape = XC_sb_v_double_arrow;
       break;
     case spray:
     case mix:
       shape = XC_spraycan;
       break;
     case fill:
       shape = XC_draft_small;
       break;
     default:
       shape = XC_left_ptr;
       break;
    }
 ChangeCursor(drawAreaWin, shape, True);
 if (XtIsRealized (zoomDrawAreaW))
  ChangeCursor(XtWindow (zoomDrawAreaW), shape, True);
}

/* set exclusive event handler on main and zoom window */
DDRAW void setExclEventHandler(XtEventHandler NewEH, EventMask eventMask, Boolean onZoomWin)
{
 if (ApData.CurrEventHandler)
  {
   XtRemoveEventHandler(drawAreaWID, ApData.eventMask, FALSE,
    (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 0);
   if (XtIsRealized (zoomDrawAreaW))
    XtRemoveEventHandler(zoomDrawAreaW, ApData.eventMask, FALSE,
     (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 1);
  }
 ApData.CurrEventHandler = NewEH;
 ApData.eventMask = eventMask;
 if (NewEH)
  {
   XtAddEventHandler (drawAreaWID, ApData.eventMask, FALSE,
    (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 0);
   if ((XtIsRealized(zoomDrawAreaW)) && (onZoomWin))
    XtAddEventHandler (zoomDrawAreaW, ApData.eventMask, FALSE,
     (XtEventHandler) ApData.CurrEventHandler, (XtPointer) 1);
  }
}
