#define FILES
#include <Xm/Xm.h>
#include <X11/Xmu/Drawing.h>
#include <Xm/FileSB.h>
#include <Xm/Scale.h>
#include <Xm/List.h>
#include <Xm/ToggleB.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include "paint.h"

static Boolean ClearAfterSave = False; /* clear picture after saving? */
static Widget MainWindowW; /* for setting and changing the title */
static int zoomPerc = 100; /* zoom factor for printing */

static XImage *tmpXImage = NULL;
static unsigned char *tmpbuff = NULL;
static int tmpbpl, lType;
static unsigned char lcmap[3*256];
static XImage *tmpPic = NULL, *truePic = NULL;
static char dirData[FILENAME_MAX]="";
static Boolean motifS = False;
static int sizeXmm, sizeYmm;
static Boolean reversedLines;
static unsigned long colHist[MAX_CELLS];
static int fileSelectionBoxLoad=FALSE;
static int whenSavedCopyToFloppy;

static void saveHistoryList();
static void addItemToHistoryList(char *fileName);
static int loadMotif();

#define NFFORCE 1
#define NFFINISH 2
#define NFSTART 3
DFILES void NewFileCB(Widget w, XtPointer clientData, XtPointer callData)
{
 switch (atoi((char *) clientData))
  {
   case NFSTART:
     if (ApData.pic_change)
      {
       finalDraw(ApData.pixy);
       ClearAfterSave = True;
       manageWarningDialog(getTxtI8N (PICTURE_CHANGED_SAVE_MSG),
	saveDrawingCB, "2", getTxtI8N (YES_MSG), NewFileCB, "1", getTxtI8N (NO_MSG),
	NULL, NULL, getTxtI8N (HELP_MSG));
      }
     else
      NewFileCB(w, "1", NULL);
     break;
   case NFFORCE:
     InitChSizeDlgCB(NULL, "5", NULL);
     break;
   case NFFINISH:
     {
      XPoint *newSize = (XPoint *) callData;

      repeats.width = repeats.crop.width = ApData.DAwidth = ApData.width = newSize->x;
      repeats.height = repeats.crop.height = ApData.DAheight = ApData.height = newSize->y;
      repeats.crop.x = repeats.crop.y = 0;
     }

     ClearAfterSave = False;
     freeMainImage (&ApData.img);
     ApData.img = createMainImage(ApData.width, ApData.height);
     ApData.fname[0] = 0;
     ApData.clipWin.x = ApData.clipWin.y = 0;
     ApData.tool = none;

     ApData.pic_change = False;
     XtVaSetValues(drawAreaWID, XmNwidth, ApData.width, XmNheight, ApData.height, NULL);

     disableUndo();
     Select.rect_on = False;

     refreshDrawingArea (True, True);
     XtVaSetValues (CopyBW, XmNsensitive, False, NULL);
     XtVaSetValues (CutBW, XmNsensitive, False, NULL);
     if (Select.selected)
       {
        if (Select.imgAllocated)
	 XtVaSetValues(PasteBW, XmNsensitive, True, NULL);
	else
	 disableSelection(True);
       }
     changeTitleCB (w, "", NULL);
     toolBarCB(curveDW, "0", NULL);
     zoomDraw.newImage = True;

     ApData.aspectRatio = (float) (1.0*ApData.width)/ApData.height;
     repeats.currNum = repeats.num = 1;
     XmUpdateDisplay (drawAreaWID);
     zoomMapWindow();
     break;
  }
}

static void cleanUpAndExit()
{
 if (ApData.img)
  freeMainImage(&ApData.img);
 clearSHM();
 allocateBuffers(FALSE);
 ApData.pic_change = False;
 exit(0);
}

#define EX_NORMAL 0
#define EX_FORCE 1
#define EX_CANCEL 2
/* Final exit procedure */
DFILES void reallyExitCB(Widget w, XtPointer clientData, XtPointer callData)
{
 switch (atoi((char *) clientData))
  {
   case EX_NORMAL:
    if (ApData.pic_change)
     {
      finalDraw (ApData.pixy);
      disableUndo();
      XMapRaised (xinfo.dpy, XtWindow (XtParent (mainW)));
      manageWarningDialog (getTxtI8N (PICTURE_CHANGED_SAVE_MSG),
      saveDrawingCB, "2", getTxtI8N (YES_MSG),
      reallyExitCB, "1", getTxtI8N (NO_MSG),
      reallyExitCB, "2", getTxtI8N (CANCEL_MSG));
     }
    else
     cleanUpAndExit();
    break;
   case EX_FORCE:
    cleanUpAndExit();
    break;
   case EX_CANCEL:
    XtUnmanageChild(warnDialogW);
    break;
  }
}

/* load bitmap picture and transform it to pixmap */
DFILES Pixmap loadBitmapFile(char *filename, Display *dpy, Window win,
 unsigned int *width, unsigned int *height, Pixel fore, Pixel back)
{
 int result, dummy;
 Pixmap bitmapPic, newPic;

 result = XReadBitmapFile(dpy, win, filename, width, height, &bitmapPic, &dummy, &dummy);
 if (result == BitmapSuccess)
  {
   newPic = XmuCreatePixmapFromBitmap(xinfo.dpy, win, bitmapPic,
     *width, *height, xinfo.depth, fore, back);
   XFreePixmap (xinfo.dpy, bitmapPic);
   return (newPic);
  }
 else
  return(0);
}

static void restoreInitialPixmap()
{
 ApData.width = initWidth;
 ApData.height = initHeight;
 ApData.img = createMainImage (ApData.width, ApData.height);
 if (ApData.img == NULL)
  reallyExitCB (NULL, "1", NULL);
}

static int initImgLoad(int loadType, int xpixels, int ypixels,
 int xdpi, int ydpi, char reverse, int numInterlace, unsigned char *cmap,
 unsigned char **data, int *bpl, int *bandh, int *progress)
{
 lType = loadType;
 *bandh = (numInterlace==1) ? ypixels : MIN(ypixels, (((lType==LOAD_BW) ? 8 : 1)*200000)/xpixels);

 tmpXImage = doXImage (8, xpixels, ypixels);
 if (tmpXImage == NULL)
   return (LOAD_NO_MEMORY);

 if ((tmpXImage)&&(loadType==LOAD_TRUE_COL))
  {/* so X window will not do a 32 bpp image! */
   truePic = doXImage(8, xpixels*3, ypixels);
   if (truePic == NULL)
    return(LOAD_NO_MEMORY);
  }
 reversedLines = reverse;
 initHist(colHist);
 switch (lType)
  {
   case LOAD_BW:
    tmpbpl = (xpixels + 7) / 8;
    tmpbuff = malloc (tmpbpl * (*bandh));
    if (tmpbuff == NULL)
     {
      undoXImage(&tmpXImage);
      return(LOAD_NO_MEMORY);
     }
    *data = tmpbuff;
    *bpl = tmpbpl;
    *progress = False;
    break;
   case LOAD_CMAP:
    memcpy(lcmap, cmap, 256*3);
    tmpbpl = xpixels;
    *data = (unsigned char *) tmpXImage->data;
    *bpl = tmpXImage->bytes_per_line;
    *progress = True;
    break;
   case LOAD_TRUE_COL:
    *data = (unsigned char *) truePic->data;
    *bpl = truePic->bytes_per_line;
    *progress = True;
    break;
  }
 return (LOAD_OK);
}

static int bandImgLoad(unsigned char *data, int lines, int bpl, int ypos)
{
 int x, y;
 unsigned char *data2;

 switch (lType)
  {
   case LOAD_BW:
    {
     unsigned char black, white;

     black = (unsigned char) xinfo.allocCells[1].pixel;
     white = (unsigned char) xinfo.allocCells[0].pixel;
     for (y = 0; y < lines; y++)
      {
       if (reversedLines)
	 data2 = (unsigned char *) &tmpXImage->data[(tmpXImage->height-ypos-lines+y)*tmpXImage->bytes_per_line];
       else
	 data2 = (unsigned char *) &tmpXImage->data[(ypos + y) * tmpXImage->bytes_per_line];
       for (x = 0; x < tmpXImage->width; x++)
	 *data2++ = (data[bpl * y + (x >> 3)] & (128 >> (x & 7))) ? black : white;
      }
    }
    break;
   case LOAD_CMAP:
    addToHist (&colHist[0], data, 0, 0, tmpXImage->width, lines, bpl);
    if (reversedLines)
      mirrorImgVertical (tmpXImage, tmpXImage->height);
    break;
   case LOAD_TRUE_COL:
     /*nothing to be done - need the whole picture for color quantization */
    break;
  }
 return (LOAD_OK);
}

static int endImgLoad(int status)
{
 if (lType == LOAD_BW)
  free (tmpbuff);
 if (lType == LOAD_TRUE_COL)
  {
   int requestColors = 200;

   status = colorQuantize(truePic->data, truePic->bytes_per_line, tmpXImage->data, tmpXImage->bytes_per_line,
    tmpXImage->width, tmpXImage->height, lcmap, 0, &requestColors);
   undoXImage(&truePic);
   addToHist(colHist, tmpXImage->data, 0, 0, tmpXImage->width, tmpXImage->height, tmpXImage->bytes_per_line);
   /* we are done with true color image now, so we treat it as colormap one */
   lType = LOAD_CMAP;  
  }

 if (status != LOAD_OK)
  {
   if (truePic)
    undoXImage(&truePic);
   undoXImage(&tmpXImage);
  }
 else
  loadMotif();

 if (tmpXImage != NULL)
   undoXImage (&tmpXImage);
 tmpXImage = NULL;
 return(status);
}

static int initLoad(int loadType, int xpixels, int ypixels, int xdpi, int ydpi, char reverse,
 int numInterlace, unsigned char *cmap, unsigned char **data, int *bpl, int *bandh, int *progress)
{
 int status = LOAD_OK;

 if (ApData.viewMode != normalView)
   XmToggleButtonSetState(normalViewTBW, True, True);
 finalDraw(ApData.pixy);
 disableUndo();
 toolBarCB(curveDW, "0", NULL);

 freeMainImage(&ApData.img); /* free it now, so we will have more memory */
 tmpPic = createMainImage(xpixels, ypixels);
 if (tmpPic == NULL)
   status = LOAD_NO_MEMORY;
 if ((status==LOAD_OK)&&(loadType==LOAD_TRUE_COL))
  {/* so X window will not do a 32 bpp image! */
   truePic = doXImage(8, xpixels*3, ypixels);
   if (truePic == NULL)
    status = LOAD_NO_MEMORY;
  }
 *bandh = (numInterlace==1) ? ypixels : MIN(ypixels, 200000/xpixels);
 lType = loadType;
 reversedLines = reverse;
 initHist(colHist);

 if ((xdpi!=0) && (ydpi!=0))
  {
   ApData.imageXdpi = xdpi/1000.0;
   ApData.imageYdpi = ydpi/1000.0;
  }
 aspectRatioCB(NULL, "9", NULL); /* update interface values */

 if (status==LOAD_OK)
  switch (lType)
   {
    case LOAD_BW:
     tmpbpl = (xpixels + 7) / 8;
     tmpbuff = malloc (tmpbpl * (*bandh));
     if (tmpbuff == NULL)
       status = LOAD_NO_MEMORY;
     *data = tmpbuff;
     *bpl = tmpbpl;
     *progress = False;
     break;
    case LOAD_CMAP:
     memcpy(lcmap, cmap, 256 * 3);
     tmpbpl = xpixels;
     *data = (unsigned char *) tmpPic->data;
     *bpl = tmpPic->bytes_per_line;
     *progress = True;
     break;
    case LOAD_TRUE_COL:
     *data = (unsigned char *) truePic->data;
     *bpl = truePic->bytes_per_line;
     *progress = True;
     break;
   }

 ApData.height = ypixels;
 ApData.width = xpixels;
 if (status == LOAD_NO_MEMORY)
  {
   if (tmpPic)
     freeMainImage(&tmpPic);
   if (truePic)
    undoXImage(&truePic);
   if (tmpbuff)
     free(tmpbuff);
   restoreInitialPixmap();
   return(LOAD_NO_MEMORY);
  }
 else
  return(LOAD_OK);
}

static int bandLoad(unsigned char *data, int lines, int bpl, int ypos)
{
 int x, y;
 unsigned char *data2;

 switch (lType)
  {
   case LOAD_BW:
    {
     unsigned char black, white;

     black = (unsigned char) xinfo.allocCells[1].pixel;
     white = (unsigned char) xinfo.allocCells[0].pixel;
     for (y = 0; y < lines; y++)
      {
       if (reversedLines)
	 data2 = (unsigned char *) &tmpPic->data[
	   (tmpPic->height - ypos - lines + y) * tmpPic->bytes_per_line];
       else
	 data2 = (unsigned char *) &tmpPic->data[(ypos + y) * tmpPic->bytes_per_line];
       for (x = 0; x < tmpPic->width; x++)
	 *data2++ = (data[bpl * y + (x >> 3)] & (128 >> (x & 7))) ? black : white;
      }
    }
   break;
  case LOAD_CMAP:
   addToHist(colHist, data, 0, 0, tmpPic->width, lines, bpl);
   if (reversedLines)
     mirrorImgVertical (tmpPic, tmpPic->height);
   break;
  case LOAD_TRUE_COL: /* nothing to be done - need to whole picture for color quantization */
   break;
  }
 return(LOAD_OK);
}

static int endLoad(int status)
{
 if (lType == LOAD_BW)
  free(tmpbuff);
 if (status != LOAD_OK)
  {
   if (truePic)
    undoXImage(&truePic);
   freeMainImage(&tmpPic);
   restoreInitialPixmap();
  }
 else
  {
   int picColors;
   int requestColors = 200;
   unsigned char trTbl[256];

   if (lType == LOAD_TRUE_COL)
    {
     status = colorQuantize(truePic->data, truePic->bytes_per_line, tmpPic->data, tmpPic->bytes_per_line,
      tmpPic->width, tmpPic->height, lcmap, 0, &requestColors);
     undoXImage(&truePic);
     addToHist(colHist, tmpPic->data, 0, 0, tmpPic->width, tmpPic->height, tmpPic->bytes_per_line);
     lType = LOAD_CMAP; /* we are done with true color image now, so we treat it as colormap one */
    }
   picColors = countHistColors(colHist);
   if (picColors > maxColors)
     maxColors += allocateMoreColors (picColors - maxColors, maxColors);
   else
     freeColors (maxColors - picColors);
   if (lType == LOAD_BW)
    {
     xinfo.red[0] = xinfo.green[0] = xinfo.blue[0] = 255;
     xinfo.red[1] = xinfo.green[1] = xinfo.blue[1] =
       xinfo.allocCells[1].red = xinfo.allocCells[1].green = xinfo.allocCells[1].blue = 0;
     xinfo.allocCells[0].red = xinfo.allocCells[0].green = xinfo.allocCells[0].blue = 255 << 8;
     xinfo.allocCells[0].flags = xinfo.allocCells[1].flags = DoRed | DoGreen | DoBlue;
     ApData.ColorNo = 2;
    }
   else
    if (lType == LOAD_CMAP)
     quantize(tmpPic, maxColors, &ApData.ColorNo, lcmap, xinfo.allocCells,
      &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0], trTbl, &colHist[0]);
   if (xinfo.trueColor)
    calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
     &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
   else
    XStoreColors(xinfo.dpy, xinfo.cmap, &xinfo.allocCells[0], ApData.ColorNo);
   ApData.img = tmpPic;
  }
 tmpPic = NULL;
 XtVaSetValues(drawAreaWID, XmNheight, ApData.height, XmNwidth, ApData.width, NULL);
 ApData.TranspColNo = ApData.ProtColNo = 0;
 ApData.specialPutImg = False;
 if (Select.imgAllocated)
  {
   unsigned char trTbl[MAX_CELLS], cmap[256*3];
   int i, p;

   memset (cmap, 0, 256 * 3);
   for (i = 0; i < MAX_CELLS; i++)
    {
     p = Select.allocCells[i].pixel;
     cmap[p * 3 + RED] = Select.red[i];
     cmap[p * 3 + GREEN] = Select.green[i];
     cmap[p * 3 + BLUE] = Select.blue[i];
    }

   initHist (&colHist[0]);
   addToHist (&colHist[0], (unsigned char *) Select.img->data, 0, 0,
      Select.img->width, Select.img->height, Select.img->bytes_per_line);
   Select.colorNo = countHistColors (&colHist[0]);
   maxColors += allocateMoreColors(Select.colorNo, maxColors);
   memcpy(&Select.allocCells[0], &xinfo.allocCells[0], MAX_CELLS * sizeof (XColor));
   quantize(Select.img, maxColors, (int *) &Select.colorNo, cmap, &Select.allocCells[0],
	   &Select.red[0], &Select.green[0], &Select.blue[0], &trTbl[0], &colHist[0]);
   quantizeOnCmap(Select.img, Select.colorNo, &Select.red[0], &Select.green[0], &Select.blue[0], &trTbl[0]);
   memcpy(&Select.allocCells[0], &xinfo.allocCells[0], sizeof(XColor)*MAX_CELLS);
   Select.colorCons = True;
   Select.extraColors = True;
   Select.rect_on = False;
   if (xinfo.trueColor)
    {
     calcTrueColorPixels(&Select.allocCells[0], &Select.pixelTC[0],
	     &Select.red[0], &Select.green[0], &Select.blue[0]);
     calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
	     &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
    }
  }
 else
  disableSelection(True);
 ChangeCursor (drawAreaWin, XC_left_ptr, True);
 if (XtIsRealized (zoomDrawAreaW))
   ChangeCursor (XtWindow (zoomDrawAreaW), XC_left_ptr, True);
 zoomDraw.newImage = True;
 repeats.width = repeats.crop.width = ApData.DAwidth = ApData.width;
 repeats.height = repeats.crop.height = ApData.DAheight = ApData.height;
 repeats.crop.x = repeats.crop.y = 0;

 if (status == LOAD_OK)
  {
   initColorSelectionArea(ApData.ColorNo, 0, 0);
   ColorSelectConseq(Button4, 0);
   ColorSelectConseq(Button1, 1);
   if (EQUAL(dirName(dirData), "/tmp/floppy"))
    {
     strcpy(ApData.fname, ApData.fname[0] ? dirName(ApData.fname) : ApData.imgPath);
     strcat(ApData.fname, "/");
     strcat(ApData.fname, baseName(dirData));
    }
   else
    strcpy(ApData.fname, dirData);
   changeTitleCB(NULL, ApData.fname, NULL);
   ApData.aspectRatio = (float) (1.0*ApData.width)/ApData.height;
   ApData.pic_change = False;
   repeats.num = repeats.currNum = 1;
  }
 else
  {
   ApData.fname[0] = 0;
   changeTitleCB (NULL, "", NULL);
  }
 initHistoryList(FALSE);
 addItemToHistoryList(dirData);
 saveHistoryList();
 refreshDrawingArea(True, True);
 zoomMapWindow();
 return(status);
}

/* load motif function */
static int loadMotif()
{
 unsigned char trTbl[256];

 if (lType == LOAD_BW)
  {
   int black = 3*xinfo.allocCells[1].pixel;
   int white = 3*xinfo.allocCells[0].pixel;

   Select.colorNo = 2;
   memset(colHist, 0, sizeof(unsigned long)*MAX_CELLS);
   colHist[black/3]=colHist[white/3]=1; // keep only the first two colors
   lcmap[white]=lcmap[white+1]=lcmap[white+2]=255;
   lcmap[black]=lcmap[black+1]=lcmap[black+2]=0;
  }
 else if (lType == LOAD_CMAP)
  Select.colorNo = countHistColors(&colHist[0]);

 maxColors += allocateMoreColors(Select.colorNo, maxColors);
 memcpy(Select.allocCells, xinfo.allocCells, MAX_CELLS * sizeof (XColor));
 quantize(tmpXImage, maxColors, (int *) &Select.colorNo, lcmap, &Select.allocCells[0],
  &Select.red[0], &Select.green[0], &Select.blue[0], &trTbl[0], &colHist[0]);
 quantizeOnCmap(tmpXImage, Select.colorNo, &Select.red[0], &Select.green[0],
  &Select.blue[0], &trTbl[0]);
 memcpy(&Select.allocCells[0], &xinfo.allocCells[0], sizeof (XColor) * MAX_CELLS);
 if (xinfo.trueColor)
  {
   calcTrueColorPixels(&Select.allocCells[0], &Select.pixelTC[0],
    &Select.red[0], &Select.green[0], &Select.blue[0]);
   calcTrueColorPixels(&xinfo.allocCells[0], &xinfo.pixelTC[0],
    &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
  }
 strcpy(Select.fname, dirData);

 Select.selected = Select.imgAllocated = True;
 setButtonIconCB(unselectBW, "unselect.xbm", NULL);
 Select.img = tmpXImage;
 Select.width = tmpXImage->width;
 Select.height = tmpXImage->height;
 Select.aspectR = 1.0 * Select.width / Select.height;
 Select.colorCons = True;
 Select.extraColors = True;
 Select.npoints = 0;

 tmpXImage = NULL;
 setBoundaryPtsR(&Select.npoints, Select.x, Select.y,
  Select.width, Select.height);
 Select.rectangular = True;
 XtVaSetValues(PasteBW, XmNsensitive, True, NULL);
 XtVaSetValues(CopyBW, XmNsensitive, False, NULL);
 XtVaSetValues(CutBW, XmNsensitive, False, NULL);
 initColorSelectionArea(maxColors, 0, 0);

 disableUndo(); 
 ApData.tool = Undo.tool = pasteA;
 refreshDrawingArea(True, False);
 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
  zoomDraw.clipWin11.width + 1, zoomDraw.clipWin11.height + 1, True);
 pastemoveCB(NULL, NULL, NULL);

 highlightButton(NULL);
 ChangeCursor(XtWindow(mainW), XC_left_ptr, True);
 setToolEventHandler(ApData.tool, 0);
 return(LOAD_OK);
}

#define LOADFSPICTURE 0
#define LOADFSMOTIF 1
#define LOADFSGO 2

DFILES void loadPictureCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static Boolean motifL = False;
 Boolean fileOK;
 char *tmp;

 switch (atoi((char *) clientData))
  {
   case LOADFSMOTIF:
    motifL = True;	/* continue */
   case LOADFSPICTURE:
    fileOK = False;
    tmp = xmStringToCstring(((XmFileSelectionBoxCallbackStruct *) callData)->value);
    strcpy(dirData, tmp);
    XtFree(tmp);
    if (fileExists(dirData))
     if (strlen(strrchr(dirData, '/')) > 1)
      {
       fileOK = True;
       loadPictureCB(NULL, "2", NULL);
       return;
      }
    if (!fileOK)
     {
      XBell(xinfo.dpy, 0);
      XUndefineCursor(xinfo.dpy, XtWindowOfObject(fileSelectionW));
     }
    break;
   case LOADFSGO: /* FileSelectionBox now stays mapped! */   
    if (motifL)
     {
      disableSelection(FALSE);
      loadImage(dirData, initImgLoad, bandImgLoad, endImgLoad, LOAD_CMAP);
     }
    else
     loadImage(dirData, initLoad, bandLoad, endLoad, LOAD_CMAP);
    motifL = False;
    break;
  }
}

/* return -1 if process can't be created, 0 on failure, 1 on success */
DFILES int executeCommand(char *command)
{
 int status;

 status = system(command);
 if(status<0)
  return(-1);
 if (WIFEXITED(status))
  if (WEXITSTATUS(status)==0)
   return(TRUE);
  else
   return(FALSE);
 else
  return(FALSE);
}

/* reads the floppy in MSDOS or Staubli format and copies the files to /tmp/floppy */
DFILES void readFloppyCB(Widget w, XtPointer clientData, XtPointer callData)
{
 char command[100], *newFloppy="/tmp/floppy2";
 int haveDOS = FALSE, haveStaubli = FALSE;

 /* first delete the old floppy2 */
 strcpy(command, "/bin/rm -rf ");
 strcat(command, newFloppy);
 executeCommand(command);

 /* create new floppy file */
 strcpy(command, "/bin/mkdir ");
 strcat(command, newFloppy);
 executeCommand(command);

 /* now copy from DOS floppy */
 strcpy(command, copy2dosCommand);
 strcat(command, " a: ");
 strcat(command, newFloppy);
 haveDOS = executeCommand(command);   

 if (!haveDOS) /* then try with Staubli */
  {
   strcpy(command, "../common/staubli -s2u");
   haveStaubli = executeCommand(command);
  }

 /* delete old /tmp/floppy directory and move /tmp/floppy2 into it */
 if (haveDOS || haveStaubli)
  {
   /* first delete the /tmp/floppy */
   strcpy(command, "/bin/rm -rf ");
   strcat(command, newFloppy);
   command[strlen(command)-1]=0; /* cutoff last */
   executeCommand(command);
   /* create new floppy file */
   strcpy(command, "/bin/mkdir ");
   strcat(command, newFloppy);
   command[strlen(command)-1]=0; /* cutoff last */
   executeCommand(command);
   /* now do "cp /tmp/floppy2 /tmp/floppy" */
   strcpy(command, "/bin/cp ");
   strcat(command, newFloppy);
   strcat(command, "/* ");
   strcat(command, newFloppy);
   command[strlen(command)-1]=0; /* cutoff last */
   executeCommand(command);
   /* give read/write permission to everyone for the newly read files */
   strcpy(command, "/bin/chmod -R 777 ");
   strcat(command, newFloppy);
   command[strlen(command)-1]=0; /* cutoff last */
   executeCommand(command);
  }
 /* and again delete the old floppy2 */
 strcpy(command, "/bin/rm -rf ");
 strcat(command, newFloppy);
 executeCommand(command);
}

#define LOADINIT	0
#define LOADPICTURE	1
#define LOADMOTIF	2
#define LOADFORCE	3
#define LOADMOTIFFORCE	4
#define LOADPICFLOPPY	5
#define FIX_FSB_FOCUS	10

DFILES void loadPictButtonCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int param = atoi ((char *) clientData);

 switch (param)
   {
    case FIX_FSB_FOCUS:
     XmProcessTraversal(XmFileSelectionBoxGetChild(fileSelectionW,
      fileSelectionBoxLoad ? XmDIALOG_LIST : XmDIALOG_TEXT), XmTRAVERSE_CURRENT);
    return;
   break;

    case LOADINIT:
     {
      char *filesS = getTxtI8N(FILE_SELECTION_FILES_MSG), *noMatchS = getTxtI8N(FILE_SELECTION_NO_MATCH_MSG),
       *dirListS = getTxtI8N(FILE_SELECTION_DIRECTORY_MSG), *filterS = getTxtI8N(FILE_SELECTION_FILTER_MSG),
       *selectionS = getTxtI8N(FILE_SELECTION_SELECTION_MSG), *helpS = getTxtI8N(FILE_SELECTION_HELP_MSG),
       *cancelS = getTxtI8N(FILE_SELECTION_CANCEL_MSG), *searchS = getTxtI8N(FILE_SELECTION_SEARCH_MSG),
       *OKS = getTxtI8N(FILE_SELECTION_OK_MSG);

      fileSelectionBoxLoad = TRUE;
      XtUnmanageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_SELECTION_LABEL));
      XtUnmanageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_TEXT));
      XtUnmanageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_SEPARATOR));
      xtVaSetString(w, XmNdirectory, ApData.imgPath);
      xtVaSetString(w, XmNnoMatchString, noMatchS);
      xtVaSetString(w, XmNdirListLabelString, dirListS);
      xtVaSetString(w, XmNfileListLabelString, filesS);
      xtVaSetString(w, XmNfilterLabelString, filterS);
      xtVaSetString(w, XmNselectionLabelString, selectionS);
      xtVaSetString(w, XmNhelpLabelString, helpS);
      xtVaSetString(w, XmNcancelLabelString, cancelS);
      xtVaSetString(w, XmNapplyLabelString, searchS);
      xtVaSetString(w, XmNokLabelString, OKS);
     }
     break;
    case LOADMOTIF:
     finalDraw(ApData.pixy);
     disableUndo();
     if (Select.rect_on)
       Select.rect_on = False;
     if (Select.imgAllocated)
       manageWarningDialog(getTxtI8N (SELECTION_EXISTS_REPLACE_MSG),
	loadPictButtonCB, "4", getTxtI8N (YES_MSG),
	NULL, "", getTxtI8N (NO_MSG), NULL, NULL, getTxtI8N (HELP_MSG));
     else
      loadPictButtonCB (w, "4", NULL);
     break;
    case LOADPICFLOPPY:
    case LOADPICTURE:
     if (ApData.pic_change)
      {
       finalDraw(ApData.pixy);
       disableUndo();
       manageWarningDialog(getTxtI8N (PICTURE_CHANGED_SAVE_MSG),
	saveDrawingCB, "2", getTxtI8N (YES_MSG), loadPictButtonCB, "3", getTxtI8N (NO_MSG),
	NULL, NULL, getTxtI8N (HELP_MSG));
       break;
      }	/* else continue */
    case LOADMOTIFFORCE:
    case LOADFORCE:
     {
      char *newTitle;

      xtVaSetString(fileSelectionW, XmNdirectory, (param==LOADPICFLOPPY) ? "/tmp/floppy" :
       (ApData.fname[0] ? dirName(ApData.fname) : ApData.imgPath));
      if (param == LOADMOTIFFORCE)
	newTitle = getTxtI8N (LOAD_MOTIF_MSG);
      else
       {
	newTitle = getTxtI8N (LOAD_PICTURE_MSG);
	ApData.pic_change = False;
       }
      setDialogTitle(fileSelectionW, newTitle);
      XtManageChild (fileSelectionW);
     }
     break;
   }
}

/*save current picture in selected format */
static Boolean savePicture(char *filename)
{
 int res, i, j;
 unsigned char saveCmap[256 * 3];

 if ((ApData.width == 0) || (ApData.height == 0))
   return (False);
 res = (Boolean) False;
 memset(saveCmap, 0, 256*3); /* clear colormap */
 for (i = 0; i < 256; i++) /* get the global colormap */
  {
   XColor myColor;

   myColor.pixel = (unsigned long) i;
   XQueryColor (xinfo.dpy, xinfo.cmap, &myColor);
   saveCmap[i*3 + RED] = (unsigned char) (myColor.red >> 8);
   saveCmap[i*3 + GREEN] = (unsigned char) (myColor.green >> 8);
   saveCmap[i*3 + BLUE] = (unsigned char) (myColor.blue >> 8);
  }
 /* override server colors - for servers with less than 8 bits per RGB */
 for (i = 0; i < 256; i++)
  for (j = 0; j < ApData.ColorNo; j++)
   if (i == (int) xinfo.allocCells[j].pixel)
    {
     saveCmap[i*3 + RED] = xinfo.red[j];
     saveCmap[i*3 + GREEN] = xinfo.green[j];
     saveCmap[i*3 + BLUE] = xinfo.blue[j];
    }
 res = saveImage(filename, ApData.img, "PNG", saveCmap, False, FILE_TO_OTHER);
 return((Boolean) res);
}

static void copyToFloppyWhenSavedCB(XtPointer clientData)
{
 if (widgetIsViewable(printProgressSW)) /* still saving, try again later */
  XtAddTimeOut(500, (XtTimerCallbackProc) *copyToFloppyWhenSavedCB, NULL);
 else
  {
   char command[500];
   /* now copy from DOS floppy */
   sprintf(command, "%s ""%s"" a:", copy2dosCommand, ApData.fname);
   if (executeCommand(command)==1)
    manageInfoDialog(getTxtI8N(SAVING_TO_FLOPPY_OK_MSG));
   else
    manageErrorDialog(getTxtI8N(UNABLE_TO_SAVE_PICTURE_MSG), NULL, NULL, getTxtI8N(OK_MSG));
   whenSavedCopyToFloppy=FALSE;
  }
}

/* required only because of the XtAddTimeOut procedure type */
static void SaveFileTimeCB(XtPointer clientData, XtIntervalId *id)
{
 savePictureCB(NULL, clientData, NULL);
}

#define SAVE_DOIT	0
#define	SAVE_SAME_NAME	1
#define	SAVE_NORMAL	2
#define	SAVE_SAME_YES	3
DFILES void savePictureCB(Widget w, XtPointer clientData, XtPointer callData)
{
 Boolean result;

 switch (atoi((char *) clientData))
  {
   case SAVE_DOIT:
    {
     char *tmp = xmStringToCstring(((XmFileSelectionBoxCallbackStruct *) callData)->value);

     XtUnmanageChild(fileSelectionW);
     strcpy(dirData, tmp);
     XtFree(tmp);
     if ((!strstr(dirData, ".png")) && (!strstr(dirData, ".PNG")))
      strcat(dirData, ".png");
     XtAddTimeOut(300, (XtTimerCallbackProc) *SaveFileTimeCB, "2");
     return;
    }
    break;
   case SAVE_SAME_NAME: strcpy(dirData, ApData.fname); break;
   case SAVE_NORMAL:
    if (fileExists(dirData))
     {
      manageWarningDialog(getTxtI8N(FILE_EXISTS_OVERWRITE_MSG), savePictureCB, "3", getTxtI8N(YES_MSG),
       saveDrawingCB, "4", getTxtI8N(NO_MSG), NULL, NULL, getTxtI8N(HELP_MSG));
      return;
     } /* no break on purpose !!! */
   case SAVE_SAME_YES: 
    strcpy(ApData.fname, dirData);
    changeTitleCB(w, dirData, NULL);
    break;
  }

 finalDraw(ApData.pixy);
 disableUndo();
 result = savePicture(dirData);
 if (result)
  {
   ApData.pic_change = False;
   if (ClearAfterSave)
    NewFileCB(w, "1", NULL);
  }
 else
  manageErrorDialog(getTxtI8N(UNABLE_TO_SAVE_PICTURE_MSG), NULL, NULL, getTxtI8N(OK_MSG));
 changeTitleCB(w, dirData, NULL);
 strcpy(ApData.fname, dirData);

 initHistoryList(FALSE);
 addItemToHistoryList(dirData);
 saveHistoryList();
 /* this will be executed while image will still not be saves, so we nee to wait until it is... */
 if (whenSavedCopyToFloppy)
  XtAddTimeOut(500, (XtTimerCallbackProc) *copyToFloppyWhenSavedCB, NULL);
}

static void saveMotif()
{
 int res, i, j;
 unsigned char saveCmap[256 * 3];

 memset(saveCmap, 0, 256 * 3);
 if (!Select.rectangular)
  {
   unsigned char backgroundC = (unsigned char) ApData.drawinfo.backgrndC;
   unsigned char rndCo[512 * 3], myCmap[256 * 3];
   unsigned long diffCo[512];

   for (i = 0; i < 256; i++)
    {
     myCmap[3 * i + RED] = xinfo.red[i];
     myCmap[3 * i + GREEN] = xinfo.green[i];
     myCmap[3 * i + BLUE] = xinfo.blue[i];
    }
   /* calculate random color and its difference from the others */
   for (i = 0; i < 512; i++)
    {
     rndCo[3 * i + RED] = (unsigned char) (drand48() * 255);
     rndCo[3 * i + GREEN] = (unsigned char) (drand48() * 255);
     rndCo[3 * i + BLUE] = (unsigned char) (drand48() * 255);
     diffCo[i] = 0;
     for (j = 0; j < ApData.ColorNo; j++)
       diffCo[i] += getDistance (&rndCo[i], &myCmap[j]);
    }
   i = 0; /* find index of the color with max difference */
   for (j = 1; j < 512; j++)
     if (diffCo[i] < diffCo[j])
       i = j;
   /* set all points not in region to the WhitePixel value */
   for (j = 0; j < Select.height; j++)
     for (i = 0; i < Select.width; i++)
       if (!XPointInRegion (Select.reg, i + Select.x, j + Select.y))
	 Select.img->data[Select.img->bytes_per_line * j + i] = backgroundC;
  }
 strcpy(Select.fname, dirData);
 if ((!strstr(dirData, ".png")) && (!strstr(dirData, ".PNG")))
  strcat(Select.fname, ".png");
 for (i = 0; i < 256; i++)
  for (j = 0; j < ApData.ColorNo; j++)
   if (i == (int) xinfo.allocCells[j].pixel)
    {
     saveCmap[i*3 + RED] = xinfo.red[j];
     saveCmap[i*3 + GREEN] = xinfo.green[j];
     saveCmap[i*3 + BLUE] = xinfo.blue[j];
    }
 res = saveImage(Select.fname, Select.img, "PNG", saveCmap, False, FILE_TO_OTHER);
 motifS = False;
}

#define SAVE_OLD		1
#define SAVE_AS			2
#define SAVE_MOTIF		3
#define SAVE_CANCEL		4
#define SAVE_DO_MOTIF		5
#define SAVE_FORCE_MOTIF 	6
#define SAVE_OLD_AND_TO_FLOPPY	7
DFILES void saveDrawingCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int param = atoi ((char *) clientData);
 char *title;

 whenSavedCopyToFloppy = FALSE;
 fileSelectionBoxLoad = FALSE;
 XtManageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_SELECTION_LABEL));
 XtManageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_TEXT));
 XtManageChild(XmFileSelectionBoxGetChild(fileSelectionW, XmDIALOG_SEPARATOR));
 switch (param)
  {
   case SAVE_OLD_AND_TO_FLOPPY: whenSavedCopyToFloppy = TRUE;
   case SAVE_OLD:
    if (ApData.fname[0] == 0)
     saveDrawingCB(w, "2", NULL);
    else /* call save procedure */
     XtAddTimeOut (300, (XtTimerCallbackProc) *SaveFileTimeCB, "1");
    break;
   case SAVE_MOTIF:
    if (!Select.selected)
     {
      manageErrorDialog(getTxtI8N(NO_SELECTION_MADE_MSG), NULL, NULL, getTxtI8N(OK_MSG));
      return;
     }
    if (!Select.imgAllocated)
     {
      Select.img = MyGetImage(ApData.img, Select.x, Select.y,
       Select.width, Select.height, ApData.viewMode);
      if (!checkXImage (Select.img))
	return;
      Select.imgAllocated = True;
     }
    if (strlen(Select.fname)>5) /* then it is probably a valid filename... */
     {
      xtVaSetString(fileSelectionW, XmNdirectory, dirName(Select.fname));
      /* MOTIF 2.0 FSB with relative pathMode */
      xtVaSetString(fileSelectionW, XmNdirSpec, (XmVERSION>1) ? baseName(Select.fname) : Select.fname);
     }

    motifS = True; /* continue !! */
   case SAVE_AS:
    if (param==SAVE_AS)
     {
      /* NOTE! changing XmNdirectory clears XmNdirSpec, so order is important! */
      xtVaSetString(fileSelectionW, XmNdirectory, (param==LOADPICFLOPPY) ? "/tmp/floppy" :
       (ApData.fname[0] ? dirName(ApData.fname) : ApData.imgPath));
      if (strlen(ApData.fname)>5) /* then it is probably a valid filename... */
       /* MOTIF 2.0 FSB with relative pathMode */
       xtVaSetString(fileSelectionW, XmNdirSpec, (XmVERSION>1) ? baseName(ApData.fname) : ApData.fname);
     }
   case SAVE_CANCEL:
    if (param == SAVE_MOTIF)
     title = getTxtI8N(SAVE_MOTIF_MSG);
    else
     {
      title = getTxtI8N(SAVE_PICTURE_MSG);
      motifS = False;
     }
    if (param == SAVE_CANCEL)
     {
      RestoreCursor(XtWindow(mainW));
      RestoreCursor(drawAreaWin);
      if (XtIsRealized(zoomDrawAreaW))
	RestoreCursor(XtWindow(zoomDrawAreaW));
     }
    if (XtHasCallbacks (fileSelectionW, XmNokCallback) == XtCallbackHasSome)
     XtRemoveAllCallbacks (fileSelectionW, XmNokCallback);
    if (XtHasCallbacks (fileSelectionW, XmNhelpCallback) == XtCallbackHasSome)
     XtRemoveAllCallbacks (fileSelectionW, XmNhelpCallback);
    if (motifS)
     XtAddCallback (fileSelectionW, XmNokCallback, saveDrawingCB, "5");
    else
     XtAddCallback (fileSelectionW, XmNokCallback, savePictureCB, "0");
    XtAddCallback (fileSelectionW, XmNhelpCallback, helpCB, "Save file");
    setDialogTitle(fileSelectionW, title);
    XtManageChild (fileSelectionW);
    break;
   case SAVE_DO_MOTIF:
    {
     char *tmp;

     tmp = xmStringToCstring(((XmFileSelectionBoxCallbackStruct *) callData)->value);
     strcpy(dirData, tmp);
     XtFree(tmp);
     XtUnmanageChild(fileSelectionW);
     if (fileExists(dirData))
      manageWarningDialog(getTxtI8N(FILE_EXISTS_OVERWRITE_MSG), saveDrawingCB, "6", getTxtI8N(YES_MSG),
       saveDrawingCB, "3", getTxtI8N(NO_MSG), NULL, NULL, getTxtI8N(HELP_MSG));
     else
      saveMotif();
    }
    break;
   case SAVE_FORCE_MOTIF:
    saveMotif();
    break;
  }
}

/* returns a band of image to be printed */
static int returnPictureImg(int xpixels, int xdpi, int ydpi, int ypos, int ytotal,
 char desiredImageType, char **data, int *bpl, int *lines, char *returnImageType, unsigned char *cmap)
{
 XImage *bandXIM;
 int bandh;
 int i, j;
 int depth = 1;

 if (desiredImageType == IMG_BW)
  return (PRT_WRONG_IMG);

 bandh = MIN(ytotal - ypos, 0x20000 / xpixels);
 if (ytotal == bandh)
  bandh /= 2;
 if (bandh == 0)
  bandh++;
 if (desiredImageType == IMG_TRUE_COL)
  depth = 3;

 bandXIM = doXImage(8, xpixels*depth, bandh);
 if (bandXIM == NULL)
   return (PRT_NO_MEMORY);
 *data = bandXIM->data;
 *bpl = bandXIM->bytes_per_line;
 *lines = bandXIM->height;
 *returnImageType = (depth==1) ? IMG_CMAP : IMG_TRUE_COL;

 for (i = 0; i < maxColors; i++)
  {
   j = (unsigned char) xinfo.allocCells[i].pixel;
   cmap[j*3] = xinfo.red[i];
   cmap[j*3+1] = xinfo.green[i];
   cmap[j*3+2] = xinfo.blue[i];
  }
 renderRealImg(bandXIM->data, bandXIM->bytes_per_line,
  depth, 0, ypos, xpixels, bandh, ytotal, xdpi, ydpi);

 /* will only destroy Ximage structure, image data is used to print */
 bandXIM->data = NULL;
 XDestroyImage (bandXIM);
 return (PRT_OK);
}

static void printPicture(int widthP, int heightP, int leftM, int topM)
{
 int wid, hei, xdpi, ydpi, nonyoff, dummy, nonxoff;
 int currw = widthP, currh = heightP, currxoff = leftM, curryoff = topM;

 getPrinterLimits (&wid, &hei, &xdpi, &ydpi, &nonyoff, &dummy, &nonxoff);
 currxoff -= nonxoff;
 curryoff -= nonyoff;
 setPageSettings(currw*xdpi/25400, currh*ydpi/25400, currxoff*xdpi/25400, curryoff*ydpi/25400, TRUE);
 if ((currw > 0) && (currh > 0))
  printPage(*returnPictureImg);
}

static float calcPerc(int zoomPerc, int sizePage, int sizeMM)
{ return(1.0 * sizePage / ((zoomPerc * sizeMM) / 100)); }

#define P_INIT_INFO	1
#define P_INIT_TOP	2
#define P_INIT_LEFT	3
#define P_INIT_WIDTH	4
#define P_INIT_HEIGHT	5
#define P_FIT_PAGE	6
#define P_NEW_VALUE	8
#define P_PRINT		9
#define P_PRN_INFO	10
#define P_MARGIN_SET	11
#define P_FULL_PAGE	12
DFILES void printPictureCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static Boolean print2fit = False;
 static inpNumStruct *topM, *leftM, *widthP, *heightP;
 static Widget prnInfoW;
 static int pageWidth = 0, pageHeight = 0;
 static int wid, hei, xdpi, ydpi, nonyoff, nonxoff;
 char tmpstr[100];
 float widthPerc, heightPerc;

 switch (atoi((char *) clientData))
  {
   case P_INIT_INFO: prnInfoW = w; break;
   case P_FULL_PAGE:
    widthP->value=widthP->upperB; setInpArrVal(widthP, widthP->upperB);
    heightP->value=heightP->upperB; setInpArrVal(heightP, heightP->upperB);
    leftM->value=nonxoff; setInpArrVal(leftM, nonxoff);
    topM->value=nonyoff; setInpArrVal(topM, nonyoff);
    printPictureCB(zoomScaleW, "8", NULL);
    break;
   case P_INIT_TOP: 
    topM = initInputText(w, 0, 0, 10); 
    XtAddCallback(w, XmNvalueChangedCallback, (XtCallbackProc) printPictureCB, "11");
    break;
   case P_INIT_LEFT: 
    leftM = initInputText(w, 0, 0, 10);
    XtAddCallback(w, XmNvalueChangedCallback, (XtCallbackProc) printPictureCB, "11");
    break;
   case P_INIT_WIDTH: 
    pageWidth = 999;
    widthP = initInputText(w, pageWidth, 0, pageWidth);
    XtAddCallback(w, XmNvalueChangedCallback, (XtCallbackProc) printPictureCB, "8");
    break;
   case P_INIT_HEIGHT: 
    pageHeight = 999;
    heightP = initInputText(w, pageHeight, 0, pageHeight);
    XtAddCallback(w, XmNvalueChangedCallback, (XtCallbackProc) printPictureCB, "8");
    break;
   case P_FIT_PAGE: 
    XtVaGetValues(w, XmNset, &print2fit, NULL);
    if (print2fit)
     {
      zoomPerc = (int) ((pageWidth * 100) / sizeXmm);
      heightPerc = calcPerc (zoomPerc, pageHeight, sizeYmm);
      if (heightPerc < 1)
	zoomPerc = (int) ((pageHeight * 100) / sizeYmm);
      XtVaSetValues(zoomScaleW, XmNvalue, zoomPerc, NULL);
      printPictureCB(zoomScaleW, "8", NULL);
      setZoomFactorCB(zoomScaleW, "0", NULL);
     }
    XtVaSetValues(zoomScaleW, XmNsensitive, !print2fit, NULL);
    break;
   case P_NEW_VALUE: 
    XmScaleGetValue(zoomScaleW, &zoomPerc);
    /* these are not inchified, so we need to "undo" the inch support */
    sizeXmm = getImageXmm()*(metricSystem ? 1.0 : 25.4);
    sizeYmm = getImageYmm()*(metricSystem ? 1.0 : 25.4);
    pageWidth = widthP->value;
    pageHeight = heightP->value;
    widthPerc = calcPerc(zoomPerc, pageWidth, sizeXmm);
    sprintf (tmpstr, "%s: %.2f", getTxtI8N(WIDTH_REPEAT_NUMBER_MSG), widthPerc);
    setLabelString(widthPrnLW, tmpstr);
    heightPerc = calcPerc(zoomPerc, pageHeight, sizeYmm);
    sprintf (tmpstr, "%s: %.2f", getTxtI8N(HEIGHT_REPEAT_NUMBER_MSG), heightPerc);
    setLabelString(heightPrnLW, tmpstr);
    sprintf(tmpstr, "%s (%%) %.2f x %.2f %s",  getTxtI8N(ZOOM_MSG),
     getImageXmm()*zoomPerc*(metricSystem ? 0.1 : 1)/100.0,
     getImageYmm()*zoomPerc*(metricSystem ? 0.1 : 1)/100.0, metricSystem ? "cm" : "in");
    setTitleString(zoomScaleW, tmpstr);
    break;
   case P_PRINT: 
    finalDraw(ApData.pixy);
    disableUndo();
    printPicture(widthP->value, heightP->value, leftM->value, topM->value);
    break;
   case P_PRN_INFO: 
    {
     char *prnstr;
     int dummy;

     prnstr = returnPrtSettingText();
     setLabelString(prnInfoW, prnstr);
     getPrinterLimits(&wid, &hei, &xdpi, &ydpi, &nonyoff, &dummy, &nonxoff);
     leftM->lowerB = nonxoff;
     if (leftM->value < nonxoff)
       leftM->value = nonxoff;
     leftM->upperB = (wid * 25400) / xdpi;
     setInpArrVal(leftM, leftM->value);
     topM->lowerB = nonyoff;
     topM->upperB = (hei * 25400) / ydpi;
     if (topM->value < nonyoff)
       topM->value = nonyoff;
     setInpArrVal(topM, topM->value);
     widthP->upperB = (wid * 25400) / xdpi;
     setInpArrVal (widthP, widthP->value);
     heightP->upperB = (hei * 25400) / ydpi;
     setInpArrVal(heightP, heightP->value);
     pageWidth = widthP->value;
     pageHeight = heightP->value;
    }
    break;
   case P_MARGIN_SET: 
    {
     int diff;

     if ((xdpi == 0) || (ydpi == 0))
       return;
     diff = leftM->value - leftM->lowerB +
       widthP->value - widthP->upperB;
     if (diff > 0)
      {
       widthP->upperB = (wid * 25400) / xdpi - diff;
       setInpArrVal (widthP, widthP->value);
      }
     diff = topM->value - topM->lowerB +
      heightP->value - heightP->upperB;
     if (diff > 0)
      {
       heightP->upperB = (hei * 25400) / ydpi - diff;
       setInpArrVal (heightP, heightP->value);
      }
    }
  }
}

DFILES void changeTitleCB(Widget w, XtPointer clientData, XtPointer callData)
{
 char s1[300], s2[100];

 if (strcmp ((char *) clientData, "Init") == 0)
  MainWindowW = w;
 else
  {
   sprintf(s1, "%s: %s(%d,%d)", ApData.version, baseName(ApData.fname)/*baseName((char *) clientData)*/, ApData.width, ApData.height);
   if (Select.selected)
    {
     sprintf(s2, " [%dx%d", Select.width, Select.height);
     strcat(s1,s2);
     if (Select.rect_on)
      {
       sprintf(s2, " @ %dx%d", Select.x, ApData.height-Select.y-Select.height);
       strcat(s1,s2);
      }
     strcat(s1,"]");
    }
   XtVaSetValues (MainWindowW, XmNtitle, s1, NULL);
  }
}

/* init file history list in file menu */
DFILES void initHistoryList(int first)
{
 int i;
 char buf[512], *p;
 FILE *fp;
 unsigned char *num;

 historyList.num = 0;
 num = &historyList.num;
 strcpy (historyList.fileN, getenv ("HOME"));
 strcat (historyList.fileN, "/.apaint_history");
 if ((fp = fopen (historyList.fileN, "r")))
  {
   while ((*num < MAX_HISTORY_LIST) && (!ferror (fp)))
    {
     memset (buf, 0, 512);	/*clear buffer */
     p = fgets (buf, 511, fp);	/*read line from file */
     if (p == NULL)
       break;
     if ((!first) && historyList.name[*num])
      free(historyList.name[*num]);
     if ((historyList.name[*num] = calloc (strlen (buf), sizeof (char))) == NULL)
      break;
     strncpy (historyList.name[*num], buf, strlen (buf) - 1);
     setLabelString(historyList.id[*num], baseName (historyList.name[*num]));
     (*num)++;
    }
   fclose(fp);
  }
 for (i = *num; i < MAX_HISTORY_LIST; i++)
  XtUnmanageChild(historyList.id[i]);
}

/* at the end of work saveHistoryList */
static void saveHistoryList()
{
 FILE *fp;
 int i;

 if ((fp = fopen (historyList.fileN, "w")) == NULL)
  return;
 i = 0;
 while ((i < historyList.num) && (!ferror (fp)))
  {
   fputs (historyList.name[i], fp);
   fputc ('\n', fp);
   i++;
  }
 fclose(fp);
}

/* add new item to file name history list: check for duplicates and rearrange if necessary */
static void addItemToHistoryList(char *fileName)
{
 int i, ind;

 for (i = 0; i < historyList.num; i++)
  if (strcmp (fileName, historyList.name[i]) == 0)
   break;
 if (i < historyList.num) /* item exists ! */
  {
   char *tmp;

   if (i == 0)
     return;
   ind = i;
   tmp = historyList.name[ind];
   for (i = ind; i > 0; i--)
     historyList.name[i] = historyList.name[i - 1];
   historyList.name[0] = tmp;
  }
 else if (historyList.num < MAX_HISTORY_LIST) /* add new item to the list */
  {
   for (i = historyList.num; i > 0; i--)
     historyList.name[i] = historyList.name[i - 1];
   historyList.name[0] = malloc (strlen (fileName) + 1);
   strcpy (historyList.name[0], fileName);
   XtManageChild (historyList.id[historyList.num]);
   historyList.num++;
  }
 else /* add item to the list by first freeing the oldest one */
  {
   free (historyList.name[MAX_HISTORY_LIST - 1]);
   for (i = historyList.num - 1; i > 0; i--)
     historyList.name[i] = historyList.name[i - 1];
   historyList.name[0] = malloc (strlen (fileName) + 1);
   strcpy (historyList.name[0], fileName);
  }
 for (i = 0; i < historyList.num; i++)
  setLabelString(historyList.id[i], baseName (historyList.name[i]));
}

DFILES void historyListCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);
 int ind;
 static char tmp[10];

 switch (prm)
  {
   case 1:
   case 2:
   case 3:
   case 4:
   case 5:
     historyList.id[prm - 1] = w;
     break;
   case 101:
   case 102:
   case 103:
   case 104:
   case 105:
     sprintf (&tmp[0], "%d", prm + 100);
     if (ApData.pic_change)
       {
	finalDraw(ApData.pixy);
	disableUndo();
	manageWarningDialog(getTxtI8N (PICTURE_CHANGED_SAVE_MSG),
	 saveDrawingCB, "2", getTxtI8N (YES_MSG), historyListCB, &tmp[0],
	 getTxtI8N(NO_MSG), NULL, NULL, getTxtI8N (HELP_MSG));
       }
     else
       historyListCB (NULL, &tmp[0], NULL);
     break;
   case 201:
   case 202:
   case 203:
   case 204:
   case 205:
     ind = prm - 201;
     if (fileExists (historyList.name[ind]))
      {
       strcpy(dirData, historyList.name[ind]);
       loadPictureCB(NULL, "2", NULL);
      }
     else /* file no longer exists - remove it from the list */
      {
       int i;

       XBell (xinfo.dpy, 0);
       free (historyList.name[ind]);
       historyList.name[ind] = NULL;
       for (i = ind; i < (historyList.num - 1); i++)
	{
	 historyList.name[i] = historyList.name[i + 1];
	 setLabelString(historyList.id[i], baseName (historyList.name[i]));
	}
       historyList.num--;
       XtUnmanageChild (historyList.id[historyList.num]);
      }
     break;
   }
}
