#define FILL
#include <Xm/Xm.h>
#include <string.h>
#include "paint.h"

#define CALC_MINMAX 1

static XImage *fillimg;

typedef struct
{
 int x, y, ind;
 XImage *pattern;
} fillPrmStruct;

static fillPrmStruct fillPrm;

#ifdef CALC_MINMAX
static struct
{
 int left, right, top, bottom;
}
minmax;
#endif
/* A Seed Fill Algorithm by Paul Heckbert from "Graphics Gems", Academic Press, 1990
   fill.c : simple seed fill program Paul Heckbert; 13 Sept 1982, 28 Jan 1987 */

typedef struct
{		/* window: a discrete 2-D rectangle */
 int x0, y0;	/* xmin and ymin */
 int x1, y1;	/* xmax and ymax (inclusive) */
} FillWindow;

typedef unsigned long FPixel;

typedef struct
{ short y, xl, xr, dy; } Segment;
/* Filled horizontal segment of scanline y for xl<=x<=xr.
   Parent segment was on line y-dy.  dy=1 or -1 */

static Segment *stack_limit;

#define MAX_STACK 100000	/* max depth of stack */

#define PUSH(Y, XL, XR, DY)	/* push new segment on stack */ \
    if ((sp<stack_limit) && ((Y)+(DY)>=win->y0) && (((Y)+(DY))<=win->y1)) \
    {sp->y = Y; sp->xl = XL; sp->xr = XR; sp->dy = DY; sp++;}

#define POP(Y, XL, XR, DY)	/* pop segment off stack */ \
    {sp--; Y = sp->y+(DY = sp->dy); XL = sp->xl; XR = sp->xr;}

/* fill: set the pixel at (x,y) and all of its 4-connected neighbors
 * with the same pixel value to the new pixel value nv.
 * A 4-connected neighbor is a pixel above, below, left, or right of a pixel.
 * if pattern is not NULL then fill area with the image. */
static void fillRegion(int x, int y, FillWindow * win, FPixel nv)
{
 int l, x1, x2, dy;
 unsigned char ov;		/* old pixel value */
 unsigned char *yline; 
 Segment stack[MAX_STACK], *sp = stack;	/* stack of filled segments */

 stack_limit = stack + MAX_STACK * sizeof(Segment);
 ov = fillimg->data[fillimg->bytes_per_line * y + x];
 if (x < win->x0 || x > win->x1 || y < win->y0 || y > win->y1)
   return;
 if (ov == nv)
   return;
 PUSH (y, x, x, 1);		/* needed in some cases */
 PUSH (y + 1, x, x, -1);	/* seed segment (popped 1st) */

 while (sp > stack)
  {
   /* pop segment off stack and fill a neighboring scan line */
   POP (y, x1, x2, dy);
   yline = &fillimg->data[fillimg->bytes_per_line*y];
#ifdef CALC_MINMAX
   if (minmax.left > x1)
     minmax.left = x1;
   if (minmax.right < x2)
     minmax.right = x2;
   if (minmax.top > y)
     minmax.top = y;
   if (minmax.bottom < y)
     minmax.bottom = y;
#endif
   /* segment of scan line y-dy for x1<=x<=x2 was previously filled,
      now explore adjacent pixels in scan line y */
   for (x = x1; (x >= win->x0) && (yline[x] == ov); x--)
     yline[x] = (unsigned char) nv;
   if (x >= x1)
     goto skip;
   l = x + 1;
   if (l < x1)
     PUSH (y, l, x1 - 1, -dy);	/* leak on left? */
   x = x1 + 1;
   do
    {
     for (; (x <= win->x1) && ((yline[x]) == ov); x++)
      yline[x] = nv;
     PUSH (y, l, x - 1, dy);
     if (x > x2 + 1)
       PUSH (y, x2 + 1, x - 1, -dy);	/* leak on right? */
   skip:for (x++; (x <= x2) && (yline[x] != ov); x++);
     l = x;
    }
   while (x <= x2);
  }
}

static void fill_It(int x, int y, FillWindow *win, FPixel nv, XImage *pattern)
{
  unsigned char blackPx=0;

  if (pattern == NULL)
    fillRegion (x, y, win, nv);
  else
    { /* !! fill area with black pixel first, then replace black with pattern */
     if (xinfo.trueColor)
      {
       if (maxColors < MAX_CELLS)
	{
	 int i;
	 Boolean notFound;

	 blackPx = 0;
	 notFound = True;
	 while (notFound)
	   {
	    i = 0;
	    while (i < maxColors)
	      if (xinfo.allocCells[i].pixel == blackPx)
		i = maxColors + 1;
	      else
		i++;
	    if (i == (maxColors + 1))
	      blackPx++;
	    else
	      notFound = False;
	   }
	}
       else
	{
	 manageErrorDialog(getTxtI8N (NOT_ENOUGH_COLORS_MSG), NULL, NULL, getTxtI8N (OK_MSG));
	 return;
	}
      }
     else
      blackPx = xinfo.black;
     fillRegion(x, y, win, blackPx);
    }
#ifdef CALC_MINMAX
  if (minmax.left == 0xffff)
    opBounds.x = 0;
  else
    opBounds.x = minmax.left;
  if (minmax.top == 0xffff)
    opBounds.y = 0;
  else
    opBounds.y = minmax.top;

  if (minmax.right == 0)
    opBounds.width = 0;
  else
    opBounds.width = minmax.right - opBounds.x + 1;
  if (minmax.bottom == 0)
    opBounds.height = 0;
  else
    opBounds.height = minmax.bottom - opBounds.y + 1;
#else
  opBounds.x = opBounds.y = 0;
  opBounds.width = bBox.width;
  opBounds.height = bBox.height;
#endif
  if (pattern) /* replace black pixels with pattern */
   {
    int j, endH;
    unsigned char *dataP, *endP, *startP;

    endH = opBounds.height + opBounds.y;
    for (j = opBounds.y; j < endH; j++)
     {
      startP=endP=dataP = &fillimg->data[fillimg->bytes_per_line*j + opBounds.x];
      endP += opBounds.width;
      while (dataP < endP)
       {
	if (*dataP == blackPx)
	  *dataP = pattern->data[pattern->bytes_per_line *
	  (j % pattern->height) + ((dataP - startP) % pattern->width)];
	dataP++;
       }
     }
   }
}

static void doFillArea(fillPrmStruct *fillPrm)
{
 FillWindow fillwin;
 Rectangle bBox;
 int ind;
 XPoint fillPt, orR;

 bBox.x = bBox.y = 0;
 bBox.width = ApData.width;
 bBox.height = ApData.height;
 Undo.tool = fill;

 fillPt.x = fillPrm->x;
 fillPt.y = fillPrm->y;
 ind = fillPrm->ind;
 if (ApData.viewMode==repeatView)
  {
   orR.x = repeats.crop.width;
   orR.y = repeats.crop.height;
   movePts(&fillPt, 1, -repeats.r[ind].x, -repeats.r[ind].y);
   flipPts(&fillPt, 1, repeats.r[ind].rType, &orR);
  }

 Undo.img = SubImage(ApData.img, bBox.x, bBox.y, bBox.width, bBox.height);
 enableUndo (False);
 Undo.imgAllocated = True;
 if (!checkXImage (Undo.img))
  {
   disableUndo();
   RestoreCursor(XtWindow (mainW));
   ChangeCursor(drawAreaWin, XC_draft_small, True);
   ChangeCursor(XtWindow (mainW), XC_draft_small, True);
   if (XtIsRealized(zoomDrawAreaW))
     ChangeCursor(XtWindow(zoomDrawAreaW), XC_draft_small, True);
   return;
  }
 fillimg = ApData.img;

 fillwin.x0 = fillwin.y0 = 0;
 fillwin.x1 = bBox.width - 1;
 fillwin.y1 = bBox.height - 1;

#ifdef CALC_MINMAX
 minmax.left = 0xffff;
 minmax.right = 0;
 minmax.top = 0xffff;
 minmax.bottom = 0;
#endif
 fill_It(fillPt.x - bBox.x, fillPt.y - bBox.y, &fillwin,
	  xinfo.allocCells[ColorMatrix.FrgColorInd].pixel, fillPrm->pattern);
/* opBounds.x += bBox.x; opBounds.y += bBox.y; */
 Undo.points[Undo.npoints].x = (short) opBounds.x;
 Undo.points[Undo.npoints++].y = (short) opBounds.y;
 Undo.points[Undo.npoints].x = (short) opBounds.width;
 Undo.points[Undo.npoints++].y = (short) opBounds.height;
 Undo.area.x = Undo.area.y = 0;
 Undo.area.width = Undo.img->width;
 Undo.area.height = Undo.img->height;
/*  CopyAreaImg (fillimg, ApData.img, NULL, 0, 0,
	      opBounds.x - bBox.x, opBounds.y - bBox.y,
	      opBounds.width, opBounds.height,
	      opBounds.x, opBounds.y);*/
 fillimg = NULL;
 RestoreCursor (XtWindow (mainW));
 ChangeCursor (drawAreaWin, XC_draft_small, True);
 if (XtIsRealized (zoomDrawAreaW))
   ChangeCursor (XtWindow (zoomDrawAreaW), XC_draft_small, True);
 LastDrawOnAllWin(False);  
}

/* fill area algorithm */
DFILL void fillArea(int x, int y, XImage *pattern)
{
 int ind;
 unsigned char initPoint;

 ind = pointInRepeat(x, y, repeats.currNum);
 if (ind == -1)
   return;
 initPoint=ApData.img->data[ApData.img->bytes_per_line*(y%ApData.img->height) + (x%ApData.img->width)];
 /* if drawing with some other color gives same color, it is protected, so do not fill */
 if (ApData.transfTbl[(initPoint+1)&255][initPoint]==initPoint)
  return; 

 fillPrm.x = x;
 fillPrm.y = y;
 fillPrm.ind = ind;
 fillPrm.pattern = pattern;
 doFillArea(&fillPrm);
}

