/* print.c */
#define PRINT
#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/TextF.h>
#include <Xm/List.h>
#include <Xm/Scale.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include "paint.h"

static void selectPrinter(int printer);
static void selectMode(int mode);
static void selectDensityX(int dens);
static void selectDensityY(int dens);

static getPrinterInfoProc getPrinterInfo;
static initPrintProc initPrint;
static printBandProc printBand;
static endPrintProc endPrint;
static getNextBandProc getNextBand;

static int numPrinters; /* number of supported printers in print library */
static int savingImage=FALSE; /* determines if we are saving image or printing */
static int doRemove=TRUE; /* should it remove printed files */
static int freeSavedImage; /* free saved image at the end of saving */
static char *saveFname;	/* to which filename should the image be saved */
static XImage *saveXIM=NULL; /* image which we are saving */
static unsigned char cmap[256*3]; /* permanent cmap for colormap images */

/* public variable for printer drivers modules */
static char *prtTxt[MAX_PTXT]; /* localized strings for printer interface */

/* printer capabilities & current settings */
static int printType; /* IMG_BW ... IMG_TRUE_COL */
static int maxxpixels;
static int maxypixels;
static int xdpi; /* horizontal dots per inch *1000 */
static int ydpi; /* vertical dots per inch *1000 */
static int nonprintableTop; /* nonprintable mm on top for paper handling */
static int nonprintableBottom; /* nonprintable mm on bottom */
static int nonprintableLeft; /* nonprintable mm on left */
static int dpivariable; /* user can vary the density */
static char *printerName;
static char *profileName;

/* settings for the page to be printed */
static int xpixels;
static int ypixels;
static int xmargin;
static int ymargin;
static int complexImage; /* TRUE or FALSE */
static int printToFile = FALSE;
static char *printToFilename = NULL;
static char *prtCommand = NULL;

static int prtStatus; /* if everything is proceeding smooth within this print */
static int nowPrinting = FALSE;

static int selectedPrinter;
static int numModes;
static int selectedMode; /* currently selected printer mode */
static printerMode prtModes[5]; /* list of printer modes for selected printer */
static int numDensitX;
static printerDensity prtDensityX[20]; /* list of printer X densities for selected printer */
static int numDensitY;
static printerDensity prtDensityY[20]; /* list of printer Y densities for selected printer */

static int imagePrintDst; /* where do we want images to be finally stored */

/* selects saving to image in the passed format, or reverts to previous state */
static void initendSaveImage(int init, char *imgFormat);

static void selectItCB(Widget w, XtPointer clientData, XtPointer callData)
{
 XmListCallbackStruct *listCB;
 
 listCB = (XmListCallbackStruct *) callData;
 switch ((int) clientData)
  {
   case 0:
    selectPrinter(listCB->item_position - 1);
    break;
   case 1:
    selectMode(listCB->item_position - 1);
    break;
   case 2:
    selectDensityX(listCB->item_position - 1);
    break;
   case 3:
    selectDensityY(listCB->item_position - 1);
    break;
  }
}

/* print mode index to localized string description */
static char *prtMode2string(int i)
{
 switch (i)
  {
   case IMG_BW: return(prtTxt[PTXT_MODE_BW]);
   case IMG_CMAP: return(prtTxt[PTXT_MODE_CMAP]);
   case IMG_TRUE_COL_DITH: return(prtTxt[PTXT_MODE_COLOR_DITHER]);
   case IMG_TRUE_COL: return(prtTxt[PTXT_MODE_COLOR_TRUE]);
   default: return("");
  }
}

/* set the value of print to file and destination */
DPRINT void setPrintToFileCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int reason;
 
 reason = atoi((char *) clientData);
 if (!reason)
  {
   printToFile = !printToFile;
   setState(printToFileTBW, printToFile);
   setSensitive(printToFileIW, printToFile);
   setSensitive(printToFileDstW, printToFile);
  }
 else
  selectedPrintDst = reason-1;
}

static int returnMatchingPrinter(char *findPrinter)
{
 int i, matchIndex = 0;

 for (i=0; i<numPrinters; i++) 
  {
   selectPrinter(i);  
   if (EQUAL(findPrinter, printerName))
    matchIndex = i;
  }
 return(matchIndex);
}

/* customize print interface */
DPRINT void initPrtInterfaceCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int i;
 Widget listW;
 XmString xms;

 /*************** BEGIN customization */
 prtCommand = printCommand;
 selectedPrintDst = FILE_TO_A;

 XmTextFieldSetString(printToFileIW, baseName(getTxtI8N(NEW_PICTURE_MSG)));

 prtTxt[PRT_OK] = getTxtI8N(MESSAGE_FILE_IN_PRINT_QUEUE_OK_MSG);
 prtTxt[PRT_USER_ABORT] = getTxtI8N(PRINT_CANCELLED_MSG);
 prtTxt[PRT_NO_MEMORY] = getTxtI8N(NOT_ENOUGH_MEMORY_MSG);
 prtTxt[PRT_IO_ERROR] = getTxtI8N(ERROR_IO_MSG);
 prtTxt[PRT_INTERNAL_ERROR] = getTxtI8N(ERROR_INTERNAL_CALL_SUPPORT_MSG);
 prtTxt[PRT_EXEC_ERROR] = getTxtI8N(ERROR_PROCESS_CREATION_MSG);
 prtTxt[PRT_WRONG_IMG] = getTxtI8N(ERROR_WRONG_IMAGE_FOR_PRINT_MSG);
 prtTxt[PTXT_WAIT] = getTxtI8N(WAIT_MSG);
 prtTxt[PTXT_PRINTING_WAIT] = getTxtI8N(PRINTING_WAIT_MSG);
 prtTxt[PTXT_CANCEL_PRINT] = getTxtI8N(CANCEL_PRINT_MSG);
 prtTxt[PTXT_SAVING_WAIT] = getTxtI8N(SAVING_WAIT_MSG);
 prtTxt[PTXT_CANCEL_SAVE] = getTxtI8N(CANCEL_SAVE_MSG);
 prtTxt[PTXT_PRINTER] = getTxtI8N(PRINTER_MSG);
 prtTxt[PTXT_PRINT_MODE] = getTxtI8N(PRINT_MODE_MSG);
 prtTxt[PTXT_DENSITY_X] = getTxtI8N(PRINT_DENSITY_X_MSG);
 prtTxt[PTXT_DENSITY_Y] = getTxtI8N(PRINT_DENSITY_Y_MSG);
 prtTxt[PTXT_PAPER_WARNING] = getTxtI8N(WARNING_PAPER_MSG);
 prtTxt[PTXT_PRINT_TO_FILE] = getTxtI8N(PRINT_TO_FILE_MSG);
 prtTxt[PTXT_PRINT_TO_FILE_OK] = getTxtI8N(PRINT_TO_FILE_OK_MSG);

 prtTxt[PTXT_MODE_BW] = getTxtI8N(PRINT_MODE_BLACK_AND_WHITE_MSG);
 prtTxt[PTXT_MODE_CMAP] = getTxtI8N(PRINT_MODE_COLORMAP_MSG);
 prtTxt[PTXT_MODE_COLOR_DITHER] = getTxtI8N(PRINT_MODE_COLOR_DITHER_MSG);
 prtTxt[PTXT_MODE_COLOR_TRUE] = getTxtI8N(PRINT_MODE_COLOR_TRUE_MSG);
 /*************** END customization */

 setLabelString(printToFileTBW, prtTxt[PTXT_PRINT_TO_FILE]);
 setLabelString(printingInfoLW, prtTxt[PTXT_PRINTING_WAIT]);
 setDialogTitle(printingTW, prtTxt[PTXT_WAIT]);
 setLabelString(printCancelW, prtTxt[PTXT_CANCEL_PRINT]);
 xtVaSetString(printerSelBW, XmNselectionLabelString, prtTxt[PTXT_PRINTER]);
 xtVaSetString(printModeSelBW, XmNselectionLabelString, prtTxt[PTXT_PRINT_MODE]);
 xtVaSetString(printXdpiSelBW, XmNselectionLabelString, prtTxt[PTXT_DENSITY_X]);
 xtVaSetString(printYdpiSelBW, XmNselectionLabelString, prtTxt[PTXT_DENSITY_Y]);

 listW = XmSelectionBoxGetChild (printYdpiSelBW, XmDIALOG_LIST);
 XtAddCallback(listW, XmNsingleSelectionCallback, selectItCB, (XtPointer) 3);
 XtVaSetValues(listW, XmNselectionPolicy, XmSINGLE_SELECT, NULL);

 listW = XmSelectionBoxGetChild(printXdpiSelBW, XmDIALOG_LIST);
 XtAddCallback(listW, XmNsingleSelectionCallback, selectItCB, (XtPointer) 2);
 XtVaSetValues(listW, XmNselectionPolicy, XmSINGLE_SELECT, NULL);

 listW = XmSelectionBoxGetChild(printModeSelBW, XmDIALOG_LIST);
 XtAddCallback(listW, XmNsingleSelectionCallback, selectItCB, (XtPointer) 1);
 XtVaSetValues(listW, XmNselectionPolicy, XmSINGLE_SELECT, NULL);

 listW = XmSelectionBoxGetChild(printerSelBW, XmDIALOG_LIST);
 XtAddCallback(listW, XmNsingleSelectionCallback, selectItCB, (XtPointer) 0);
 XtVaSetValues(listW, XmNselectionPolicy, XmSINGLE_SELECT, NULL);

 numPrinters = initPrintLib(pathCommon);
 for (i=0; i<numPrinters; i++) 
  {
   selectPrinter(i);  
   xms = XmStringCreateSimple(printerName);
   XmListAddItem(listW, xms, 0);
   XmStringFree(xms);
  }
 selectPrinter(returnMatchingPrinter(defaultPrinter));
 selectMode(0); selectDensityX(0); selectDensityY(0);
}

/* returns localized string containing current printer settings */
DPRINT char *returnPrtSettingText(void)
{
 static char setting[1000];

 if (dpivariable)
  {
   xdpi = 1000*getIntFromWidget(printXdpiSelBW, 10, 3000);
   ydpi = 1000*getIntFromWidget(printYdpiSelBW, 10, 3000);
  }
 sprintf(setting, "%s: %s\n%s: %s\n%s: %d\n%s: %d\n", prtTxt[PTXT_PRINTER], printerName,
  prtTxt[PTXT_PRINT_MODE], prtMode2string(prtModes[selectedMode].printType),
  prtTxt[PTXT_DENSITY_X], xdpi/1000, prtTxt[PTXT_DENSITY_Y], ydpi/1000);

 if (printToFile)
  {
   char tmpstr[300];

   if (printToFilename)
    XtFree(printToFilename);
   printToFilename = XmTextFieldGetString(printToFileIW);
   if (!strlen(printToFilename)) /* do not allow  empty filenames */
    {
     XtFree(printToFilename);
     printToFilename = strdup(getTxtI8N(NEW_PICTURE_MSG));
    }
   sprintf(tmpstr, "%s %s%s",  prtTxt[PTXT_PRINT_TO_FILE],
    (selectedPrintDst) ? ((selectedPrintDst==1) ? "B:" : "/tmp/") : "A:",
    printToFilename);
   strcat(setting, tmpstr);
  }
 else
  strcat(setting, prtTxt[PTXT_PAPER_WARNING]);

 return(setting);
}

/* select printer x density */
static void selectDensityX(int dens)
{
 char numString[20];

 if (dens >= numDensitX)
   return;
 maxxpixels = prtDensityX[dens].pixels;
 xdpi = prtDensityX[dens].dpi;
 sprintf(numString, "%d", xdpi/1000);
 XmTextFieldSetString(XmSelectionBoxGetChild(printXdpiSelBW, XmDIALOG_TEXT), numString);
 XmListSelectPos(XmSelectionBoxGetChild(printXdpiSelBW, XmDIALOG_LIST), dens + 1, FALSE);
}

/* select printer y density */
static void selectDensityY(int dens)
{
 char numString[20];

 if (dens >= numDensitY)
   return;
 maxypixels = prtDensityY[dens].pixels;
 ydpi = prtDensityY[dens].dpi;
 sprintf(numString, "%d", ydpi/1000);
 XmTextFieldSetString(XmSelectionBoxGetChild(printYdpiSelBW, XmDIALOG_TEXT), numString);
 XmListSelectPos(XmSelectionBoxGetChild(printYdpiSelBW, XmDIALOG_LIST), dens + 1, FALSE);
}

/* select printer mode */
DPRINT void selectMode(int mode)
{
 int i;

 if (mode >= numModes)
   return;
 selectedMode = mode;
 printType = prtModes[mode].printType;

 XmTextFieldSetString(XmSelectionBoxGetChild(printModeSelBW, XmDIALOG_TEXT),
  prtMode2string(prtModes[mode].printType));
 XmListSelectPos(XmSelectionBoxGetChild(printModeSelBW, XmDIALOG_LIST), mode+1, FALSE);

 i = numDensitX - 1;
 while ((i > 0) && (xdpi != prtDensityX[i].dpi))
   i--;
 selectDensityX(i);
 i = numDensitY - 1;
 while ((i > 0) && (ydpi != prtDensityY[i].dpi))
   i--;
 selectDensityY(i);
}

/* select printer */
static void selectPrinter(int printer)
{
 int i;
 Widget listW, textW;
 XmString xms;
 char numStr[20];

 if (printer >= numPrinters)
  return;
 setPrinterFunctions(printer);
 getPrinterInfo(prtModes, &numModes, prtDensityX, &numDensitX, prtDensityY, &numDensitY,
  &nonprintableTop, &nonprintableBottom, &nonprintableLeft, &dpivariable, &printerName, &profileName);
 if (!savingImage)
  {
   selectedPrinter = printer;
   XmTextFieldSetString(XmSelectionBoxGetChild(printerSelBW, XmDIALOG_TEXT), printerName);

   listW = XmSelectionBoxGetChild(printModeSelBW, XmDIALOG_LIST);
   XmListDeleteAllItems(listW);
   for (i = 0; i < numModes; i++)
    {
     xms = XmStringCreateSimple(prtMode2string(prtModes[i].printType));
     XmListAddItem(listW, xms, 0);
     XmStringFree(xms);
    }
   listW = XmSelectionBoxGetChild(printXdpiSelBW, XmDIALOG_LIST);
   XmListDeleteAllItems(listW);
   for (i = 0; i < numDensitX; i++)
    {
     sprintf(numStr, "%d", prtDensityX[i].dpi/1000);
     xms = XmStringCreateSimple(numStr);
     XmListAddItem(listW, xms, 0);
     XmStringFree(xms);
    }
   listW = XmSelectionBoxGetChild(printYdpiSelBW, XmDIALOG_LIST);
   XmListDeleteAllItems(listW);
   for (i = 0; i < numDensitY; i++)
    {
     sprintf(numStr, "%d", prtDensityY[i].dpi/1000);
     xms = XmStringCreateSimple(numStr);
     XmListAddItem(listW, xms, 0);
     XmStringFree(xms);
    }
   textW = XmSelectionBoxGetChild(printXdpiSelBW, XmDIALOG_TEXT);
   XtVaSetValues(textW, XmNsensitive, dpivariable, NULL);
   textW = XmSelectionBoxGetChild(printYdpiSelBW, XmDIALOG_TEXT);
   XtVaSetValues(textW, XmNsensitive, dpivariable, NULL);
   if (dpivariable != printToFile)
    setPrintToFileCB(textW, "0", NULL);
   setSensitive(printToFileTBW, !dpivariable);

   i = numModes - 1;
   while ((i > 0) && (printType != prtModes[i].printType))
     i--;
   selectMode(i);
  }
}

/* get the printable limits of current printer */
DPRINT void getPrinterLimits(int *maxxpix, int *maxypix, int *prtxdpi,
 int *prtydpi, int *nonprtTop, int *nonprtBottom, int *nonprtLeft)
{
 *maxxpix = maxxpixels;
 *maxypix = maxypixels;
 *prtxdpi = xdpi;
 *prtydpi = ydpi;
 *nonprtTop = nonprintableTop;
 *nonprtBottom = nonprintableBottom;
 *nonprtLeft = nonprintableLeft;
}

/* set page settings */
DPRINT void setPageSettings(int pxpixels, int pypixels,
 int pxmargin, int pymargin, int pcomplexImage)
{
 xpixels = pxpixels;
 ypixels = pypixels;
 xmargin = pxmargin;
 ymargin = pymargin;
 complexImage = pcomplexImage;
}

/* registers printer with printer driver */
void registerPrinter(getPrinterInfoProc getPrtInfo, initPrintProc initPrt,
 printBandProc prtBand, endPrintProc endPrt)
{
 getPrinterInfo = getPrtInfo;
 initPrint = initPrt;
 printBand = prtBand;
 endPrint = endPrt;
}

/* user wishes to cancel the print */
DPRINT void cancelPrintCB(Widget w, XtPointer clientData, XtPointer callData)
{
 prtStatus = PRT_USER_ABORT;
}

static void dpyFinalInfo(int prtStatus)
{
 if (prtStatus <= PRT_USER_ABORT)
  {
   if ((printToFile) && (prtStatus == PRT_OK))
     manageInfoDialog(prtTxt[PTXT_PRINT_TO_FILE_OK]);
   else
     manageInfoDialog(prtTxt[prtStatus]);
  }
 else
  manageErrorDialog(prtTxt[prtStatus], NULL, NULL, getTxtI8N(OK_MSG));
}

static void processBand(XtPointer position)
{
 int ypos;
 char *data = NULL;
 int lines, bpl;
 char returnImageType;

 ypos = (int) position;

 XmScaleSetValue(printProgressSW, (ypos*100)/ypixels);

 if ((prtStatus != PRT_OK) || (ypos >= ypixels))
  {
   XtUnmanageChild(printingTW);
   prtStatus = endPrint(prtStatus);
   if ((prtStatus != PRT_OK) || (!savingImage))
    dpyFinalInfo(prtStatus);
   if (savingImage)
    initendSaveImage(FALSE, NULL);
   nowPrinting = FALSE;
   return;
  }
 if (ypos < ypixels)	/* OK for 1 and 2 prints */
   prtStatus = getNextBand(xpixels, xdpi, ydpi, ypos, ypixels,
    printType, &data, &bpl, &lines, &returnImageType, cmap);

 if (returnImageType > printType)
   prtStatus = PRT_WRONG_IMG;
 if (prtStatus == PRT_OK)
   prtStatus = printBand(xpixels, data, bpl, ypos, lines, returnImageType, cmap);

 if ((data) && (!savingImage))
   free(data);

 ypos += lines;
 XtAddTimeOut(adaptiveTimeout(), (XtTimerCallbackProc) *processBand, (XtPointer) ypos);
}

static int prtToFileOK(char *pfile)
{
 if ((printToFile) || (savingImage))
  if (access(pfile, W_OK) != 0)
    {
     FILE *testf;

     if ((testf = fopen(pfile, "w")) == NULL)
      {
       manageErrorDialog(prtTxt[PRT_IO_ERROR], NULL, NULL, getTxtI8N(OK_MSG));
       return(FALSE);
      }
     fclose(testf);
    }
 return(TRUE);
}

/* prints the page according to previous settings */
DPRINT void printPage(getNextBandProc getNextBandP)
{
 getNextBand = getNextBandP;
 if (!savingImage)
  {
   setLabelString(printingInfoLW, prtTxt[PTXT_PRINTING_WAIT]);
   setLabelString(printCancelW, prtTxt[PTXT_CANCEL_PRINT]);
   setDialogTitle(printingTW, prtTxt[PTXT_WAIT]);
  }
 if (printToFile)
  if (!prtToFileOK(printToFilename))
   return;
 XmScaleSetValue(printProgressSW, 0);
 /* only one print job at a time */
 if (nowPrinting)
  return;
 nowPrinting = TRUE;
 prtStatus = initPrint(printType, complexImage, xpixels, ypixels, xdpi, ydpi, xmargin, ymargin);
 if (prtStatus == PRT_OK)
  if (xpixels * ypixels < 640000) /* do images below 800x800 in one pass */
   processBand(0);
  else
   { /* give it a little more time to pop up */
    XtAddTimeOut(adaptiveTimeout() * 10, (XtTimerCallbackProc) *processBand, 0);
    XtManageChild(printingTW);
   }
 else
   { endPrint(prtStatus); nowPrinting = FALSE; }
}

/* put the file in the print queue of the currently selected printer or copy it to */
/* the selected destination */
DPRINT int enqueuePrintFile(char *fn)
{
 int status, copies;
 char command[500], tmp[500];

 XtVaGetValues(copiesW, XmNvalue, &copies, NULL);
 if ( ((!savingImage) && printToFile) || (savingImage && (selectedPrintDst!=FILE_TO_OTHER)) )
  {
   sprintf(tmp, "/tmp/%s", (savingImage) ? baseName(saveFname) : printToFilename);
   if (selectedPrintDst==FILE_TO_TMP)
    return((rename(fn, tmp)==0) ? PRT_OK : PRT_IO_ERROR);
   else /* (selectedPrintDst==FILE_TO_A) || (selectedPrintDst==FILE_TO_B) */
    {
     sprintf(command, "%s \"%s\" %c:\"%s\"", copy2dosCommand, fn,
      (selectedPrintDst==FILE_TO_A) ? 'a' : 'b', baseName(tmp));
     doRemove=TRUE;   
    }
  }
 else
  { 
   if (savingImage)
    {
     if (rename(fn, saveFname)==0)
      return(PRT_OK);
     else /* rename fails across filesystems, so use mv command */
      sprintf(command, "/bin/mv -f \"%s\" \"%s\"", fn, saveFname);
    }
   else
    {
     sprintf(tmp, prtCommand, copies);
     sprintf(command, "%s \"%s\"", tmp, fn);
     doRemove=TRUE;
    }
  }

 /* if we did not return till now, we need to execute this command */
 status = executeCommand(command);
 if (status<0)
  return(PRT_INTERNAL_ERROR);
 else
  return((status==0) ? PRT_IO_ERROR : PRT_OK);
}

/* cleanup of temporary files */
DPRINT void removeFileFreeFilename(char *filename)
{
 if ((!savingImage) || (savingImage && (prtStatus!=PRT_OK)))
  if (access(filename, F_OK) == 0)
   {
    if (doRemove)
     remove(filename);
    free(filename);
   }
}

/* decompose XImage for saving */
static int returnImg(int xpixels, int xdpi, int ydpi,
 int ypos, int ytotal, char desiredImageType, char **data,
 int *bpl, int *lines, char *returnImageType, unsigned char *cmap)
{
 if ((desiredImageType == IMG_BW) || (saveXIM == NULL))
  return(PRT_WRONG_IMG);
 *data = &saveXIM->data[saveXIM->bytes_per_line*ypos];
 *bpl = saveXIM->bytes_per_line;
 *lines = min(ytotal-ypos, 200000/xpixels);
 *returnImageType = IMG_CMAP;
 /* and colormap has already been filled with the right values */
 return(PRT_OK);
}

/* selects saving to image in the passed format, or reverts to previous state */
static void initendSaveImage(int init, char *imgFormat)
{
 static int oldPrintType, oldPrintDst;
 static int oldXdpi, oldYdpi;

 if (init)
  {
   char setting[200];
   
   oldPrintDst = selectedPrintDst;
   selectedPrintDst = imagePrintDst;
   oldPrintType = printType;
   oldXdpi = xdpi;
   oldYdpi = ydpi;
   selectPrinter(returnMatchingPrinter(imgFormat));
   if (saveXIM->depth==1)
    printType = IMG_BW;
   else
    printType = IMG_CMAP;
   setPageSettings(saveXIM->width, saveXIM->height, 0, 0, FALSE);
   sprintf(setting, "%s\n%d x %d", prtTxt[PTXT_SAVING_WAIT],
    saveXIM->width, saveXIM->height);
   setLabelString(printingInfoLW, setting);
   setLabelString(printCancelW, prtTxt[PTXT_CANCEL_SAVE]);
   xdpi =  (int) (ApData.imageXdpi*1000);
   ydpi =  (int) (ApData.imageYdpi*1000);
  }
 else
  {
   if (prtStatus==PRT_OK)
    chmod(saveFname, 00666);
   if (freeSavedImage)
    undoXImage(&saveXIM);
   saveXIM=NULL;
   free(saveFname);
   /* unregistring temporary settings that were used in saving */
   /* to get in sync with the settings in interface */
   xdpi = oldXdpi;
   ydpi = oldYdpi;

   selectedPrintDst = oldPrintDst;
   selectPrinter(selectedPrinter);
   printType = oldPrintType;
   freeSavedImage = savingImage = FALSE;
   setLabelString(printingInfoLW, prtTxt[PTXT_PRINTING_WAIT]);
   setLabelString(printCancelW, prtTxt[PTXT_CANCEL_PRINT]);
  }
}

/* save an image in the passed format with print interface; return false if file not writeable */
DPRINT int saveImage(char *fname, XImage *xim, char *imgFormat, unsigned char *imgcmap,
 int freeWhenOver, int destination)
{
 savingImage = TRUE;
 imagePrintDst = destination;
 if (!prtToFileOK(fname))
  return(FALSE);
 freeSavedImage = freeWhenOver; 
 saveFname = strdup(fname);
 memcpy(cmap, imgcmap, 256*3); /* copy the colormap to my place */
 saveXIM=xim;
 initendSaveImage(TRUE, imgFormat);
 printPage(*returnImg);
 return(TRUE);
}
