#define SELECT
#include <Xm/Xm.h>
#include <Xm/ToggleB.h>
#include <stdlib.h>
#include "paint.h"

#define minWidth 2
#define minHeight 2

static Rectangle last_r;
static Boolean SelectPolygon;
static Boolean firstClick;

static void makeSelection(Widget w, XtPointer clientData, XEvent *event);
static void makePolygonSelection(Widget w, XtPointer clientData, XEvent *event);
static void HandlePasteMotion(Widget w, XtPointer clientData, XEvent *event);
static void resizeSelection(Widget w, XtPointer clientData, XEvent * event);
static int countSelectionColors();

#define SELECT_RECT 0
#define SELECT_POLYGON 1
#define SELECT_ALL 2
#define UNSELECT 3
/* selection button callback */
DSELECT void selectCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case SELECT_POLYGON:
    ApData.tool = selectA;
    SelectPolygon = False;
    setToolEventHandler(selectPolygonA, 0);
    SelectPolygon = True;
    ApData.tool = selectPolygonA;
    highlightButton(w);
    break;
   case SELECT_ALL:
    ApData.tool = selectA;
    SelectPolygon = False;
    finalDraw(ApData.pixy);
    disableSelection(True);

    Select.x = Select.y = 0;
    Select.width = ApData.width;
    Select.height = ApData.height;
    Select.state = undef;

    setBoundaryPtsR(&Select.npoints, Select.x, Select.y,
	Select.width, Select.height);

    Select.rect_on = Select.selected = True;
    Select.rectangular = True;

    drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
    drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW), zoomDraw.factorX,
	zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
    disableUndo();
    ApData.tool = Undo.tool = selectA;
    XtVaSetValues(CopyBW, XmNsensitive, True, NULL);
    XtVaSetValues(CutBW, XmNsensitive, True, NULL);

    setToolEventHandler(selectA, 0);
    highlightButton(NULL);
    setButtonIconCB(unselectBW, "unselect.xbm", NULL);
    break;
   case UNSELECT:
    finalDraw(ApData.pixy);
    opBounds.x = Select.x - 2;
    opBounds.y = Select.y - 2;
    opBounds.width = Select.width + 4;
    opBounds.height = Select.height + 4;
    disableSelection(True);
    disableUndo();
    refreshDrawingArea(False, False);
    refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
    setToolEventHandler(none, 0);
    break;
   case SELECT_RECT:
    ApData.tool = selectA;
    SelectPolygon = False;
    setToolEventHandler(selectA, 0);
    highlightButton(w);
    break;
  }
 changeTitleCB (NULL, "", NULL); /* to display the selection parameters in window title */
}

/* move selection rectangle around; if floating, move the whole block, if still in image */
/* move the selection boundaries */
DSELECT void moveSelection(int changeX, int changeY)
{
 if (!select)
  return;

 drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
 drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW), zoomDraw.factorX,
  zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
 if (Select.rectangular)
  {
   Select.npoints=0; 
   Select.x += changeX;
   Select.y += changeY;
   if ((!Select.imgAllocated)&&(ApData.viewMode == normalView))
    {
     if ((Select.x<0) || (Select.x+Select.width>ApData.width))
      Select.x -= changeX; /* then undo the change */
     if ((Select.y<0) || (Select.y+Select.height>ApData.height))
      Select.y -= changeY; /* then undo the change */
    }
   setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);
  }
 else
  {
   movePts(&Select.points[0], Select.npoints, changeX, changeY);
   if (Select.reg)
    XDestroyRegion(Select.reg);
   Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
  }
 if (Select.imgAllocated)
  {
   refreshDrawingArea(True, True);
   if (widgetIsViewable(zoomDrawAreaW))
    {
     refreshZoomWin(zoomDraw.clipWin11.x-2, zoomDraw.clipWin11.y-2,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
     finalDraw(ApData.scrPix);
     refreshZoomWin(zoomDraw.clipWin11.x-2, zoomDraw.clipWin11.y-2,
      zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, False);
    }
  }
 else
  {
   refreshDrawingArea(True, True);
   refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
    zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);  
  }
}

/* Button Press event handler for making selection */
DSELECT void SelectBtnPress(Widget w, XtPointer clientData, XEvent * event)
{
 if (event->xbutton.button == Button1)
  {
   int zmXF = 1, zmYF = 1, zoX = 0, zoY = 0;
   int drawZoom = FALSE;

   if ((int) clientData != 0)
    {
     zmXF = zoomDraw.factorX;
     zmYF = zoomDraw.factorY;
     zoX = zoomDraw.clipWin.x;
     zoY = zoomDraw.clipWin.y;
     drawZoom=TRUE;
    }
   ChangeCursor(event->xbutton.window, XC_hand2, True);
   finalDraw(ApData.pixy);
   disableUndo();
   XSetFunction(xinfo.dpy, Select.gc, GXinvert);
   if (Select.reg)
    {
     XDestroyRegion(Select.reg);
     Select.reg = NULL;
     XSetClipMask(xinfo.dpy, Select.moveGC, None);
    }

   Undo.tool = ApData.tool = selectA;
   disableSelection(True);
   Undo.points[0].x = Select.x = ApData.orig_xy.x = (event->xbutton.x + zoX)/zmXF;
   Undo.points[0].y = Select.y = ApData.orig_xy.y = (event->xbutton.y + zoY)/zmYF;
   Undo.points[1].x = Undo.points[1].y = Select.width = Select.height = 0;
   Undo.npoints = 2;

   if (drawZoom)
    {
     XGCValues values;
     char dash_list[2];

     values.line_width = max(2, (zmXF + zmYF)/3);
     dash_list[0] = dash_list[1] = 2 * values.line_width;
     XChangeGC (xinfo.dpy, Select.gc, GCLineWidth, &values);
     XSetDashes (xinfo.dpy, Select.gc, 0, dash_list, 2);
    }
   if (SelectPolygon)
    {
     Select.points[0].x = Select.points[1].x = Undo.points[0].x;
     Select.points[0].y = Select.points[1].y = Undo.points[0].y;
     Select.npoints = 2;
     setExclEventHandler((XtEventHandler) makePolygonSelection, PointerMotionMask | ButtonReleaseMask, True);
    }
   else
    setExclEventHandler((XtEventHandler) makeSelection, PointerMotionMask | ButtonReleaseMask, True);
  }
}

/*set cursor shape inside selection area and set selection state accordingly */
static void SetSelCursor(Window win, int x, int y)
{
 int state = 0;
 unsigned int shape = XC_fleur;
 SelectState oldstate;
 int resizeBorderW, resizeBorderH;


 if (win==XtWindow(zoomDrawAreaW))
  {
   resizeBorderW = max(1, 5/ApData.zoomX);
   resizeBorderH = max(1, 5/ApData.zoomY);
  }
 else
  {
   resizeBorderW = resizeBorderH = 5;
  }
 oldstate = Select.state;
 Select.state = move;

 if (Select.rectangular)
  {
   if ((x < Select.x) || (x > (Select.x + Select.width)) ||
       (y < Select.y) || (y > (Select.y + Select.height)))
    state = 32;
   else
    {
     if (abs (x - Select.x) < resizeBorderW)
       state = state | 1;
     if (abs (y - Select.y) < resizeBorderH)
       state = state | 4;
     if (abs (Select.x + Select.width - x) <= resizeBorderW)
       state = state | 8;
     if (abs (Select.y + Select.height - y) <= resizeBorderH)
       state = state | 16;
    }
  }
 else if (XPointInRegion (Select.reg, x, y))
   Select.state = move;
 else
   state = 32;
 if (state != 0)
   switch (state)
    {
     case 1:
       shape = XC_left_side;
       Select.state = resizeLeft;
       break;
     case 4:
       shape = XC_top_side;
       Select.state = resizeUp;
       break;
     case 8:
       shape = XC_right_side;
       Select.state = resizeRight;
       break;
     case 16:
       shape = XC_bottom_side;
       Select.state = resizeDown;
       break;
     case 5:
       shape = XC_top_left_corner;
       Select.state = resizeUpLeft;
       break;
     case 17:
       shape = XC_bottom_left_corner;
       Select.state = resizeDownLeft;
       break;
     case 12:
       shape = XC_top_right_corner;
       Select.state = resizeUpRight;
       break;
     case 24:
       shape = XC_bottom_right_corner;
       Select.state = resizeDownRight;
       break;
     case 32:
       Select.state = undef;
       shape = XC_left_ptr;
       break;
    }
 if (oldstate != Select.state)
   ChangeCursor (win, shape, True);
}

/* set resizing directions */
static void setResizeDirections()
{
 Select.resDown = Select.resUp = Select.resLeft = Select.resRight = Select.preserveAspectR = False;
 switch (Select.state)
  {
   case resizeLeft: Select.resLeft = True; break;
   case resizeRight: Select.resRight = True; break;
   case resizeUp: Select.resUp = True; break;
   case resizeDown: Select.resDown = True; break;
   case resizeUpLeft: Select.resUp = Select.resLeft = Select.preserveAspectR = True; break;
   case resizeUpRight: Select.resUp = Select.resRight = Select.preserveAspectR = True; break;
   case resizeDownLeft: Select.resLeft = Select.resDown = Select.preserveAspectR = True; break;
   case resizeDownRight: Select.resRight = Select.resDown = Select.preserveAspectR = True; break;
   default: break;
  }
}

/* check motions and actions inside selection */
DSELECT void handleSelection(Widget w, XtPointer clientData, XEvent * event)
{
 int zmXF = (int) clientData, zmYF = 1, x, y, zoX, zoY;

 if (zmXF)
  {
   zmXF = zoomDraw.factorX;
   zmYF = zoomDraw.factorY;
   zoX = zoomDraw.clipWin.x;
   zoY = zoomDraw.clipWin.y;
  }
 else
  {
   zmXF = zmYF;
   zoX = zoY = 0;
  }
 x = (event->xbutton.x + zoX) / zmXF;
 y = (event->xbutton.y + zoY) / zmYF;

 if ((x >= Select.x) && (x <= (Select.x + Select.width)) &&
     (y >= Select.y) && (y <= (Select.y + Select.height)) &&
     (Select.width >= minWidth) && (Select.height >= minHeight))
   {
     switch (event->type)
       {
       case MotionNotify:
	 SetSelCursor (event->xmotion.window, x, y);
	 break;
       case ButtonPress:
	 {
	   setExclEventHandler (NULL, 0, True);	/* remove EH */
	   switch (Select.state)
	     {
	     case move:
	       if (Undo.tool == pasteMulA)
		 finalDraw (ApData.pixy);
	       disableUndo ();
	       if (!Select.imgAllocated)
		 {
		   Select.img = MyGetImage(ApData.img,
		    Select.x, Select.y, Select.width, Select.height, ApData.viewMode);

		   if (!checkXImage (Select.img))
		     {
		       setExclEventHandler ((XtEventHandler) handleSelection,
			       PointerMotionMask | ButtonPressMask, False);
		       return;
		     }
		   Select.imgAllocated = True;

		   XtVaSetValues (PasteBW, XmNsensitive, True, NULL);
		   enableUndo (True);
		 }
	       drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);

	       opBounds.width = Select.width;
	       opBounds.height = Select.height;
	       opBounds.x = Select.x;
	       opBounds.y = Select.y;

	       last_r.x = x;
	       last_r.y = y;
	       last_r.width = (unsigned int) Select.width;
	       last_r.height = (unsigned int) Select.height;
	       /*store original coordinates for calculation of offset */
	       Select.points[Select.npoints].x = (short) Select.x;
	       Select.points[Select.npoints].y = (short) Select.y;
	       ApData.tool = none;
	       LastDrawOnAllWin(False);

	       drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW),
		zoomDraw.factorX, zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
	       ApData.tool = moveA;

	       XSetFunction(xinfo.dpy, Select.gc, GXcopy);
	       ChangeCursor(XtWindow(w), XC_fleur, True);
	       getClipWinDims (clipWin, scrollBX, scrollBY, ApData.DAwidth,
			       ApData.DAheight, &ApData.clipWin);

	       firstClick = True;

	       setExclEventHandler((XtEventHandler) HandlePasteMotion, Button1MotionMask | ButtonReleaseMask, True);
	       break;
	     case undef:
	       setExclEventHandler((XtEventHandler) handleSelection, PointerMotionMask | ButtonPressMask, True);
	       break;
	     default:
	       /* before resizing draw permanently the last operation */
	       if (event->xbutton.button == Button1)
		 {
		   finalDraw (ApData.pixy);
		   drawSelectionBoundaries(XtWindow (w), zmXF, zmYF, zoX, zoY);
		 }
	       else if (XtWindow (w) != drawAreaWin)
		 drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);

	       /*resize make two copies of the selection parms */
	       Undo.npoints = 0;
	       Undo.points[Undo.npoints].x = (short) Select.x;
	       Undo.points[Undo.npoints++].y = (short) Select.y;
	       Undo.points[Undo.npoints].x = (short) Select.width;
	       Undo.points[Undo.npoints++].y = (short) Select.height;

	       ApData.orig_xy.x = (short) Select.x;
	       ApData.orig_xy.y = (short) Select.y;

	       Undo.points[Undo.npoints].x = (short) Select.x;	/* boundary */
	       Undo.points[Undo.npoints++].y = (short) Select.y;
	       Undo.points[Undo.npoints].x = (short) Select.width;
	       Undo.points[Undo.npoints].y = (short) Select.height;

	       switch (Select.state)
		 {
		 case resizeLeft:
		 case resizeUp:
		 case resizeUpLeft:
		   ApData.orig_xy.x += Select.width;
		   ApData.orig_xy.y += Select.height;
		   Undo.points[Undo.npoints].x *= -1;
		   Undo.points[Undo.npoints].y *= -1;
		   break;

		 case resizeUpRight:
		   ApData.orig_xy.y += Select.height;
		   Undo.points[Undo.npoints].y *= -1;
		   break;

		 case resizeDownLeft:
		   ApData.orig_xy.x += Select.width;
		   Undo.points[Undo.npoints].x *= -1;
		   break;
		 default:
		   break;
		 }
	       Undo.npoints++;
	       setResizeDirections ();
	       XSetFunction (xinfo.dpy, Select.gc, GXinvert);
	       if (XtWindow (w) != drawAreaWin)
		 {
		  char dash_list[2];
		  XGCValues values;

		  values.line_width = max(2, (zoomDraw.factorX + zoomDraw.factorY)/3);
		  dash_list[0] = dash_list[1] = 2 * values.line_width;
		  XChangeGC(xinfo.dpy, Select.gc, GCLineWidth, &values);
		  XSetDashes(xinfo.dpy, Select.gc, 0, dash_list, 2);
		 }
	       if (event->xbutton.button == Button1)
		 Select.preserveAspectR = False;
	       setExclEventHandler((XtEventHandler) resizeSelection, PointerMotionMask | ButtonReleaseMask, True);
	       break;
	     }
	 }
       }
   }
 else /* mouse event outside selection area */
   {
    if (ApData.cursorShape != XC_left_ptr)
     ChangeCursor (XtWindow (w), XC_left_ptr, True);
    Select.state = undef;
   }
}

/* event handler for selecting rectangular areas */
static void makeSelection(Widget w, XtPointer clientData, XEvent * event)
{
  int dx, dy, xl, yt, np, cd = (int) clientData, zmXF = 1, zmYF = 1, zoX = 0,
    zoY = 0;
  int drawZoom = FALSE;

  if (cd == 1)
    {
     zmXF = zoomDraw.factorX;
     zmYF = zoomDraw.factorY;
     zoX = zoomDraw.clipWin.x;
     zoY = zoomDraw.clipWin.y;
     drawZoom = TRUE;
    }

  switch (event->type)
    {
    case ButtonRelease:
      if (Select.imgAllocated)
	{
	 undoXImage(&Select.img);
	 Select.imgAllocated = False;
	}
      XtVaSetValues(PasteBW, XmNsensitive, False, NULL);
      XtVaSetValues(CopyBW, XmNsensitive, False, NULL);
      XtVaSetValues(CutBW, XmNsensitive, False, NULL);

      ChangeCursor (event->xbutton.window, XC_left_ptr, True);
      setExclEventHandler (NULL, 0, True);

      Undo.tool = ApData.tool;
      memcpy (&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));

      /*check width + x, height + y < Apdata.width, height */
      np = Undo.npoints - 2;
      XDrawRectangle(xinfo.dpy, event->xbutton.window, Select.gc,
	Undo.points[np].x * zmXF - zoX, Undo.points[np].y * zmYF - zoY,
	Undo.points[np + 1].x * zmXF, Undo.points[np + 1].y * zmYF);

      if (Undo.points[np].x + Undo.points[np + 1].x > ApData.DAwidth)
	Undo.points[np + 1].x = ApData.DAwidth - Undo.points[np].x;
      if (Undo.points[np].y + Undo.points[np + 1].y > ApData.DAheight)
	Undo.points[np + 1].y = ApData.DAheight - Undo.points[np].y;

      if (Undo.points[np].x < 0)
	{
	  Undo.points[np + 1].x += Undo.points[np].x;
	  Undo.points[np].x = 0;
	}
      if (Undo.points[np].y < 0)
	{
	  Undo.points[np + 1].y += Undo.points[np].y;
	  Undo.points[np].y = 0;
	}
      if (drawZoom)
	{
	  XGCValues values;
	  char dash_list[2];

	  values.line_width = 2;
	  dash_list[0] = dash_list[1] = 4;
	  XChangeGC (xinfo.dpy, Select.gc, GCLineWidth, &values);
	  XSetDashes (xinfo.dpy, Select.gc, 0, dash_list, 2);
	}
      if ((Undo.points[np + 1].x >= minWidth) ||
	  (Undo.points[np + 1].y >= minHeight))
	{
	  int ind;
	  XPoint orR, *pt;

	  pt = &Undo.points[Undo.npoints - 2];
	  ind = pointInRepeat (pt->x, pt->y, repeats.currNum);
	  if (ind == -1)
	    return;
	  orR.x = repeats.crop.width;
	  orR.y = repeats.crop.height;
	  pt[1].x += pt[0].x;
	  pt[1].y += pt[0].y;
	  movePts (pt, 2, -repeats.r[ind].x, -repeats.r[ind].y);
	  flipPts (pt, 2, repeats.r[ind].rType, &orR);
	  findMinMax (pt, 2, &pt[0].x, &pt[0].y, &pt[1].x, &pt[1].y);
	  pt[1].x -= pt[0].x;
	  pt[1].y -= pt[0].y;

	  Select.npoints = 0;
	  Select.selected = Select.rect_on = True;
	  Select.x = pt[0].x;
	  Select.y = pt[0].y;
	  Select.width = pt[1].x;
	  Select.height = pt[1].y;
	  Select.rectangular = True;
	  Select.aspectR = 1.0 * Select.width / Select.height;

	  setBoundaryPtsR(&Select.npoints, pt[0].x, pt[0].y, pt[1].x, pt[1].y);

	  drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
	  drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW), zoomDraw.factorX,
		zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);

	  XtVaSetValues(CopyBW, XmNsensitive, True, NULL);
	  XtVaSetValues(CutBW, XmNsensitive, True, NULL);
	  enableUndo(True);

	  highlightButton(NULL);
	  Select.state = undef;
	  setExclEventHandler((XtEventHandler) handleSelection,
	   PointerMotionMask | ButtonPressMask, True);
	  setButtonIconCB(unselectBW, "unselect.xbm", NULL);
	}
      else /* to small selection */
	setExclEventHandler((XtEventHandler) SelectBtnPress, ButtonPressMask, False);
      break;

    case MotionNotify:
      np = Undo.npoints - 2;
      XDrawRectangle(xinfo.dpy, event->xbutton.window, Select.gc,
	Undo.points[np].x * zmXF - zoX, Undo.points[np].y * zmYF - zoY,
	Undo.points[np + 1].x * zmXF, Undo.points[np + 1].y * zmYF);
      dx = ApData.orig_xy.x - (event->xbutton.x + zoX) / zmXF;
      dy = ApData.orig_xy.y - (event->xbutton.y + zoY) / zmYF;
      xl = ApData.orig_xy.x;
      yt = ApData.orig_xy.y;
      if (dx > 0)
	xl -= dx;
      if (dy > 0)
	yt -= dy;
      XDrawRectangle(xinfo.dpy, event->xbutton.window, Select.gc,
       xl * zmXF - zoX, yt * zmYF - zoY, abs(dx) * zmXF, abs(dy) * zmYF);
      Undo.points[np].x = (short) xl;
      Undo.points[np].y = (short) yt;
      Undo.points[np + 1].x = (short) abs(dx);
      Undo.points[np + 1].y = (short) abs(dy);
      break;
    }
 changeTitleCB(NULL, "", NULL); /* to display the selection parameters in window title */
}

/*event handler for polygon selecting */
static void makePolygonSelection(Widget w, XtPointer clientData, XEvent * event)
{
 int i, zmXF = 1, zmYF = 1, zoX = 0, zoY = 0;
 XRectangle rect;
 int ind;
 XPoint orR;
 int drawZoom = FALSE;

 if ((int) clientData != 0)
  {
   zmXF = zoomDraw.factorX;
   zmYF = zoomDraw.factorY;
   zoX = zoomDraw.clipWin.x;
   zoY = zoomDraw.clipWin.y;
   drawZoom = TRUE;
  }

 switch (event->type)
  {
   case ButtonRelease:
    if (Select.imgAllocated)
     {
      undoXImage (&Select.img);
      Select.imgAllocated = False;
     }
    XtVaSetValues(PasteBW, XmNsensitive, False, NULL);
    XtVaSetValues(CopyBW, XmNsensitive, False, NULL);
    XtVaSetValues(CutBW, XmNsensitive, False, NULL);
    ChangeCursor(event->xbutton.window, XC_left_ptr, True);
    setExclEventHandler (NULL, 0, True);
    Undo.tool = ApData.tool;
    memcpy(&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));
    for (i = Select.npoints - 1; i > 0; i--)
     XDrawLine(xinfo.dpy, event->xbutton.window, Select.gc,
      Select.points[i - 1].x * zmXF - zoX, Select.points[i - 1].y * zmYF - zoY,
      Select.points[i].x * zmXF - zoX, Select.points[i].y * zmYF - zoY);
    /*check x,y < 0 & x,y < width, height */
    for (i = 0; i < Select.npoints; i++)
     {
      if (Select.points[i].x < 0)
       Select.points[i].x = 0;
      if (Select.points[i].y < 0)
       Select.points[i].y = 0;
      if (Select.points[i].x > ApData.DAwidth)
       Select.points[i].x = (short) ApData.DAwidth;
      if (Select.points[i].y > ApData.DAheight)
       Select.points[i].y = (short) ApData.DAheight;
     }
    if ((Select.points[Select.npoints - 1].x != Select.points[0].x)
	|| (Select.points[Select.npoints - 1].x != Select.points[0].x))
     {
      Select.points[Select.npoints].x = Select.points[0].x;
      Select.points[Select.npoints++].y = Select.points[0].y;
     }

    ind = pointInRepeat (Select.points[0].x, Select.points[0].y, repeats.currNum);
    if (ind == -1)
     return;
    orR.x = repeats.crop.width;
    orR.y = repeats.crop.height;
    movePts (&Select.points[0], Select.npoints, -repeats.r[ind].x, -repeats.r[ind].y);
    flipPts (&Select.points[0], Select.npoints, repeats.r[ind].rType, &orR);

    Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
    XClipBox(Select.reg, &rect);

    if (drawZoom)
      {
       XGCValues values;
       char dash_list[2];

       values.line_width = 2;
       dash_list[0] = dash_list[1] = 4;
       XChangeGC(xinfo.dpy, Select.gc, GCLineWidth, &values);
       XSetDashes(xinfo.dpy, Select.gc, 0, dash_list, 2);
      }
    if ((rect.width >= minWidth) || (rect.height >= minHeight))
     {
      Select.selected = Select.rect_on = True;
      Select.x = rect.x;
      Select.y = rect.y;
      Select.width = rect.width;
      Select.height = rect.height;
      Select.rectangular = False;
      Select.aspectR = 1.0 * Select.width / Select.height;
      Select.state = undef;

      drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
      drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW), zoomDraw.factorX,
       zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
      XtVaSetValues(CopyBW, XmNsensitive, True, NULL);
      XtVaSetValues(CutBW, XmNsensitive, True, NULL);
      enableUndo(True);
      highlightButton(NULL);
      XSetRegion(xinfo.dpy, Select.moveGC, Select.reg);
      setExclEventHandler((XtEventHandler) handleSelection, PointerMotionMask | ButtonPressMask, True);
      setButtonIconCB(unselectBW, "unselect.xbm", NULL);
     }
    else /* selection too small */
     setExclEventHandler((XtEventHandler) SelectBtnPress, ButtonPressMask, True);
    break;

   case MotionNotify:
    if (Select.npoints < Select.allocPoints)
     {
      XPoint *xp = &Select.points[Select.npoints];

      if (Select.npoints == Select.allocPoints)
	{ XBell (xinfo.dpy, 0); return; }
      xp->x = (event->xbutton.x + zoX) / zmXF;
      xp->y = (event->xbutton.y + zoY) / zmYF;
      if (((xp->x) == ((xp - 1)->x)) && ((xp->y) == ((xp - 1)->y)))
       return;

      XDrawLine(xinfo.dpy, event->xbutton.window, Select.gc,
       (xp - 1)->x * zmXF - zoX, (xp - 1)->y * zmYF - zoY,
       xp->x * zmXF - zoX, xp->y * zmYF - zoY);
      Select.npoints++;
     }
    break;
  }
}

#define CUT_AREA 1
#define COPY_AREA 2
/* cut & copy button callback */
DSELECT void cutCopyCB(Widget w, XtPointer clientData, XtPointer callData)
{
  int prm = atoi ((char *) clientData);

  if (!Select.selected)
    return;

  switch (prm)
    {
    case CUT_AREA:
      ApData.tool = cutA;
      if (Select.imgAllocated)
	ApData.tool = cutA;
      else
	ApData.tool = cutClearA;
      break;
    case COPY_AREA:
      finalDraw(ApData.pixy);
      drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
      drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW), zoomDraw.factorX,
		  zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
      ApData.tool = copyA;
      if (Select.imgAllocated)
       return; /* just draw the image; we already have the selection */
      break;
    }

  disableUndo();
  last_r.x = Select.x;
  last_r.y = Select.y;
  last_r.width = (unsigned int) Select.width;
  last_r.height = (unsigned int) Select.height;
  if ((ApData.tool == copyA) || (!Select.imgAllocated))
   {
    Select.img = MyGetImage(ApData.img, Select.x, Select.y,
      Select.width, Select.height, ApData.viewMode);
    if (!checkXImage (Select.img))
      return;
    Select.imgAllocated = True;
   }
  memcpy(&Select.red[0], &xinfo.red[0], MAX_CELLS);
  memcpy(&Select.green[0], &xinfo.green[0], MAX_CELLS);
  memcpy(&Select.blue[0], &xinfo.blue[0], MAX_CELLS);
  memcpy(&Select.allocCells[0], &xinfo.allocCells[0], sizeof (XColor) * MAX_CELLS);
  memcpy(&Select.pixelTC[0], &xinfo.pixelTC[0], sizeof (unsigned long) * MAX_CELLS);
  Select.colorNo = countSelectionColors();
  Select.colorCons = True;
  XtVaSetValues (PasteBW, XmNsensitive, True, NULL);
  enableUndo (True);
  Undo.tool = ApData.tool;

  if ((ApData.tool == cutA) || (ApData.tool == cutClearA))
   {
    drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
    opBounds.x = Select.x;
    opBounds.y = Select.y;
    opBounds.width = Select.width;
    opBounds.height = Select.height;
    segmentDrawOp ();
    Undo.points[Undo.npoints].x = (short) Select.x;
    Undo.points[Undo.npoints++].y = (short) Select.y;
    Undo.points[Undo.npoints].x = (short) Select.width;
    Undo.points[Undo.npoints++].y = (short) Select.height;
    calcOpBounds ();
    if (ApData.tool == cutClearA)
     { /* normal cut-fill area with background color */
      Boolean fillInf;
      XGCValues values;

      fillInf = ApData.drawinfo.fill;
      ApData.drawinfo.fill = True;
      if (Select.reg)
	XSetRegion (xinfo.dpy, xinfo.gc1bit, Select.reg);
      values.line_width = 0;
      XChangeGC (xinfo.dpy, xinfo.gc1bit, GCLineWidth, &values);
      DrawRectangleWrapped(ApData.pixy, xinfo.gc1bit, ApData.drawinfo, &Undo.points[0], 0, 0);
      ApData.drawinfo.fill = fillInf;
      draw2Image(ApData.img, ApData.pixy, Shm.alloc, opBounds.x, opBounds.y, opBounds.width,
       opBounds.height, ColorMatrix.BckColorInd);
      XSetClipMask(xinfo.dpy, xinfo.gc1bit, None);
     }
    Select.rect_on = False;
    XtVaSetValues (PasteBW, XmNsensitive, True, NULL);
    XtVaSetValues (CopyBW, XmNsensitive, False, NULL);
    XtVaSetValues (CutBW, XmNsensitive, False, NULL);

    setToolEventHandler(none, 0);
    refreshDrawingArea(False, True);
    refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
   }
 changeTitleCB(NULL, "", NULL); /* to display the selection parameters in window title */
}

/* paste move callback */
DSELECT void pastemoveCB(Widget w, XtPointer clientData, XtPointer callData)
{
 if (!Select.selected)
  return;
 setExclEventHandler(NULL, 0, True);
 if (!Select.imgAllocated)
  return;
 finalDraw(ApData.pixy);

 ApData.tool = Undo.tool = pasteA;
 if (!Select.colorCons)
  {
   unsigned char trTbl[MAX_CELLS], cmap[256 * 3];
   unsigned long colHist[MAX_CELLS];
   int i, p;

   memset (cmap, 0, 256 * 3);
   for (i = 0; i < MAX_CELLS; i++)
     {
      p = Select.allocCells[i].pixel;
      cmap[p*3 + RED] = Select.red[i];
      cmap[p*3 + GREEN] = Select.green[i];
      cmap[p*3 + BLUE] = Select.blue[i];
     }
   initHist (&colHist[0]);
   addToHist (&colHist[0], (unsigned char *) Select.img->data, 0, 0,
   Select.img->width, Select.img->height, Select.img->bytes_per_line);

   Select.colorNo = countHistColors (&colHist[0]);
   maxColors += allocateMoreColors (Select.colorNo, maxColors);

   quantize (Select.img, maxColors, (int *) &Select.colorNo, cmap, &Select.allocCells[0],
	     &Select.red[0], &Select.green[0], &Select.blue[0], trTbl, &colHist[0]);
   quantizeOnCmap (Select.img, Select.colorNo, &Select.red[0],
		   &Select.green[0], &Select.blue[0], trTbl);
   memcpy (&Select.allocCells[0], &xinfo.allocCells[0], MAX_CELLS * sizeof (XColor));
   Select.colorCons = True;
   Select.extraColors = True;
   calcTrueColorPixels (&Select.allocCells[0], &Select.pixelTC[0],
		  &Select.red[0], &Select.green[0], &Select.blue[0]);
   calcTrueColorPixels (&xinfo.allocCells[0], &xinfo.pixelTC[0],
		     &xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
   initColorSelectionArea (ApData.ColorNo, 0, 0);
  }
 getClipWinDims(clipWin, scrollBX, scrollBY, ApData.DAwidth, ApData.DAheight, &ApData.clipWin);

 moveBoundaryPts(0, Select.npoints, ApData.clipWin.x + 10 - Select.x, ApData.clipWin.y + 10 - Select.y);

 Undo.points[Undo.npoints].x = opBounds.x = Select.x = ApData.clipWin.x + 10;
 Undo.points[Undo.npoints++].y = opBounds.y = Select.y = ApData.clipWin.y + 10;
 Undo.points[Undo.npoints].x = opBounds.width = (short) Select.width;
 Undo.points[Undo.npoints++].y = opBounds.height = (short) Select.height;

 if (!Select.rectangular)
  {
   if (Select.reg)
    {
     XDestroyRegion (Select.reg);
     XSetClipMask (xinfo.dpy, Select.moveGC, None);
    }
   Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
   XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
  }
 Select.rect_on = True;
 LastDrawOnAllWin (False);
 Select.state = undef;
 XtVaSetValues(CutBW, XmNsensitive, True, NULL);

 setToolEventHandler(ApData.tool, 0);
 highlightButton(NULL);
}

static void HandlePasteMotion(Widget w, XtPointer clientData, XEvent * event)
{ /* if press copy to screen and replace event handlers, otherwise move */
 int dx, dy, start, distance, dest, zmXF = 1, zmYF = 1, zoX = 0, zoY = 0;

 if ((int) clientData != 0)
  {
   zmXF = zoomDraw.factorX;
   zmYF = zoomDraw.factorY;
   zoX = zoomDraw.clipWin.x;
   zoY = zoomDraw.clipWin.y;
  }
 if (event->type == ButtonRelease)
  {
   setExclEventHandler((XtEventHandler) handleSelection, PointerMotionMask | ButtonPressMask, True);
   XSetFunction (xinfo.dpy, xinfo.gc, GXcopy);
   Select.rect_on = True;
   Undo.tool = ApData.tool;
   XCopyGC(xinfo.dpy, xinfo.gc, GCFunction | GCForeground | GCBackground |
    GCLineWidth | GCLineStyle | GCDashList, Undo.gc);
   Undo.points[Undo.npoints].x = opBounds.x = (short) Select.x;
   Undo.points[Undo.npoints++].y = opBounds.y = (short) Select.y;
   Undo.points[Undo.npoints].x = opBounds.width = (short) Select.width;
   Undo.points[Undo.npoints++].y = opBounds.height = (short) Select.height;

   segmentDrawOp ();
   calcOpBounds ();

   if (w==drawAreaWID)
    {
     XPoint orR;

     orR.x = repeats.crop.width;
     orR.y = repeats.crop.height;
     moveBoundaryPts(0, Select.npoints, Select.x - Select.points[Select.npoints].x,
		      Select.y - Select.points[Select.npoints].y);
     moveBoundaryPts(0, Select.npoints, -repeats.r[repeats.firstDraw].x,
		      -repeats.r[repeats.firstDraw].y);
     flipPts (&Select.points[0], Select.npoints,
	      repeats.r[repeats.firstDraw].rType, &orR);
     findMinMax (&Select.points[0], Select.npoints, (short *) &Select.x,
		 (short *) &Select.y, (short *) &Select.width, (short *) &Select.height);
     Select.width -= Select.x;
     Select.height -= Select.y;
     SetSelCursor(event->xmotion.window, (event->xbutton.x + zoX) / zmXF,
		   (event->xbutton.y + zoY) / zmYF);
    }
   LastDrawOnAllWin (False);
   enableUndo (True);
   ApData.pic_change = True;
   XtVaSetValues (PasteBW, XmNsensitive, True, NULL);
   XtVaSetValues (CopyBW, XmNsensitive, True, NULL);
   XtVaSetValues (CutBW, XmNsensitive, True, NULL);
   if (Select.colorNo == 0)
    {
     Select.colorNo = countSelectionColors ();
     memcpy (&Select.red[0], &xinfo.red[0], MAX_CELLS);
     memcpy (&Select.green[0], &xinfo.green[0], MAX_CELLS);
     memcpy (&Select.blue[0], &xinfo.blue[0], MAX_CELLS);
     memcpy (Select.allocCells, xinfo.allocCells, MAX_CELLS * sizeof (XColor));
    }
   Select.colorCons = True;
  }
 else
  {
   Rectangle winR, areaR, resR;

   dx = (event->xbutton.x + zoX) / zmXF - last_r.x;
   dy = (event->xbutton.y + zoY) / zmYF - last_r.y;
   if (zmXF == 1)
     {
       memcpy (&winR, &ApData.clipWin, sizeof (Rectangle));
       if (Select.rectangular)
	{
	 if (dx >= 0)
	  {
	   dest = start = Select.x;
	   distance = dx;
	  }
	 else
	  {
	   dest = start = Select.x + Select.width + dx;
	   distance = -dx;
	  }
	 areaR.x = start;
	 areaR.y = Select.y;
	 areaR.width = distance;
	 areaR.height = Select.height;
	 if (intersectRect (&winR, &areaR, &resR))
	   exposeRepeat(drawAreaWin, resR.x, resR.y,
	    resR.width, resR.height, resR.x, resR.y, repeats.drawGrid, False);
	 if (dy >= 0)
	  {
	   dest = start = Select.y;
	   distance = dy;
	  }
	 else
	  {
	   dest = start = Select.y + Select.height + dy;
	   distance = -dy;
	  }
	 areaR.x = Select.x;
	 areaR.y = start;
	 areaR.width = Select.width;
	 areaR.height = distance;
	 if (intersectRect (&winR, &areaR, &resR))
	   exposeRepeat(drawAreaWin, resR.x, resR.y,
	    resR.width, resR.height, resR.x, resR.y, repeats.drawGrid, False);
	}
       else
	{
	 areaR.x = Select.x;
	 areaR.y = Select.y;
	 areaR.width = Select.width;
	 areaR.height = Select.height;
	 if (intersectRect(&winR, &areaR, &resR))
	   exposeRepeat(drawAreaWin, resR.x, resR.y,
	    resR.width, resR.height, resR.x, resR.y, repeats.drawGrid, False);
	}
      }
     last_r.x = (event->xbutton.x + zoX) / zmXF;
     last_r.y = (event->xbutton.y + zoY) / zmYF;
     Select.x += dx;
     Select.y += dy;

     if (w==drawAreaWID)
       {
	if (!Select.rectangular)
	  XSetClipOrigin(xinfo.dpy, Select.moveGC,
	   Select.x - Select.points[Select.npoints].x, Select.y - Select.points[Select.npoints].y);
	MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x, Select.y,
	 Select.width, Select.height, Select.reg);
       }
     else
       {
	if (firstClick)
	 firstClick = False;
	else
	 drawSelectionBoundaries(XtWindow (zoomDrawAreaW), zmXF, zmYF, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
	moveBoundaryPts(0, Select.npoints, dx, dy);
	drawSelectionBoundaries(XtWindow (zoomDrawAreaW), zmXF, zmYF, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
       }
   }
}

/* event handler resizing the selection area */
static void resizeSelection(Widget w, XtPointer clientData, XEvent * event)
{
  int dx, dy, newW, newH, np, zmXF = 1, zmYF = 1, zoX = 0, zoY = 0;

  if ((int) clientData != 0)
   {
    zmXF = zoomDraw.factorX;
    zmYF = zoomDraw.factorY;
    zoX = zoomDraw.clipWin.x;
    zoY = zoomDraw.clipWin.y;
   }

  switch (event->type)
    {
    case ButtonRelease:
      setExclEventHandler (NULL, 0, True);

      /*check x,y are > 0 and width+x, height+y < Apdata.width, height */
      np = Undo.npoints - 2;
      Undo.points[np + 1].x = (short) abs (Undo.points[np + 1].x);
      Undo.points[np + 1].y = (short) abs (Undo.points[np + 1].y);
      XDrawRectangle(xinfo.dpy, XtWindow (w), Select.gc,
	Undo.points[np].x * zmXF - zoX, Undo.points[np].y * zmYF - zoY,
	Undo.points[np + 1].x * zmXF, Undo.points[np + 1].y * zmYF);
      if (XtWindow (w) != drawAreaWin)
	{
	  char dash_list[2];
	  XGCValues values;

	  values.line_width = 2;
	  dash_list[0] = dash_list[1] = 4;
	  XChangeGC (xinfo.dpy, Select.gc, GCLineWidth, &values);
	  XSetDashes (xinfo.dpy, Select.gc, 0, dash_list, 2);
	}

      if (Undo.points[np].x < 0)
	{
	  Undo.points[np + 1].x += Undo.points[np].x;
	  Undo.points[np].x = 0;
	}
      if (Undo.points[np].y < 0)
	{
	  Undo.points[np + 1].y += Undo.points[np].y;
	  Undo.points[np].y = 0;
	}
      if (Undo.points[np].x + Undo.points[np + 1].x > ApData.DAwidth)
	Undo.points[np + 1].x = ApData.DAwidth - Undo.points[np].x;
      if (Undo.points[np].y + Undo.points[np + 1].y > ApData.DAheight)
	Undo.points[np + 1].y = ApData.DAheight - Undo.points[np].y;

      Select.aspectR = 1.0 * Undo.points[np + 1].x / Undo.points[np + 1].y;
      if (event->xbutton.button == Button2)	/*resize selected area */
	{
	  Undo.tool = ApData.tool = resizeA;
	  if (!scaleSelectionArea (Undo.npoints - 1))
	    manageErrorDialog (getTxtI8N (UNABLE_TO_RESIZE_SELECTED_AREA_MSG),
			       NULL, NULL, getTxtI8N (OK_MSG));
	}
      else
	/*change selection dimensions */
	{
	  Undo.tool = ApData.tool = selectA;
	  Select.x = opBounds.x = Undo.points[np].x;
	  Select.y = opBounds.y = Undo.points[np].y;
	  Select.width = opBounds.width = Undo.points[np + 1].x;
	  Select.height = opBounds.height = Undo.points[np + 1].y;

	  if (Select.imgAllocated)
	    {
	      undoXImage (&Select.img);
	      Select.imgAllocated = False;
	    }
	}
      Select.npoints = 0;
      setBoundaryPtsR (&Select.npoints, Select.x, Select.y,
		       Select.width, Select.height);
      Undo.npoints -= 2;	/* point to the original dimensions */
      Undo.drawinfo = ApData.drawinfo;

      drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
      drawSelectionBoundariesWrapped (XtWindow (zoomDrawAreaW), zoomDraw.factorX,
		  zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);

      Select.state = undef;
      setExclEventHandler ((XtEventHandler) handleSelection,
			   PointerMotionMask | ButtonPressMask, True);

      XtVaSetValues (CopyBW, XmNsensitive, True, NULL);
      XtVaSetValues (CutBW, XmNsensitive, True, NULL);
      enableUndo (False);
      ApData.pic_change = True;
      break;

    case MotionNotify:
      np = Undo.npoints - 2;
      XDrawRectangle(xinfo.dpy, XtWindow (w), Select.gc,
       Undo.points[np].x * zmXF - zoX, Undo.points[np].y * zmYF - zoY,
       abs(Undo.points[np + 1].x) * zmXF, abs(Undo.points[np + 1].y) * zmYF);

      dx = (event->xbutton.x + zoX) / zmXF - ApData.orig_xy.x;
      dy = (event->xbutton.y + zoY) / zmYF - ApData.orig_xy.y;

      if (Select.preserveAspectR)
	{
	  newH = abs (dx) / Select.aspectR;
	  if (abs (newH - abs (dy)) < 5)
	    {
	      newW = dx;
	      newH *= SGN (dy);
	    }
	  else
	    {
	      newW = SGN (dx) * abs (dy) * Select.aspectR;
	      newH = dy;
	    }
	}
      else
	{
	  newW = Undo.points[np + 1].x;
	  newH = Undo.points[np + 1].y;
	  if ((Select.resLeft) || (Select.resRight))
	    newW = dx;
	  if ((Select.resUp) || (Select.resDown))
	    newH = dy;
	}
      if (newW < 0)
	Undo.points[np].x = ApData.orig_xy.x + newW;
      else
	Undo.points[np].x = ApData.orig_xy.x;
      if (newH < 0)
	Undo.points[np].y = ApData.orig_xy.y + newH;
      else
	Undo.points[np].y = ApData.orig_xy.y;
      Undo.points[np + 1].x = (short) newW;
      Undo.points[np + 1].y = (short) newH;

      XDrawRectangle(xinfo.dpy, XtWindow (w), Select.gc,
       Undo.points[np].x * zmXF - zoX, Undo.points[np].y * zmYF - zoY,
       abs(Undo.points[np + 1].x) * zmXF, abs(Undo.points[np + 1].y) * zmYF);
      break;
    }
}

/* change coordinates of all points determining selection area by dx and dy */
DSELECT void moveBoundaryPts(int First, int Last, int dx, int dy)
{
 int i;

 for (i = First; i < Last; i++)
  {
   Select.points[i].x += dx;
   Select.points[i].y += dy;
  }
}

/*set boundary points of selection area */
DSELECT void setBoundaryPtsR(int *First, int x, int y, int width, int height)
{
 int i = *First;

 Select.points[i].x = (short) x;
 Select.points[i++].y = (short) y;
 Select.points[i].x = (short) x + width;
 Select.points[i++].y = (short) y;
 Select.points[i].x = (short) x + width;
 Select.points[i++].y = (short) y + height;
 Select.points[i].x = (short) x;
 Select.points[i++].y = (short) y + height;
 Select.points[i].x = (short) x;
 Select.points[i++].y = (short) y;
 *First = i;
}

/* draw boundaries of the selection area */
DSELECT void drawSelectionBoundaries(Window win, int zoomX, int zoomY, int offsX, int offsY)
{
 int i;
 XPoint *xp;
 XGCValues values;
 char dash_list[2];

 if ((!Select.rect_on) || (win == 0))
  return;
 XSetFunction(xinfo.dpy, Select.gc, GXinvert);
 xp = &Select.points[Select.npoints];
 for (i = 0; i < Select.npoints; i++)
  {
   xp--;
   xp->x *= zoomX;
   xp->x -= offsX;
   xp->y *= zoomY;
   xp->y -= offsY;
  }
 values.line_width = max(2, (zoomX + zoomY)/3);
 dash_list[0] = dash_list[1] = 2 * values.line_width;
 XChangeGC (xinfo.dpy, Select.gc, GCLineWidth, &values);
 XSetDashes(xinfo.dpy, Select.gc, 0, dash_list, 2);
 XDrawLines(xinfo.dpy, win, Select.gc, Select.points, Select.npoints, CoordModeOrigin);
 values.line_width = 2;
 dash_list[0] = dash_list[1] = 4;
 XChangeGC (xinfo.dpy, Select.gc, GCLineWidth, &values);
 XSetDashes (xinfo.dpy, Select.gc, 0, dash_list, 2);

 for (i = 0; i < Select.npoints; i++)
  {
   xp->x += offsX;
   xp->x /= zoomX;
   xp->y += offsY;
   xp->y /= zoomY;
   xp++;
  }
 changeTitleCB (NULL, "", NULL); /* to display the selection parameters in window title */
}

DSELECT void drawSelectionBoundariesWrapped(Window win, int zoomX, int zoomY, int offsX, int offsY)
{
 int i;
 XPoint orR, newOr;

 if ((!Select.rect_on) || (win == 0))
   return;
 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;

 for (i = 0; i < repeats.currNum; i++)
  {
   newOr.x = repeats.r[i].x;
   newOr.y = repeats.r[i].y;
   flipPts (&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
   movePts (&Select.points[0], Select.npoints, newOr.x, newOr.y);
   drawSelectionBoundaries (win, zoomX, zoomY, offsX, offsY);
   movePts (&Select.points[0], Select.npoints, -newOr.x, -newOr.y);
   flipPts (&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
  }
}

/* disable selected area by freeing all structures; */
/* if disableButtons then deactivate Cut, Copy, Paste menu items */
/* if Select.extraColors checks if there are any colors used exclusively by */
/* selection and free them */
DSELECT void disableSelection(Boolean disableButtons)
{
 if (Select.reg)
  {
   XDestroyRegion(Select.reg);
   Select.reg = NULL;
   XSetClipMask (xinfo.dpy, Select.moveGC, None);
  }
 setButtonIconCB(unselectBW, "unselectNone.xbm", NULL);

 Select.rect_on = Select.selected = Select.colorCons = False;
 Select.npoints = Select.colorNo = 0;
 Select.aspectR = 1.0;
 Select.rectangular = True;

 if (Select.imgAllocated)
  {
   if (Select.extraColors)
    {
      unsigned long colHistSel[MAX_CELLS], colHistMain[MAX_CELLS];
      unsigned char freeCol[MAX_CELLS], freeColArr[MAX_CELLS];
      int i, col2free, ind;

      initHist(&colHistSel[0]);
      addToHist (&colHistSel[0], (unsigned char *) Select.img->data, 0, 0,
      Select.img->width, Select.img->height, Select.img->bytes_per_line);
      initHist (&colHistMain[0]);
      addToHist(&colHistMain[0], (unsigned char *) ApData.img->data, 0, 0,
      ApData.img->width, ApData.img->height, ApData.img->bytes_per_line);
      col2free = 0;
      for (i = 0; i < MAX_CELLS; i++)
	freeCol[i] = 0;
      for (i = 0; i < ApData.ColorNo; i++)
	{
	  ind = xinfo.allocCells[i].pixel;
	  if ((colHistSel[ind]) && (colHistMain[ind] == 0))
	    {
	      freeCol[i] = 1;
	      freeColArr[col2free++] = i;
	    }
	}
      if (col2free)
	{
	  int firstFree, moveIt, j;
	  unsigned char remap[256];
	  Boolean doRemap, foundIR;
	  Pixel bckColor, frgColor;

	    /* make protected and transparent array consistent */
	  for (i = 0; i < col2free; i++)
	    {
	      ind = freeColArr[i];
	      foundIR = False;
	      for (j = 0; j < ApData.TranspColNo; j++)
		{
		 if (ApData.TranspColors[j] == ind)
		   foundIR = True;
		if (foundIR)
		  ApData.TranspColors[j] = ApData.TranspColors[j + 1];
		}
	      if (foundIR)
		ApData.TranspColNo--;

	      foundIR = False;
	      for (j = 0; j < ApData.ProtColNo; j++)
		{
		 if (ApData.ProtectColors[j] == ind)
		   foundIR = True;
		if (foundIR)
		  ApData.ProtectColors[j] = ApData.ProtectColors[j + 1];
		}
	      if (foundIR)
		ApData.ProtColNo--;
	    }
	  for (i = 0; i < MAX_CELLS; i++)
	    remap[i] = i;
	  doRemap = False;
	  moveIt = ApData.ColorNo - 1;
	  frgColor = xinfo.allocCells[ColorMatrix.FrgColorInd].pixel;
	  bckColor = xinfo.allocCells[ColorMatrix.BckColorInd].pixel;
	  for (i = 0; i < col2free; i++)
	    {
	      firstFree = freeColArr[i];
	      while ((moveIt > 0) && freeCol[moveIt])
		moveIt--;
	      if (moveIt >= (ApData.ColorNo - col2free))
	       {
		XColor tmp;

		remap[xinfo.allocCells[moveIt].pixel] = xinfo.allocCells[firstFree].pixel;
		doRemap = True;
		SWAP(xinfo.allocCells[moveIt].pixel, xinfo.allocCells[firstFree].pixel);
		memcpy(&tmp, &xinfo.allocCells[firstFree], sizeof(XColor));
		memcpy(&xinfo.allocCells[firstFree], &xinfo.allocCells[moveIt], sizeof(XColor));
		memcpy(&xinfo.allocCells[moveIt], &tmp, sizeof(XColor));
		xinfo.red[firstFree] = xinfo.red[moveIt];
		xinfo.green[firstFree] = xinfo.green[moveIt];
		xinfo.blue[firstFree] = xinfo.blue[moveIt];
		freeCol[firstFree] = 0;
	       }
	      moveIt--;
	    }
	  if (doRemap)
	    remapData ((unsigned char *) ApData.img->data, &remap[0],
		     ApData.img->bytes_per_line * ApData.img->height);
	  freeColors (col2free);
	  ApData.ColorNo -= col2free;
	  if (xinfo.trueColor)
	    calcTrueColorPixels (&xinfo.allocCells[0], &xinfo.pixelTC[0],
			&xinfo.red[0], &xinfo.green[0], &xinfo.blue[0]);
	  else
	    XStoreColors (xinfo.dpy, xinfo.cmap, &xinfo.allocCells[0], ApData.ColorNo);
	  /* restore active fore/back ground colors */
	  if ((remap[bckColor] != bckColor) || (remap[frgColor] != frgColor))
	    {
	     int cInd;
	     cInd = 0;
	     for (i = 0; i < ApData.ColorNo; i++)
	       if (xinfo.allocCells[i].pixel == frgColor)
		 {
		  cInd = i;
		  break;
		 }
	     ColorSelectConseq(Button1, cInd);
	     cInd = 1;
	     for (i = 0; i < ApData.ColorNo; i++)
	       if (xinfo.allocCells[i].pixel == bckColor)
		 {
		  cInd = i;
		  break;
		 }
	     ColorSelectConseq(Button4, cInd);
	    }
	  initColorSelectionArea(ApData.ColorNo, 0, 0);
	  refreshDrawingArea(True, True);
	  refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
	   zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);

	  ApData.pic_change = True;
	}
    }
   undoXImage (&Select.img);
   Select.imgAllocated = Select.extraColors = False;
  }
 if (disableButtons)
  {
   XtVaSetValues(PasteBW, XmNsensitive, False, NULL);
   XtVaSetValues(CopyBW, XmNsensitive, False, NULL);
   XtVaSetValues(CutBW, XmNsensitive, False, NULL);
  }
 calcColTransfTbl(); /* fix the transparent/protected color table */
}

/* BAD SYTLE: change constants to capital letters some day */
#define MPinitDlg 0
#define MPinitX 1
#define MPinitY 2
#define MPinitRepX 3
#define MPinitRepY 4
#define MPmap 5
#define MPblock 6
#define MPbrick 7
#define MPdrop 8
#define MPAllowResize 10
#define MPOK 11
#define MP_SET_DROP_OFFSET 12
#define MPinitOptX 21
#define MPinitOptY 22
#define MPmenuFocusX 23
#define MPmenuFocusY 24
#define MPtopOption 25
#define MPinitOffsX 26
#define MPinitOffsY 27
#define MPinitMotifW 28
#define MPinitMotifH 29
#define MPpreview 30
#define MPinitFrameH 31
#define MPinitFrameV 32
#define MPinitAlignH 33
#define MPinitAlignV 34
#define MPCancel 35

static void doMultiplePaste(int offsX, int offsY, int repX, int repY, int marginX, int marginY,
 int fractX, int fractY, int distX, int distY, int frameOffsX, int frameOffsY)
{
 int i, j, posX, posY, start, limit, advance;

 /* store rectangle coordinates of maximum mul-paste area  to clip the exccess parts when necessary */
 Undo.points[Undo.npoints].x = (short) marginX + frameOffsX;
 Undo.points[Undo.npoints++].y = (short) marginY + frameOffsY;

 Undo.points[Undo.npoints].x = (short) distX *repX;
 if (distX > Select.width)
   Undo.points[Undo.npoints].x -= (short) distX - Select.width;
 Undo.points[Undo.npoints].y = (short) distY *repY;
 if (distY > Select.height)
   Undo.points[Undo.npoints].y -= (short) distY - Select.height;

 Undo.npoints++;
 if (offsY == 0)
  {
   advance = distX;
   posY = marginY + frameOffsY;
   limit = marginX + repX * advance;
   start = marginX + frameOffsX;
   for (i = 0; i < repY; i++)
    {
     posX = start;
     j = 0;
     while (posX < limit)
       {
	if (((posX + distX) > 0) && (posX < ApData.width))
	  {
	   Undo.points[Undo.npoints].x = (short) posX;
	   Undo.points[Undo.npoints++].y = (short) posY;
	  }
	 j++;
	 if ((fractX) && ((j % fractX) == 0))
	  posX = start + j * advance;
	 else
	  posX += advance;
       }
     if (fractX)
       if (((i + 1) % fractX) == 0)
	 start = marginX + frameOffsX;
       else
	 start += advance / fractX;
     else
       start += offsX;
     while (start > (marginX + frameOffsX))
       start -= advance;
     posY += distY;
    }
  }
 else /* pillar */
   {
    advance = distY;
    posX = marginX + frameOffsX;
    limit = marginY + repY * advance;
    start = marginY + frameOffsY;
    for (i = 0; i < repX; i++)
      {
	posY = start;
	j = 0;
	while (posY < limit)
	  {
	    if (((posY + distY) > 0) && (posY < ApData.height))
	      {
		Undo.points[Undo.npoints].x = (short) posX;
		Undo.points[Undo.npoints++].y = (short) posY;
	      }
	    j++;
	    if ((fractY) && ((j % fractY) == 0))
	      posY = start + j * advance;
	    else
	      posY += advance;
	  }
	if (fractY)
	  if (((i + 1) % fractY) == 0)
	    start = marginY + frameOffsY;
	  else
	    start += advance / fractY;
	else
	  start += offsY;
	while (start > (marginY + frameOffsY))
	  start -= advance;
	posX += distX;
      }
   }
 findMinMax(&Undo.points[2], Undo.npoints - 2, &opBounds.x, &opBounds.y, &opBounds.width, &opBounds.height);
 opBounds.width -= opBounds.x - Select.width;
 opBounds.height -= opBounds.y - Select.height;
 Undo.tool = pasteMulA;
 Select.rect_on = False;
 LastDrawOnAllWin (False);
 enableUndo (True);
 ApData.pic_change = True;
}

DSELECT void multiplePasteCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static inpNumStruct *xInd, *yInd, *repXInd, *repYInd, *xOffs, *yOffs, *frameV, *frameH;
 static Widget dialogW = 0, optX, optY, firstOpt, motifW, motifH, alignH, alignV;
 static char repeatMode = 0, optMenuFocus = 5;
 static int optXValue = 0, optYValue = 0, drawnOnce = 0;
 static short distType = 0;
 static int *frameDistW, *frameDistH;
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case MPinitX: xInd = initInputText(w, 0, 0, 100); break;
   case MPinitY: yInd = initInputText(w, 0, 0, 100); break;
   case MPinitRepX: repXInd = initInputText(w, 1, 1, 10); break;
   case MPinitRepY: repYInd = initInputText(w, 1, 1, 10); break;
   case MPinitOffsX: xOffs = initInputText(w, 0, 0, 100); break;
   case MPinitOffsY: yOffs = initInputText(w, 0, 0, 100); break;
   case MPinitFrameH: frameH = initInputText(w, 0, 0, 1000); break;
   case MPinitFrameV: frameV = initInputText(w, 0, 0, 1000); break;
   case MPinitAlignH: alignH = w; break;
   case MPinitAlignV: alignV = w; break;
   case MPmap:
     {
       char buf[512];

       if (ApData.viewMode != normalView)
	 XmToggleButtonSetState (normalViewTBW, True, True);

       xInd->upperB = Select.width < ApData.width ? Select.width : ApData.width;
       yInd->upperB = Select.height < ApData.height ? Select.height : ApData.height;

       xOffs->upperB = ApData.width - 1;
       yOffs->upperB = ApData.height - 1;
       repXInd->upperB = (ApData.width + Select.width - 1) / Select.width;
       repYInd->upperB = (ApData.height + Select.height - 1) / Select.height;
       setInpArrVal (repXInd, repXInd->value);
       setInpArrVal (repYInd, repYInd->value);
       frameH->upperB = ApData.width < Select.width ? Select.width : ApData.width;
       frameV->upperB = ApData.height < Select.height ? Select.height : ApData.height;

       if (distType)
	{
	 frameH->lowerB = frameV->lowerB = 0;
	 setInpArrVal (frameH, 0);
	 setInpArrVal (frameV, 0);
	 frameDistW = &Select.width;
	 frameDistH = &Select.height;
	}
       else
	{
	 frameH->lowerB = Select.width;
	 frameV->lowerB = Select.height;
	 setInpArrVal (frameH, Select.width);
	 setInpArrVal (frameV, Select.height);
	 frameDistW = &frameH->value;
	 frameDistH = &frameV->value;
	}

       sprintf (buf, "%s: %d", getTxtI8N (MOTIF_WIDTH_MSG), Select.width);
       setLabelString(motifW, buf);
       sprintf (buf, "%s: %d", getTxtI8N (MOTIF_HEIGHT_MSG), Select.height);
       setLabelString(motifH, buf);
       drawnOnce = 0;
       highlightButton (NULL);
     }
     break;
   case MPinitDlg: dialogW = w; break;
   case MPinitOptX: optX = w; break;
   case MPinitOptY: optY = w; break;
   case MPmenuFocusX: optMenuFocus = 0; break;
   case MPmenuFocusY: optMenuFocus = 1; break;
   case MPtopOption: firstOpt = w; break;
   case MPinitMotifW: motifW = w; break;
   case MPinitMotifH: motifH = w; break;
   case MP_SET_DROP_OFFSET:
    {
     int fract = getMenuIndexFromWidget(w)+1, value = 0;
     Boolean Bval = False;
     inpNumStruct *ind = NULL;

     if (fract==1)
      {
       fract = 0;
       Bval = True;
      }
     if (optMenuFocus == 0)
      {
       ind = xInd;
       optXValue = fract;
       if (fract)
	 value = *frameDistW / fract;
      }
     else if (optMenuFocus == 1)
      {
       ind = yInd;
       optYValue = fract;
       if (fract)
	 value = *frameDistH / fract;
      }
     if (ind)
      {
       ind->enableChange = Bval;
       XtVaSetValues (ind->inpField, XmNsensitive, Bval, NULL);
       setInpArrVal (ind, value);
      }
    }
    break;
   case MPblock:
     repeatMode = optXValue = optYValue = 0;
     xInd->enableChange = False;
     yInd->enableChange = False;
     setInpArrVal (xInd, 0);
     setInpArrVal (yInd, 0);
     XtVaSetValues (xInd->inpField, XmNsensitive, False, NULL);
     XtVaSetValues (optX, XmNsensitive, False, XmNmenuHistory, firstOpt, NULL);
     XtVaSetValues (yInd->inpField, XmNsensitive, False, NULL);
     XtVaSetValues (optY, XmNsensitive, False, XmNmenuHistory, firstOpt, NULL);
     break;
   case MPbrick:
     repeatMode = 1;
     optYValue = 0;
     xInd->enableChange = True;
     yInd->enableChange = False;
     setInpArrVal (xInd, *frameDistW / 2);
     setInpArrVal (yInd, 0);
     XtVaSetValues (xInd->inpField, XmNsensitive, True, NULL);
     XtVaSetValues (optX, XmNsensitive, True, XmNmenuHistory, firstOpt, NULL);
     XtVaSetValues (yInd->inpField, XmNsensitive, False, NULL);
     XtVaSetValues (optY, XmNsensitive, False, XmNmenuHistory, firstOpt, NULL);
     break;
   case MPdrop:
     repeatMode = 2;
     optXValue = 0;
     yInd->enableChange = True;
     xInd->enableChange = False;
     setInpArrVal (yInd, *frameDistH / 2);
     setInpArrVal (xInd, 0);
     XtVaSetValues (yInd->inpField, XmNsensitive, True, NULL);
     XtVaSetValues (optY, XmNsensitive, True, XmNmenuHistory, firstOpt, NULL);
     XtVaSetValues (xInd->inpField, XmNsensitive, False, NULL);
     XtVaSetValues (optX, XmNsensitive, False, XmNmenuHistory, firstOpt, NULL);
     break;
   case MPAllowResize:
     {
      XmToggleButtonCallbackStruct *TBS = (XmToggleButtonCallbackStruct *) callData;

      distType = TBS->set == True;
      if (distType)
       {
	XtUnmanageChild (XtParent (alignH));
	XtUnmanageChild (XtParent (alignV));
	frameH->lowerB = frameV->lowerB = 0;
	setInpArrVal (frameH, 0);
	setInpArrVal (frameV, 0);
	frameDistW = &Select.width;
	frameDistH = &Select.height;
       }
      else
       {
	XtManageChild (XtParent (alignH));
	XtManageChild (XtParent (alignV));
	frameH->lowerB = Select.width;
	frameV->lowerB = Select.height;
	setInpArrVal (frameH, Select.width);
	setInpArrVal (frameV, Select.height);
	frameDistW = &frameH->value;
	frameDistH = &frameV->value;
       }
     }
     break;
   case MPCancel:
     if (drawnOnce > 0)
       undoCB (NULL, NULL, NULL);
     XtUnmanageChild (dialogW);
     break;
   case MPOK:
     XtUnmanageChild (dialogW);	/*!! no break - on purpose */
     setToolEventHandler(none, 0);
   case MPpreview:
     if (drawnOnce > 0)
       undoCB(NULL, NULL, NULL);
     else
      {
       if ((Undo.tool != cutA) && (Undo.tool != pasteA) && (Undo.tool != rotateA) &&
	   (Undo.tool != mirrorHor) && (Undo.tool != mirrorVer) && (Undo.tool != resizeA))
	 finalDraw(ApData.pixy);
       disableUndo();
      }
     yInd->value = getIntFromWidget(yInd->inpField, yInd->lowerB, yInd->upperB);
     xInd->value = getIntFromWidget(xInd->inpField, xInd->lowerB, xInd->upperB);
     yOffs->value = getIntFromWidget(yOffs->inpField, yOffs->lowerB, yOffs->upperB);
     xOffs->value = getIntFromWidget(xOffs->inpField, xOffs->lowerB, xOffs->upperB);
     /* correct the upper borders based on entered values above */
     repXInd->upperB = (ApData.width - xOffs->value + Select.width - 1) / max(Select.width, *frameDistW);
     repYInd->upperB = (ApData.height - yOffs->value + Select.height - 1) / max(Select.height, *frameDistH);

     repXInd->value = getIntFromWidget(repXInd->inpField, repXInd->lowerB, repXInd->upperB);
     repYInd->value = getIntFromWidget(repYInd->inpField, repYInd->lowerB, repYInd->upperB);
      {				/* get alignment type of motif inside the frame */
       short i;
       Widget tb;
       int vertShift = 0, horShift = 0;
       int sizeX = *frameDistW, sizeY = *frameDistH;

       XtVaGetValues (alignH, XmNmenuHistory, &tb, NULL);
       XtVaGetValues (tb, XmNpositionIndex, &i, NULL);
       if (distType == 0)
	{
	 if (i == 1)
	   horShift = (*frameDistW - Select.width) / 2;
	 else
	  if (i == 2)
	   horShift = *frameDistW - Select.width;
	}
       XtVaGetValues (alignV, XmNmenuHistory, &tb, NULL);
       XtVaGetValues (tb, XmNpositionIndex, &i, NULL);
       if (distType == 0)
	{
	 if (i == 1)
	   vertShift = (*frameDistH - Select.height) / 2;
	 else if (i == 2)
	   vertShift = *frameDistH - Select.height;
	}

       if (distType)
	 {
	   sizeX += frameH->value;
	   sizeY += frameV->value;
	 }
       doMultiplePaste(xInd->value, yInd->value, repXInd->value, repYInd->value,
	xOffs->value, yOffs->value, optXValue, optYValue, sizeX, sizeY, horShift, vertShift);
      }
     drawnOnce++;
     break;
  }
}

static int countSelectionColors()
{
  unsigned long colHist[MAX_CELLS];

  if (!Select.imgAllocated)
    return (0);
  initHist (&colHist[0]);
  addToHist (&colHist[0], (unsigned char *) Select.img->data, 0, 0, Select.img->width,
	     Select.img->height, Select.img->bytes_per_line);
  return (countHistColors (&colHist[0]));
}

/* paste motif on all repeats in drawable */
DSELECT void pasteMotifWrapped(Drawable drawable)
{
 int i, offsX, offsY;
 XPoint r[2], orR, newOr;
 Rectangle boundR, winR, clipR;

 if ((!Select.imgAllocated)||(!Select.img))
   return;
 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;
 offsX = offsY = 0;

 if (drawable == ApData.pixy)
   {
    winR.x = winR.y = 0;
    winR.width = ApData.width;
    winR.height = ApData.height;
   }
 else if (drawable == ApData.scrPix) /* ABC double buffer problem */
   {
    offsX = winR.x = zoomDraw.clipWin11.x;
    offsY = winR.y = zoomDraw.clipWin11.y;
    winR.width = zoomDraw.clipWin11.width + 2;
    winR.height = zoomDraw.clipWin11.height + 2;
   }
 else
   memcpy (&winR, &ApData.clipWin, sizeof (Rectangle));

 XSetFunction (xinfo.dpy, Select.moveGC, GXcopy);
 for (i = 0; i < repeats.currNum; i++)
  {
   newOr.x = repeats.r[i].x;
   newOr.y = repeats.r[i].y;
   if (drawable == ApData.pixy)
    {
     movePts(&Select.points[0], Select.npoints, -newOr.x, -newOr.y);
     flipPts(&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
    }
   else
    {
     flipPts(&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
     movePts(&Select.points[0], Select.npoints, newOr.x, newOr.y);
    }
   findMinMax(&Select.points[0], Select.npoints, &r[0].x, &r[0].y, &r[1].x, &r[1].y);
   r[1].x -= r[0].x;
   r[1].y -= r[0].y;
   boundR.x = r[0].x;
   boundR.y = r[0].y;
   boundR.height = r[1].y;
   boundR.width = r[1].x;
   if (intersectRect (&boundR, &winR, &clipR))
    {
     if (!Select.rectangular)
      {
       XDestroyRegion (Select.reg);
       XSetClipMask (xinfo.dpy, Select.moveGC, None);
       Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
       XOffsetRegion (Select.reg, -offsX, -offsY);
       XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
      }
     repeatCopyArea (drawable, Select.img, Select.moveGC,
      clipR.x - boundR.x, clipR.y - boundR.y, clipR.width, clipR.height,
      clipR.x - offsX, clipR.y - offsY, ApData.specialPutImg, repeats.r[i].rType,
      Select.img->width, Select.img->height, offsX, offsY, Select.reg);
    }
   if (drawable == ApData.pixy)
    {
     flipPts(&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
     movePts(&Select.points[0], Select.npoints, newOr.x, newOr.y);
    }
   else
    {
     movePts(&Select.points[0], Select.npoints, -newOr.x, -newOr.y);
     flipPts(&Select.points[0], Select.npoints, repeats.r[i].rType, &orR);
    }
  }
 if (!Select.rectangular)
  {
   XDestroyRegion (Select.reg);
   XSetClipMask (xinfo.dpy, Select.moveGC, None);
   XSetClipOrigin (xinfo.dpy, Select.moveGC, 0, 0);
   Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
   XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
  }
 if (drawable == ApData.pixy)
  {
   refreshDrawingArea (False, False);
   drawSelectionBoundariesWrapped (XtWindow (zoomDrawAreaW), zoomDraw.factorX,
	       zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
  }
}
