#define TOOLS
#include <Xm/Xm.h>
#include <stdlib.h>
#include <math.h>
#include "paint.h"

#define GUESS_WEAVE_ACTION 1
#define GUESS_WEAVE_DLG 2
#define GUESS_WEAVE_INIT_X 3
#define GUESS_WEAVE_INIT_Y 4
#define GUESS_WEAVE_PREVIEW 5
#define GUESS_WEAVE_OK	6
#define TEST_REPEAT	7

#define MAX_SELECT MAX_IMAGE/2 /* maximal size of selection in grid detection */
#define MIN_GRID   4 /* minial grid size in auto detection */
#define MAX_GRID   MAX_SELECT/MIN_GRID /* the biggest grid we are able to detect */

static Widget GW_dialogW;
static inpNumStruct *GWwidthInd, *GWheightInd;
static int minSquareX=MIN_GRID, minSquareY=MIN_GRID; /* minimal size of one grid square in pixels for automatic detection */
unsigned char allocMap[MAX_CELLS];

static void guess_weave()
{
 int sx, sy, swidth, sheight;
 int i,j,k, d1,d2,d3, foundX, foundY;
 int maxwid, maxhei, bpl;
 unsigned short diffTable[0xFFFF]; /* for pre-calculation of color differences of all color pairs */
 unsigned char *data;
 unsigned short midIndex[MAX_SELECT], gridX[MAX_GRID];
 unsigned long gridError[MAX_GRID];

 for (i=0;i<256;i++)
  allocMap[xinfo.allocCells[i].pixel]=i;

  if (!Select.selected)
   {
    manageErrorDialog (getTxtI8N (NO_SELECTION_MADE_MSG), NULL, NULL, getTxtI8N (OK_MSG));
    return;
   }
  finalDraw(ApData.pixy);
  highlightButton(NULL);

  sx = Select.x;
  sy = Select.y;
  swidth = Select.width;
  sheight = Select.height;
  disableSelection(False);
  Select.x = sx;
  Select.y = sy;
  Select.width = swidth;
  Select.height = sheight;
  ApData.tool = Undo.tool = guessWeave;
  
  if (!Select.imgAllocated)
   {
    Select.img = MyGetImage(ApData.img, Select.x, Select.y, Select.width, Select.height, ApData.viewMode);
    if (!checkXImage (Select.img))
      return;
    Select.imgAllocated = True;
    memcpy(Select.red, xinfo.red, MAX_CELLS);
    memcpy(Select.green, xinfo.green, MAX_CELLS);
    memcpy(Select.blue, xinfo.blue, MAX_CELLS);
    memcpy(Select.allocCells, xinfo.allocCells, sizeof (XColor) * MAX_CELLS);
    memcpy(Select.pixelTC, xinfo.pixelTC, sizeof (unsigned long) * MAX_CELLS);
   }
  Select.extraColors = False;
  Select.rect_on = Select.selected = Select.rectangular = Select.colorCons = True;
  Select.npoints=0;
  setBoundaryPtsR (&Select.npoints, Select.x, Select.y, Select.width, Select.height);

 for (i=0;i<256;i++)
  for (j=0;j<256;j++)
   { /* calculate the difference based on allocCells, so we do not do it in the main loop! */
    d1=(int)Select.red[allocMap[i]]-(int)Select.red[allocMap[j]];
    d2=(int)Select.green[allocMap[i]]-(int)Select.green[allocMap[j]];
    d3=(int)Select.blue[allocMap[i]]-(int)Select.blue[allocMap[j]];
    diffTable[(i<<8)+j] = (unsigned short) sqrt(d1*d1+d2*d2+d3*d3);
   }
 maxwid = swidth/minSquareX;
 maxhei = sheight/minSquareY;
 memset(gridError, 0, sizeof(unsigned long)*maxwid);
 bpl = Select.img->bytes_per_line;
 /* find the horizontal grid */
 /* testing hypothesis of grid width k */
 for (k=MIN_GRID; k<maxwid; k++)
  {
   for (i=0; i<swidth; i++)
    midIndex[i]= (unsigned short) (((i*k)/swidth)*((swidth*1.0)/k)+((swidth*1.0)/(k*2)));
   for (j=0; j<sheight; j++)
    {
     data = (unsigned char *) &Select.img->data[j*bpl];
     for (i=swidth-1; i>=0; i--)
      gridError[k]+= diffTable[(((int)data[i])<<8)+data[midIndex[i]]];
    }
  }
  /* search for horizontal minimum */
  foundX=MIN_GRID;
  for (i=MIN_GRID; i<maxwid; i++)
   if (gridError[foundX]>gridError[i])
    foundX=i;
  /* now find vertical grid, but look only at points at center of horizontal grid */
  for (i=0; i<foundX; i++)
   gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
  memset(gridError, 0, sizeof(unsigned long)*maxhei);
  /* find the vertical grid */
  /* testing hypothesis of grid width k */
  for (k=MIN_GRID; k<maxhei; k++)
   {
    for (i=0; i<sheight; i++)
     midIndex[i]= (unsigned short) (((i*k)/sheight)*((sheight*1.0)/k)+((sheight*1.0)/(k*2)));
    for (j=0; j<foundX; j++)
     {
      data = (unsigned char *) &Select.img->data[gridX[j]];
      for (i=sheight-1; i>=0; i--)
       gridError[k]+= diffTable[(((int)data[i*bpl])<<8)+data[midIndex[i]*bpl]];
     }
   }
  /* search for vertical minimum */
  foundY=MIN_GRID;
  for (i=MIN_GRID; i<maxhei; i++)
   if (gridError[foundY]>gridError[i])
    foundY=i;
  setInpArrVal(GWwidthInd, foundX);
  setInpArrVal(GWheightInd, foundY);
  XtManageChild(GW_dialogW);
}

static void guess_preview()
{
 int sx, sy, swidth, sheight, foundX, foundY;
 int i,j, bpl;
 unsigned short gridX[MAX_GRID], gridY[MAX_GRID];
 unsigned char *data;
 int gridOffX, gridLenX, gridOffY, gridLenY;

 if (!Select.imgAllocated)
  return;
 foundX = GWwidthInd->value;
 foundY = GWheightInd->value;
 swidth = Select.width;
 sheight = Select.height;
 sx = Select.x;
 sy = Select.y;
 if ((foundX*MIN_GRID>swidth)||(foundY*MIN_GRID>sheight))
  { XBell(xinfo.dpy, 0); return; }
 bpl = Select.img->bytes_per_line;
 data = (unsigned char *) Select.img->data;

 /* it is useful to still see the grid in the preview, so only draw squares of odd size */
 gridLenX=swidth/foundX;
 if (!(gridLenX&1))
  gridLenX--;
 gridOffX=gridLenX/2;
 gridLenY=sheight/foundY;
 if (!(gridLenY&1))
  gridLenY--;
 gridOffY=gridLenY/2;

 for (i=0; i<foundX; i++)
  gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
 for (i=0; i<foundY; i++)
  gridY[i]= (unsigned short) ((sheight*i*1.0)/foundY + (sheight*1.0)/(foundY*2));
 for (i=0; i<foundX; i++)
  for (j=0; j<foundY; j++)
   {
    XSetForeground(xinfo.dpy, xinfo.gc, (xinfo.depth==8) ? data[bpl*gridY[j]+gridX[i]] :
     Select.pixelTC[allocMap[data[bpl*gridY[j]+gridX[i]]]] );
    XFillRectangle(xinfo.dpy, XtWindow(drawAreaWID), xinfo.gc, sx+gridX[i]-gridOffX, sy+gridY[j]-gridOffY, gridLenX, gridLenY);
   }
}

static void guess_weaveOK()
{
 int sx, sy, swidth, sheight, foundX, foundY;
 int i,j, bpl, bpl2;
 unsigned short gridX[MAX_GRID], gridY[MAX_GRID];
 unsigned char *data, *data2;
 XImage *img2;

 if (!Select.imgAllocated)
  return;
 foundX = GWwidthInd->value;
 foundY = GWheightInd->value;
 swidth = Select.width;
 sheight = Select.height;
 sx = Select.x;
 sy = Select.y;
 if ((foundX*MIN_GRID>swidth)||(foundY*MIN_GRID>sheight))
  { XBell(xinfo.dpy, 0); return; }
 minSquareX = max(MIN_GRID, swidth/foundX-1); /* so it will try to find it in the same range on the next guess */
 minSquareY = max(MIN_GRID, sheight/foundY-1);

 /* the new destination selection image */ 
 img2 = MyGetImage(ApData.img, 0, 0, foundX, foundY, ApData.viewMode);

 bpl = Select.img->bytes_per_line;
 data = (unsigned char *) Select.img->data;
 bpl2 = img2->bytes_per_line;
 data2 = (unsigned char *) img2->data;

 for (i=0; i<foundX; i++)
  gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
 for (i=0; i<foundY; i++)
  gridY[i]= (unsigned short) ((sheight*i*1.0)/foundY + (sheight*1.0)/(foundY*2));
 for (i=0; i<foundX; i++)
  for (j=0; j<foundY; j++)
   data2[bpl2*j+i] = data[bpl*gridY[j]+gridX[i]];
 undoXImage(&Select.img);

 Select.img = img2;
 Select.width = foundX;
 Select.height = foundY;
 Select.rect_on = Select.selected = Select.rectangular = Select.colorCons = True;
 Select.npoints=0;
 setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);

 drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
 drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW), zoomDraw.factorX,
  zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
 setButtonIconCB(unselectBW, "unselect.xbm", NULL);
 XtVaSetValues(PasteBW, XmNsensitive, True, NULL);
 ApData.tool = Undo.tool = pasteA;
 LastDrawOnAllWin(False);
}

static void restoreNormalPictureCB(void)
{
 refreshDrawingArea(TRUE, TRUE);
}

static void testRepeat(void)
{
 if ((!Select.selected) || (!Select.rect_on))
  {
   manageErrorDialog (getTxtI8N (NO_SELECTION_MADE_MSG), NULL, NULL, getTxtI8N (OK_MSG));
   return;
  }
 if (!Select.rectangular)
  { XBell(xinfo.dpy, 0); return; }
 if (!Select.imgAllocated)
  {
   Select.img = MyGetImage(ApData.img, Select.x, Select.y, Select.width, Select.height, ApData.viewMode);
   if (!checkXImage (Select.img))
    return;
   Select.imgAllocated = TRUE;
  }
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x - Select.width, Select.y - Select.height, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x - Select.width, Select.y, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x - Select.width, Select.y + Select.height, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x, Select.y - Select.height, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x, Select.y, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x, Select.y + Select.height, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x + Select.width, Select.y - Select.height, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x + Select.width, Select.y, Select.width, Select.height, Select.reg);
 MyPutImage(drawAreaWin, Select.moveGC, Select.img, 0, 0, Select.x + Select.width, Select.y + Select.height, Select.width, Select.height, Select.reg);
 XtAddTimeOut(2000, (XtTimerCallbackProc) *restoreNormalPictureCB, NULL);
}

DTOOLS void toolsCB(Widget w, XtPointer clientData, XtPointer callData)
{
  switch (atoi((char *) clientData))
   {
    case GUESS_WEAVE_ACTION: guess_weave(); break;
    case GUESS_WEAVE_DLG: GW_dialogW = w; break;
    case GUESS_WEAVE_INIT_X: GWwidthInd = initInputText(w, 2, 2, maxWidth); break;
    case GUESS_WEAVE_INIT_Y: GWheightInd = initInputText(w, 2, 2, maxHeight); break;
    case GUESS_WEAVE_PREVIEW: guess_preview(); break;
    case GUESS_WEAVE_OK: guess_weaveOK(); break;
    case TEST_REPEAT: testRepeat(); break;
   }
}
