#define ZOOMDRAW
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <stdlib.h>
#include "paint.h"

static int zoomLineColor = 1; /* line colors, 0 black 1 gray, 2 white */

static void adjustSB(Widget w, int *sliderVal, int *winSize, int maxVal, int zoomF)
{
 int sliderSize = *winSize, incr = zoomF, pageIncr = *winSize - zoomF;

 if (!zoomF)
   zoomF = 1;
 pageIncr -= pageIncr % zoomF;
 if ((!maxVal) || (*sliderVal < 0))
  {
   *sliderVal = 0;
   return;
  }
 if (*winSize > maxVal)	/* clip window greater than zoomed picture */
  {
    maxVal = *sliderVal + *winSize;
    *winSize = maxVal;
  }
 if ((*sliderVal + sliderSize) > maxVal)
  *sliderVal = maxVal - sliderSize;
 XtVaSetValues(w, XmNmaximum, maxVal, XmNvalue, *sliderVal, XmNsliderSize, sliderSize,
  XmNpageIncrement, pageIncr, XmNincrement, incr, NULL);
}

/*adjust zoom scroll bars according to the provided data */
static void setZmScrollBars(int valueX, int valueY, int width, int height)
{
 zoomDraw.clipWin.x = valueX;
 adjustSB(scrollX, &zoomDraw.clipWin.x, &width, zoomDraw.width, zoomDraw.factorX);
 zoomDraw.clipWin.width = width;
 zoomDraw.clipWin.y = valueY;
 adjustSB(scrollY, &zoomDraw.clipWin.y, &height, zoomDraw.height, zoomDraw.factorY);
 zoomDraw.clipWin.height = height;
 zoomDraw.clipWin11.x = zoomDraw.clipWin.x / zoomDraw.factorX;
 zoomDraw.clipWin11.y = zoomDraw.clipWin.y / zoomDraw.factorY;
 zoomDraw.clipWin11.width = zoomDraw.clipWin.width / zoomDraw.factorX;
 zoomDraw.clipWin11.height = zoomDraw.clipWin.height / zoomDraw.factorY;
}

static void drawGrayRect(int x, int y, int width, int height)
{
 if ((width <= 0) || (height <= 0))
   return;
 XSetForeground(xinfo.dpy, xinfo.gc, GrayBckGround);
 XFillRectangle(xinfo.dpy, XtWindow(zoomDrawAreaW), xinfo.gc, x, y, width, height);
 XSetForeground(xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
}

/* perform an integer zoom of the image within the same image */
/* Image must be big enough to hold the magnified image! */
static void resizePictureZoom(XImage *srcdst, int width, int height, int zoomX, int zoomY)
{
 unsigned char *src8, *dst8, *data;
 unsigned short *src16, *dst16;
 unsigned long *src32, *dst32;
 int x, y, i, bpl, bytesPerPixel;

 if (!srcdst)
  return;
 if (((zoomX==1)&&(zoomY==1)) || ((width*zoomX)>srcdst->width) || ((height*zoomY)>srcdst->height))
  return;
 bpl = srcdst->bytes_per_line;
 bytesPerPixel = srcdst->bits_per_pixel/8;
 data = srcdst->data;
 /* we render bottom up to avoid destroying the source */
 for (y=height-1; y>=0; y--)
  {
   src8 = &data[bpl*y];
   src16 = (unsigned short *) &data[bpl*y];
   src32 = (unsigned long *) &data[bpl*y];
   dst8 = &data[bpl*y*zoomY];
   dst16 = (unsigned short *) &data[bpl*y*zoomY];
   dst32 = (unsigned long *) &data[bpl*y*zoomY];
   if (zoomX>1)
    switch (bytesPerPixel)
     {
      case 1:
       for (x=width-1; x>=0; x--)
        for (i=0; i<zoomX; i++)
	 dst8[x*zoomX+i] = src8[x];
       break;
      case 2:
       for (x=width-1; x>=0; x--)
        for (i=0; i<zoomX; i++)
	 dst16[x*zoomX+i] = src16[x];
       break;
      case 3:
       for (x=width-1; x>=0; x--)
        for (i=0; i<zoomX; i++) /* hope the optimizing compiler will take care of this... */
	 { /* anyway, 24bpp arhitecture is braindead */
	  dst8[(x*zoomX+i)*3] = src8[x];
	  dst8[(x*zoomX+i)*3+1] = src8[x+1];
	  dst8[(x*zoomX+i)*3+1] = src8[x+2];
	 }
       break;
      case 4:
       for (x=width-1; x>=0; x--)
        for (i=0; i<zoomX; i++)
	 dst32[x*zoomX+i] = src32[x];
       break;
     }
   else
    memcpy(dst8, src8, width*bytesPerPixel*zoomX);  
   for (i=1; i<zoomY; i++) /* valid for all depths */
    memcpy(&dst8[i*bpl], dst8, width*bytesPerPixel*zoomX);  
  }
}

static void redrawZoomArea(int x, int y, int width, int height)
{
 Rectangle rect, srcR, dstR, clipR, zmPixR, boundR;
 Rectangle *cw = &zoomDraw.clipWin;
 int modX, modY, diffX, diffY;

 /* if clip window is smaller than drawing area draw the excessive part in background color */
 diffX = x + width - zoomDraw.width;
 if (diffX > 0)
  {
   if (x > zoomDraw.width)
    { drawGrayRect(x - cw->x, y - cw->y, width, height); return; }
   else
    { width -= diffX; drawGrayRect(x - cw->x + width, y - cw->y, diffX, height); }
  }
 diffY = y + height - zoomDraw.height;
 if (diffY > 0)
  {
   if (y > zoomDraw.height)
    { drawGrayRect(x - cw->x, y - cw->y, width, height); return; }
   else
     { height -= diffY; drawGrayRect(x - cw->x, y - cw->y + height, width, diffY); }
  }
 rect.x = x;
 rect.y = y;
 rect.width = width;
 rect.height = height;

 srcR.x = rect.x / zoomDraw.factorX;
 srcR.y = rect.y / zoomDraw.factorY;
 zmPixR.width = srcR.width =(int) ((rect.width + zoomDraw.factorX - 1) / zoomDraw.factorX);
 zmPixR.height = srcR.height = (int) ((rect.height + zoomDraw.factorY - 1) / zoomDraw.factorY);
 zmPixR.x = srcR.x - cw->x / zoomDraw.factorX;
 zmPixR.y = srcR.y - cw->y / zoomDraw.factorY;

 if ((int) srcR.width > ApData.DAwidth)
   srcR.width = (unsigned int) ApData.DAwidth;
 if ((int) srcR.height > (int) ApData.DAheight)
   srcR.height = (unsigned int) ApData.DAheight;

 modX = rect.x % zoomDraw.factorX;
 modY = rect.y % zoomDraw.factorY;

 if ((modX > 0) && ((int) (srcR.width + srcR.x) < ApData.DAwidth))
   srcR.width++;
 if ((modY > 0) && ((int) (srcR.height + srcR.y) < ApData.DAheight))
   srcR.height++;

 boundR.x = boundR.y = 0;
 boundR.width = cw->width / zoomDraw.factorX;
 boundR.height = cw->height / zoomDraw.factorY;

 if ((!intersectRect (&boundR, &zmPixR, &clipR)) || (zmPixR.x < 0) ||
     (zmPixR.y < 0) || (srcR.width <= 0) || (srcR.height <= 0))
   return;

 if (Select.selected && Select.rect_on) /* undraw dashed selection rectangle */
   drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW), zoomDraw.factorX,
    zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);

 dstR.width = srcR.width * zoomDraw.factorX;
 dstR.height = srcR.height * zoomDraw.factorY;

 if (!pixmapIsXimage) /* the megaslow pixmap to ximage data transfer */
  XGetSubImage(xinfo.dpy, ApData.scrPix, zmPixR.x, zmPixR.y,
   srcR.width, srcR.height, AllPlanes, ZPixmap, ApData.scrImg, 0, 0);
 /* the smart resize with source=destination */
 resizePictureZoom(ApData.scrImg, srcR.width, srcR.height, zoomDraw.factorX, zoomDraw.factorY);

 if (zoomDraw.drawMesh)
  {
   int i, j, ymod = zoomDraw.factorY - 1, thickX, thickY, byteNo, grid2X, grid2Y;
   unsigned long black;
   unsigned char *b1, *d1;
   unsigned short *d2;
   unsigned long *d4;
   unsigned char gridmask[xinfo.dpyw];

   if (zoomLineColor==0)
    black = xinfo.black;
   else
    black = (zoomLineColor==1) ? GrayBckGround : xinfo.white;
   byteNo = (ApData.scrImg->bits_per_pixel) / 8;
   b1 = (unsigned char *) &black;
   grid2X = (zoomDraw.factorX > 3);
   grid2Y = (zoomDraw.factorY > 3);
   thickX = zoomDraw.thickX - srcR.x % zoomDraw.thickX;
   thickY = srcR.y % zoomDraw.thickY;
   memset(gridmask, 0, xinfo.dpyw);
   if (grid2X)
    {
     for (j = zoomDraw.factorX - 1; j < dstR.width; j += zoomDraw.factorX)
      gridmask[j] = TRUE;
     for (j = zoomDraw.factorX * thickX - 2; j < dstR.width;
	  j += zoomDraw.factorX * zoomDraw.thickX)
      gridmask[j] = TRUE;
    }
   else
    for (j = zoomDraw.factorX * thickX - 1; j < dstR.width;
	 j += zoomDraw.factorX * zoomDraw.thickX)
     gridmask[j] = TRUE;
   /* draw grid */
   for (i = 0; i < dstR.height; i++)
     { /*horizontal lines */
      if ( (((i % zoomDraw.factorY) == ymod) && grid2Y) || ( (!grid2Y) && (((i+srcR.y*zoomDraw.factorY) % (zoomDraw.thickY*zoomDraw.factorY)) == 0)))
       {
	switch (byteNo)
	 {
	  case 1:
	   memset(&ApData.scrImg->data[ApData.scrImg->bytes_per_line*i], black, dstR.width);
	   break;
	  case 2:
	   d2 = (unsigned short *) &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	   for (j = 0; j < dstR.width; j++)
	     *d2++ = (unsigned short) black;
	   break;
	  case 3:
	   d1 = (unsigned char *) &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	   for (j = 0; j < dstR.width; j++)
	     { *d1++ = b1[0]; *d1++ = b1[1]; *d1++ = b1[2]; }
	   break;
	  case 4:
	   d4 = (unsigned long *) &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	   for (j = 0; j < dstR.width; j++)
	     *d4++ = black;
	   break;
	 }
  	thickY++;
	if (grid2Y)
	 if ((thickY % zoomDraw.thickY) == (ApData.height%zoomDraw.thickY))
	  memcpy (&ApData.scrImg->data[ApData.scrImg->bytes_per_line * (i - 1)],
	   &ApData.scrImg->data[ApData.scrImg->bytes_per_line * i], dstR.width * byteNo);
	}
       else /* vertical lines */
	{
	 switch (byteNo)
	  {
	   case 1:
	     d1 = &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	     for (j=0; j < dstR.width; j++, d1++)
	      if (gridmask[j])
	       *d1 = (unsigned char) black;
	     break;
	   case 2:
	     d2 = (unsigned short *) &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	     for (j=0; j < dstR.width; j++, d2++)
	      if (gridmask[j])
	       *d2 = (unsigned short) black;
	     break;
	   case 3:
	     d1 = &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	     for (j=0; j < dstR.width; j++, d1+=3)
	      if (gridmask[j])
	       { d1[0]=b1[0]; d1[1]=b1[1]; d1[2]=b1[2]; }
	     break;
	   case 4:
	     d4 = (unsigned long *) &ApData.scrImg->data[ApData.scrImg->bytes_per_line*i];
	     for (j=0; j < dstR.width; j++, d4++)
	      if (gridmask[j])
	       *d4=black;
	     break;
	  }
	}
     }
  }

 if (pixmapIsXimage)
  xPutBuffer(XtWindow(zoomDrawAreaW), modX, modY, srcR.x*zoomDraw.factorX + modX - cw->x,
   srcR.y*zoomDraw.factorY + modY - cw->y, dstR.width - modX, dstR.height - modY);
 else
  MyPutImage(XtWindow(zoomDrawAreaW), xinfo.gc, ApData.scrImg,
   modX, modY, srcR.x*zoomDraw.factorX + modX - cw->x,
   srcR.y*zoomDraw.factorY + modY - cw->y, dstR.width - modX, dstR.height - modY, NULL);

 if (Select.selected && Select.rect_on) /* undraw dashed selection rectangle */
  drawSelectionBoundariesWrapped(XtWindow(zoomDrawAreaW), zoomDraw.factorX,
   zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
}

/* expose DrawingArea Callback */
DZOOMDRAW void redrawZoomDAreaCB(Widget w, XtPointer clientData, XtPointer callData)
{
 Rectangle *cw = &zoomDraw.clipWin;

 XmDrawingAreaCallbackStruct *callS = (XmDrawingAreaCallbackStruct *) callData;

 refreshZoomWin((cw->x+callS->event->xexpose.x)/zoomDraw.factorX, (cw->y+callS->event->xexpose.y)/zoomDraw.factorY,
  callS->event->xexpose.width+1, callS->event->xexpose.height+1, TRUE);
 /*
 redrawZoomArea(callS->event->xexpose.x + cw->x, callS->event->xexpose.y + cw->y,
  callS->event->xexpose.width, callS->event->xexpose.height);
 */
}

DZOOMDRAW void refreshZoomWin(short x, short y, unsigned short width, unsigned short height,
 Boolean refreshFromSource)
{
 Rectangle clipR, boundR, *cw = &zoomDraw.clipWin;
 int zmXF = zoomDraw.factorX, zmYF = zoomDraw.factorY;
 int m = MAX(zoomDraw.factorX, zoomDraw.factorY);

 if (!widgetIsViewable(zoomDrawAreaW))
   return;

 boundR.x = x*zmXF - m;
 boundR.y = y*zmYF - m;
 boundR.width = width*zmXF + 2*m;
 boundR.height = height*zmYF + 2*m;
 if (boundR.x < 0)
   {
     if ((-boundR.x) > boundR.width)
       boundR.width = 0;
     else
       boundR.width += boundR.x;
     boundR.x = 0;
   }
 if (boundR.y < 0)
   {
     if ((-boundR.y) > boundR.height)
       boundR.height = 0;
     else
       boundR.height += boundR.y;
     boundR.y = 0;
   }

 if (intersectRect(&boundR, cw, &clipR))
   {
    if (refreshFromSource)
     {
      if (ApData.viewMode==repeatView)
       exposeRepeat(ApData.scrPix, clipR.x/zmXF, clipR.y/zmYF,
	clipR.width/zmXF + 2, clipR.height/zmYF + 2, clipR.x/zmXF - cw->x/zmXF,
	clipR.y/zmYF - cw->y/zmYF, repeats.drawGrid, True);
      else
       MyPutImage(ApData.scrPix, xinfo.gc, ApData.img,
	clipR.x/zmXF, clipR.y/zmYF, clipR.x/zmXF - cw->x/zmXF,
	clipR.y/zmYF - cw->y/zmYF, clipR.width/zmXF + 2, clipR.height/zmYF + 2, NULL);
     }
    if (pixmapIsXimage)
     XSync(xinfo.dpy, FALSE);
    redrawZoomArea(clipR.x, clipR.y, clipR.width, clipR.height);
   }
}

DZOOMDRAW void mousePosZoom(Widget w, XtPointer unused, XEvent * event)
{
 static int oldX = 0, oldY = 0;
 static Boolean cursOut = False;
 int x, y;

 switch (event->type)
  {
   case MotionNotify:
    x = (event->xmotion.x + zoomDraw.clipWin.x) / (int) zoomDraw.factorX;
    y = (event->xmotion.y + zoomDraw.clipWin.y) / (int) zoomDraw.factorY;
    x++;
    if ((event->xmotion.x > zoomDraw.width) || (event->xmotion.y > zoomDraw.height))
     {
      if (cursOut == False)
       {
	setLabelString(posXYZm, getTxtI8N(HELP_MSG));
	cursOut = True;
       }
      return;
     }
    if ((x != oldX) || (y != oldY))
     {
      char pos[20], *ptr = pos;

      ptr += sprintf(pos, "%4d", x);
      sprintf(ptr, "%5d", ApData.DAheight - y);
      setLabelString(posXYZm, pos);
     }
    oldX = x;
    oldY = y;
    cursOut = False;
    break;
   case LeaveNotify:
    setLabelString(posXYZm, getTxtI8N(HELP_MSG));
    oldX = oldY = -1;
    break;
   case KeyRelease:
    {
     KeySym key;
     char buf[5];
     int bufSize = 5, len;

     len = XLookupString((XKeyEvent *) event, buf, bufSize, &key, NULL);
     if (key == XK_Left)
      moveSelection(-1, 0);
     if (key == XK_Right)
      moveSelection(1, 0);
     if (key == XK_Up)
      moveSelection(0, -1);
     if (key == XK_Down)
      moveSelection(0, 1);
     if (len == 0)
      return;

     if ((oldX != -1) && (key == XK_space))
      changeColorKbd(XtWindow (w), (oldX - 1) * zoomDraw.factorX - zoomDraw.clipWin.x,
       oldY * zoomDraw.factorY - zoomDraw.clipWin.y, Button1);
    }
    break;
  }
}

static void zoomFactorChanged(int zoomX, int zoomY)
{
 int new_orig_x, new_orig_y;
 int cww_2 = zoomDraw.clipWin.width / 2, cwh_2 = zoomDraw.clipWin.height / 2;
 char label[200];

 new_orig_x = (zoomDraw.clipWin.x + cww_2) / zoomDraw.factorX;	/* store old values */
 new_orig_y = (zoomDraw.clipWin.y + cwh_2) / zoomDraw.factorY;

 zoomDraw.factorX = zoomX;
 zoomDraw.factorY = zoomY;

 zoomDraw.width = ApData.DAwidth * zoomX;
 zoomDraw.height = ApData.DAheight * zoomY;

 new_orig_x = new_orig_x - (cww_2 / zoomX);
 new_orig_y = new_orig_y - (cwh_2 / zoomY);
 new_orig_x *= zoomX;
 new_orig_y *= zoomY;

 if ((new_orig_x + zoomDraw.scrollW) > zoomDraw.width)
   new_orig_x = zoomDraw.width - zoomDraw.scrollW;
 if ((new_orig_y + zoomDraw.scrollH) > zoomDraw.height)
   new_orig_y = zoomDraw.height - zoomDraw.scrollH;

 if (new_orig_x < 0)
   new_orig_x = 0;
 if (new_orig_y < 0)
   new_orig_y = 0;
 setZmScrollBars(new_orig_x, new_orig_y, zoomDraw.scrollW, zoomDraw.scrollH);
 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
  zoomDraw.clipWin11.width + 1, zoomDraw.clipWin11.height + 1, True);
 if (Undo.tool != none)
  {
   finalDraw(ApData.scrPix);
   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
  }
 sprintf(label, "%s %dx%d",  getTxtI8N(ZOOM_MSG), zoomX, zoomY);
 setDialogTitle(zoomDrawAreaW, label);
}

/* zoom map window function - initialize the necessary variables if necessary */
DZOOMDRAW void zoomMapWindow()
{
 if (!widgetIsViewable(zoomDrawAreaW))
  return;
 if (ApData.viewMode == densityView)
  {
   XtPopdown(GetShellParent(zoomDrawAreaW));
   return;
  }
 if (zoomDraw.newImage)
  {
   Rectangle *cw = &zoomDraw.clipWin;
   int zmX = zoomDraw.factorX, zmY = zoomDraw.factorY;

   zoomDraw.newImage = False;
   zoomDraw.width = ApData.DAwidth * zoomDraw.factorX;
   zoomDraw.height = ApData.DAheight * zoomDraw.factorY;

   setZmScrollBars(0, 0, zoomDraw.scrollW, zoomDraw.scrollH);
   if (ApData.viewMode==repeatView)
    exposeRepeat(ApData.scrPix, 0, 0, (cw->width + zmX - 1) / zmX,
     (cw->height + zmY - 1) / zmY, 0, 0, repeats.drawGrid, True);
   else
    MyPutImage(ApData.scrPix, xinfo.gc, ApData.img,
     0, 0, 0, 0, (cw->width + zmX - 1) / zmX, (cw->height + zmY - 1) / zmY, NULL);
   finalDraw(ApData.scrPix);
    if (Shm.alloc)
     XSync(xinfo.dpy, FALSE);
   redrawZoomArea(cw->x, cw->y, zoomDraw.scrollW, zoomDraw.scrollH);
  }
 else
  finalDraw(ApData.scrPix);
 refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
}

#define ZOOMMAP 0
#define ZOOMBUTTON_INIT 1
#define ZOOMINITSCROLLWIN 3
#define ZOOMINITPOSITON 4
#define ZOOMRESIZE 5
#define ZOOMSCROLLBAR_VC_X 6
#define ZOOMSCROLLBAR_VC_Y 7
#define ZOOMSCROLLBAR_INIT_X 8
#define ZOOMSCROLLBAR_INIT_Y 9
#define ZOOMMESH_ON_OFF 10

DZOOMDRAW void handleZoomDrawCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi((char *) clientData);
 static Widget sbW;
 static int sbValue;

 switch (prm)
  {
   case ZOOMBUTTON_INIT:
    {
     static int num=1;
     char numStr[10];

     sprintf(numStr, "%d", num);
     setLabelString(w, numStr);
     if (num>9)
      num=0;
     sprintf(numStr, "<Key>%d", num);
     XtVaSetValues(w, XmNaccelerator, numStr, NULL);
     num++;
     break;
    }
   case ZOOMSCROLLBAR_INIT_X:
     blockZoomRefreshes=FALSE;
     scrollX = w;
     sbW = NULL;
     sbValue = -1;
     break;
   case ZOOMSCROLLBAR_INIT_Y:
     scrollY = w;
     break;
   case ZOOMINITSCROLLWIN:
     zoomDraw.clipWin.width = zoomDraw.clipWin.height = zoomDraw.clipWin.x =
       zoomDraw.clipWin.y = 0;
     zoomDraw.clipWin11.x = zoomDraw.clipWin11.y = zoomDraw.clipWin11.width =
       zoomDraw.clipWin11.height = 0;
     zoomDraw.width = ApData.DAwidth * zoomDraw.factorX;
     zoomDraw.height = ApData.DAheight * zoomDraw.factorY;
     XtVaSetValues(zoomDrawAreaW, XmNwidth, xinfo.dpyw / 4, XmNheight, xinfo.dpyh / 4, NULL);
     XmScrolledWindowSetAreas(w, scrollX, scrollY, zoomDrawAreaW);
     break;
   case ZOOMRESIZE:
     {
      unsigned int oldWidth = zoomDraw.clipWin.width, oldHeight = zoomDraw.clipWin.height;
      Rectangle *cw = &zoomDraw.clipWin;

      XtVaGetValues(w, XmNwidth, &cw->width, XmNheight, &cw->height, NULL);
      zoomDraw.scrollW = cw->width;
      zoomDraw.scrollH = cw->height;
      setZmScrollBars(cw->x, cw->y, cw->width, cw->height);
      if ((oldWidth < cw->width) || (oldHeight < cw->height))
       {
	zoomDraw.clipWin11.width = zoomDraw.clipWin.width / zoomDraw.factorX;
	zoomDraw.clipWin11.height = zoomDraw.clipWin.height / zoomDraw.factorY;
	refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
	 zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
	finalDraw(ApData.scrPix);
	refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
       }
     }
     break;
   case ZOOMSCROLLBAR_VC_X:
   case ZOOMSCROLLBAR_VC_Y:
     {
      XmScrollBarCallbackStruct *CB = (XmScrollBarCallbackStruct *) callData;
      Rectangle *cw = &zoomDraw.clipWin, srcR;

      if ((w == sbW) && (CB->value == sbValue))
       return; /* after drag callback, the valueChangeCB is useless */
      sbW = w;
      sbValue = CB->value;
      if (prm == ZOOMSCROLLBAR_VC_X)
       zoomDraw.clipWin.x = CB->value;
      else
       zoomDraw.clipWin.y = CB->value;
      srcR.width = (cw->width + zoomDraw.factorX - 1) / zoomDraw.factorX + 1;
      srcR.height = (cw->height + zoomDraw.factorY - 1) / zoomDraw.factorY + 1;
      srcR.x = zoomDraw.clipWin11.x = cw->x / zoomDraw.factorX;
      srcR.y = zoomDraw.clipWin11.y = cw->y / zoomDraw.factorY;
      if (blockZoomRefreshes)
       return;
      if (ApData.viewMode==repeatView)
       exposeRepeat(ApData.scrPix, srcR.x, srcR.y, srcR.width, srcR.height,
	0, 0, repeats.drawGrid, True);
      else
       MyPutImage(ApData.scrPix, xinfo.gc, ApData.img,
	srcR.x, srcR.y, 0, 0, srcR.width, srcR.height, NULL);
      finalDraw(ApData.scrPix);
      if (Shm.alloc)
	XSync(xinfo.dpy, False);
      redrawZoomArea(cw->x, cw->y, zoomDraw.scrollW, zoomDraw.scrollH);
     }
     break;
   case ZOOMMAP:
     zoomMapWindow();
     break;
   case ZOOMINITPOSITON:
     posXYZm = w;
     break;
   case ZOOMMESH_ON_OFF:
     {
      XmToggleButtonCallbackStruct *CBS = (XmToggleButtonCallbackStruct *) callData;
      zoomDraw.drawMesh = CBS->set;
refreshZoomWin(zoomDraw.clipWin.x/zoomDraw.factorX, zoomDraw.clipWin.y/zoomDraw.factorX,
       zoomDraw.clipWin.width, zoomDraw.clipWin.height, TRUE);
/*
      redrawZoomArea(zoomDraw.clipWin.x, zoomDraw.clipWin.y,
       zoomDraw.clipWin.width, zoomDraw.clipWin.height);*/
     }
     break;
   default:
     {
      int i = getMenuIndexFromWidget(w) + 1;
      int newFactorX, newFactorY;

      if (ApData.imageXdpi<ApData.imageYdpi)
       { newFactorX = (ApData.imageYdpi*i)/ApData.imageXdpi+0.5; newFactorY = i; }
      else
       { newFactorX = i; newFactorY = (ApData.imageXdpi*i)/ApData.imageYdpi+0.5; }
      zoomFactorChanged(newFactorX, newFactorY);
      setZoomFactorCB(w, "7", NULL); // update interface values
     }
     break;
   }
}

#define MAX_RAND 4096
#define MAX_RAND_MASK 0xfff
/* unified function to generate real view image for both screen and print in any depth */
DZOOMDRAW void renderRealImg(unsigned char *data, int bpl, int bytesPerPixel,
 int xpos, int ypos, int width, int height, int firstWeftPix, int xdpi, int ydpi)
{
 static unsigned short xrandList[MAX_RAND], yrandList[MAX_RAND];
 static int oldxfac = 0, oldyfac = 0;
 int xfactor, yfactor, x, y;
 unsigned char *dst8=NULL, *src;
 unsigned short *dst16=NULL;
 unsigned long *dst32=NULL;
 unsigned long srcy, srcx; /* on big zoom-outs it can go out of bounds, but it must not be negative (modulo %!)*/
 int imageX, imageY, i, valr, valg, valb;
 int randInd=0, srcbpl = ApData.img->bytes_per_line;
 unsigned char cmap[256*3], *cmapPtr;
 unsigned char *srcPtr[MAX_IMAGE];

 imageX = ApData.width;
 imageY = ApData.height;
 xfactor = (int) ((819200.0 * 1000 * ApData.imageXdpi ) / (1.0 * xdpi * ApData.zoomX));
 yfactor = (int) ((819200.0 * 1000 * ApData.imageYdpi ) / (1.0 * ydpi * ApData.zoomY));

 if ((oldxfac != xfactor) || (oldyfac != yfactor))
  for (x=0; x<MAX_RAND; x++)
   { xrandList[x] = (unsigned short) (drand48()*xfactor);
     yrandList[x] = (unsigned short) (drand48()*yfactor); }
 oldxfac = xfactor; oldyfac = yfactor;
 src = (unsigned char *) ApData.img->data;

 /* fill in the cmap for faster access on 16/24/32 bpp */
 cmapPtr = cmap;
 for (x=0; x<256; x++)
  {
   *cmapPtr++ = xinfo.red[x];
   *cmapPtr++ = xinfo.green[x];
   *cmapPtr++ = xinfo.blue[x];
  }
 /* prepare a table for inverse bottom up rendering and bpl calculation */
 for (y=0; y<imageY; y++)
  srcPtr[y] = &src[srcbpl*(imageY - y -1)];
 /* render from bottom  up! */
 for (y=ypos+height-1; y>=ypos; y--)
  {
   srcy = (firstWeftPix - y)*yfactor;
   dst8 = data + (y-ypos)*bpl;
   dst16 = (unsigned short *) (data + (y-ypos)*bpl);
   dst32 = (unsigned long *) (data + (y-ypos)*bpl);
   switch (bytesPerPixel)
    {
     case 1:
      for (x = 0; x < width; x++)
       {
	*dst8++ = *(srcPtr[((srcy + yrandList[randInd&MAX_RAND_MASK]) >> 13)%imageY] +
	 ((((x+xpos)*xfactor + xrandList[randInd&MAX_RAND_MASK]) >> 13)%imageX));
	randInd++;
       }
      break;  
     case 2:
      for (x = 0; x < width; x++)
       {
	valr=valg=valb=0;
	srcx = (x+xpos)*xfactor;
	for (i=8; i; i--)
         {
	  cmapPtr = &cmap[3* *(srcPtr[((srcy + yrandList[randInd&MAX_RAND_MASK]) >> 13)%imageY] +
	   (((srcx + xrandList[(randInd++)&MAX_RAND_MASK]) >> 13)%imageX))];
	  valr+= *cmapPtr++; valg+=*cmapPtr++; valb+=*cmapPtr;
   	 }
	valr>>=6; valg>>=5; valb>>=6;
	*dst16++ = (valr<<11) | (valg<<5) | valb; 
       }
      break;  
     case 3:
      for (x = 0; x < width; x++)
       {
	valr=valg=valb=0;
	srcx = (x+xpos)*xfactor;
	for (i=8; i; i--)
         {
	  cmapPtr = &cmap[3* *(srcPtr[((srcy + yrandList[randInd&MAX_RAND_MASK]) >> 13)%imageY] +
	   (((srcx + xrandList[(randInd++)&MAX_RAND_MASK]) >> 13)%imageX))];
	  valr+= *cmapPtr++; valg+=*cmapPtr++; valb+=*cmapPtr;
   	 }
	*dst8++ = valr>>3; *dst8++ = valg>>3; *dst8++ = valb>>3;
       }
      break;  
     case 4:
      for (x = 0; x < width; x++)
       {
	valr=valg=valb=0;
	srcx = (x+xpos)*xfactor;
	for (i=8; i; i--)
         {
	  cmapPtr = &cmap[3* *(srcPtr[((srcy + yrandList[randInd&MAX_RAND_MASK]) >> 13)%imageY] +
	   (((srcx + xrandList[(randInd++)&MAX_RAND_MASK]) >> 13)%imageX))];
	  valr+= *cmapPtr++; valg+=*cmapPtr++; valb+=*cmapPtr;
   	 }
	valr>>=3; valg>>=3; valb>>=3;
	*dst32++ = (valr<<16) | (valg<<8) | valb; 
       }
      break;  
    } 
  }
}

DZOOMDRAW void exposeTkanina(int x, int y, int width, int height, Widget w)
{
 int ypos;

 if (ApData.viewMode != densityView)
  return;
 for (ypos=0; ypos<height; ypos+=IMG_BUF_Y)
  {
   renderRealImg(ApData.scrImg->data, ApData.scrImg->bytes_per_line,
    ApData.scrImg->bits_per_pixel/8, x, y+ypos, width, min(IMG_BUF_Y, height-ypos),
    REAL_VIEW_SIZE, (25400*xinfo.dpyw)/screenWidthMM, (25400*xinfo.dpyh)/screenHeightMM);
   xPutBuffer(XtWindow(w), 0, 0, x, y+ypos, width, min(IMG_BUF_Y, height-ypos));
  }
}

/* in the main window show position of the area viewed in the zoom window */
static void drawEnlargedAreaCB(XtPointer clientData, XtIntervalId * id)
{
 static short num = 0;
 Rectangle *mainWR = (Rectangle *) clientData;

 XFillRectangle(xinfo.dpy, drawAreaWin, Select.gc, mainWR->x, mainWR->y,
  mainWR->width, mainWR->height);
 if (num < 5)
  {
   num++;
   XtAddTimeOut(500, (XtTimerCallbackProc) *drawEnlargedAreaCB, clientData);
  }
 else
  {
   num = 0;
   XtPopup(GetShellParent(zoomDrawAreaW), XtGrabNone);
  }
}

/* Event Handler for setting the upper left point in zoom window */
static void getCoordEH(Widget w, XtPointer unused, XEvent *event)
{
 if (event->type == ButtonPress)
  centerZoomWindowToMouse(event);
}

#define ZM_VIEW_NEW_VALUE 0
#define ZM_VIEW_SHOW_POSITION 2
#define ZM_VIEW_ZOOM 3
#define ZM_VIEW_RESTORE_ZOOM 4
#define ZM_VIEW_INIT_MESH_X 5
#define ZM_VIEW_INIT_MESH_Y 6
#define ZM_VIEW_MESHDIAG_MAP 7
#define ZM_VIEW_MESHOK 8
#define ZM_VIEW_MESH_INIT 9
#define ZM_VIEW_THICK_X 10
#define ZM_VIEW_THICK_Y 11
#define ZM_VIEW_MESH_COLOR 13
DZOOMDRAW void setZoomFactorCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static EventMask oldEM = 0;
 static XtEventHandler oldEH = NULL;
 static inpNumStruct *meshX, *meshY, *thickX, *thickY;
 static Widget meshDialogW;
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case ZM_VIEW_NEW_VALUE:
    {
     int val, xpos, ypos;
     Dimension hCW, wCW;

     XmScaleGetValue(zoomScaleW, &val);
     ApData.zoomX = ApData.zoomY = val;
     XtVaGetValues(scrollBX, XmNvalue, &xpos, NULL);
     XtVaGetValues(scrollBY, XmNvalue, &ypos, NULL);
     XtVaGetValues(clipWin, XmNheight, &hCW, XmNwidth, &wCW, NULL);
     ApData.clipWin.width = (int) wCW;
     ApData.clipWin.height = (int) hCW;
     ApData.clipWin.x = xpos;
     ApData.clipWin.y = ypos;

     printPictureCB(w, "8", NULL); /* update the size information */
     exposeTkanina(ApData.clipWin.x, ApData.clipWin.y,
      ApData.clipWin.width, ApData.clipWin.height, drawAreaWID);
   }
    break;
   case ZM_VIEW_SHOW_POSITION:
    {
     int newX, newY, valX, valY, sliderSize, incr, pageIncr;
     static Rectangle origRect;

     origRect.x = newX = zoomDraw.clipWin11.x = zoomDraw.clipWin.x / zoomDraw.factorX;
     origRect.y = newY = zoomDraw.clipWin11.y = zoomDraw.clipWin.y / zoomDraw.factorY;

     XmScrollBarGetValues(scrollBX, &valX, &sliderSize, &incr, &pageIncr);
     valX = ((newX + sliderSize) > ApData.DAwidth) ? ApData.DAwidth - sliderSize : newX;
     XmScrollBarSetValues(scrollBX, valX, sliderSize, incr, pageIncr, True);

     XmScrollBarGetValues(scrollBY, &valY, &sliderSize, &incr, &pageIncr);
     valY = ((newY + sliderSize) > ApData.DAheight) ? ApData.DAheight - sliderSize : newY;
     XmScrollBarSetValues(scrollBY, valY, sliderSize, incr, pageIncr, True);

     patchScrollBarPos(scrollBX, scrollBY, valX, valY);

     origRect.width = zoomDraw.clipWin11.width;
     origRect.height = zoomDraw.clipWin11.height;

     XSetFunction(xinfo.dpy, Select.gc, GXinvert);
     XtPopdown(GetShellParent(zoomDrawAreaW));
     XtAddTimeOut(800, (XtTimerCallbackProc) *drawEnlargedAreaCB, &origRect);
    }
    break;
   case ZM_VIEW_ZOOM:
    oldEM = ApData.eventMask;
    oldEH = ApData.CurrEventHandler;
    setExclEventHandler((XtEventHandler) getCoordEH, ButtonPressMask, False);
    break;
   case ZM_VIEW_RESTORE_ZOOM:
    setExclEventHandler((XtEventHandler) oldEH, oldEM, True);
    break;
   case ZM_VIEW_INIT_MESH_X:
    meshX = initInputText(w, 2, 1, 30);
    break;
   case ZM_VIEW_INIT_MESH_Y:
    meshY = initInputText(w, 2, 1, 30);
    break;
   case ZM_VIEW_THICK_X:
    thickX = initInputText(w, zoomDraw.thickX , 4, 30);
    break;
   case ZM_VIEW_THICK_Y:
    thickY = initInputText(w, zoomDraw.thickY, 4, 30);
    break;
   case ZM_VIEW_MESHDIAG_MAP:
    setInpArrVal(meshX, zoomDraw.factorX);
    setInpArrVal(meshY, zoomDraw.factorY);
    setInpArrVal(thickX, zoomDraw.thickX);
    setInpArrVal(thickY, zoomDraw.thickY);
    break;
   case ZM_VIEW_MESHOK:
    zoomDraw.thickX = thickX->value;
    zoomDraw.thickY = thickY->value;
    zoomFactorChanged(meshX->value, meshY->value);
    break;
   case ZM_VIEW_MESH_INIT:
    meshDialogW = w;
    break;
   case ZM_VIEW_MESH_COLOR:
    zoomLineColor = getMenuIndexFromWidget(w);
    refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
     zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, False);
    break;
  }
}
