/* ljpeg.c load JPEG images */
#include "loadimg.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include "jpeglib.h"

/* default 24bpp output image if normal and 8 bit if fgrayscale */

/* for special needs enable one of defines below */

//#define DITHER200
	// 200 color Floyd/Steinberg dithered image
//#define REDUCECMAP
	// reduce to colormap and number of colors entered
	// as cmap; 2 equal colors in sequence mean end of colors
	// REDUCECMAP does not work on grayscale JPEG images


/* if requesed image is LOAD_CMAP then do a 128 color no dither image */

static FILE *fp;
static int didStartDecompress=FALSE;

struct my_error_mgr
{
 struct jpeg_error_mgr pub;	/* "public" fields */
 jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

static struct jpeg_decompress_struct cinfo;
static struct my_error_mgr jerr;

/* Here's the routine that will replace the standard error_exit method: */
METHODDEF(void) my_error_exit (j_common_ptr cinfo)
{
 /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
 my_error_ptr myerr = (my_error_ptr) cinfo->err;
 /* Always display the message. */
 /* We could postpone this until after returning, if we chose. */
 (*cinfo->err->output_message) (cinfo);
 /* Return control to the setjmp point */
 longjmp(myerr->setjmp_buffer, 1);
}

extern void registerLoader(tryLoadProc tLoad, loadBandProc lBand, endLoadProc eLoad);

static int tryLoad(char *filename, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int realLoad)
{
 int i;
 int c1, c2;

 if ((fp = fopen(filename, "rb" )) == NULL)
  return(LOAD_IO_ERROR);

 /* before doing anything, check the start of file marker */
 c1=fgetc(fp); c2=fgetc(fp);
 if ((c1!=0xFF) || (c2!=0xD8))
  return(LOAD_FILETYPE_UNKNOWN);

 /* It's OK, now let's go from the start */
 rewind(fp);
 /* We set up the normal JPEG error routines, then override error_exit. */
 cinfo.err = jpeg_std_error(&jerr.pub);
 jerr.pub.error_exit = my_error_exit;
 /* Establish the setjmp return context for my_error_exit to use. */
 if (setjmp(jerr.setjmp_buffer))
  {
   jpeg_destroy_decompress(&cinfo);
   fclose(fp);
   return(LOAD_FILETYPE_UNKNOWN);
  }
 /* Now we can initialize the JPEG decompression object. */
 jpeg_create_decompress(&cinfo);

 /* specify data source (eg, a file) */
 jpeg_stdio_src(&cinfo, fp);

 (void) jpeg_read_header(&cinfo, TRUE);
 /* here I set what I want */
 /* first true color default version */
 cinfo.quantize_colors = FALSE;
 cinfo.dither_mode = JDITHER_NONE;
 if (cinfo.out_color_space==JCS_GRAYSCALE)
  {
   cinfo.quantize_colors = TRUE;
   cinfo.two_pass_quantize = FALSE;
   cinfo.desired_number_of_colors = 256;
  }
 else
  cinfo.out_color_space = JCS_RGB;

#ifdef DITHER200
 cinfo.quantize_colors = TRUE;
 cinfo.two_pass_quantize = TRUE;
 cinfo.dither_mode = JDITHER_FS;
 cinfo.desired_number_of_colors = 200;
#endif
 if (*loadType==LOAD_CMAP)
  {
   cinfo.quantize_colors = TRUE;
   cinfo.two_pass_quantize = TRUE;
   cinfo.dither_mode = JDITHER_NONE;
   cinfo.desired_number_of_colors = 128;
  }
#ifdef REDUCECMAP
 if ((realLoad) && (cinfo.out_color_space!=JCS_GRAYSCALE))
  {
   i=0;
   while ( (i<255) && (
    (cmap[i*3]!=cmap[(i+1)*3]) ||
    (cmap[i*3+1]!=cmap[(i+1)*3+1]) ||
    (cmap[i*3+2]!=cmap[(i+1)*3+2]))) i++;
   cinfo.two_pass_quantize = FALSE;
   cinfo.desired_number_of_colors = i;
   cinfo.actual_number_of_colors = i;
   cinfo.colormap = (*cinfo.mem->alloc_sarray)
		    ((j_common_ptr) &cinfo, JPOOL_IMAGE, i, 3);
   for (i=0; i<cinfo.desired_number_of_colors; i++)
    {
     cinfo.colormap[0][i] = cmap[i*3];
     cinfo.colormap[1][i] = cmap[i*3+1];
     cinfo.colormap[2][i] = cmap[i*3+2];
    }
  }
#endif

 /* since we need colormap, this will actually do the whole uncompress */
 /* and we do not want it if we are just browsing in file selection box */
 if (realLoad)
  (void) jpeg_start_decompress(&cinfo);
 didStartDecompress = realLoad;

 *xpixels = cinfo.image_width;
 *ypixels = cinfo.image_height;

 sprintf(info, "JPEG");
 *loadType = ((cinfo.quantize_colors)||(cinfo.out_color_space==JCS_GRAYSCALE)) ?
   LOAD_CMAP : LOAD_TRUE_COL;

 if (cinfo.colormap)
  for (i=0; i<cinfo.desired_number_of_colors; i++)
   {
    if (cinfo.out_color_space==JCS_GRAYSCALE)
     cmap[i*3] = cmap[i*3+1] = cmap[i*3+2] = cinfo.colormap[0][i];
    else
     {
      cmap[i*3] = cinfo.colormap[0][i];
      cmap[i*3+1] = cinfo.colormap[1][i];
      cmap[i*3+2] = cinfo.colormap[2][i];
     }
   }
 if (cinfo.density_unit==1) /* dots per inch */
  {
   *xdpi= (int)(cinfo.X_density*1000);
   *ydpi= (int)(cinfo.Y_density*1000);
  }
 else
  *xdpi = *ydpi = 0; /* means we do not have a value here */
 *reverse = FALSE;
 *numInterlace = 1; /* no interlace */
 return(LOAD_OK);
}

/* loader reads a band */
static int loadBand(unsigned char *data, int  lines, int  bpl)
{
 JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
 int i;

 if (setjmp(jerr.setjmp_buffer))
  return(LOAD_INTERNAL_ERROR);

 for (i=0; i<lines; i++)
  {
   row_pointer[0]= &data[i*bpl];
   (void) jpeg_read_scanlines(&cinfo, row_pointer, 1);
  }
 return(LOAD_OK);
}

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 if (setjmp(jerr.setjmp_buffer)); /* do nothing on purpose */
 else
  {
   if (didStartDecompress)
    (void) jpeg_finish_decompress(&cinfo);
   didStartDecompress=FALSE;
   jpeg_destroy_decompress(&cinfo);
  }
 fclose(fp);
 return(LOAD_OK);
}

/* register jpeg image loader */
void registerLoadJPEG(void)
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
