/* loadimg.h */
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#ifndef min
#define max(a,b) 	((a)<(b) ? (b) : (a))
#define min(a,b) 	((a)>(b) ? (b) : (a))
#endif

#define LOAD_BW            0 /* 1=black */
			     /* 128 = 10000000 binary = leftmost pixel black */
#define LOAD_CMAP          1 /* colormapped image; up to 256 colors */
#define LOAD_TRUE_COL      2 /* rgbrgbrgb... probably should be quantized */

#define LOAD_OK               0
#define LOAD_USER_ABORT       1
#define LOAD_NO_MEMORY        2
#define LOAD_IO_ERROR         3
#define LOAD_INTERNAL_ERROR   4
#define LOAD_WRONG_IMG        5
#define LOAD_FILETYPE_UNKNOWN 6
#define LOAD_FILE_CORRUPT     7
#define LOAD_NO_FREE_COLORS   8

/* defines 0-6 are equal to PRT_OK .. PRT_WRONG_IMG */
#define LTXT_WAIT		10
#define LTXT_LOADING_WAIT	11
#define LTXT_CANCEL_LOAD	12

#define MAX_LTXT 		20

/* checks if image is in selected format and reads the header */
/* TRUE on success, FALSE otherwise */
typedef int (*tryLoadProc) (
	char *filename,
	int *loadType, /* LOAD_BW ... LOAD_TRUE_COL */
	int *xpixels,
	int *ypixels,
	int *xdpi, /* horizontal dots per inch *1000 */
	int *ydpi, /* vertical dots per inch *1000 */
	unsigned char *cmap, /* caller pre-allocated return colormap */
	char *info, /* caller preallocated loader info only 500 bytes!!! */
	char *reverse, /* return TRUE if bottom line is first in this format */
	int *numInterlace, /* number of interlace passes, 1 for no interlace */
	int realLoad /* TRUE if we are really loading, FALSE if we just want info */
	);

/* loader reads a band */
typedef int (*loadBandProc) (
	unsigned char *data, /* put the data in here */
	int  lines, /* number of lines in this band */
        int  bpl /* bytes per line */
	);

/* closes the input file and releases any temporary storage */
typedef int (*endLoadProc) ();

/* initializes the destination data for loading */
/* TRUE on success, FALSE otherwise */
typedef int (*initConsumeProc) (
	int loadType, /* LOAD_BW ... LOAD_TRUE_COL */
	int xpixels,
	int ypixels,
	int xdpi, /* horizontal dots per inch *1000 */
	int ydpi, /* vertical dots per inch *1000 */
	char reverse, /* TRUE if producer wants to feed image in reverse y order */
	int numInterlace, /* # of interlace passes; if different than 1 then band height must be whole image */
	unsigned char *cmap, /* caller pre-allocated colormap */
	unsigned char **data, /* allocate destination data */
	int *bpl, /* bytes per line */
	int *bandh, /* height of a band */
	int *progress /* advance in the data buffer or stay put */
	);

/* consume a band */
typedef int (*consumeBandProc) (
	unsigned char *data, /* get data from here */
	int  lines, /* number of lines in this band */
        int  bpl, /* bytes per line */
        int  ypos /* y position */
	);

/* frees any temporary storage */
typedef int (*endConsumeProc) (int status);

/* P U B L I C   F U N C T I O N S */

/* if numLoader is negative, returns number of loader functions, otherwise selects the */
/* one which is passed as parameter */
int returnLoaderFunctions(int numLoader);

/* tuneable color quantization function: trueData is input data in RGB, cmapData is colormapped
   output image. You set the requested number of colors, and the program returns the actual
   number. You can previously put in the colormap the colors you want to have
   in the image (regardless of histogram), and indicate this by setting the number of fixColors>0.
   If cmapData is NULL, then we just want back the reduced colormap, without remapping the image.
   If trueData is NULL, then we are doing quantization of colormapped image; do not actually
   change the data, ony return the colormap and actual colors; not yet implemented */
int colorQuantize(unsigned char *trueData, int trueBPL, unsigned char *cmapData, int cmapBPL,
 int sizeX, int sizeY, unsigned char *cmap, int fixColors, int *requestColors);
