/* lpng.c */
#include "loadimg.h"
#include <stdio.h>
#include <stdlib.h>
#include "png.h"

static FILE *fp;
static png_structp   png_ptr;
static png_infop     info_ptr;
static int lType, npasses, ytotal, readSoFar;

extern void registerLoader(tryLoadProc tLoad, loadBandProc lBand, endLoadProc eLoad);

static int tryLoad(char *filename, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int realLoad)
{
 char header[8];
 int           i;
 png_uint_32 width, height;
 int bit_depth, color_type, interlace_type, num_palette;
 png_colorp palette;

 if ((fp = fopen(filename, "rb" )) == NULL)
  return(LOAD_IO_ERROR);
 fread(header, 1, 8, fp);
 if (!png_check_sig(header, 8))
  { fclose(fp); return(LOAD_FILETYPE_UNKNOWN); }

 png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
 if (!png_ptr)
  return(LOAD_NO_MEMORY);

 info_ptr = png_create_info_struct(png_ptr);
 if (!info_ptr)
  {
   png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
   return(LOAD_NO_MEMORY);
  }

 if (setjmp(png_ptr->jmpbuf))
  {
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
   fclose(fp);
   return(LOAD_IO_ERROR);
  }

 png_init_io(png_ptr, fp);
 /* tell the library how many bytes of signature have we already read */
 png_set_sig_bytes(png_ptr, 8);
 png_read_info(png_ptr, info_ptr);
 png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
     &interlace_type, NULL, NULL);

 if (png_get_valid(png_ptr, info_ptr, PNG_INFO_PLTE))
  {
   png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette);
   for (i = 0;  i < num_palette;  i++)
    {
     cmap[i*3]=palette[i].red;
     cmap[i*3+1]=palette[i].green;
     cmap[i*3+2]=palette[i].blue;
    }
  }

 if (png_get_valid(png_ptr, info_ptr, PNG_INFO_pHYs))
  {
   png_uint_32 res_x, res_y;
   int unit_type;

   png_get_pHYs(png_ptr, info_ptr, &res_x, &res_y, &unit_type);
   *xdpi = (int) (res_x*25.4);
   *ydpi = (int) (res_y*25.4);
  }
 else
  *xdpi = *ydpi = 0;
 
 /* our treatment of black and white is inverted with respect to PNG so tell it */
 png_set_invert_mono(png_ptr);

 /* documentation says it will return 1 if it is not interlaced */
 npasses = png_set_interlace_handling(png_ptr);
 switch (color_type)
  {
   case PNG_COLOR_TYPE_PALETTE:
    *loadType=LOAD_CMAP;
    if (bit_depth < 8)
     png_set_packing(png_ptr);
    break;

   case PNG_COLOR_TYPE_RGB:
    if (bit_depth == 16)
     png_set_strip_16(png_ptr);
    *loadType=LOAD_TRUE_COL;
    break;

   case PNG_COLOR_TYPE_GRAY_ALPHA:
   case PNG_COLOR_TYPE_GRAY:
    if ((bit_depth < 8)&&(bit_depth>1))
     png_set_packing(png_ptr);

    if (color_type==PNG_COLOR_TYPE_GRAY_ALPHA)
     png_set_strip_alpha(png_ptr);

    *loadType=LOAD_CMAP;
    if (bit_depth == 16)
     { png_set_strip_16(png_ptr); bit_depth = 8; }
    switch (bit_depth)
     {
      case 1:
       cmap[0]=cmap[1]=cmap[2]=0;
       cmap[3]=cmap[4]=cmap[5]=255;
       *loadType=LOAD_BW;
       break;
      case 2:
       cmap[0]=cmap[1]=cmap[2]=0;
       cmap[3]=cmap[4]=cmap[5]=85; /* 255/3 */
       cmap[6]=cmap[7]=cmap[8]=170;
       cmap[9]=cmap[10]=cmap[11]=255;
       break;
      case 4:
       for (i = 0;  i < 16;  ++i)
	cmap[i*3]=cmap[i*3+1]=cmap[i*3+2]= i*17;
       break;
      case 8:
       for (i = 0;  i < 256;  ++i)
	cmap[i*3]=cmap[i*3+1]=cmap[i*3+2]= i;
       break;
     }
    break;

   case PNG_COLOR_TYPE_RGB_ALPHA:
    if (bit_depth == 16)
     png_set_strip_16(png_ptr);
    png_set_strip_alpha(png_ptr);
    *loadType=LOAD_TRUE_COL;
    break;
  }

 png_read_update_info(png_ptr, info_ptr);

 strcpy(info, "PNG");
 lType = *loadType;
 *xpixels = width;
 *ypixels = ytotal = height;
 *reverse = FALSE;
 *numInterlace = interlace_type;
 readSoFar=0;
 return(LOAD_OK);
}

/* loader reads a band */
static int loadBand(unsigned char *data, int  lines, int  bpl)
{
 png_bytep row_pointers[lines+1000];
 int i;

 /* need to set it for every function which calls routines from PNG library */
 if (setjmp(png_ptr->jmpbuf))
  return(LOAD_IO_ERROR); /* the acutal destroy code will be called in endLoad */

 for (i=0;  i < lines;  i++)
  row_pointers[i] = (png_bytep) &data[i*bpl];

 for (i=0; i<npasses; i++)
  png_read_rows(png_ptr, row_pointers, NULL, lines);
 readSoFar+=lines;
 /* cannot call this at endLoad, since sometimes endLoad did not read the whole image */
 if (readSoFar==ytotal)
  png_read_end(png_ptr, info_ptr);
 return(LOAD_OK);
}

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
 fclose(fp);
 return(LOAD_OK);
}

/* register PNG image loader */
void registerLoadPNG()
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
