/* print.h */
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#ifndef min
#define max(a,b) 		((a)<(b) ? (b) : (a))
#define min(a,b) 		((a)>(b) ? (b) : (a))
#endif

#define MAX_PRINT 20000 /* maximal x or y printable pixels for all printers */

#define IMG_BW            0 /* 1=black */
			    /* 128 = 10000000 binary = leftmost pixel black */
#define IMG_CMAP          1 /* colormapped image; up to 256 colors */
#define IMG_TRUE_COL_DITH 2 /* rgbrgbrgb... will be dithered when printed */
#define IMG_TRUE_COL      3 /* rgbrgbrgb... will print on trueColor system */

#define PRT_OK             0
#define PRT_USER_ABORT     1
#define PRT_NO_MEMORY      2
#define PRT_IO_ERROR       3
#define PRT_INTERNAL_ERROR 4
#define PRT_EXEC_ERROR     5
#define PRT_WRONG_IMG      6

/* defines 0-6 are equal to PRT_OK .. PRT_WRONG_IMG */
#define PTXT_WAIT		10
#define PTXT_PRINTING_WAIT	11
#define PTXT_CANCEL_PRINT	12
#define PTXT_SAVING_WAIT	13
#define PTXT_CANCEL_SAVE	14
#define PTXT_PRINTER		15
#define PTXT_PRINT_MODE		16
#define PTXT_DENSITY_X		17
#define PTXT_DENSITY_Y		18
#define PTXT_PAPER_WARNING	19
#define PTXT_PRINT_TO_FILE	20
#define PTXT_PRINT_TO_FILE_OK	21
#define PTXT_MODE_BW		22
#define PTXT_MODE_CMAP		23
#define PTXT_MODE_COLOR_DITHER	24
#define PTXT_MODE_COLOR_TRUE	25

#define MAX_PTXT 		30

typedef struct {
	int printType; /* IMG_BW ... IMG_TRUE_COL */
} printerMode;

typedef struct {
	int pixels; /* number of printable pixels */
	int dpi; /* dots per inch *1000 */
} printerDensity;

/* returns printer info */
typedef void (*getPrinterInfoProc) (
	printerMode *printerMod, /* fill in the preallocated printer modes */
	int *numModes, /* number of modes returned */
	printerDensity *prtDensityX, /* fill in the preallocated X density table */
        int *numDensitX, /* number of X densities returned */
	printerDensity *prtDensityY, /* fill in the preallocated Y density table */
        int *numDensitY, /* number of Y densities returned */
	int *nonprintableTop, /* nonprintable mm on top for paper handling */
	int *nonprintableBottom, /* nonprintable mm on bottom */
	int *nonprintableLeft, /* nonprintable mm on left */
	int *dpiVariable, /* print density is user variable */
        char **prtName, /* string containing printer's brand name */
        char **profileName /* string containing printer's profile name */
	);

/* application returns next band to be printed */
typedef int (*getNextBandProc) (
	int xpixels, /* number of horizontal pixels */
	int xdpi, /* horizontal dots per inch *1000 */
	int ydpi, /* vertical dots per inch *1000 */
	int ypos, /* requested y position of band within ytotal */
	int ytotal, /* total length of image to be printed in pixels */
	char desiredImageType, /* IMG_BW ... IMG_TRUE_COL */
	char **data, /* return data; to be freed by caller!!! */
        int  *bpl, /* bytes per line */
	int  *lines, /* number of return lines */
	char *returnImageType, /* IMG_BW ... IMG_TRUE_COL */
	unsigned char *cmap /* [256*3]; rgbrgbrgb; caller should preallocate it */
	);

/* printer driver processes the band */
typedef int (*printBandProc) (
	int xpixels, /* number of horizontal pixels */
	char *data, /* data; processBand must free it */
        int  bpl, /* bytes per line */
        int  ypos, /* current y position on page */
	int  lines, /* number of lines in this band */
	char bandType, /* IMG_BW ... IMG_TRUE_COL */
	unsigned char *cmap /* [256*3]; rgbrgbrgb */
	);

/* initializes the printer to the desired print type and page size */
/* opens oputput temporary files */
/* TRUE on success, FALSE otherwise */
typedef int (*initPrintProc) (
	int printType, /* IMG_BW ... IMG_TRUE_COL */
	int complexImage, /* TRUE or FALSE */
	int xpixels,
	int ypixels,
	int xdpi, /* horizontal dots per inch *1000 */
	int ydpi, /* vertical dots per inch *1000 */
	int xmargin, /* in pixels from top of printable page */
	int ymargin /* in pixels from left printable border */
	);

/* emits printer dependent end of page/eject page code, */
/* closes temporary files, enques the files in the print queue, removes them; */
/* returns TRUE if enqueued OK, FALSE othervise */
typedef int (*endPrintProc) (
	int normal /* PRT_OK ... PRT_WRONG_IMG */
	);

/* P U B L I C   F U N C T I O N S */

/* returns the number of printer drivers, gets the location of download files */
int initPrintLib(char *path);

/* sets the functions of currently selected printer */
void setPrinterFunctions(int numPrinter);

