/* register.c */

/*   plib - generic unifed library for printing and saving in graphical file formats
 *   distributed as part of ArahPaint
 *
 *   Copyright (C) 1993-2001  Arahne d.o.o.
 *   http://www.arahne.si
 *   E-mail: arahne@arahne.si
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "print.h"

char *matrixPath;

extern void registerPNG(void);
extern void registerJPEG(void);
extern void registerPbmPlus(void);
extern void registerPostScript(void);
extern void registerTIFF(void);
extern void registerStylus(void);
extern void registerStylusPro(void);
extern void registerStylusProXL(void);
extern void registerStylus800(void);
extern void registerStylus850(void);
extern void registerStylus1520(void);
extern void registerStylus3000(void);
extern void registerStylusPhoto(void);
extern void registerStylusPhoto700(void);
extern void registerStylusPhotoEX(void);
extern void registerStylus5000(void);
extern void registerStylusPhoto750(void);
extern void registerStylusPhoto1200(void);
extern void registerStylusPhoto870(void);
extern void registerStylusPhoto1270(void);

#define NUM_PRINTERS	20 /* must match the number of functions above */

/* returns the number of printer drivers, gets the location of dithering matrix */
int initPrintLib(char *path)
{
 matrixPath = path;
 return(NUM_PRINTERS);
}

/* sets the functions of currently selected printer */
void setPrinterFunctions(int numPrinter)
{
 switch (numPrinter)
  {
   case 0: registerPNG(); break;
   case 1: registerJPEG(); break;
   case 2: registerPbmPlus(); break;  
   case 3: registerPostScript(); break;  
   case 4: registerTIFF(); break;  
   case 5: registerStylus(); break;  
   case 6: registerStylusPro(); break;  
   case 7: registerStylusProXL(); break;  
   case 8: registerStylus800(); break;  
   case 9: registerStylus850(); break;  
   case 10: registerStylus1520(); break;  
   case 11: registerStylus3000(); break;  
   case 12: registerStylusPhoto(); break;  
   case 13: registerStylusPhoto700(); break; 
   case 14: registerStylusPhotoEX(); break;  
   case 15: registerStylus5000(); break;  
   case 16: registerStylusPhoto750(); break;  
   case 17: registerStylusPhoto1200(); break;  
   case 18: registerStylusPhoto870(); break;  
   case 19: registerStylusPhoto1270(); break;  
 }
}

#define MATS  65536
/* allocates and returns the dither matrix of size 256*256 and distribution */
/* of values ranging from 0 to 255 according to the value of distr[256] */
/* array. The sum of all values in the distribution array must be 65536 */
unsigned char *makeDither(unsigned short *distr)
{
 unsigned char *dith;
 unsigned char mat[MATS*2], *matx, *maty;
 int i, distrPos;
 char filen[100];
 FILE *fp;

 matx=mat;
 maty=&mat[MATS];
 strcpy(filen, matrixPath);
 strcat(filen, "/matrix");
 if ((fp = fopen(filen, "r" )) == NULL)
  { printf("open %s failed!\n", filen); return(NULL); }
 fread(mat, 1, MATS*2, fp);
 fclose(fp);

 dith = malloc(MATS);
 if (dith==NULL)
  return(NULL);

 distrPos=0;
 for (i=0; i<MATS; i++)
  {
   while ((distr[distrPos]==0)&&(distrPos<255)) distrPos++;
   dith[(maty[i]<<8) + matx[i]] = (unsigned char) distrPos;
   distr[distrPos]--;
  }
 return(dith);
}
